;;; -*- Mode: Lisp; Package: ystok.odbc -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2005 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Access database - Specialization for ODBC
;;; Fatal BUGs
;;; 1) Driver odbcjt32.dll 4.0.6200.0 and msjet40.dll 4.0.6218.0 (Jet 4.0 Service Pack 6)
;;;    SQLFetch() and SQLFetchScroll.
;;;    For block cursor (size > 1) and forward-only (tested), the driver sometimes
;;;	- sets SQL_ATTR_ROWS_FETCHED_PTR incorrectly,
;;;	- fills in array SQL_ATTR_ROW_STATUS_PTR incorrectly.
;;;    All these worked fine with Jet 3.5.
;;;    Workaround: Always use rowset-size = 1 (default)!

(in-package :ystok.odbc)

(yl:export-toplevel ()
  #:*access-driver*
  #:access-odbc-database
  #:access-3.5-database
  ;#:create-database
  #:compact-database
  #:*excel-driver*
  #:excel-odbc-database
  #:repair-database
)

;;; Jet 3.5 database format
(defclass access-3.5-database (access-odbc-database) ())

(defmethod db-lisp-to-c-type ((database desktop-mixin) (lisp-type (eql 'boolean)))
  SQL_C_BIT)

(defmethod db-describe-columns ((database access-mixin) table
                                &key column catalog schema)
  ;; Access NB:
  ;;   1)   REMARKS    (  )
  ;;        \0  , ..  ,   .
  ;;   2)   Null  . COLUMN_DEF The default value of the column.
  ;;   3)  Jet4   Unicode-      out-len-ind:
  ;;      - -65536  SQL_NULL_DATA  TABLE_SCHEM
  ;;	  - 0    COLUMN_NAME.
  (declare (ignore schema))
  (mapcar (lambda (description)
            (when-bind (remarks (nth 11 description))
              (setf (nth 11 description) (subseq remarks 0 (position #\Null remarks))))
            description)
          (with-cursor-handle-results (handle database)
            (%sql-columns handle (name catalog) nil
                          (ordinary-argument table "%")
                          (ordinary-argument column "%")))))

(defmethod db-list-columns ((database access-mixin) table schema &key catalog)
  (declare (ignore schema))
  (let ((cursor (obtain-statement database 'odbc-cursor)))
    (unwind-protect       
        (progn
          (setf (expression cursor) '%sql-columns)
          (%initialize-cursor cursor :forward-only)
          (%sql-columns (handle cursor) (name catalog) nil (ordinary-argument table) "%")
          (%initialize-rowset cursor :column-numbers '(4)) ; COLUMN_NAME 4
          (db-fetch-rows cursor :flatp t))
      (db-close cursor))))

(defmethod db-describe-foreign-keys ((database access-mixin) pk-table fk-table
                                     &key pk-catalog pk-schema fk-catalog fk-schema)
 ;;; Simulate SQLForeignKeys() by means of selecting from MSysRelationships table.
  ;; NB: Requires an explicit Read Data permission on the MSysRelationships table for
  ;;  the user connected; the permission inherited implicitly via groups does not work.
  (declare (ignore pk-catalog pk-schema fk-catalog fk-schema))
  (values (map-query 'list
                     (lambda (ccolumn grbit icolumn szColumn szObject
                              szReferencedColumn szReferencedObject szRelationship)
                       (declare (ignore ccolumn))
                       (list nil nil szReferencedObject szReferencedColumn
                             nil nil szObject szColumn
                             (1+ icolumn)
                             (if (logand grbit 256)	; dbRelationUpdateCascade
                                 SQL_CASCADE
                                 SQL_NO_ACTION)
                             (if (logand grbit 4096)	; dbRelationDeleteCascade
                                 SQL_CASCADE
                                 SQL_NO_ACTION)
                             szRelationship		; for simplicity
                             nil
                             SQL_NOT_DEFERRABLE))
                     (format nil			; <> dbRelationDontEnforce
                             "SELECT * FROM MSysRelationships WHERE (grbit <> 2)~
 ~:[~; AND (szReferencedObject LIKE '~A')~]~:[~; AND (szObject LIKE '~A')~]~
 ORDER BY szReferencedObject,szObject,icolumn"
                             pk-table (ordinary-argument pk-table)
                             fk-table (ordinary-argument fk-table))
                     :database database)
          '("PKTABLE_CAT" "PKTABLE_SCHEM" "PKTABLE_NAME" "PKCOLUMN_NAME"
            "FKTABLE_CAT" "FKTABLE_SCHEM" "FKTABLE_NAME" "FKCOLUMN_NAME"
            "KEY_SEQ" "UPDATE_RULE" "DELETE_RULE" "FK_NAME" "PK_NAME" "DEFERRABILITY")
       ))

(defmethod db-describe-primary-keys ((database access-mixin) table &key catalog schema)
  ;; As is not supported by MS Access driver, use db-describe-special-columns instead.
  ;; NB: Constraint name "PrimaryKey" can be an inappropriate guess.
  (declare (ignore schema))
  (loop with table-name = (ordinary-argument table)
        for (nil column-name nil nil nil nil nil pseudo-column)
         in (db-describe-special-columns database table
                                         :type SQL_BEST_ROWID
                                         :scope SQL_SCOPE_CURROW
                                         :nullable nil :catalog catalog)
        and position fixnum upfrom 1
        when (eql pseudo-column SQL_PC_NON_PSEUDO)
        collect (list (name catalog) nil table-name column-name position "PrimaryKey")))

(defmethod db-describe-statistics ((database access-mixin) table
                                   &key unique (reserved 0) catalog schema)
  (declare (ignore schema))					; don't pass schema
  (call-next-method database table :unique unique :reserved reserved
                    :catalog catalog :schema nil))

(defmethod db-list-tables ((database access-mixin) schema &key table catalog)
  (let ((cursor (obtain-statement database 'odbc-cursor)))
    (unwind-protect       
        (progn
          (setf (expression cursor) '%sql-tables)
          (%sql-tables (handle cursor)
                       (name catalog) nil
                       (ordinary-argument table "%")
                       (case schema
                         (:sys "SYSTEM TABLE")
                         (:all "TABLE,SYSTEM TABLE")   ; "SYNONYM,VIEW" ?
                         (:view "VIEW")
                         (otherwise "TABLE")))
          (%initialize-rowset cursor :column-numbers '(3)) ; TABLE_NAME 3
          (db-fetch-rows cursor :flatp t))
      (db-close cursor))))

(defmethod db-list-views ((database access-mixin) schema &key view catalog)
  (let ((cursor (obtain-statement database 'odbc-cursor)))
    (unwind-protect       
        (progn
          (setf (expression cursor) '%sql-tables)
          (%sql-tables (handle cursor)
                       (name catalog) nil
                       (ordinary-argument view "%") "VIEW")
          (%initialize-rowset cursor :column-numbers '(3)) ; TABLE_NAME 3
          (db-fetch-rows cursor :flatp t))
      (db-close cursor))))

(defmethod db-drop-view ((database access-odbc-database) view
                         &key (if-does-not-exist :error))
 ;;; Neither Jet 3.5 not 4.0 supports DROP VIEW (though doc on 4.0 mentions it).
  (db-drop-table database view :if-does-not-exist if-does-not-exist))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; As connection specification, pathname accompanied by explicit database-type is allowed:
;;;  (sql:connect pathname :database-type :access)
;;; In long form, driver specification string can be used in pattern like this:
;;;  (sql:connect nil :driver-av-pairs `((:dbq ,pathname) (:driver ,odbc:*access-driver*)))

(defparameter *access-driver* "Microsoft Access Driver (*.mdb)")

(defmethod db-name-from-spec ((db-type (eql :access)) (connection-spec list))
  (declare (ignore db-type))
  (db-name-from-spec :odbc connection-spec))

(defmethod db-user-name-from-spec ((db-type (eql :access)) (connection-spec list))
  (declare (ignore db-type))
  (db-user-name-from-spec :odbc connection-spec))

(defmethod driver-connect ((db odbc-database) (be-type (eql :access))
                           (driver-av-pairs list) &rest args &key dsn)
 ;;; Provide pre-system and -user driver attribute completion.
  ;; Args: driver-av-pairs - List of (attribute value) pairs
  (declare (ignore be-type))
  (unless (assoc #1="DRIVER" driver-av-pairs :test #'string-equal)
    (push (list #1# *access-driver*) driver-av-pairs))
  (when (and (pathnamep dsn) (not (assoc #2="DBQ" driver-av-pairs :test #'string-equal)))
    (push (list #2# (namestring dsn)) driver-av-pairs)
    (setq args (remove-properties args '(:dsn))))
  (apply #'driver-connect db :odbc driver-av-pairs args))

(defmethod db-create ((type (eql :access)) pathname &rest args &key version sort-order)
 ;;; Args: systemdb
  (without-properties (args args :version :sort-order)
    (apply #'%sql-config-data-source ODBC_CONFIG_DSN *access-driver*
           (format nil "CREATE_DB~[~;~;V2~;V3~;V4~]=~A~@[ ~A~]"
                   (case version (2 2)
                                 ((97 1997 3) 3)
                                 ((2000 4) 4)
                                 (otherwise (if (eql version 3.5) 3 0)))
                   (namestring (first-or-self pathname))
                   (and sort-order (string sort-order)))
           args)))

(defmethod db-create ((db-type (eql 'access-odbc-database)) pathname &rest args &key)
  (apply #'db-create :access pathname args))

(defmethod db-create ((db-type (eql 'access-3.5-database)) pathname &rest args &key version)
  (apply #'db-create :access pathname
         :version (or version 3)
         (remove-properties args '(:version))))


(defmethod db-destroy ((db-type (eql :access)) pathname &key)
  (delete-file (first-or-self pathname)))


(defun compact-database (pathname 
                         &rest args &key destination sort-order &allow-other-keys)
  (let ((namestring (namestring pathname)))
    (without-properties (args args :destination :sort-order)
      (apply #'%sql-config-data-source ODBC_CONFIG_DSN *access-driver*
             (format nil "COMPACT_DB=~A ~A~@[ ~A~]"
                    namestring 
                    (if destination (namestring destination) namestring)
                    (and sort-order (string sort-order)))
             args))))

(defun repair-database (pathname &rest args)
  (apply #'%sql-config-data-source ODBC_CONFIG_DSN *access-driver* nil
         :repair_db (namestring pathname)
         args))

(defvar *excel-driver* "Microsoft Excel Driver (*.xls)")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  ERROR HANDLING  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Signal another condition instead of the poorly informative odbc-error.
;;; According to Microsoft Jet Database Engine Programmer's Guide,
;;; the native database error codes should be different!
;;;
;;;   sec-error-id  MSJDBEPG value  Message
;;;   ------------  --------------  --------------------------------
;;;	-1605	    3022	    see below
;;;	-1612	    3200	    see below
;;;	-3701	    3314	    see below
;;;	-3703	    3316	    <Table-level validation text>
;;;	?	    3317	    One or more values are prohibited by the validation
;;;				    rule '<rule>' set for '<name>'. Enter a value that
;;;				    the expression for this field can accept.

(defmethod odbc-error ((database access-odbc-database) sql-state
                       (secondary-error-id (eql -1605))
                       &key expression message)
  ;;	3022	The changes you requested to the table were not successful because
  ;;		they would create duplicate values in the index, primary key, or
  ;;		relationship.
  ;;		Change the data in the field or fields that contain duplicate data,
  ;;		remove the index, or redefine the index to permit duplicate entries
  ;;		and try again.
  (if (string= sql-state "23000")
      (error 'sql-unique-error
             :table (name (table expression))
             :columns (mapcar 'name (columns expression))
             :database database :expression expression
             :sql-state sql-state  :secondary-error-id secondary-error-id
             :message message)
      (call-next-method)))

;;; Instead of sql-state HY000 (General error), the driver should return one of:
;;;   23000	Integrity constraint violation
;;;   40002	Integrity constraint violation (on committing transaction).

(defmethod odbc-error ((database access-odbc-database) sql-state
                       (secondary-error-id (eql -3701))
                       &key expression message)
;;;	3314    The field '[<table-name>.]<name>' cannot contain a
;;;		Null value because the Required property for this
;;;		field is set to True. Enter a value in this field. 
  (if (string= sql-state "HY000")
      (let* ((start1 (position #\' message))
             (end1 (and start1 (position #\. message :start (incf start1))))
             (start2 (if end1 (1+ end1) start1))
             (end2 (and start2 (position #\' message :start start2))))
        (error 'sql-not-nullable-error
               :table (if end1 (subseq message start1 end1) nil)
               :column (if end2 (subseq message start2 end2) nil)
               :database database :expression expression
               :sql-state sql-state  :secondary-error-id secondary-error-id
               :message message))
      (call-next-method)))

(defmethod odbc-error ((database access-odbc-database) sql-state
                       (secondary-error-id (eql -1612))
                       &key expression message)
 ;;; Signal another condition instead of the poorly informative odbc-error.
  ;;	3200	The record cannot be deleted or changed because table '<name>' includes
  ;;		related records.
  ;;	3201?	You can't add or change a record because a related record is required
  ;;		in table '<name>'.
  (if (string= sql-state "HY000")
      (let* ((start (position #\' message))
             (end (and start (position #\' message :start (1+ start)))))
        (error 'sql-foreign-key-error
               :table (if end (subseq message (1+ start) end) nil)
               :database database :expression expression
               :sql-state sql-state  :secondary-error-id secondary-error-id
               :message message))
      (call-next-method)))

  
#||
(create-database #P"E:/Lisp/Test/sam2.mdb" :version 3 :sort-order :russian)
(compact-database #P"E:/Lisp/Test/sam2.mdb"
                  :destination #P"E:/Lisp/Test/sam3.mdb" :sort-order "Russian")
(compact-database #P"E:/Projects/Test/test2K.mdb" :sort-order "Russian")

(repair-database #P"E:/Lisp/Test/sam2.mdb")
(compact-database #P"E:/Projects/Test/YCATABG.MDB" :hwnd t
                  :systemdb (namestring #P"E:/Projects/divanov.mdw") :uid 'divanov)

(%sql-installer-error 2)

(db-describe-foreign-keys *default-database* nil nil)
||#
