;;; -*- Mode: Lisp; Package: ystok.odbc -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2005 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Oracle database - Specialization for ODBC
;;;
;;; (1) Schema names seem to be case-sensitive in Ora10 driver (created uppercase
;;;	in server).

(in-package :ystok.odbc)
(in-syntax *sql-readtable*)

(yl:export-toplevel ()
  #:*oracle-driver*
)

#+local-time
(defmethod output-sql ((expr local-time:local-time) (database oracle-odbc-database))
 ;;; Oracle driver seems to correctly return 0 as maximum scale for SQL_TYPE_TIMESTAMP
  ;; type and does not pass fractional seconds from {ts...} literals to the server.
  ;; Work around the driver and specify 3 as fractional second precision.
  (write-string "TO_TIMESTAMP('" *sql-stream*)
  (let ((*decimal-separator* #\.))
    (local-time:format-timestring *sql-stream* expr nil nil 4 4 #\- #\: #\Space))
  (write-string "','YYYY-MM-DD HH24:MI:SS.FF3')" *sql-stream*))

(defmethod db-sql-to-c-type ((database oracle-odbc-database) (sql-type (eql #.SQL_DECIMAL))
                             column-size decimal-digits)
 ;;; Ora8: Avoid unsupported SQL_C_NUMERIC
  ;; Ora10: Dispite the declaration of fixed Bug2945992 in driver version 10.1.0.2.0,
  ;;        SQL_C_NUMERIC seems neither to work against 10g nor 8i server.
  (cond ((/= decimal-digits 0)				; scale /= 0
         ;SQL_C_NUMERIC
         (if (<= column-size 15) SQL_C_DOUBLE		; precision <= 15            
                                 SQL_C_CHAR)
        )
        ((< column-size 5) SQL_C_SSHORT)		; scale = 0, precision < 5
        ((< column-size 10) SQL_C_SLONG) 		; scale = 0, precision < 10
        (t (values SQL_C_CHAR
                   (display-size sql-type column-size decimal-digits)))))
                               ;(%get-type-descriptor database sql-type))))

(defmethod db-lisp-to-c-type ((database oracle-odbc-database) (lisp-type (eql 'boolean)))
  (values SQL_C_CHAR 1))

(defmethod db-describe-columns ((database oracle-odbc-database) table
                                &key column catalog schema)
 ;;; All drivers put garbage into SQL_DATETIME_SUB (15) for char columns.
  ;; Both 8 and 10 neither returns REMARKS (12) nor COLUMN_DEF (13) (always "" for both)
  ;; Query REMARKS from
  ;; - either USER_COL_COMMENTS => ("TABLE_NAME" "COLUMN_NAME" "COMMENTS")
  ;; - or ALL_COL_COMMENTS => ("OWNER" "TABLE_NAME" "COLUMN_NAME" "COMMENTS") 
  ;; We could also extract the COLUMN_DEF by means of quering
  ;;	SELECT COLUMN_NAME, DEFAULT_LENGTH, DATA_DEFAULT FROM USER_TAB_COLUMNS 
  ;;	WHERE TABLE_NAME='MYTAB';
  ;; or
  ;;	SELECT CNAME, DEFAULTVAL FROM COL WHERE TNAME='MYTAB';
  (let* ((view (sql-expression :table (if schema "ALL_COL_COMMENTS" "USER_COL_COMMENTS")
                               :catalog catalog))
         (where (apply #'sql-operation 'and [not [null [comments]]]
                       (append (when (and schema (neq schema :all))
                                 (list [= [owner] (ordinary-argument schema)]))
                               (when table
                                 (list [= [table-name] (ordinary-argument table)]))
                               (when column
                                 (list [= [column-name] (ordinary-argument column)])))))
         (col-comments (sql:select [*] :from view :where where :database database)))
    (multiple-value-bind (list columns)
        (with-cursor-handle-results (handle database)
          (%sql-columns handle (name catalog)
                        (cond ((eq schema :all)
                               SQL_ALL_SCHEMAS)
                              ((ordinary-argument schema))
                              (t (string-upcase (user-name database))))
                        (ordinary-argument table "%")
                        (ordinary-argument column "%")))
      (dolist (description list)
        (let ((owner (second description))
              (table-name (third description))
              (column-name (fourth description)))
          (when-bind (tuple (find-if (if schema
                                         (lambda (tuple)
                                           (and (equal (first tuple) owner)
                                                (equal (second tuple) table-name)
                                                (equal (third tuple) column-name)))
                                         (lambda (tuple)
                                           (and (equal (first tuple) table-name)
                                                (equal (second tuple) column-name))))
                                     col-comments))
            (setf (nth 11 description) (first (last tuple))))))
      (values list columns))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; As connection specification, pathname accompanied by explicit database-type is allowed:
;;;  (sql:connect pathname :database-type :oracle)

(defparameter *oracle-driver* "Oracle ODBC Driver") ; "Oracle in instantclient10_2"

(defmethod db-name-from-spec ((db-type (eql :oracle)) (connection-spec list))
  (declare (ignore db-type))
  (db-name-from-spec :odbc connection-spec))

(defmethod db-user-name-from-spec ((db-type (eql :oracle)) (connection-spec list))
  (declare (ignore db-type))
  (db-user-name-from-spec :odbc connection-spec))

(defmethod driver-connect ((db odbc-database) (be-type (eql :oracle))
                           (driver-av-pairs list) &rest args)
 ;;; Provide pre-system and -user driver attribute completion.
  ;; Args: driver-av-pairs - List of (attribute value) pairs, must include:
  ;;		DBQ=TNS Service Name
  (declare (ignore be-type))
  (unless (assoc :driver driver-av-pairs :test #'string-equal)
    (push (list :driver *oracle-driver*) driver-av-pairs))
  (apply #'driver-connect db :odbc driver-av-pairs args))

#|
(defmethod db-create ((database-type (eql :oracle)) pathname &rest args))
(defmethod db-create ((database-type (eql 'oracle-odbc-database)) pathname &rest args)
  (apply #'db-create :oracle pathname args))

(defmethod db-destroy ((type (eql :oracle)) pathname &key))
  (delete-file (first-or-self pathname)))
|#
