;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2006 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Oracle database - Specialization for SQL expressions and types
;;; (tested on Oracle 8.1.7, 10.1)
;;; CLSQL maps bigint, universal-time to CHAR(20)?!
;;; Foreign key NB: 10.1 only supports ON DELETE SET NULL or CASCASE, but not ON UPDATE!

(in-package :ystok.sql.sys)

;;; Types

(defmethod db-sql-type-specifier ((database oracle-mixin) (lisp-type (eql 'boolean))
                                  &rest args)
  (declare (ignore database lisp-type args))
  "CHAR(1)")

(defmethod db-sql-type-specifier ((database oracle-mixin) (lisp-type (eql 'string))
                                  &rest args)
  (declare (ignore database lisp-type))
  (if args
      (let ((length (first args)))
        (if (<= length 4000)
            (format nil "VARCHAR(~A)" length)
            "LONG"))
      "VARCHAR(4000)"))			; as max length is required.

(defmethod db-sql-type-specifier ((database oracle-mixin) (lisp-type (eql 'bignum))
                                  &rest args)
  (declare (ignore database lisp-type args))
  "INTEGER")				; actually means "DECIMAL(38)"

(defmethod db-sql-type-specifier ((database oracle-mixin) (lisp-type (eql 'integer))
                                  &rest args)
 ;;; Try to fit into SQL_INTEGER
  (declare (ignore database lisp-type))
  (format nil "DECIMAL(~D)"
          (if args
              (destructuring-bind (low high) args
                (max (length (write-to-string (abs low) :base 10 :radix nil))
                     (length (write-to-string (abs high) :base 10 :radix nil))))
              10)))

(defmethod db-sql-type-specifier ((database oracle-mixin)
                                  (lisp-type (eql 'unsigned-byte)) &rest args)
  (declare (ignore database lisp-type))
  (format nil "DECIMAL(~D)" (if (null args) 3 (ceiling (first args) 0.30103))))

(defmethod db-sql-type-specifier ((database oracle-mixin) (lisp-type (eql 'bit-vector))
                                  &rest args)
  (declare (ignore database lisp-type))
  (if args (format nil "RAW(~A)" (ceiling (first args) 8)) "RAW"))

(defmethod db-sql-type-specifier ((database oracle-mixin)
                                  (lisp-type (eql 'simple-bit-vector)) &rest args)
  (apply #'db-sql-type-specifier database 'bit-vector args))

(defmethod db-sql-type-specifier ((database oracle-mixin) (lisp-type (eql 'vector))
                                  &rest args)
  (declare (ignore database lisp-type))
  (if args
      (destructuring-bind (type length) args
        (ecase (if (consp type) (first type) type)
          ((signed-byte unsigned-byte)
           (when (and (consp type) (second type))
             (setq length (* length (ceiling (second type) 8))))
           (if (<= length 2000)
               (format nil "RAW(~A)" length)
               "LONG ROW"))))
      "RAW"))

#|;;; Do we need this? Leaving TIMESTAMP as is works for 10g at least.
(defmethod db-sql-type-specifier ((database oracle-mixin) (lisp-type (eql :timestamp))
                                  &rest args)
  (declare (ignore lisp-type args))
  (if (string<= (odbc:db-get-info database odbc::SQL_DBMS_VER) ; "10.01.0020", "08.01.0741"
                "08" :end1 2)
      "DATE"
      (call-next-method))|#

(defmethod db-sql-type-specifier ((database oracle-mixin) (lisp-type (eql 'universal-time))
                                  &rest args)
  (declare (ignore database lisp-type args))
  "TIMESTAMP(0)")

;;; "TIMESTAMP WITH LOCAL TIME ZONE" seems to be stored as BLOB, at least Oracle ODBC
;;; driver reports as SQL_LONGVARBINARY (-4) and is buggy.
#+local-time
(defmethod db-sql-type-specifier ((database oracle-mixin)
                                  (lisp-type (eql 'local-time:local-time)) &rest args)
  (declare (ignore database lisp-type args))
  "TIMESTAMP(3)")			; only milliseconds are stored in local-time

;;; Boolean types are represented by 'F' and 'T'
(defmethod output-sql ((expr (eql :false)) (database oracle-mixin))
  (declare (ignore database))
  (write-string "'F'" *sql-stream*))
(defmethod output-sql ((expr (eql :true)) (database oracle-mixin))
  (declare (ignore database))
  (write-string "'T'" *sql-stream*))

(defmethod output-sql ((expr sql-like-exp) (database oracle-mixin))
  (with-slots (components plist) expr
    (case (getf plist :case)
      ((:upcase :ignore)
       (write-string "(UPPER(" *sql-stream*)
       (output-sql (first components) database)
       (write-string ") LIKE UPPER(" *sql-stream*)
       (output-sql (second components) database)
       (%output-plist (remove-properties plist '(:case)) database)
       (write-string "))" *sql-stream*))
      (otherwise
       (call-next-method)))))

;;; Functions

(defsql-function (:case t) NVL)

(defconstant +oracle-function-replacements+
   ;; Q: As CURRENT_TIMESTAMP returns a value of datatype TIMESTAMP WITH TIME ZONE,
   ;;    replace it with LOCALTIMESTAMP[(fractional_second_precision)]?
  '(;(current-timestamp . localtimestamp)
    (round . round)))		; Scalar function escape {fn ROUND(...)} does not work

(yl:keep-symbol-names (mapcar #'cdr +oracle-function-replacements+))

(defmethod output-sql ((expr sql-function-exp) (database oracle-mixin))
 ;;; Replace some standard SQL-99 or ODBC scalar functions by builtin Oracle's ones.
  (if-bind (fn (cdr-assoc (name expr) +oracle-function-replacements+ :test #'string=))
    (%output-function-exp fn (slot-value expr 'args) database)
    (call-next-method)))

(defmethod output-sql ((expr sql-cast-exp) (database oracle-mixin))
  (%output-cast-exp expr database))

;;; Sentences

(defmethod output-sql ((expr sql-set-query-exp) (database oracle-mixin))
 ;;; EXCEPT is not supported - use MINUS instead.
  (with-slots (operator components modifier) expr
    (let ((*in-subselect* nil))
      (dolist (arg components)
        (write-char #\Space *sql-stream*)
        (write-string (let ((string (string operator)))
                        (if (string-equal string "EXCEPT") "MINUS" string))
                      *sql-stream*)
        (when modifier
          (write-char #\Space *sql-stream*)
          (write-string (string modifier) *sql-stream*))
        (write-char #\Space *sql-stream*)
        (output-sql arg database)))))

;;; DB-methods

(defmethod db-type-initialize ((db-type (eql :oracle)) &optional version)
 ;;; Simulate type-subtype relationship between :odbc and :oracle
  ;; and map them both to the odbc-environment.
  (or (find-environment :odbc :errorp nil)
      (db-type-initialize :odbc version)))

(defmethod db-connect ((db-type (eql :oracle)) connection-spec 
                       &rest args &key &allow-other-keys)
  ;; Push the specific back-end type into args for possible retreival via driver-connect
  (apply #'db-connect :odbc connection-spec  :database-type :oracle  args))


#||
(defmethod output-sql ((expr sql-drop-index-exp) (database oracle-mixin))
 ;;; Ignore ON and IF-DOES-NOT-EXISTS whatsoever
  (with-slots (name) expr
    (write-string "DROP INDEX " *sql-stream*)
    (output-sql-identifier name database)))


(defmethod db-drop-index ((database oracle-mixin) index &rest args
                          &key on if-does-not-exist)
 ;;; Ignore ON and IF-DOES-NOT-EXISTS whatsoever.
  (declare (ignore on if-does-not-exist))
  (execute-command (write-drop-exp-to-string :index index database
                     (remove-properties args '(:on :if-does-not-exist)))
                   :database database))
||#