;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2006 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; PostgreSQL database - Specialization for SQL expressions and types

(in-package :ystok.sql.sys)

;;; Type specifiers

(defmethod db-sql-type-specifier ((database postgresql-mixin)
                                  (lisp-type (eql 'unsigned-byte)) &rest args)
  (declare (ignore database lisp-type))
  (if (or (null args) (<= (first args) 16))
      "SMALLINT"			; there is no byte or int1 type in PostgreSQL
      "INTEGER"))

(defmethod db-sql-type-specifier ((database postgresql-mixin) (lisp-type (eql :clob))
                                  &rest args)
  (declare (ignore database lisp-type args))
  "text")

(defmethod db-sql-type-specifier ((database postgresql-mixin) (lisp-type (eql :blob))
                                  &rest args)
  (declare (ignore database lisp-type args))
  "bytea")


(defmethod output-sql ((expr string) (database postgresql-mixin))
 ;;; Construct character string literal and escape backslashes
  ;; Q: Shall we support deprecated C-style backslash escapes
  ;;	- \b  backspace, \f  form feed, \n  newline, \r  carriage return, \t  tab;
  ;;    - \digits, where digits represents an octal byte value;
  ;;    - \xhexdigits, where hexdigits represents a hexadecimal byte value
  ;;   or E'\ooo' syntax?
  (write-char #\' *sql-stream*)
  (loop for char across expr
        if (char= char #\') do (write-char #\' *sql-stream*)
        else if (char= char #\\) do (write-char #\\ *sql-stream*)
        do (write-char char *sql-stream*))
  (write-char #\' *sql-stream*))

(defmethod output-sql ((expr sql-like-exp) (database postgresql-mixin))
  (with-slots (components plist) expr
    (case (getf plist :case)
      (:ignore
       (%output-relational-exp "ILIKE" components database
                               (remove-properties plist '(:case))))
      (otherwise
       (call-next-method)))))


;;; Functions

(defsql-function () trunc)

(defconstant +postgresql-function-replacements+
  '((truncate . trunc)))

(yl:keep-symbol-names (mapcar #'cdr +postgresql-function-replacements+))

(defmethod output-sql ((expr sql-function-exp) (database postgresql-mixin))
 ;;; Replace some standard SQL-99 or ODBC scalar functions by builtin Oracle's ones.
  (if-bind (fn (cdr-assoc (name expr) +postgresql-function-replacements+ :test #'string=))
    (%output-function-exp fn (slot-value expr 'args) database)
    (call-next-method)))

(defmethod output-sql ((expr sql-cast-exp) (database postgresql-mixin))
 ;;; PostgreSQL has terrific extensible repository of type conversions, or casts,
  ;; and CREATE CAST command!
  (%output-cast-exp expr database))

(defmethod output-sql ((expr sql-concat-exp) (database postgresql-mixin))
 ;;; Instead of scalar function escapes, write standard SQL syntax || 
  (with-slots (components) expr
    (%output-relational-exp "||" components database)))

;;; DB-methods

(defmethod db-type-initialize ((db-type (eql :postgresql)) &optional version)
 ;;; Only PostgreSQL ODBC is supported
  (or (find-environment :odbc :errorp nil)
      (db-type-initialize :odbc version)))

;;; Create or destroy database has two variants:
;;; 1) without connection (make sure PosgreSQL/**/bin/ is in PATH)
;;; 2) when connection is already established.
;;; Args: ARGS is a plist with the following indicators:
;;;       :owner | :template | :encoding | :tablespace | :connection\ limit

(defmethod db-create ((db-type (eql :postgresql)) connection-spec &rest args &key host)
 ;;; Create database when no database is open yet.
  (declare (ignore db-type))
  (let* ((user-name (db-user-name-from-spec db-type connection-spec))
         (password (db-password-from-spec db-type connection-spec))
         (host (if host
                   (prog1 host (setq args (remove-properties args '(:host))))
                   "localhost"))
         (command (format nil "createdb --host ~a~:[ --username ~a~;~*~]~{ --~(~a~) ~a~} ~a"
                          host (zerop (length user-name)) user-name args
                          (db-name-from-spec :postgresql connection-spec))))
    (multiple-value-bind (status message)
        (command-output command :extra-environ (if password `(("PGPASSWORD" . ,password))))
      (if (or (not (zerop status))
              (search "database creation failed: ERROR:" message))
          (error 'sql-user-error
                 :message (format nil "createdb failed for PostgreSQL backend.
Connection spec ~s, arguments ~s~%~a"
                                  connection-spec args message))
          t))))

(defmethod db-destroy ((db-type (eql :postgresql)) connection-spec &rest args &key host)
  (declare (ignore db-type))
  (let* ((user-name (db-user-name-from-spec db-type connection-spec))
         (password (db-password-from-spec db-type connection-spec))
         (host (if host
                   (prog1 host (setq args (remove-properties args '(:host))))
                   "localhost"))
         (command (format nil "dropdb --host ~a~:[ --username ~a~;~*~]~{ --~(~a~) ~a~} ~a"
                          host (zerop (length user-name)) user-name args
                          (db-name-from-spec :postgresql connection-spec))))
    (multiple-value-bind (status message)
        (command-output command :extra-environ (if password `(("PGPASSWORD" . ,password))))
      ;; The message from console can be localized, e.g.
      ;;  dropdb:     : ERROR:  database "..." does not exist
      (if (or (not (zerop status))
              (search "database removal failed: ERROR:" message))
          (error 'sql-user-error
                 :message (format nil "dropdb failed for PostgreSQL backend
Connection spec ~s, arguments ~s~%~a"
                                  connection-spec args message))
          t))))

(defmethod db-create ((database postgresql-mixin) connection-spec &rest args)
  (execute-command
   (with-sql-string-stream ()
     (write-string "CREATE DATABASE " *sql-stream*)
     (output-sql-identifier (db-name-from-spec :postgresql connection-spec) database t)
     (%output-plist args database))
   :database database))

(defmethod db-destroy ((database postgresql-mixin) connection-spec &key)
  (execute-command
   (write-drop-exp-to-string :database (db-name-from-spec :postgresql connection-spec)
                             database)
   :database database))


(defmethod db-connect ((db-type (eql :postgresql)) connection-spec 
                       &rest args &key &allow-other-keys)
  ;; Push the specific back-end type into args for possible retreival via driver-connect
  (apply #'db-connect :odbc connection-spec  :database-type :postgresql  args))

;;; SQL-schema: roles, users, and privileges

(defmethod db-create-user ((database postgresql-mixin) name &rest args &key group
                                                            &allow-other-keys)
 ;;; Args: args - useful options ::=
 ;;;	   [[UN]ENCRYPTED] PASSWORD 'password'
  ;;	   | CONNECTION LIMIT connlimit
  ;;       | VALID UNTIL 'timestamp'
  (execute-command
   (with-sql-string-stream ()
     (write-string "CREATE ROLE " *sql-stream*)
     (output-sql-identifier name database t)
     (write-string " LOGIN " *sql-stream*)
     (when group
       (write-string " IN ROLE " *sql-stream*)
       (if (consp group)
           (output-sql-identifiers group database t nil)
           (output-sql-identifier group database t)))
     (%output-plist (retain-properties args '(:password)) database))
   :database database))


