;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2008 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Sequence implementations
;;; 1) For the databases that support sequence natively, is based on SQL standard;
;;; 2) For sequenceless-mixin, via a helper table.

(in-package :ystok.sql.sys)
(in-syntax sql:*sql-readtable*)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  SQL-2003  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; <sequence generator definition> ::=
;;;   CREATE SEQUENCE <sequence generator name> [ <sequence generator option> ... ]
;;; <sequence generator option> ::= <sequence generator data type option>
;;; 				| <common sequence generator option> ...
;;; <common sequence generator option> ::= <sequence generator start with option>
;;; 				| <basic sequence generator option>
;;; <basic sequence generator option> ::=
;;; 	<sequence generator increment by option>
;;; 	| <sequence generator maxvalue option>
;;; 	| <sequence generator minvalue option>
;;; 	| <sequence generator cycle option>
;;; <sequence generator data type option> ::= AS <data type>
;;; <sequence generator start with option> ::= START WITH <signed numeric literal>
;;; <sequence generator increment by option> ::= INCREMENT BY <signed numeric literal>
;;; <sequence generator maxvalue option> ::= MAXVALUE <signed numeric literal>
;;;					| NO MAXVALUE
;;; <sequence generator minvalue option> ::= MINVALUE <signed numeric literal>
;;;					| NO MINVALUE
;;; <sequence generator cycle option> ::= CYCLE | NO CYCLE

(defmethod db-create-sequence ((database database) name &rest plist)
  (execute-command
   (with-sql-string-stream ()
     (write-string "CREATE SEQUENCE " *sql-stream*)
     (output-sql-identifier name database)
     (%output-plist plist database))
   :database database))

;;; DROP SEQUENCE <sequence generator name> <drop behavior>

(defmethod db-drop-sequence ((database database) name)
  (execute-command
   (string-append "DROP SEQUENCE " (write-sql-identifier-to-string name database))
   :database database))

#|
;;; <alter sequence generator statement> ::=
;;;   ALTER SEQUENCE <sequence generator name> <alter sequence generator option> ...
;;; <alter sequence generator option> ::= <alter sequence generator restart option>
;;; 			| <basic sequence generator option>
;;; <alter sequence generator restart option> ::= RESTART WITH <signed numeric literal>

(defmethod db-alter-sequence ((database oracle-database) name position )
  (without-interrupts
   (let* ((next (database-sequence-next name database))
	  (incr (- position next)))
     (unless (zerop incr)
       (database-execute-command
	(format nil "ALTER SEQUENCE ~A INCREMENT BY ~D" name incr)
	database))
     (database-sequence-next name database)
     (database-execute-command
      (format nil "ALTER SEQUENCE ~A INCREMENT BY 1" name)
      database))))


;;; <next value expression> ::= NEXT VALUE FOR <sequence generator name>

(defmethod db-sequence-next ((database database) name)
  (caar (select name :from
	 (concatenate 'string "SELECT "
		      (sql-escape sequence-name)
		      ".NEXTVAL FROM dual")
	 database :auto nil)))

;; Unspecified in SQL-2003

(defmethod db-sequence-current ((database database) name)
  (caar (database-query
	 (concatenate 'string "SELECT "
		      (sql-escape sequence-name)
		      ".CURRVAL FROM dual")
	 database :auto nil)))

(defmethod db-list-sequences ((database oracle-database) &key schema catalog)
  (db-query database 
            (case schema
              ((nil)
               "SELECT sequence_name FROM user_sequences")
              (:all
               "SELECT sequence_name FROM all_sequences")
              (otherwise
               (format nil
                       "SELECT user_sequences.sequence_name FROM user_sequences,all_sequences WHERE user_sequences.sequence_name=all_sequences.sequence_name AND all_sequences.sequence_owner='~:@(~A~)'"
                       (write-sql-identifier-to-string schema database t nil))))
            :flatp t))
|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  SEQUENCESLESS mixin  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Implementation Note: The schema parameter is ignored.

#+ys-product (keep-symbol-names 'YSQL_SEQUENCELESS)

(defparameter %sequences-table% (sql-expression :table 'YSQL_SEQUENCELESS))

(defun %sequences-table% (name &optional (catalog (getf (plist name) :catalog)))
 ;;; If catalog is specified within a sequence name, embed it into a table expression.
  (if catalog
      (sql-expression :table 'YSQL_SEQUENCELESS :catalog catalog)
      %sequences-table%))

(defun %create-sequences-table-exp (&optional catalog)
  (make-instance 'sql-create-table-exp
   :name (%sequences-table% nil catalog)
   :columns
    `((,[sequence_name] (string 64) :primary-key)
      (,[current_value] integer :not-null)
      (,[increment_by] integer :not-null)
      (,[min_value] integer)
      (,[is_issued] boolean))))	  ; if true, sequence-next should advance current_value

(defun %sequence-name-to-value (database name)
  (write-sql-identifier-to-string name database t nil))
  ;(declare (ignore database))
  ;(string (name name)))

(defun %signal-sequence-does-not-exist-error (database name expression)
  (error 'sql-database-error :database database
         :message (format nil "Sequence ~A does not exist" name)
         :expression expression))

(defmethod db-create-sequence ((database sequenceless-mixin) name
                               &key (min 1) (start min) (by 1))
 ;;; If the generator is already exists, the insert statement itself should signal
  ;; about duplication of primary key.
  (let ((catalog (getf (plist name) :catalog)))
    (unless (table-exists-p %sequences-table% :catalog catalog)
      (execute-command (%create-sequences-table-exp catalog) :database database))
    (insert-records :into (%sequences-table% nil catalog)
                    :columns `(,[sequence_name]
                               ,[current_value] ,[increment_by] ,[min_value] ,[is_issued])
                    :values `(,(%sequence-name-to-value database name)
                              ,start ,by ,min :false)
                    :database database)))

(defmethod db-drop-sequence ((database sequenceless-mixin) name)
  (when (zerop (nth-value 1
                (delete-records :from (%sequences-table% name)
                                :where [= [sequence_name]
                                          (%sequence-name-to-value database name)]
                                :database database)))
    (%signal-sequence-does-not-exist-error database name 'drop-sequence)))

(defmethod db-alter-sequence ((database sequenceless-mixin) name start &key by min)
  (when (zerop (nth-value 1
                (update-records (%sequences-table% name)
                  :cv-alist `((,[current_value] . ,start)
                              (,[is_issued] . :true)
                              ,@(when by `((,[increment_by] . ,by)))
                              ,@(when min `((,[min_value] . , min))))
                  :where [= [sequence_name] (%sequence-name-to-value database name)]
                  :database database)))
    (%signal-sequence-does-not-exist-error database name 'alter-sequence))
  start)

(defmethod db-list-sequences ((database sequenceless-mixin) &key schema catalog)
  (declare (ignore schema))
  (when (table-exists-p %sequences-table% :catalog catalog)
    (select [sequence_name] :from (%sequences-table% nil catalog)
            :database database :flatp t)))

(defmethod db-sequence-next ((database sequenceless-mixin) name)
  (ffc:without-scheduling
   (let* ((table (%sequences-table% name))
          (where [= [sequence_name] (%sequence-name-to-value database name)])
          (tuple (first (select [current_value] [is_issued :type boolean] [increment_by]
                                :from table
                                :where where
                                :database database))))
     (cond ((null tuple)
            (%signal-sequence-does-not-exist-error database name 'sequence-next))
           ((second tuple)				; already issued - increase
            (let ((next (+ (first tuple) (third tuple))))
              (update-records table :columns [current_value] :values next
                              :where where
                              :database database)
              next))
           (t						; not issued - just return
            (update-records table :columns [is_issued] :values :true
                            :where where
                            :database database)
            (first tuple))))))
	     
(defmethod db-sequence-last ((database sequenceless-mixin) name)
  (if-bind (tuple (select [current_value] :from (%sequences-table% name)
                   :where [= [sequence_name] (%sequence-name-to-value database name)]
                   :database database
                   :flatp t))
    (first tuple)
    (%signal-sequence-does-not-exist-error database name 'sequence-last)))


