;;; -*- Mode: Lisp; Package: ystok.sql.sys -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2006 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; SQLite3 database - Specialization for SQL expressions and types
;;; TODO: Simulate join clauses.

(in-package :ystok.sql.sys)

#|;;; Type specifiers

(defmethod db-sql-type-specifier ((database sqlite3-mixin) (lisp-type (eql 'string))
                                  &rest args)
  (declare (ignore database lisp-type))
  (if args
      (let ((length (first args)))
        (if (<= length 255)
            (format nil "VARCHAR(~A)" length)
            "LONGTEXT"))
      "VARCHAR"))

(defmethod db-sql-type-specifier ((database sqlite3-mixin) (lisp-type (eql 'boolean))
                                  &rest args)
  (declare (ignore database lisp-type args))
  "BIT")
|#

;;; Boolean types are represented by 0 and 1
(defmethod output-sql ((expr (eql :false)) (database sqlite3-mixin))
  (declare (ignore database))
  (write-string "0" *sql-stream*))
(defmethod output-sql ((expr (eql :true)) (database sqlite3-mixin))
  (declare (ignore database))
  (write-string "1" *sql-stream*))

;;; Functions

(defsql-function () substr)

(defconstant +sqlite-function-replacements+
  '((substring . substr)))

(yl:keep-symbol-names (mapcar #'cdr +sqlite-function-replacements+))

(defmethod output-sql ((expr sql-function-exp) (database sqlite3-mixin))
 ;;; Replace some standard SQL-99 or ODBC scalar functions by builtin SQLite's ones.
  (if-bind (fn (cdr-assoc (name expr) +sqlite-function-replacements+ :test #'string=))
    (%output-function-exp fn (slot-value expr 'args) database)
    (call-next-method)))

#|
(defmethod output-sql ((expr sql-join-exp) (database sqlite3-mixin))
  (declare (ignore database))
  (with-slots (operator components on) expr
    (write-char #\( *sql-stream*)
    (output-sql (first components) database)
    (write-char #\Space *sql-stream*)
    (write-string (symbol-name operator) *sql-stream*)
    (write-char #\Space *sql-stream*)
    (output-sql (second components) database)
    (write-string " ON " *sql-stream*)
    (output-sql on database)
    (write-char #\) *sql-stream*)))
|#

;;; Sentences

(defmethod output-sql ((expr sql-create-view-exp) (database sqlite3-mixin))
  (%output-create-view-no-columns expr database))	; columns are not supported

#|(defmethod output-sql ((expr sql-drop-index-exp) (database sqlite3-mixin))
  (with-slots (name plist) expr
    (write-string "DROP INDEX " *sql-stream*)
    (ecase (getf plist :if-does-not-exist)
      (:ignore (write-string "IF EXISTS " *sql-stream*))
      (:error))
    (output-sql-identifier name database)))|#

(defmethod output-sql ((expr sql-join-exp) (database sqlite3-mixin))
 ;;; SQLite3 cannot stand outer parantheses but seems to group tables from left to right.
  (with-slots (operator components on using) expr
    (%output-join-exp operator components on using database nil)))

;;; DB-methods

(defmethod db-type-initialize ((db-type (eql :sqlite3)) &optional version)
 ;;; Simulate type-subtype relationship between :odbc and :sqlite3
  ;; and map them both to the odbc-environment.
  (or (find-environment :odbc :errorp nil)
      (db-type-initialize :odbc version)))

(defmethod db-create ((db-type (eql :sqlite3)) pathname &rest args)
 ;;; Databases are created automatically by the SQLite3 driver provided the setting
  ;; "Don't Create Database" is not ticked.
  (declare (ignore db-type pathname args))
  t)
  ;(sys:call-system (format nil "sqlite3.exe \"~A\" .quit"
  ;                         (namestring (first-or-self pathname)))))

(defmethod db-destroy ((db-type (eql :sqlite3)) pathname &key)
  (declare (ignore db-type))
  (delete-file (first-or-self pathname)))

(defmethod db-connect ((db-type (eql :sqlite3)) connection-spec 
                       &rest args &key &allow-other-keys)
  ;; Push the specific back-end type into args for possible retreival via driver-connect
  (apply #'db-connect :odbc connection-spec  :database-type :sqlite3  args))

;;; SQLite driver returns [ODBC state HY000 error 1] instead of error 42S02.
;;; So we use a specialized SQL clause.
(defmethod db-drop-table ((database sqlite3-mixin) table &key (if-does-not-exist :error))
  (execute-command (string-append "DROP TABLE "
                                  (ecase if-does-not-exist
                                    (:ignore "IF EXISTS ")
                                    (:error  +null-string+))
                                  (write-sql-identifier-to-string table database))))

(defmethod db-drop-view ((database sqlite3-mixin) view &key (if-does-not-exist :error))
  (execute-command (string-append "DROP VIEW "
                                  (ecase if-does-not-exist
                                    (:ignore "IF EXISTS ")
                                    (:error  +null-string+))
                                  (write-sql-identifier-to-string view database))))

(defmethod db-drop-index ((database sqlite3-mixin) index &key on (if-does-not-exist :error))
  (declare (ignore on))
  (execute-command
   (with-sql-string-stream ()
     (write-string "DROP INDEX " *sql-stream*)
     (ecase if-does-not-exist
       (:ignore (write-string "IF EXISTS " *sql-stream*))
       (:error))
     (output-sql-identifier index database))
   :database database))
