;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2005 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Foreign function compatibility module 
;;;
;;; In this file, LW specific code is isolated.
;;; It consists mostly of wrapper functions and macros 
;;; around the Lisp implementation-dependent foreign function interface.

(in-package :ystok.ffc)

(defvar *foreign-module* nil)

;;; ToDo: External format should be specified as an argument to every function call
;;;       rather than globally?
(defvar *foreign-ef* nil ;#+win32 win32:*multibyte-code-page-ef* #-win32 :latin-1
  "External format used by all odbc backends for string encoding.")

(defvar *foreign-code-page* nil
  "Default code page used for encoding foreign C strings being passed to/from driver.")


(defun initialize-ffc-locale (&optional (ef #+win32 win32:*multibyte-code-page-ef*
                                            #-win32 :latin-1))
  (setq *foreign-ef* ef			;=(getf (rest win32:*multibyte-code-page-ef*) :id)
        *foreign-code-page* (if (consp ef) (EF::EF-CODED-CHARACTER-SET ef) ef)))

;(defvar *foreign-charset* (ef::find-charset *foreign-code-page*)
;  "LispWorks internal ef::charset structure")

(eval-when (:load-toplevel) (initialize-ffc-locale))
(lw:define-action "When starting image" "Initialize FFC locale" 'initialize-ffc-locale)

;;; ODBC Specific types matching SQLTYPES.H /* API declaration data types */

;typedef void *          SQLPOINTER;
;; /* function return type */
;typedef SQLSMALLINT SQLRETURN;
;; /* generic data structures */
;; #if (ODBCVER >= 0x0300)
;; #if defined(WIN32)
;typedef void*	SQLHANDLE;
;; #else
;;typedef SQLINTEGER		SQLHANDLE;
;; #endif	/* defined(WIN32) */
;typedef SQLHANDLE      SQLHENV;
;typedef SQLHANDLE      SQLHDBC;
;typedef SQLHANDLE      SQLHSTMT;
;typedef SQLHANDLE      SQLHDESC;
;; #else
;; #if defined(WIN32)
;typedef void*		SQLHENV;
;typedef void*		SQLHDBC;
;typedef void*		SQLHSTMT;
;#else
;typedef SQLINTEGER     SQLHENV;
;typedef SQLINTEGER     SQLHDBC;
;typedef SQLINTEGER     SQLHSTMT;
;#endif /* defined(WIN32) */
;#endif /* ODBCVER >= 0x0300 */

;; Used with arguments of indeterminate type: addresses of buffers or fixed values
(eval-when (:compile-toplevel :load-toplevel :execute)
(fli:define-c-typedef SQL-POINTER (:one-of :pointer :unsigned-long :signed-long))
(fli:define-c-typedef SQL-RETURN :signed-short)
(fli:define-c-typedef SQL-HANDLE (:pointer :void))
(fli:define-c-typedef SQL-HANDLE-PTR (:pointer SQL-HANDLE))
)

#-lispworks4.4
(defconstant *null-pointer* (load-time-value (fli:make-pointer :address 0 :type :void)))
(defconstant +SQL-NULL-HANDLE+                                       ;= SQL_NULL_HANDLE
  (load-time-value (fli:make-pointer :pointer-type 'SQL-HANDLE :address 0)))

(defmacro %null-ptr-p (ptr)
 "Value: T if ptr eql to nil or its address is 0."
  `(fli:null-pointer-p ,ptr))

(defmacro %get-ptr (ptr) `(fli:dereference ,ptr))

(defun %new-ptr (type &optional (element-count 1))
 ;;; Generaly used to allocate a new FLI object of the scalar type given.
 ;(fli:allocate-foreign-object :type (if element-count (list type element-count) type)))
  (fli:allocate-foreign-object :type type :nelems element-count))

(defun %dispose-ptr (ptr)
 ;;; Must ignore if (%null-pointer-p ptr)
  (fli:free-foreign-object ptr))

;(defun %ptr-eql (ptr1 ptr2)
;  (= (fli:pointer-address ptr1) (fli:pointer-address ptr2)))
(setf (fdefinition '%ptr-eql) (fdefinition 'fli:pointer-eq))

(defmacro %address-to-pointer (address)
  `(fli:make-pointer :address ,address :type :void))

(defmacro %pointer-to-address (ptr)
  `(fli:pointer-address ,ptr))

(defmacro %incf-ptr (ptr &optional offset)
  `(fli:incf-pointer ,ptr ,@(when offset `(,offset))))

(defmacro %decf-ptr (ptr &optional offset)
  `(fli:decf-pointer ,ptr ,@(when offset `(,offset))))

;; The below macro is deprecated: if needed, use (define-symbol-macro *null-pointer* ...)
(defmacro %null-ptr () '*null-pointer*)

(defmacro with-dynamic-pointer ((ptr &key index type) pointer &body body)
 ;;; Bind PTR to dynamic copy of POINTER, increase by INDEX in needed, 
  ;; and evaluate the BODY. Should not cons!
  ;; Args: index  Index within a coerced pointer.
  `(fli:with-coerced-pointer (,ptr ,@(when type `(:type ,type))) ,pointer
     ,@(when index `((fli:incf-pointer ,ptr ,index)))
     ,@body))
     
#+unused
(defmacro %copy-pointer (ptr &optional offset)
  (if offset
      `(fli:incf-pointer (fli:copy-pointer ,ptr) ,offset)
      `(fli:copy-pointer ,ptr)))


;; The below macro is deprecated for the two reasons:
;; 1) It is a misnomer and conflict with type SQL-POINTER;
;;    a better name could have been %with-sql-hanlde.
;; 2) Its definition is equivalent to the following:
;;	(defmacro %with-sql-pointer ((pointer-var) &body body)
;;	  `(fli:with-dynamic-foreign-objects ((,pointer-var sql-handle))
;;	    ,@body))
;;    Hence
;;	(with-sql-pointer (ph) . body) 
;;	== (with-temporary-allocation ((ph sql-handle) . body)
;(defmacro %with-sql-pointer ((pointer-var) &body body)
;  `(let ((,pointer-var (fli:allocate-foreign-object :pointer-type 'sql-handle-ptr)))
;     ,@body))

;;; Two types of array wrapping are possible.
;;; The second could be used to access scalar data-ptr as well.
;;; NB: Shall we use (sys:in-static-area (fli:allocate-foreign-object ...))?
;;;     Though according to sys:staticp documentation:
;;;	 "Foreign instantiations made by Lisp -- for example in a foreign 
;;;	  language interface program -- are made in static memory."
                             
(defmacro %new-vector (type size)
  `(fli:allocate-foreign-object :type (list :c-array ,type ,size)))
  ;`(fli:allocate-foreign-object :type ,type :nelems ,size)))	; looks quicker

(defmacro %svref (array index &key type)
 ;;; LW4.4: type  Evaluated and must produce
  ;;	(unsigned-byte 32), (signed-byte 32), (unsigned-byte 16), (signed-byte 16),
  ;;	(unsigned-byte 8) or (signed-byte 8),  double-float or single-float.
  ;; NB: Despite the doc fli:foreign-typed-aref requires byte-offset as index argument.
  #-(or lispworks4.2 lispworks4.3)
  (if type
      `(fli:foreign-typed-aref ,type ,array ,(if (eql (fli:size-of type) 1)
                                                 index
                                                 `(* ,index ,(fli:size-of type))))
      `(fli:foreign-aref ,array ,index))
  #+(or lispworks4.2 lispworks4.3)
  `(fli:foreign-aref ,array ,index))			; type and bounds checking
  ;`(fli:dereference ,array :index ,index))		; no bounds checking

(defmacro %fill-vector (pointer &key start end byte)
 ;;; Args: pointer  Pointer of an appropriate type; don't use an array - hangs up!
  ;;	   byte	    Defaults to 0; if element size >1, each of its bytes is set the same.
  ;; NB: fill-foreign-object seems to neither respect actual array size nor nelems
  (declare (ignore start))
  `(fli:fill-foreign-object ,pointer
                            ,@(when end `(:nelems ,end))
                            ,@(when byte `(:byte ,byte))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  STRINGS  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Unicode NB1:
;;;    ,   -W    
;;;     : 
;;;  -   ,   /  .
;;;  -   ,    /  ,  
;;;      , .g. SQLGetInfo.
;;; NB: Unfortunately, fli:replace-foreign-array 
;;;	- does not works with strings
;;;	- does not support :wchar-t arrays.

(eval-when (:compile-toplevel :load-toplevel :execute)
(fli:define-c-typedef SQL-CHAR (:unsigned :char)) ;typedef unsigned char SQLCHAR;
(fli:define-c-typedef SQL-WCHAR :wchar-t)	  ;typedef unsigned short SQLWCHAR;
;;#if (ODBCVER >= 0x0300)
(fli:define-c-typedef SQL-SCHAR (:signed :char))   ;typedef signed char SQLSCHAR;
;; #endif

(fli:define-c-typedef SQL-CHAR-PTR (:pointer :unsigned :char))
(fli:define-c-typedef STRING-PTR :pointer #+ignore (:pointer :char))
)

(defmacro %get-char (ptr &key (type :char) index) 
 ;;; Getting a single character.
  ;; Args: ptr  Could be a pointer of any type, including :c-array
  ;;       type Foreign character type
  ;;		::= :char (actually used :unsigned) | 'SQL-CHAR | :wchar-t | 'SQL-WCHAR |
  ;; Used by: %sql-get-info.
      ;(case (if (consp type) (second type) type)
      ; ((SQL-CHAR :char)
  (cond ((or (fli:foreign-type-equal-p type '(:unsigned :char))
             (fli:foreign-type-equal-p type :char))
         #+(and Unicode (not lispworks4.2))
         `(ef:find-external-char (fli:dereference ,ptr :type :unsigned-byte
                                              ,@(when index `(:index ,index)))
                                 *foreign-code-page*)
         #+(and Unicode lispworks4.2 win32)
         `(code-char (win32::cp-to-unicode *foreign-code-page* nil 
                                       (fli:dereference ,ptr :type :unsigned-byte
                                                        ,@(when index `(:index ,index)))))
         #-Unicode
         `(fli:dereference ,ptr :type :char ,@(when index `(:index ,index))))
        ;((SQL-WCHAR :wchar-t)
        ((fli:foreign-type-equal-p type :wchar-t)
         #+Unicode
         `(fli:dereference ,ptr :type :wchar-t ,@(when index `(:index ,index)))
         #-Unicode
         `(code-char (fli:dereference ,ptr :type :unsigned-short
                                      ,@(when index `(:index ,index)))))
        (t			; provide execute-time checking
         #+Unicode
         `(if (fli:foreign-type-equal-p ,type :wchar-t)
              (fli:dereference ,ptr :type :wchar-t ,@(when index `(:index ,index)))
              ;; one-byte char
              #-lispworks4.2
              (ef:find-external-char (fli:dereference ,ptr :type :unsigned-byte
                                                      ,@(when index `(:index ,index)))
                                     *foreign-code-page*)
              #+(and lispworks4.2 win32)
              (code-char (win32::cp-to-unicode *foreign-code-page* nil 
                                               (fli:dereference ,ptr :type :unsigned-byte
                                                ,@(when index `(:index ,index)))))
              #+(and lispworks4.2 (not win32))
              #1=`(fli:dereference ,ptr :type :char ,@(when index `(:index ,index))) )
         #-Unicode
         #1#)))

(defun %get-cstring (pointer)
 ;;;  Create new lisp string with element-type according to external format *foreign-ef*
  ;; Args: pointer  Pointer to a null-terminated C string of a proper type
  ;;		    :char or :wchar-t.
  ;; NB: &optional (start 0) was only used in odbc-streams. Now replaced by %get-string.
  (let ((length 0))
    (declare (optimize (speed 3)
                       #-debug (debug 0) #-debug (safety 0) #-debug (hcl:fixnum-safety 0))
             (fixnum length))
    (do ()						; Find foreign string length
        ((char= (fli:dereference pointer) #\Null))	; similar to %cstring-length
      (fli:incf-pointer pointer)			; but modify ptr destructively
      (incf length))
    (let ((str (make-string length :element-type (ef:external-format-type *foreign-ef*))))
      (if (fli:pointer-element-type-p pointer :wchar-t)
          (do ()
              ((zerop length))
            (fli:decf-pointer pointer)
            (setf (lw:stchar str (decf length)) (fli:dereference pointer)))
          (fli:with-coerced-pointer (ptr :type #+Unicode :unsigned-byte #-Unicode :char)
              pointer
            (do ()
                ((zerop length))
              (fli:decf-pointer ptr)
              (setf (schar str (decf length))
                    #+(and Unicode (not lispworks4.2))
                    (ef:find-external-char (fli:dereference ptr) *foreign-code-page*)
                    #+(and Unicode lispworks4.2 win32) 
                    (code-char (win32::cp-to-unicode *foreign-code-page*
                                                     nil (fli:dereference ptr)))
                    #-Unicode (fli:dereference ptr)))))
      str)))

(declaim (ftype (function (t fixnum &optional fixnum) t) %get-string
                                                         #+Unicode %get-wstring))
(defun %get-string (pointer end &optional (start 0))
 ;;; Similiar to %get-cstring, but can deal with exact start and end length
 "Args: POINTER - is not obliged to be of type :char.
 START, END - subscripts in pointer or byte-counts"
  (declare (optimize (speed 3) ;(space 0) 
                     #-debug (debug 0) #-debug (safety 0) #-debug (hcl:fixnum-safety 0))
           (fixnum start end))
  #+(and Unicode (not (or lispworks4.2 lispworks4.3)))
  (do ((str (make-string (- end start) :element-type 'lw:simple-char))
       (i 0 (1+ i))				; char index in str
       (j start (1+ j)))			; byte index in pointer
      ((>= j end) str)
    (declare (fixnum i j) (type lw:simple-text-string str))
    (setf (lw:stchar str i) (ef:find-external-char
                             (fli:foreign-typed-aref '(unsigned-byte 8) pointer j)
                             *foreign-code-page*)))
  #-(and Unicode (not (or lispworks4.2 lispworks4.3)))
  (let* ((length (- end start))
         (str (make-string length :element-type (ef:external-format-type *foreign-ef*))))
         ;(EF::EXT-TO-INT (slot-value *foreign-charset* 'EF::EXT-TO-INT))
    (declare (fixnum length))
    (fli:with-coerced-pointer (ptr :type #+Unicode :unsigned-byte #-Unicode :char)
        pointer
      (fli:incf-pointer ptr start)
      (dotimes (i length)
        (declare (fixnum i))
        (setf (schar str i)
              #+(and Unicode (not lispworks4.2))
              (ef:find-external-char (fli:dereference ptr) *foreign-code-page*)
               ;(funcall EF::EXT-TO-INT (fli:dereference ptr))
              #+(and Unicode lispworks4.2 win32)
              (code-char (win32::cp-to-unicode *foreign-code-page* nil
                                               (fli:dereference ptr)))
              #-Unicode
              (fli:dereference ptr))
        (fli:incf-pointer ptr)))
    str))

#+Unicode
(defun %get-wstring (pointer end &optional (start 0))
 ;;; Args: POINTER  Is not obliged to be of type :wchar-t.
  ;;       START, END - bytes offsets in pointer, not subscripts!
  ;; NB: Column binding should be SQL_C_WCHAR!
  (declare (optimize (speed 3) (space 0) #-debug (safety 0) #-debug (debug 0)
                     #-debug (hcl:fixnum-safety 0))
           (fixnum start end))
  #-(or lispworks4.2 lispworks4.3)
  (do ((str (make-string (ash (- end start) -1) :element-type 'lw:simple-char))
       (i 0 (1+ i))				; char index in str
       (j start (+ j 2)))			; byte index in pointer
      ((>= j end) str)
    (declare (fixnum i j) (type lw:simple-text-string str))
    (setf (lw:stchar str i) (code-char (fli:foreign-typed-aref '(unsigned-byte 16)
                                                               pointer j))))
  #+(or lispworks4.2 lispworks4.3)
  (let* ((length (ash (- end start) -1))
         (str (make-string length :element-type 'lw:simple-char)))
    (declare (fixnum length))
    (fli:with-coerced-pointer (ptr :type :wchar-t) pointer
      (fli:incf-pointer ptr (ash start -1))
      (dotimes (i length)
        (declare (fixnum i))
        (setf (lw:stchar str i) (fli:dereference ptr))
        (fli:incf-pointer ptr)))
    str))

(defmacro %new-cstring (length &optional (type :char))
 ;;; Args: length  Length in characters
  ;; ToDo: As :initial-element degrades peformance, can it be avoided?
  `(fli:allocate-foreign-object :type ,type :nelems ,length)) ; :initial-element #\Null))

(declaim (ftype (function (t t fixnum &key (start fixnum) (index fixnum)
                                           (null-terminated-p t)) fixnum)
                %put-string  #+Unicode %put-wstring))
(defun %put-string (pointer string max-octet-length &key (start 0) (index 0)
                            				 null-terminated-p)
 "Args: POINTER assumed to be pre-allocated but not obliged to be of type :char.
  START - subscript in string
  INDEX - subscript in pointer
 Values: Actual number of octets written."
  ;; NB: Works slower than fli:convert-to-foreign-string, but conses less.
  ;; Unicode NB: More elebaration is needed if ptr is of type :wchar-t??
  (declare (optimize (speed 3)
                     #-debug (debug 0) #-debug (safety 0) #-debug (hcl:fixnum-safety 0))
           (fixnum start index max-octet-length))
  #+(and Unicode (not (or lispworks4.2 lispworks4.3)))
  (do* ((size (the fixnum (min (the fixnum (- (length string) start)) max-octet-length)))
        (end (the fixnum (+ index size)))
        (i index (1+ i))				; byte index in pointer
        (j start (1+ j)))				; char index in string
       ((>= i end)
        (when null-terminated-p
          (when (= size max-octet-length)		; As the termination does not fit,
            (decf size)				; truncate the rightmost char
            (decf i))
          (setf (fli:foreign-typed-aref '(unsigned-byte 8) pointer i) 0))
        size)
    (declare (fixnum size end i j))
    (setf (fli:foreign-typed-aref '(unsigned-byte 8) pointer i)
          (ef:char-external-code (char string i) *foreign-code-page*)))

  #-(and Unicode (not (or lispworks4.2 lispworks4.3)))
  (let* ((size (the fixnum (min (the fixnum (- (length string) start)) max-octet-length)))
         (end (the fixnum (+ start size))))
    (declare (fixnum size end))
    (fli:with-coerced-pointer (ptr :type #+Unicode :unsigned-byte #-Unicode :char) pointer
      (fli:incf-pointer ptr index)
      (do ((i start (1+ i)))
          ((>= i end))
        (declare (fixnum i))
        (setf (fli:dereference ptr)
              #+Unicode        	; Can also be used for data source binding in Unicode
              (ef:char-external-code (char string i) *foreign-code-page*)
			;(win32::unicode-to-cp *foreign-code-page* (char ,string i))
              #-Unicode
              (char string i))
        (fli:incf-pointer ptr))
      (when null-terminated-p
        (when (= size max-octet-length)	; truncate for one char
          (decf size)
          (fli:decf-pointer ptr))
        (setf (fli:dereference ptr) #\Null)))
    size))

#+Unicode
(defun %put-wstring (pointer string max-octet-length &key (start 0) (index 0)
							  null-terminated-p)
 "Args: POINTER assumed to be pre-allocated but not obliged to be of type :wchar-t.
  START - byte offset in string
  INDEX - byte offset in pointer
 Values: Actual number of octets written."
  (declare (optimize (speed 3) (space 0) #-debug (debug 0)
                     #-debug (safety 0) #-debug (hcl:fixnum-safety 0))
           (fixnum start index max-octet-length)
           (type string string))
  #-(or lispworks4.2 lispworks4.3)
  (do* ((size (the fixnum (min (the fixnum (- (ash (length string) 1) start))
                               max-octet-length)))
        (end (the fixnum (+ index size)))
        (i index (+ i 2))				; byte index in pointer
        (j (ash start -1) (1+ j)))			; char index in string
       ((>= i end)
        (when null-terminated-p
          (when (= size max-octet-length)		; As the termination does not fit,
            (decf size 2)				; truncate the rightmost char
            (decf i 2))
          (setf (fli:foreign-typed-aref '(unsigned-byte 16) pointer i) 0))
        size)
    (declare (fixnum size end i j))
    (setf (fli:foreign-typed-aref '(unsigned-byte 16) pointer i) (char string j)))
  #+(or lispworks4.2 lispworks4.3)
  (let* ((size (min (- (ash (length string) 1) start) max-octet-length))
         (end (+ start size)))
    (declare (fixnum size end))
    (fli:with-coerced-pointer (ptr :type :wchar-t) pointer
      (fli:incf-pointer ptr (ash index -1))
      (loop for i fixnum upfrom (ash start -1) below (ash end -1)
            do (setf (fli:dereference ptr) (char string i))
               (fli:incf-pointer ptr))
      (when null-terminated-p
        (when (= size max-octet-length)		; As the termination does not fit,
          (decf size 2)				; truncate the rightmost char
          (fli:decf-pointer ptr))
        (setf (fli:dereference ptr) #\Null)))
    size))

(defmacro with-cstr ((ptr string &optional element-count (external-format '*foreign-ef*))
                     &body body)
 "Args: STRING : Input Lisp string or nil.
 ELEMENT-COUNT : Variable bound to resulting string length; specifying it means
	that C-string will not be terminated by the null char.
 EXTERNAL-FORMAT : :Unicode or nil (means default *foreign-ef*)."
  ;; Implementation Notes:
  ;;  1) Proper external-format guarantees that ptr will be of proper char type.
  ;;  2) Can be safely applied only to pass arguments that are always strings or nils
  ;;  3) Xanalys supports the :allow-null keyword starting from LW 4.3.
  `(fli:with-foreign-string (,ptr ,(if element-count element-count 'element-count)
                                  byte-count
                                  :external-format ,external-format
                                  :null-terminated-p ,(null element-count)
                                  :allow-null t)
       ,string
     (declare (ignore byte-count ,@(unless element-count `(element-count))))
     ,@body))

(defmacro with-cstrings (bindings &rest body)
 "Binds dynamically all pointers and element-counts from bindings like
  ((ptr string &optional element-count external-format)...).
 STRING : Input parameter that either can be a lisp string or nil.
 ELEMENT-COUNT : Output parameter that does not takes terminating null 
 char into account, i.e. equals to (length string)."
  (loop for (ptr string element-count external-format) in bindings
        and string-gensym = (gensym "string")
        collect ptr into gensyms
        when element-count collect element-count into gensyms
        collect `(,string-gensym ,string) into gensyms
        if element-count collect
            `(cond ((null ,string-gensym)
                    (setq ,ptr *null-pointer*
                          ,element-count 0))
                   ((multiple-value-setq (,ptr ,element-count)	
                        ;; return char-, not byte-count as a second value
                        (fli:convert-to-dynamic-foreign-string ,string-gensym
                            :external-format ,(or external-format '*foreign-ef*)
                            :null-terminated-p nil)) ))
                    ;(when (= (fli:pointer-element-size ,ptr) 2) ; convert to byte-count
                    ;  (incf ,byte-count ,byte-count)) ))
                   ;(decf ,element-count))) ; as takes terminating nullchar into account
            into setqs
        else collect
            `(setq ,ptr (if (null ,string-gensym)
                            *null-pointer* 
                            (fli:convert-to-dynamic-foreign-string ,string-gensym
                                :external-format ,(or external-format '*foreign-ef*)
                                :null-terminated-p t)))
            into setqs
        finally (return `(fli:with-dynamic-foreign-objects ()
                           (let ,gensyms
                             ,@setqs
                             ,@body)))))

(declaim (ftype (function (t t fixnum t &optional fixnum) fixnum) %buffer-into-vector))
(defun %buffer-into-vector (pointer type octet-length vector &optional (start 0))
 ;;; Push elements of foreign buffer pointed by pointer into the lisp vector
  ;; Args:  pointer	Pointer to buffer of arbitrary type.
  ;;	    type 	Foreign type determining kind of C elements stored in buffer.
  ;;        octet-length Number of data bytes read from buffer; doesn't count null-term.
  ;;        vector	Simple string, simple vector, or adjustable array.
  ;;        start	Starting index value in the vector.
  ;; Value: Ending subscript of the vector.
  (declare (optimize (speed 3) (space 0) #-debug (safety 0) #-debug (debug 0)
                     #-debug (hcl:fixnum-safety 0))
           (fixnum start octet-length))
  (let ((length (truncate octet-length (fli:size-of type))))
    (declare (fixnum length))
    (case type
      (SQL-CHAR
       #+(and Unicode (not (or lispworks4.2 lispworks4.3)))
       (cond ((stringp vector)
              (dotimes (i octet-length) (declare (fixnum i))
                (setf (schar vector start) #1=(ef:find-external-char
                                               (fli:foreign-typed-aref '(unsigned-byte 8)
                                                                       pointer i)
                                               *foreign-code-page*))
                (incf start)))
             (t
              (when (adjustable-array-p vector)
                (adjust-array vector (+ start length))) ; :initial-element #\?
              (dotimes (i octet-length) (declare (fixnum i))
                (setf (aref vector start) #1#)
                (incf start))))
       #-(and Unicode (not (or lispworks4.2 lispworks4.3)))
       (fli:with-coerced-pointer (ptr :type #+Unicode :unsigned-byte #-Unicode :char)
           pointer
         (cond ((stringp vector)
                (dotimes (i length) (declare (fixnum i))
                  (setf (schar vector start)
                        #+(and Unicode (not lispworks4.2))
                        (ef:find-external-char (fli:dereference ptr) *foreign-code-page*)
                        #+(and Unicode lispworks4.2 win32)
                        (code-char (win32::cp-to-unicode *foreign-code-page* nil
                                                         (fli:dereference ptr)))
                        #-Unicode (fli:dereference ptr))
                  (fli:incf-pointer ptr)
                  (incf start)))
               (t
                (when (adjustable-array-p vector)
                  (adjust-array vector (+ start length))) ; :initial-element #\?
                (dotimes (i length) (declare (fixnum i))
                  (setf (aref vector start)
                        #+(and Unicode (not lispworks4.2))
                        (ef:find-external-char (fli:dereference ptr) *foreign-code-page*)
                        #+(and Unicode lispworks4.2 win32)
                        (code-char (win32::cp-to-unicode *foreign-code-page* nil
                                                         (fli:dereference ptr)))
                        #-Unicode (fli:dereference ptr))
                  (fli:incf-pointer ptr)
                  (incf start))))))
      #+Unicode
      (SQL-WCHAR
       #-(or lispworks4.2 lispworks4.3)
       (cond ((stringp vector)
              (do ((i 0 (+ i 2))) ((>= i octet-length)) (declare (fixnum i))
                (setf (lw:stchar vector start) #2=(code-char (fli:foreign-typed-aref
                                                              '(unsigned-byte 16)
                                                              pointer i)))
                (incf start)))
             (t
              (when (adjustable-array-p vector)
                (adjust-array vector (+ start length)))
              (do ((i 0 (+ i 2))) ((>= i octet-length)) (declare (fixnum i))
                (setf (aref vector start) #2#)
                (incf start))))
       #+(or lispworks4.2 lispworks4.3)
       (fli:with-coerced-pointer (ptr :type :wchar-t) pointer
         (cond ((stringp vector)
                (dotimes (i length) (declare (fixnum i))
                  (setf (lw:stchar vector start) (fli:dereference ptr))
                  (fli:incf-pointer ptr)
                  (incf start)))
               (t
                (when (adjustable-array-p vector)
                  (adjust-array vector (+ start length)))
                (dotimes (i length) (declare (fixnum i))
                  (setf (aref vector start) (fli:dereference ptr))
                  (fli:incf-pointer ptr)
                  (incf start))))))
      (otherwise				; SQL-OCTET
       (when (adjustable-array-p vector)
         (adjust-array vector (+ start length)))
       (fli:with-coerced-pointer (ptr :type type) pointer
         #-lispworks4.2
         (fli:replace-foreign-array vector ptr :start1 start :end2 length)
         #+lispworks4.2
         (dotimes (i length)
           (declare (fixnum i))
           (setf (aref vector start) (fli:dereference ptr))
           (fli:incf-pointer ptr)
           (incf start)))) ))
  start)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; OTHER TYPES ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;typedef long            SQLINTEGER;
;typedef unsigned long   SQLUINTEGER;
;typedef short           SQLSMALLINT;
;typedef unsigned short  SQLUSMALLINT;
;; #if (ODBCVER >= 0x0300)
;typedef unsigned char   SQLNUMERIC;
;; #endif
;; #if (ODBCVER >= 0x0300)
;typedef float           SQLREAL;
;; #endif
;typedef unsigned char   SQLDECIMAL;
;typedef double          SQLDOUBLE;
;typedef double          SQLFLOAT;

(eval-when (:compile-toplevel :load-toplevel :execute)
(fli:define-c-typedef SQL-INTEGER   :long)
(fli:define-c-typedef SQL-UINTEGER  :unsigned-long)
(fli:define-c-typedef SQL-SMALLINT  :short)
(fli:define-c-typedef SQL-USMALLINT :unsigned-short)
(fli:define-c-typedef SQL-BYTE	    :byte)	    ; represents SQL_BIT and SQL_TINYINT
(fli:define-c-typedef SQL-OCTET	    :unsigned-byte) ; represents SQL_TINYINT and our own
(fli:define-c-typedef SQL-REAL      :float)
(fli:define-c-typedef SQL-DOUBLE    :double)

;;; ODBC 64-bit introduced the below types and changed in API function declarations
;;; - from SQL-INTEGER to SQL-LEN,
;;; - from SQL-UINTEGER to SQL-ULEN.
;(fli:define-c-typedef SQL-LEN	:long or :int64)
;(fli:define-c-typedef SQL-ULEN	:unsigned-long or :uint64)
)



;;; Macros hanlding function calls conforming ODBC specification.
;;; The :type argument slows down dereferencing in about 1.5 times!
;;; CAUTION: We must guarantee that ptr is of an appropriate type.
;;; Use an optional cast parameter?

(defmacro %get-integer (ptr &optional index)
  `(fli:dereference ,ptr ,@(when index `(:index ,index))))	;:type :long
;(defmacro %put-integer (ptr value) `(setf (%get-integer ,ptr) ,value))
(defmacro %get-uinteger (ptr &optional index)
  `(fli:dereference ,ptr ,@(when index `(:index ,index))))	;:type :unsigned-long
;(defmacro %put-uinteger (ptr value) `(setf (%get-uinteger ,ptr) ,value))

(defmacro %get-smallint (ptr &optional index)
  `(fli:dereference ,ptr ,@(when index `(:index ,index))))	;:type :signed-short
;(defmacro %put-smallint (ptr value) `(setf (%get-smallint ,ptr) ,value))
(defmacro %get-usmallint (ptr &optional index)
  `(fli:dereference ,ptr ,@(when index `(:index ,index))))	;:type :unsigned-short
;(defmacro %put-usmallint (ptr value) `(setf (%get-usmallint ,ptr) ,value))

;;; Macros converting between C buffers and Lisp objects. Usually perform casting.
;;; Must be consistent with db-sql-to-c-type, %read-datum, %write-datum

(defmacro %get-sshort (ptr &optional index)
  `(fli:dereference ,ptr ,@(when index `(:index ,index))))	;:type :signed-short
(defmacro %put-sshort (ptr short &optional index)
  `(setf (%get-sshort ,ptr ,index) ,short))

(defmacro %get-ushort (ptr &optional index)
  `(fli:dereference ,ptr ,@(when index `(:index ,index))))	;:type :unsigned-short
(defmacro %put-ushort (ptr word &optional index)
  `(setf (%get-ushort ,ptr ,index) ,word))

(defmacro %get-slong (ptr &optional index)
  `(fli:dereference ,ptr ,@(when index `(:index ,index))))	;:type :signed-long
(defmacro %put-slong (ptr long &optional index)
  `(setf (%get-slong ,ptr ,index) ,long))

(defmacro %get-ulong (ptr &optional index)
  `(fli:dereference ,ptr ,@(when index `(:index ,index))))	;:type :unsigned-long 
(defmacro %put-ulong (ptr long &optional index)
  `(setf (%get-ulong ,ptr ,index) ,long))

;; %get-single and %put-single are dangerous as add arbitrary digits at the end.
;; In LW, there seems no function that correcly truncates this digits.
(defmacro %get-single (ptr &optional index)
  `(fli:dereference ,ptr ,@(when index `(:index ,index))))	;:type :float
  ;#+(or :win32 :linux)
  ;`(lw:round-to-single-precision (fli:dereference ,ptr :type :float
  ;                                                ,@(when index `(:index ,index))))
  ;#+(and unix (not :linux))
(defmacro %put-single (ptr float &optional index)
  `(setf (%get-single ,ptr ,index) ,float))

(defmacro %get-double (ptr &optional index)
  `(fli:dereference ,ptr ,@(when index `(:index ,index))))	;:type :double
(defmacro %put-double (ptr double &optional index)
  `(setf (%get-double ,ptr ,index) ,double))

;; a cludge! better way?
(defmacro %get-bit (ptr &optional (index 0))
  ;`(- (char-code (fli:dereference ,ptr :type :char ,@(when index `(:index ,index))))))
  `(%get-byte ,ptr ,index))
(defmacro %put-bit (ptr bit &optional (index 0))
  ;`(- (char-code (fli:dereference ,ptr :type :char ,@(when index `(:index ,index))))))
  `(setf (%get-byte ,ptr ,index) ,bit))
       #|(cond ((numberp ,bit) (if (zerop ,bit) 0 1))
                (,bit (if (eq ,bit :false) 0 1))
                (t 0))))|#

(defmacro %get-byte (ptr index)
  `(fli:dereference ,ptr :index ,index))	;:type :signed-byte
(defmacro %put-byte (ptr byte index)
  `(setf (%get-byte ,ptr ,index) ,byte))
(defmacro %get-octet (ptr index)
  `(fli:dereference ,ptr :index ,index))	;:type :unsigned-byte
(defmacro %put-octet (ptr byte index)
  `(setf (%get-octet ,ptr ,index) ,byte))

(declaim (ftype (function (t fixnum t &optional fixnum) t) %get-binary))
(defun %get-binary (pointer octet-length result-type &optional (start 0))
 "POINTER : Preallocated of arbitrary type.
 OCTET-LENGTH : Number of bytes to read.
 RESULT-TYPE is one of vector, bit-vector, :bookmark, or string.
 START : subscript in pointer."
  ;;       , ..   
  ;;   .  ,    .
  (declare (optimize (speed 3) (space 0) #-debug (safety 0) #-debug (debug 0)
                     #-debug (hcl:fixnum-safety 0))
           (fixnum octet-length start))
  (fli:with-coerced-pointer (ptr :type :unsigned-byte) pointer
    (case result-type
      ((bit-vector simple-bit-vector)
       (fli:incf-pointer ptr start)
       (loop with vector = (make-array (ash octet-length 3) :element-type 'bit)
             for i fixnum upfrom 0 below octet-length
             for byte fixnum = (fli:dereference ptr)
             do (dotimes (j 8)
                  (declare (fixnum j))
                  (setf (sbit vector (+ (ash i 3) j)) (logand (ash byte (- j 7)) 1)))
                (fli:incf-pointer ptr)
             finally (return vector)))
      (string 					;:hex-string
       (fli:incf-pointer ptr start)
       (loop with string = (make-string (* octet-length 2) :element-type 'base-char)
             repeat octet-length
             for j fixnum upfrom 0 by 2
             for byte fixnum = (fli:dereference ptr)
             do (setf (schar string j) (yl:hex-digit-char (ash byte -4))
                      (schar string (1+ j)) (yl:hex-digit-char (logand byte #x0F)))
                (fli:incf-pointer ptr)
             finally (return string)))
      (:bookmark					
       ;; Bookmark: By assuming that the less significant byte goes first 
       ;; we often get a fixnum as a result.
       (fli:incf-pointer ptr start)
       (loop with bookmark of-type integer = 0
             for byte-position fixnum upfrom 0 by 8
             repeat octet-length
             ;for bignum of-type integer = (fli:dereference ptr :type :signed-byte)
             ;           then (logior (ash bignum 8) (fli:dereference ptr))
             do ;(setf bignum (+ (ash bignum 8) (fli:dereference ptr)))
                (setf bookmark (dpb (fli:dereference ptr)
                                    (byte 8 byte-position) bookmark))
                (fli:incf-pointer ptr)
             finally (return bookmark)))
      (otherwise 					;:unsigned-byte-vector
       #-lispworks4.2
       (fli:replace-foreign-array (make-array octet-length
                                              :element-type '(unsigned-byte 8))
                                  ptr :start2 start :end2 (+ start octet-length))
       #+lispworks4.2
       (let ((vector (make-array octet-length :element-type '(unsigned-byte 8))))
         (fli:incf-pointer ptr start)
         (loop for i fixnum upfrom 0 below octet-length
               do (setf (aref vector i) (fli:dereference ptr))
                  (fli:incf-pointer ptr)
               finally (return vector))) ))))

(defmacro %new-binary (length &optional (type :unsigned-byte))
  `(fli:allocate-foreign-object :type ,type :nelems ,length))

(defun %put-binary (pointer vector max-octet-length &key (start 0) (index 0))
 "Args: POINTER - preallocated of arbitrary type.
  START - subscript in vector.
  INDEX - subscript in pointer.
 Value: returns size in bytes"
  (declare (optimize (speed 3) (space 0) #-debug (debug 0)
                     #-debug (safety 0) #-debug (hcl:fixnum-safety 0))
           (fixnum max-octet-length))
  (fli:with-coerced-pointer (ptr :type :unsigned-byte) pointer
    (fli:incf-pointer ptr index)
    (cond ((bit-vector-p vector)		
           #+debug (assert (zerop start) (start)
                     "Bit-vector can only be put into BINARY buffer without chunking.")
           (let* ((bit-count (length vector))
                  (byte-count (ceiling bit-count 8)))
             (declare (fixnum bit-count byte-count))
             ;(when (and max-octet-length (> byte-count max-octet-length))
             ;  (error "bit vector of length ~d is longer than max-octet-length: ~d"
             ;         bit-count (* max-octet-length 8)))
             (loop for i fixnum upfrom 0 below byte-count
                   and byte fixnum = 0 
                   do
                   (loop for j fixnum upfrom 0 below 8
                         for index fixnum = (+ (ash i 3) j)
                         while (< index bit-count)
                         do (setf byte (logior byte (ash (sbit vector index) (- 7 j)))))
                   (setf (fli:dereference ptr) byte)
                   (fli:incf-pointer ptr))
             byte-count))
          ((vectorp vector)
           #-lispworks4.2
           (let ((byte-count (the fixnum (min (the fixnum (- (length vector) start))
                                              max-octet-length))))
             (declare (fixnum byte-count))
             (fli:replace-foreign-array ptr vector
                                        :start1 index :end1 (+ index byte-count)
                                        :start2 start)
             byte-count)
           #+lispworks4.2
           (loop with byte-count of-type fixnum
                      = (the fixnum (min (the fixnum (- (length vector) start))
                                         max-octet-length))
                 with end = (+ start byte-count)
                 for i fixnum upfrom start below end
                 do (setf (fli:dereference ptr) (aref vector i))
                    (fli:incf-pointer ptr)
                 finally (return byte-count)))
          ((integerp vector)
           ;; Bookmark: we must fill all the reserved space with trailing zeros
           ;; because we can't change its length according to the integer length.
           ;(assert (zerop start) (start)
           ;  "Bookmarks can only be put into BINARY buffer without chunking.")
           (loop repeat max-octet-length
                 for byte-position fixnum upfrom 0 by 8 ;below (integer-length vector)
                 do (setf (fli:dereference ptr) (ldb (byte 8 byte-position) vector))
                    (fli:incf-pointer ptr))
                 ; sum 1))
           max-octet-length)
          (t (error "not implemented")))))

;typedef unsigned char   SQLDATE;
;; #if (ODBCVER >= 0x0300)
;typedef unsigned char   SQLTIME;
;typedef unsigned char   SQLTIMESTAMP;
;typedef unsigned char   SQLVARCHAR;
;; #endif

(defmacro %make-struct (type &optional (element-count 1))
  `(fli:allocate-foreign-object :type ',type
      ,@(unless (eql element-count 1) `(:nelems ,element-count))))

(defmacro allocate-dynamic-string (length &optional (type :char))
 ;;; Args: type  one of :char or :wchar-t
  ;; Value: Dynamic foreign string which is not null-terminated.
  ;; ToDo: Do we need to initialize it with #\Space?
  `(fli:allocate-dynamic-foreign-object :type ,type :nelems ,length))
				        ;:initial-element #\Space

(defmacro with-temporary-allocation (bindings &body body)
  (let ((simple-types ())
        (strings ())) 
    (dolist (binding bindings)
      (case (second binding)
        (:string   
         (push `(,(first binding) (allocate-dynamic-string ,(third binding)))
               strings))
        (:ef-wc-string
         (push `(,(first binding) (allocate-dynamic-string ,(third binding) :wchar-t))
               strings))
        (otherwise
         (push `(,(first binding) ,(second binding)) simple-types))))
    `(fli:with-dynamic-foreign-objects ,simple-types
       ,@(if strings
             `((let ,strings
               ,@body))
             body))))

#-lispworks4.2
(defmacro define-foreign-function (c-name args result-type &key documentation module
                                                           (no-check nil nc-supplied-p))
  (let* ((lisp-name (intern (string-upcase c-name)))
         (name-list (list lisp-name c-name :source)))  ; :dbcs
    `(dspec:def (define-foreign-function ,lisp-name)
       (fli:define-foreign-function
           ,name-list
           ,args
         :result-type ,result-type
         :language :ansi-c
         :documentation ,documentation
         :module (or ,module ,*foreign-module*)
         ,@(when nc-supplied-p `(:no-check ,no-check)) ))))
#-lispworks4.2
(dspec:define-form-parser define-foreign-function (c-name)
  `(,define-foreign-function ,(cond ((symbolp c-name) c-name)
                                    ((consp c-name) (first c-name))
                                    (t (intern (string-upcase c-name))))))
;=> DEFINE-FOREIGN-FUNCTION-FORM-PARSER

#+lispworks4.2
(defmacro define-foreign-function (c-name args result-type &key documentation module
                                                           (no-check nil nc-supplied-p))
  (let* ((lisp-name (intern (string-upcase c-name)))
         (name-list (list lisp-name c-name :source)))  ; :dbcs
    `(lw:top-level-form (define-foreign-function ,lisp-name)
       (fli:define-foreign-function
           ,name-list
           ,args
         :result-type ,result-type
         :language :ansi-c
         :documentation ,documentation
         :module (or ,module ,*foreign-module*)
         ,@(when nc-supplied-p `(:no-check ,no-check)) ))))

#+lispworks4.2 
(defun parse-define-foreign-function (stream definer-name)
  (let* ((pair (read stream))
         ;(name (first pair)))
         (name (intern (string-upcase pair))))
    `(,definer-name ,name)))
#+lispworks4.2
(setf (editor::get-parser 'define-foreign-function) 'parse-define-foreign-function)

#-lispworks4.2
(defmacro define-foreign-struct (name &rest slots)
 `(dspec:def (define-foreign-struct ,name)
   (fli:define-c-struct ,name ,@slots)))
#-lispworks4.2
(dspec:define-dspec-alias define-foreign-struct (name) `(fli:define-foreign-type ,name))

#+lispworks4.2
(defmacro define-foreign-struct (name &rest slots)
  `(fli:define-c-struct ,name ,@slots))

(defmacro %foreign-slot (ptr structure-type slot &key type)
 ;;; Args: structure-type, slot, and type are not evaluated.
  ;; NB: Specifying structure-type results in compiled code that (for both read and write)
  ;;	 1.9 times larger
  ;;	 5.0 times is faster
 `(fli:foreign-slot-value ,ptr ',slot
                          ,@(when structure-type `(,:object-type ',structure-type))
                          ,@(when type `(,:type ',type))))	; value type

(defmacro %foreign-slot-pointer (ptr structure-type slot &key type)
 ;;; Args: structure-type, slot, and type are not evaluated.
 `(fli:foreign-slot-pointer ,ptr ',slot
                          ,@(when structure-type `(,:object-type ',structure-type))
                          ,@(when type `(,:type ',type))))	; value type

(defmacro without-scheduling (&body body)
  `(mp:without-preemption ,@body))

(defun %ensure-hwnd (hwnd)
 "Convert implementation and platform specific parent window argument into pointer or
 address suitable for passing to SQLDriverConnect(), SQLManageDataSource() and the like"
  (if hwnd
      (%address-to-pointer
       (cond 
        #+(and capi win32)
        ((eq hwnd t)					; No window may exist at all
         (if ww::*active-window* (win32::r-window-hwnd ww::*active-window*) 0))
                     ;(yl:if-bind (window (ww::find-topmost-window nil nil))
                     ;  (capi:simple-pane-handle window) 0)
        #+(and capi (not unix))
        ((eq hwnd :podium)
         (capi:simple-pane-handle (capi:locate-interface 'lw-tools::lispworks-podium)))
         ;(win32::r-window-hwnd win32::*main-representation*)
        #+capi
        ((typep hwnd 'capi:simple-pane)		; interface or arbitrary pane
         #-lispworks4.2 (capi:simple-pane-handle hwnd)
         #+lispworks4.2 (capi-library:representation-handle
                         (capi::representation (capi:top-level-interface hwnd))))
        ((integerp hwnd)
         hwnd)
        (t
         (error "HWND is not NIL, T, :PODIUM, a simple-pane, or an integer")) ))
      *null-pointer*))


#|;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  BENCHMARKING  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; fli:convert-from-foreign-string vs. %get-string
;;; - allocates standard : 1.5	times more (len=50)
;;; - allocates conses	 : 2.25	times more - 99 vs. 44 (0 for lw44, len independent)
;;; - user time		 : 2.5-5 times slower (len=50, for lw<=43 the longer, the faster)
;;; NB: For LW44, we don't look at *foreign-ef* and always allocate a text string.

(setf len 50
      p1 (fli:allocate-foreign-object :type :char
                                       :initial-element 196
                                       :nelems len))
(setf (fli:dereference p1 :index (1- len)) 0)

(time (fli:convert-from-foreign-string p1
        :external-format win32:*multibyte-code-page-ef*
        :null-terminated-p nil
        :length len))		;=> 160 bytes standard / 99 bytes conses
(defun test-fli (ptr len)
  (declare (optimize (speed 3) (debug 0) (safety 0) (hcl:fixnum-safety 0)))
  (dotimes (n 100000)
    (declare (fixnum n))
    (fli:convert-from-foreign-string ptr
        :external-format win32:*multibyte-code-page-ef*
        :null-terminated-p nil
        :length len)))
(time (test-fli p1 len))

(time (%get-string p1 len))	;=> 120 bytes standard / 44 (0 for lw44!) bytes conses
(defun test-%get-string (ptr end)
  (declare (optimize (speed 3) (debug 0) (safety 0) (hcl:fixnum-safety 0))
           (fixnum end))
  (dotimes (n 100000)
    (declare (fixnum n))
    (%get-string ptr end)))
(time (test-%get-string p1 len))

;;; Paul's original %get-cstring without external format handling is very fast!
(time (%get-cstring p1)) 	;=> 32 bytes standard / 0 bytes conses
(defun test-%get-cstring (ptr)
  (declare (optimize (speed 3) (debug 0) (safety 0) (hcl:fixnum-safety 0)))
  (dotimes (n 100000)
    (declare (fixnum n))
    (%get-cstring ptr)))
(time (test-%get-cstring p1))

(fli:free-foreign-object p1)

;;; fli:convert-to-foreign-string vs. %put-string get-string
;;; - allocates standard : 40 vs. 0 (len independent)
;;; - allocates conses	 : 121 vs. 0 (len independent)
;;; - user time		 : 1.5	times faster (len=50, the longer, the faster)
;;; But for fli:convert-to-foreign-string
;;; - :null-terminated-p nil  does not work!
;;; - we must pass pointer to :c-array, not to :char

(setq str (make-string len :initial-element #\)
      ar (fli:allocate-foreign-object :type `(:c-array :char ,(1+ len))))
(fli:fill-foreign-object ar :byte 9)

(time (fli:convert-to-foreign-string str :into ar
        :external-format win32:*multibyte-code-page-ef*
        :null-terminated-p t		; nil - does not work!
        :limit (1+ len)))		;=> 40 bytes standard / 99 bytes conses
(defun test-put-fli (str ar len)
  (declare (optimize (speed 3) (debug 0) (safety 0) (hcl:fixnum-safety 0)))
  (declare (fixnum len))
  (let ((limit (1+ len)))
    (declare (fixnum limit))
    (dotimes (n 100000)
      (declare (fixnum n))
      (fli:convert-to-foreign-string str :into ar
        :external-format win32:*multibyte-code-page-ef*
        :null-terminated-p t		; nil - does not work!
        :limit limit))))		;=> 40 bytes standard / 99 bytes conses
(hcl:extended-time (test-put-fli str ar len))


(time (%put-string p1 str len))	;=> 0 bytes standard / 0 bytes conses
(defun test-%put-string (str ptr len)
  (declare (optimize (speed 3) (debug 0) (safety 0) (hcl:fixnum-safety 0)))
  (declare (fixnum len))
  (dotimes (n 100000)
    (declare (fixnum n))
    (%put-string ptr str len)))
(hcl:extended-time (test-%put-string str p1 len))


;;; UNICODE fli:convert-from-foreign-string vs. %get-wstring
;;; - allocates standard : eql
;;; - allocates conses	 : 66 vs. 0 (len independent)
;;; - user time		 : 3.5-30 times slower (len=50; for lw<=43:the longer, the faster)
(setf len 50
      p2 (fli:allocate-foreign-object :type :wchar-t
                                      :initial-element #.(char-code #\)
                                      :nelems len))
(time (fli:convert-from-foreign-string p2
        :external-format :unicode
        :null-terminated-p nil
        :length len))		;=> 120 bytes standard / 66 bytes conses
(defun test-fli-get-wchar (ptr len)
  (declare (optimize (speed 3) (debug 0) (safety 0) (hcl:fixnum-safety 0))
           (fixnum len))
  (dotimes (n 100000)
    (declare (fixnum n))
    (fli:convert-from-foreign-string ptr :external-format :unicode
        :null-terminated-p nil :length len)))
(hcl:extended-time (test-fli-get-wchar p2 len))

(time (%get-wstring p2 (* len 2)))	;=> 120 bytes standard / 0 bytes conses
(defun test-%get-wstring (ptr end)
  (declare (optimize (speed 3) (debug 0) (safety 0) (hcl:fixnum-safety 0))
           (fixnum end))
  (dotimes (n 100000)
    (declare (fixnum n))
    (%get-wstring ptr end)))
(hcl:extended-time (test-%get-wstring p2 (* len 2)))

;;; UNICODE fli:convert-to-foreign-string vs. %put-wstring get-string
;;; - allocates standard : 0
;;; - allocates conses	 : 66 vs. 0 (len independent)
;;; - user time		 : 1.5-5 times slower (len=50, lw<=43: the longer, the faster)
;;; But for fli:convert-to-foreign-string
;;; - :null-terminated-p nil  does not work!
;;; - we must pass pointer to :c-array, not to :char

(setq str (make-string len :initial-element #\)
      p2 (fli:allocate-foreign-object :nelems len :type 'SQL-WCHAR)
      a2 (fli:allocate-foreign-object :type `(:c-array :wchar-t ,(1+ len))))

(time (fli:convert-to-foreign-string str :into a2
        :external-format :unicode
        :null-terminated-p t		; nil - does not work!
        :limit (1+ len)))		;=> 0 bytes standard / 66 bytes conses
(defun test-fli-put-wchar (str ar len)
  (declare (optimize (speed 3) (debug 0) (safety 0) (hcl:fixnum-safety 0))
           (fixnum len))
  (let ((limit (1+ len)))
    (declare (fixnum limit))
    (dotimes (n 100000)
      (declare (fixnum n))
      (fli:convert-to-foreign-string str :into ar
        :external-format :unicode
        :null-terminated-p t		; nil - does not work!
        :limit limit))))
(hcl:extended-time (test-fli-put-wchar str a2 len))

(time (%put-wstring p2 str (* 2 len)))	;=> 0 bytes standard / 0 bytes conses
(defun test-%put-wstring (str ptr len)
  (declare (optimize (speed 3) (debug 0) (safety 0) (hcl:fixnum-safety 0)))
  (let ((octet-len (* 2 len)))
    (declare (fixnum octet-len))
    (dotimes (n 100000)
      (declare (fixnum n))
      (%put-wstring ptr str octet-len))))
(hcl:extended-time (test-%put-wstring str p2 len))

(setf ;size 50
      dp (%make-struct odbc::sql-c-date 1)
      (%foreign-slot dp odbc::sql-c-date odbc::day) 31
      (%foreign-slot dp odbc::sql-c-date odbc::month) 3
      (%foreign-slot dp odbc::sql-c-date odbc::year) 2005)

(time (odbc::date-to-universal-time dp))
(defun test-date-time (p)
  (declare (optimize (speed 3) (debug 0) (safety 0) (hcl:fixnum-safety 0)))
  (dotimes (n 10000000)
    (declare (fixnum n))
    (setf (%foreign-slot p odbc::sql-c-date odbc::day) 31
          (%foreign-slot p odbc::sql-c-date odbc::month) 3
          (%foreign-slot p odbc::sql-c-date odbc::year) 2005)))
    #|(list (%foreign-slot p odbc::sql-c-date odbc::day :type SQL-USMALLINT)
          (%foreign-slot p odbc::sql-c-date odbc::month :type SQL-USMALLINT)
          (%foreign-slot p odbc::sql-c-date odbc::year :type SQL-SMALLINT))))|#
    ;(odbc::date-to-universal-time p)))
(hcl:extended-time (test-date-time dp))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; OTHER TESTING

(setf len 50
       ;:initial-element 194))
      p1 (fli:allocate-foreign-object :nelems len :type '(:unsigned :char))
      p2 (fli:allocate-foreign-object :nelems len :type 'SQL-WCHAR)
      pb (fli:allocate-foreign-object :nelems len :type :unsigned-byte)
      ar (fli:allocate-foreign-object :type `(:c-array (:unsigned :char) ,(1+ len)))
      ar2 (fli:allocate-foreign-object :type `(:c-array :wchar-t ,(1+ len))))
(setq s  (make-string len :initial-element #\)
      ca (make-array len :element-type 'lw:simple-char :initial-element #\)
      bs (make-string len :element-type 'base-char :initial-element #\b)
      bca(make-array len :element-type 'base-char :initial-element #\b)
      v  (make-array len :element-type '(unsigned-byte 8)
                     :initial-contents (loop for i upfrom 0 below len collect i))
      av (make-array len :adjustable t :element-type '(unsigned-byte 8)
                     :initial-contents (loop for i downfrom (1- len) to 0 collect i)))

(fli:dereference ptr) :type :unsigned-byte)
(fli:replace-foreign-array p1 bs :end1 10)	; does not work for strings!
(fli:replace-foreign-array p2 ar2 :end1 10)	; does not work for :wchar-t!
(fli:replace-foreign-array ar v :end2 10)	; works
(fli:replace-foreign-array pb v :end1 len)
(fli:replace-foreign-array av pb :end2 len)	; works with adjusted.

(fli:replace-foreign-object p1 ar :nelems len)	; pointer <- array does not work!
(fli:replace-foreign-object p1 pb :nelems len)	; pointer <- array does not work!
(fli:replace-foreign-object p2 ar2 )	; works for :wchar-t!
(fli:with-dynamic-lisp-array-pointer (ptr s)	; requires static array!
  (fli:replace-foreign-object p2 ptr :nelems len))
;;; DANGEROUS: how to get sting address?
(fli:replace-foreign-object p2 (fli:make-pointer :address (+ (sys:object-address s) 16)
                                                 :type :wchar-t)
                            :nelems len)

(%fill-vector p1 :end nil :byte (char-code #\b))

(setq np (%make-struct odbc::sql-c-numeric))
(setq vp (%foreign-slot-pointer np nil odbc::value :type SQL-CHAR))
(%fill-vector vp :end 5 :byte 4)

(loop for i upfrom 0 below 17 collect (fli:dereference vp :index i))
(loop for i upfrom 0 below len collect (fli:foreign-aref ar i))

;;; Using arguments (:type :char/SQL-CHAR) is about 1.5 times slower!
(defun test-dereference (ptr)
  (declare (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0) (debug 0)))
  (fli:with-coerced-pointer (p :type :char) ptr
    (dotimes (i most-positive-fixnum i)
      (declare (fixnum i))
      (dotimes (j 10)
        (declare (fixnum j))
        (fli:dereference p :type :char))))) ;:type 'SQL-CHAR)))))
(time (test-dereference ptr))

(fli:pointer-element-type-p p ':unsigned-short)

(fli:foreign-type-equal-p ':unsigned-byte '(:unsigned :byte))


(with-cstr (p1 "") (list p1))
(with-cstr (p1 "" b1) (list p1 b1))
(with-cstrings ((p1 "" c1) (p2 nil) (p3 (lw:string-append "a" "b") c3))
 (list p1 p2 p3 c1 c3))

(defun allocate-dynamic-char (c-type &rest args)
  (fli:with-dynamic-foreign-objects ((ptr :long)) ; (char 2)))
    (setf (fli:dereference ptr) 0)
    (loop for i upfrom 0 
          and char in args
          do (setf (fli:dereference ptr :type c-type :index i) char))
    (values (%get-char ptr :type c-type)
            (%get-char ptr :type c-type :index 1)
            (%get-char ptr :index 0)
            (%get-char ptr :index 1)
            (%get-char ptr :index 2)
            (%get-char ptr :index 3))))
  
(allocate-dynamic-char :wchar-t #\a #\b)
(fli:allocate-foreign-object :type 'sql-c-timestamp)

(EXTERNAL-FORMAT:ENCODE-LISP-STRING "abcd" win32:*multibyte-code-page-ef*)
(EXTERNAL-FORMAT:ENCODE-LISP-STRING "abcd" :unicode)
(EXTERNAL-FORMAT:ENCODE-LISP-STRING "" :unicode)	;=> #(1072 1092 1073)
(EXTERNAL-FORMAT:DECODE-EXTERNAL-STRING #(255 254) win32:*multibyte-code-page-ef*)
(external-format:char-external-code #\ 1251)
(external-format:char-external-code #\a 1252)

(external-format-type win32:*multibyte-code-page-ef*)
(ef:external-format-type :latin-1)

;;;;;;;;;;;;;;;;;;;;;;;;;;;  OLD AND UNUSED STUFF ;;;;;;;;;;;;;;;;;;;;;;;;;;
#+ignore
(defmacro %get-wchar (ptr) `(fli:dereference ,ptr :type :wchar-t))

#+ignore
(declaim (ftype (function (t) fixnum) %cstring-length)
         (inline %cstring-length))
#+ignore
(defun %cstring-length (ptr) ; &optional type)
 ;;; Args: ptr   Null-terminated C string.
  ;;       type  If nil, assume default pointer element type, otherwise must be
  ;;       :char or :wchar_t.
  ;(if type
  ;    (loop with length = 0
  ;          until (char= (fli:dereference ptr :type type :index length) #\Null)
  ;          do (incf length)
  ;          finally return length)
  (declare (optimize (speed 3) (space 0) #-debug (safety 0) #-debug (debug 0)
                     #-debug (hcl:fixnum-safety 0)))
  (the fixnum (loop with length of-type fixnum = 0
                    until (char= (fli:dereference ptr :index length) #\Null)
                    do (incf length)
                    finally return length)))
#+ignore
(defun %cstring-to-keyword (ptr &optional (length (%cstring-length ptr)))
  (let ((string (make-string length :element-type 'base-char)))
    (declare (dynamic-extent string))
    (%cstring-into-string ptr string 0 length)
    (intern string (find-package :keyword))))
#+ignore
(defun %cstring-into-string (ptr string offset size-in-bytes)
    (dotimes (i size-in-bytes offset)
      (setf (char string offset) (%get-char ptr :index i))
      (incf offset)))
#+ignore
(defun %cstring-into-vector (ptr vector offset size-in-bytes)
    (dotimes (i size-in-bytes offset)
      (setf (aref vector offset) (fli:dereference ptr :index i))
      (incf offset)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;; FFC ON OTHER LISP IMPLEMENTATIONS ;;;;;;;;;;;;;;;;;;;;;;
;;; (1) Must equip define-foreign-struct with appropriate replacement of array type
;;;	specifications like (:c-array SQL-CHAR 16)

;(lispworks:without-interrupts a)
;(lispworks:without-preemption a)
#+allegro
(defmacro without-scheduling (&body body)
  `(mp:without-scheduling ,@body))

#+cormanlisp
(defmacro without-scheduling (&body body)
  `(progn ,@body))

#+pcl
(defmacro without-scheduling (&body body)
  `(pcl::without-interrupts ,@body))
 |#
