;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2005 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; ODBC interface - variables and classes

(in-package :ystok.odbc)

(pushnew :odbc3.0 *features*)	; Means that our program is an ODBC 3.0 application

(defparameter $ODBCVER 0351 ;0210
  "Environment version matching the version of Driver Manager
 determined dynamically via SQLGetInfo(SQL_ODBC_VER)")

(defparameter *unicode* nil
  "Default value of switch exposing ourselves as a Unicode UCS-2 application or not.")

;(defparameter *odbc-driver-connect-p* nil)

(defmacro with-error-signaling ((handle-type handle &key (on-info t) args)
                                odbc-call &body body)
 ;;; ToDo:    ?
  `(progn (%analyse-return-code ,odbc-call ,handle-type ,handle ,on-info ,args)
          ,@body))

(defmacro %make-int64 (high low) `(logior (ash ,high 32) ,low))
  #|(with-gensyms (value)
    `(let ((,value ,high))
       (if (zerop (ldb (byte 1 31) ,value))				; plusp
           (+ (ash ,value 32) ,low)
           (- (+ (ash ,value 32) ,low) #x10000000000000000)))))|# 	; 2^64
(defmacro %int64-low  (value) `(logand ,value #xFFFFFFFF))	 	; 2^32 - 1
(defmacro %int64-high (value) `(ash ,value -32))  ; only works correctly for signed int64!
#|(defmacro %int64-uhigh (value) (with-gensyms (variable)
                                 `(let ((,variable ,value))
                                    `(if (minusp ,variable)
                                         (logand (ash ,variable -32) #xFFFFFFFF)
                                         (ash ,variable -32)))))|#

;;; ODBC conditions

(defstruct (diag-record (:conc-name diag-)
                        (:copier nil))
 ;;; Diagnostic status records
  sql-state		;SQL_DIAG_SQLSTATE 
  secondary-error-id	;SQL_DIAG_NATIVE
  message		;SQL_DIAG_MESSAGE_TEXT
  ;; Row number in the rowset, or the parameter number in the paramset,
  ;; or SQL_NO_ROW_NUMBER(-1) if this status record is not associated with number,
  ;; or SQL_ROW_NUMBER_UNKNOWN(-2) if the driver cannot determine the number.
  row-number		;SQL_DIAG_ROW_NUMBER
  ;; Column number in the result set or the parameter number in the paramset
  ;; if the SQL_DIAG_ROW_NUMBER field is a valid row number,
  ;; or SQL_NO_COLUMN_NUMBER(-1) if this status record is not associated with any col,
  ;; or SQL_COLUMN_NUMBER_UNKNOWN(-2) if the driver cannot determine the number.
  column-number		;SQL_DIAG_COLUMN_NUMBER
			;SQL_DIAG_CONNECTION_NAME
			;SQL_DIAG_SERVER_NAME
			;SQL_DIAG_CLASS_ORIGIN	- "ISO 9075" or "ODBC 3.0" 
			;SQL_DIAG_SUBCLASS_ORIGIN - "ISO 9075" or "ODBC 3.0"
 )

(define-condition odbc-condition (sql-condition)
  (;(secondary-error-id :reader native-error :initarg :native-error)
   (diag-records       :reader diag-records :initarg :records  :initform nil)))

(defun print-diag-records (condition &optional stream)
  (when-bind (diag-records (diag-records condition))
    (fresh-line stream)
    (princ "More diagnostic records:" stream)
    (loop for record in diag-records
          and record-number upfrom 1
          ;unless (equal (diag-sql-state record) (sql-state c))
          do (format stream
                     "~%~D [SQLSTATE ~A~@[ native ~A~]~@[ row ~S~]~@[ column ~S~]] ~A"
                     record-number (diag-sql-state record) (diag-secondary-error-id record)
                     (diag-row-number record) (diag-column-number record)
                     (diag-message record))) ))

(define-condition odbc-connection-error (sql-connection-error odbc-condition)
 ((db-type :initform :odbc))
 (:report (lambda (c stream)
            (flet ((mask-password (connection-spec)
                     (if (consp connection-spec)
                         (cons (first connection-spec)
                              (remove-properties (rest connection-spec) '(:password :pwd)))
                         connection-spec)))
              (format stream "[ODBC~@[ SQLSTATE ~A~] error~@[ ~A~]] connecting to ~S.~%~A"
                      (sql-state c) (secondary-error-id c)
                      (mask-password (connection-spec c)) (message c))
              (print-diag-records c stream)))))

(define-condition odbc-connection-timeout-error (odbc-connection-error sql-timeout-error)
 ())

(define-condition odbc-error (sql-database-error odbc-condition) ()
 (:report (lambda (c stream)
            (format stream "[ODBC~@[ SQLSTATE ~a~] error~@[ ~a~]] using database ~:[#<unknown>~;~:*~a~].
~a~@[~%SQL expression: ~a~]"
                    (sql-state c) (secondary-error-id c)
                    (and (database c) (database-name (database c)))
                    (message c)
                    (if (typep (expression c) 'sql-expression)	; print cautiously
                        (ignore-errors (write-sql-to-string (expression c) (database c)))
                         (expression c)))
            (print-diag-records c stream))))

(defgeneric odbc-error (database sql-state secondary-error-id &key &allow-other-keys)
 (:documentation "Generic ODBC error signaler")
 (:method (database sql-state secondary-error-id &key expression message records)
  (error 'odbc-error :database database :expression expression
                     :sql-state sql-state  :secondary-error-id secondary-error-id
                     :message message
                     :records records)))

(define-condition odbc-invalid-handle-error (odbc-error) ()
  (:default-initargs :message "Invalid handle."))

(define-condition odbc-still-executing-error (odbc-error sql-temporary-error) ()
  (:default-initargs :message "Still executing."))

(define-condition odbc-timeout-error (odbc-error sql-timeout-error)
 ()
 (:report (lambda (c stream)
            (format stream "[ODBC~@[ SQLSTATE ~a~] error~@[ ~a~]] using database ~:[#<unknown>~;~:*~a~].
~a~@[ Timeout value (secs): ~s]~@[~%SQL expression: ~s~]"
                    (sql-state c) (secondary-error-id c)
                    (and (database c) (database-name (database c)))
                    (message c) (timeout c)
                    (if (typep (expression c) 'sql-expression)	; print cautiously
                        (ignore-errors (write-sql-to-string (expression c) (database c)))
                        (expression c)))
            (print-diag-records c stream))))

;;; Most of the warnings either come when the we have already connected or when
;;; we cannot or don't want to differentiate. So include all the baggage of slots.
(define-condition odbc-warning (sql-warning sql-database-condition odbc-condition) ()
  (:report (lambda (c stream)
	     (format stream "[ODBC~@[ SQLSTATE ~A~] info~@[ ~A~]] using database ~:[#<unknown>~;~:*~A~].
~A~@[~%SQL expression: ~S~]"
                     (sql-state c) (secondary-error-id c)
		     (and (database c) (database-name (database c)))
                     (message c)
                     ; print the expression cautiously
                     (if (typep (expression c) 'sql-expression)
                         (ignore-errors (write-sql-to-string (expression c)
                                                             (database c)))
                         (expression c)) ))))

(define-condition odbc-success-with-info-warning (odbc-warning) ())
;(define-condition odbc-still-executing-warning (odbc-warning) ()

;;; Environment, database, and statement classes

(defclass handled-object (db-object)
  ((handle :initarg :handle :accessor handle :initform nil)))

(defclass odbc-environment (handled-object environment)
  ((db-type :initform :odbc)
   (handle :accessor henv)))

;;; Caching not only speeds up processing of oftenly invoked information requests,
;;; but also decreases the possibility of simultaneous usage of the connection handle.
;;; This makes us more robust when dealing with a not thread-safe driver.

(defclass odbc-database (handled-object database)
  ((handle :accessor hdbc)
   (info :reader info :initform (make-hash-table))    ; Cache values of SQLGetInfo
   (type-info-vector :accessor type-info-vector	      ; Cache values of SQLGetTypeInfo
                     :initform nil)
   (functions-vector :initform nil)		      ; Cache values of SQLGetFunctions
  ) 
  #+cormanlisp (:metaclass cl::class-slot-class))

(defmethod henv ((db odbc-database)) (handle (environment db)))

#-mcl
(defclass access-odbc-database (access-mixin odbc-database) ())
#-mcl
(defclass excel-odbc-database (desktop-mixin odbc-database) ())
#-mcl
(defclass mssql-database (sql.sys::db-mixin odbc-database) ())		; only via ODBC
#-mcl
(defclass mysql-odbc-database (mysql-mixin odbc-database) ())
(defclass oracle-odbc-database (oracle-mixin odbc-database) ())
(defclass postgresql-odbc-database (postgresql-mixin odbc-database) ())
(defclass sqlite3-odbc-database (sqlite3-mixin odbc-database) ())	; SQLite3

(defclass odbc-statement (handled-object statement)
  ((handle :accessor hstmt)))

(defclass odbc-cursor (odbc-statement cursor)
  (;; Application or implementation row descriptor combined
   (row-descriptor :initarg :row-descriptor :accessor row-descriptor :initform nil)
  ))

(defclass odbc-param-statement (odbc-statement param-statement)
  (;; Application or implementation parameter descriptor combined
   (param-descriptor :initarg :param-descriptor :accessor param-descriptor
                     :initform nil)
  ))

(defclass odbc-param-cursor (odbc-param-statement odbc-cursor) ())


(defgeneric db-describe-columns (database table &key column catalog schema))
