;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2005 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; ODBC cursor operations

(in-package :ystok.odbc)

(defmethod initialize-instance :after ((cursor odbc-cursor) &key sql-text
                                       			    &allow-other-keys)
  (when sql-text
    (with-slots (handle active) cursor
      (%sql-exec-direct handle sql-text) ; hdbc)
      (setf active t))
    (%initialize-rowset cursor)))

(defmethod reinitialize-instance :after ((cursor odbc-cursor) &key query
                                                              &allow-other-keys)
 ;;; Requery the cursor, i.e. close it and re-execute the statement.
  ;; Args: query  Either null - use of the current expression, 
  ;;              or a new expression resulting in the same columns.
  (with-slots (handle) cursor
    (when handle
      (db-close cursor :unbind nil))
      ;(setf active t)
      (db-open-cursor cursor query :requery t)))

(defun %initialize-cursor (cursor cursor-type &optional concurrency limit timeout)
  (with-slots (handle database) cursor
    (%sql-set-attribute :hstmt handle SQL_ATTR_CURSOR_TYPE
                        (case (setf (cursor-type cursor) cursor-type)
                          (:forward-only  (setq concurrency :read-only)
                                          SQL_CURSOR_FORWARD_ONLY)
                          (:static        SQL_CURSOR_STATIC)
                          (:keyset-driven (unless concurrency
                                            (setq concurrency :optimistic))
                           		  SQL_CURSOR_KEYSET_DRIVEN)
                          (:dynamic	  (unless concurrency
                                            (setq concurrency :optimistic))
                                          SQL_CURSOR_DYNAMIC)))
    (typecase database
      (sqlite3-odbc-database
       ;; SQL_ATTR_CONCURRENCY is always set to SQL_CONCUR_LOCK.
       ;; TODO?: If LIMIT equals to :none, do not even try calling SQLSetStmtAttr 
       ;;        to work around driver nonconformity, e.g. sqlite3odbc.dll
      )
      (t
       (%sql-set-attribute :hstmt handle SQL_ATTR_CONCURRENCY
                           (case concurrency
                             (:lock      	  SQL_CONCUR_LOCK)
                             (:optimistic
                              (let ((info (db-get-info database SQL_SCROLL_CONCURRENCY)))
                                (cond ((logtest info SQL_SCCO_OPT_ROWVER)
                                       SQL_CONCUR_ROWVER)
                                      ((logtest info SQL_SCCO_OPT_VALUES)
                                       SQL_CONCUR_VALUES)
                                      (t  SQL_CONCUR_LOCK)))) ;use pessimistic lock type
                             (:rowver	  SQL_CONCUR_ROWVER)
                             (:values	  SQL_CONCUR_VALUES)
                             (otherwise	  SQL_CONCUR_READ_ONLY))) ;imply SQL_INSENSITIVE
       ;; The below attribute is intended to reduce network traffic. Conceptually, it is 
       ;; applied when the result set is created and limits the result set to the first
       ;; LIMIT rows. 
       ;; If LIMIT equals 0 (the default), the driver returns all rows. If the number
       ;; of rows in the result set is greater than LIMIT, the result set is truncated.
       (%sql-set-attribute :hstmt handle SQL_ATTR_MAX_ROWS (or limit 0))) )
    (when timeout
      (%sql-set-attribute :hstmt handle SQL_ATTR_QUERY_TIMEOUT timeout)) ))

(defmethod db-open-cursor :before ((cursor odbc-cursor) query-expression
                                   &key (cursor-type :forward-only)
                                   concurrency
                                   ;(rowset-size 1)
                                   ;(start-column 1) end-column column-numbers
                                   (limit
                                    (and (typep query-expression 'sql-query-exp)
                                         (slot-value query-expression 'limit)))
                                   requery
                                   timeout
                                   &allow-other-keys)
  (when query-expression
    (setf (expression cursor) query-expression))
  (unless requery
    (%initialize-cursor cursor cursor-type concurrency limit timeout)))

(defmethod db-open-cursor ((cursor odbc-cursor) query-expression
                           &key ;(cursor-type :forward-only) concurrency
                                (rowset-size 1)
                                (start-column 1) end-column column-numbers
                                result-types requery
                           &allow-other-keys)
 ;;; Args: start-column If zero, a bookmark column is to be created and bound
  ;; Q: Can timeout be changed when requerying?
  (with-slots (handle database expression) cursor
    (when (and (not requery) (or (eql start-column 0) (member 0 column-numbers)))
      ;; Check whether bookmarks are supported at all (none in native Oracle driver).
      (cond ((not (zerop (db-get-info database SQL_BOOKMARK_PERSISTENCE)))
             (%sql-set-attribute :hstmt handle SQL_ATTR_USE_BOOKMARKS
                                 (if (>= $ODBCVER 0300) SQL_UB_VARIABLE SQL_UB_FIXED)))
            (t
             (cerror "Do not bind the bookmark column."		; odbc-data-error?
                     'odbc-error :database database :expression query-expression
                     :message "Bookmarks are not supported by ODBC driver")
             (if (eql start-column 0)
                 (setq start-column 1)
                 (setq column-numbers (remove 0 column-numbers))))))
    (%sql-exec-direct handle (write-sql-to-string expression database)))
  (if requery
      ;(%reinitialize-rowset cursor :size rowset-size) ; size not provided with requery
      (%shared-initialize-rowset cursor)
      (%initialize-rowset cursor :size rowset-size
                                 :result-types result-types
                                 :start-column start-column :end-column end-column
                                 :column-numbers column-numbers))
  cursor)

(defmethod db-open-cursor :after ((cursor odbc-cursor) dummy &key &allow-other-keys)
  (declare (ignore dummy))
  (with-slots (active) cursor
    (setf active :open)))

(defmethod db-close :before ((cursor odbc-cursor) &key (unbind t) &allow-other-keys)
  (with-slots (handle row-descriptor active) cursor
    (case active
      ((:need-data :still-executing)
       (db-cancel cursor)		; active -> t
       (setf active :open)))
    (cond (unbind
           (when handle (%sql-free-statement handle :close :unbind))
           (when row-descriptor
             (%dispose-descriptor row-descriptor)
             (setf ;row-descriptor nil
                   active :unbound)))
          ((and handle (eq active :open))
           (%sql-close-cursor handle)
           (setf active :closed))))
  cursor)

(defmethod db-close :after ((cursor odbc-cursor) &key drop-p &allow-other-keys)
  (with-slots (row-descriptor) cursor
    (when (and drop-p row-descriptor)
      (%dispose-descriptor row-descriptor t)
      (setf row-descriptor nil))))

(defmethod db-query ((cursor odbc-cursor) query-expression
                     &key flatp result-types parameters values timeout)
  (db-open-cursor cursor query-expression :result-types result-types
                  :parameters parameters :values values :timeout timeout)
  (values (db-fetch-rows cursor :flatp flatp)
          ;; (column-names cursor)	; LMH return the column names as well
          (map 'list #'desc-name (desc-records (row-descriptor cursor)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Misc cursor operations

(defmethod db-columns-count ((cursor odbc-cursor))
  (if-bind (row-descriptor (row-descriptor cursor))
    (desc-count row-descriptor)
    (%sql-columns-count (handle cursor))))

(defmacro %decode-row-status (arg)
  `(svref #(t		;SQL_ROW_SUCCESS
            :deleted	;SQL_ROW_DELETED
            :updated	;SQL_ROW_UPDATED	; updated by this cursor or another user
            :norow	;SQL_ROW_NOROW
            :added	;SQL_ROW_ADDED
            :error	;SQL_ROW_ERROR
            :warn)	;SQL_ROW_SUCCESS_WITH_INFO
          ,arg))
  #|(case in
    ((#.SQL_ROW_SUCCESS #.SQL_ROW_SUCCESS_WITH_INFO) t)
    (#.SQL_ROW_UPDATED :updated)
    (#.SQL_ROW_ADDED   :added)
    (#.SQL_ROW_ERROR   :error)
    (#.SQL_ROW_DELETED :deleted) 
    (#.SQL_ROW_NOROW   :norow))) |#

(declaim (ftype (function (t) t) %get-row-status)
         (inline %get-row-status))
(defun %get-row-status (row-descriptor)
  (or (current-row-status row-descriptor)
      (setf (current-row-status row-descriptor)
             (let ((current-row-nr (current-row-nr row-descriptor)))
               (if (plusp current-row-nr)
                   (%decode-row-status (desc-row-status row-descriptor
                                                        (1- current-row-nr)))
                   :norow)))))
             ;(let ((current-row-nr (current-row-nr row-descriptor)))
             ;  (if (<= 1 current-row-nr (desc-array-size row-descriptor))
             ;      (%decode-row-status (desc-row-status row-descriptor
             ;                                           (1- current-row-nr)))
             ;      :norow)))))

(defmethod current-row-status ((cursor odbc-cursor))
  (%get-row-status (row-descriptor cursor)))

(defmethod current-row-number ((cursor odbc-cursor))
  ;; For the current row number, applications can:
  ;; - either keep track of the current row number 
  ;; - or call SQLGetStmtAttr with the SQL_ATTR_ROW_NUMBER attribute
  ;;   to retrieve it (doesn't work for Access).
  (with-slots (block-position current-row-nr) (row-descriptor cursor)
    (if (numberp block-position)
        (+ block-position current-row-nr)
        block-position)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Position

;;; The following macros asume number is non-negative
(defmacro %incf-block-position (number)
  `(cond ((numberp block-position)
          (incf block-position ,number))
         ((eq block-position :start)
          (setf block-position ,number))))
  
(defmacro %decf-block-position (number)
  `(cond ((numberp block-position)
          (if (minusp block-position)
              (decf block-position ,number)
              (setf block-position (max (- block-position ,number) 0))))
         ((eq block-position :end)
          (setf block-position (- ,number)))))

#|(defmethod set-cursor-position :before ((cursor odbc-cursor) orientation
                                        &key offset (retrieve-data t))
  (declare (ignore orientation offset))
  (when (active cursor)
    (%sql-set-attribute :hstmt (handle cursor) SQL_ATTR_RETRIEVE_DATA
                        (if retrieve-data SQL_RD_ON SQL_RD_OFF))))|#

(defmethod set-cursor-position ((cursor odbc-cursor) (orientation (eql :next))
                                &key offset (retrieve-data t) errorp)
  (declare (ignore offset errorp))
  (when (active cursor)
   (with-slots (handle row-descriptor) cursor
    (with-slots (block-position fetched-count last-operation
                 current-row-nr current-row-status) row-descriptor
      (cond ((< 0 current-row-nr fetched-count)
             ;(unless (eq last-operation :probe) (setf last-operation orientation))
             (cond ((eq (retrieve-data row-descriptor) retrieve-data)
                    (setf last-operation orientation
                          current-row-status nil)
                    (incf current-row-nr))
                   ((/= (%sql-fetch-scroll handle row-descriptor :relative :offset 0
                                          :retrieve-data retrieve-data)
                        SQL_NO_DATA)
                    (setf fetched-count (desc-rows-processed row-descriptor)
                          ;last-operation orientation
                          current-row-status nil)
                    (incf current-row-nr))
                   (t (setf block-position :end
                            ;fetched-count 0 current-row-nr 0
                            current-row-status :norow)
                      nil)))
            ((/= (%sql-fetch-scroll handle row-descriptor orientation
                                    :retrieve-data retrieve-data)
                 SQL_NO_DATA)
             (%incf-block-position fetched-count)
             (setf fetched-count (desc-rows-processed row-descriptor)
                   ;last-operation orientation
                   current-row-status nil
                   current-row-nr 1))			; the first within rowset
            (t (setf block-position :end
                     ;fetched-count 0 current-row-nr 0
                     current-row-status :norow)
               nil)) ))))

(defmethod set-cursor-position ((cursor odbc-cursor) (orientation (eql :prior))
                                &key offset (retrieve-data t) errorp)
  (declare (ignore offset errorp))
  (when (active cursor)
   (with-slots (handle row-descriptor) cursor
    (with-slots (block-position fetched-count last-operation
                 current-row-nr current-row-status) row-descriptor
      (cond ((<= 2 current-row-nr fetched-count)
             ;(unless (eq last-operation :probe) (setf last-operation orientation))
             (cond ((eq (retrieve-data row-descriptor) retrieve-data)
                    (setf last-operation orientation
                          current-row-status nil)
                    (decf current-row-nr))
                   ((/= (%sql-fetch-scroll handle row-descriptor :relative :offset 0
                                          :retrieve-data retrieve-data)
                        SQL_NO_DATA)
                    (setf fetched-count (desc-rows-processed row-descriptor)
                          ;last-operation orientation
                          current-row-status nil)
                    (decf current-row-nr))
                   (t (setf block-position :start
                            ;fetched-count 0 current-row-nr 0
                            current-row-status :norow)
                      nil)))
            ((/= (%sql-fetch-scroll handle row-descriptor orientation
                                    :retrieve-data retrieve-data)
                 SQL_NO_DATA)
             (%decf-block-position (desc-rows-processed row-descriptor))
             (setf fetched-count (desc-rows-processed row-descriptor)
                   ;last-operation orientation
                   current-row-status nil
                   current-row-nr (desc-rows-processed row-descriptor))) ;last in rowset
            (t (setf block-position :start
                     ;fetched-count 0 current-row-nr 0
                     current-row-status :norow)
               nil)) ))))

(defmethod set-cursor-position ((cursor odbc-cursor) (orientation (eql :relative))
                                &key offset (retrieve-data t) errorp)
  (declare (ignore errorp))
  (when (active cursor)
   (with-slots (handle row-descriptor) cursor
    (with-slots (block-position fetched-count last-operation
                 current-row-nr current-row-status) row-descriptor
      (cond ((not (integerp offset))
             (error 'sql-simple-error
                    :format-control "Bad or missing offset ~s for cursor scrolling orientation ~sS."
                    :format-arguments (list offset orientation)))
            ((<= 1 (+ current-row-nr offset) fetched-count)
             ;(unless (eq last-operation :probe) (setf last-operation orientation))
             (cond ((eq (retrieve-data row-descriptor) retrieve-data)
                    (setf last-operation orientation
                          current-row-status nil)
                    (incf current-row-nr offset))
                   ((/= (%sql-fetch-scroll handle row-descriptor :relative :offset 0
                                          :retrieve-data retrieve-data)
                        SQL_NO_DATA)
                    (setf fetched-count (desc-rows-processed row-descriptor)
                          ;last-operation orientation
                          current-row-status nil)
                    (incf current-row-nr offset))
                   (t (setf block-position :end
                            ;fetched-count 0 current-row-nr 0
                            current-row-status :norow)
                      nil)))
            ((/= (%sql-fetch-scroll handle row-descriptor orientation :offset offset
                                    :retrieve-data retrieve-data)
                 SQL_NO_DATA)
             (if (minusp offset)
                 (%decf-block-position (- offset))
                 (%incf-block-position offset))
             (setf fetched-count (desc-rows-processed row-descriptor)
                   ;last-operation orientation
                   current-row-status nil
                   current-row-nr 1)) 		; first in rowset
            (t (setf block-position (if (minusp offset) :start :end)
                     ;fetched-count 0 current-row-nr 0
                     current-row-status :norow)
               nil)) ))))

(defmethod set-cursor-position ((cursor odbc-cursor) (orientation (eql :bookmark))
                                &key offset (retrieve-data t) errorp)
 ;;; Args: offset  Bookmark value, so always position with zero FetchOffset
  (declare (ignore errorp))
  (when (>= $ODBCVER 0300)
   (when (active cursor)
    (with-slots (handle row-descriptor) cursor
     (with-slots (block-position fetched-count current-row-nr current-row-status
                  bookmark-record) row-descriptor
      (unless (and bookmark-record (integerp offset))
        (error 'sql-simple-error
               :format-control "Bad or missing bookmark ~S."
               :format-arguments (list offset)))
      (let ((ptr (bookmark-ptr row-descriptor))
            (octet-length (desc-octet-length bookmark-record)))
                                  ;(if (integerp offset)
                                  ;    (ceiling (integer-length offset) 8)
                                  ;    (length offset))))
        (unless ptr               ; check if bookmark was ever allocated
          (%sql-set-attribute :hstmt handle SQL_ATTR_FETCH_BOOKMARK_PTR 
                              (setf (bookmark-ptr row-descriptor)
                                    (setq ptr (%new-binary octet-length)))))
        (%put-binary ptr offset octet-length))
      (if (/= (%sql-fetch-scroll handle row-descriptor orientation
                                 :retrieve-data retrieve-data)
              SQL_NO_DATA)
          (setf fetched-count (desc-rows-processed row-descriptor)
                current-row-status nil
                block-position nil
                current-row-nr 1)			; the first within rowset
          (setf current-row-status :norow
                block-position nil)))))))

(defmethod set-cursor-position ((cursor odbc-cursor) orientation
                                &key offset (retrieve-data t) errorp)
 ;;; Default method used for absolute positioning.
  (declare (ignore errorp))
  (when (active cursor)
   (with-slots (handle row-descriptor) cursor
    (with-slots (block-position fetched-count current-row-nr current-row-status)
        row-descriptor
     (if-bind (row-nr			; new row number within rowset on succeeding
               (ecase orientation
                 (:first
                  (setf offset most-negative-fixnum)
                  (when (/= (%sql-fetch-scroll handle row-descriptor orientation
                                               :retrieve-data retrieve-data)
                            SQL_NO_DATA)
                    (setf block-position 0)
                    1))
                 (:last
                  (setf offset most-positive-fixnum)
                  (when (/= (%sql-fetch-scroll handle row-descriptor orientation
                                               :retrieve-data retrieve-data)
                        SQL_NO_DATA)
                    (setf block-position (- (desc-rows-processed row-descriptor)))
                    (desc-rows-processed row-descriptor)))		; first from end
                 (:absolute
                  (cond ((not (integerp offset))
                         (error 'sql-simple-error
                                :format-control "Bad or missing offset ~S for cursor scrolling orientation ~S."
                                :format-arguments (list offset orientation)))
                         ((/= (%sql-fetch-scroll handle row-descriptor orientation
                                                 :offset offset
                                                 :retrieve-data retrieve-data)
                              SQL_NO_DATA)
                          (setf block-position (1- offset))
                          1)))
                 ))
       (setf fetched-count (desc-rows-processed row-descriptor)
             current-row-status nil
             current-row-nr row-nr)
       (setf block-position (cond ((or (null offset) (zerop offset)) block-position)
                                  ((plusp offset) :end)
                                  (t :start))
             ;fetched-count 0 current-row-nr 0
             current-row-status :norow)) ))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Fetching

(defun %fetch-row (hstmt row-descriptor)
 ;;; Retrieve the next row or the current row if the previous operation was positioning.
  ;; Value: current-row-nr or nil if error or no data any more.
  ;; BUG Access driver odbcjt32.dll version 4.0.6200.0 - SQLFetch() and SQLFetchScroll.
  ;;	For block cursor (size > 1) and forward-only (tested), the driver sometimes
  ;;	- sets SQL_ATTR_ROWS_FETCHED_PTR incorrectly,
  ;;    - fills in array SQL_ATTR_ROW_STATUS_PTR incorrectly.
  ;;    All these worked fine with Jet 3.5.
  (declare (optimize (speed 3) (space 0)  #-debug (debug 0)))
                     ;(safety 1) #+(and lispworks (not debug)) (hcl:fixnum-safety 0)
  (with-accessors ((last-operation last-operation)
                   (retrieve-data retrieve-data)
                   (block-position block-position)
                   (fetched-count fetched-count)
                   (current-row-nr current-row-nr)
                   (current-row-status current-row-status)) row-descriptor
    (let ((row-nr (cond ((eq last-operation :fetch)	; Current row has been
                         (1+ current-row-nr))		; already fetched
                        (retrieve-data
                         current-row-nr)		; Not fetched but only positioned
                        ((null last-operation)
                         current-row-nr)		; The very first cursor operation
                           #|(:probe 				
                              (when (= (%sql-fetch-scroll hstmt row-descriptor :relative
                                                          :offset (1- current-row-nr))
                                       SQL_NO_DATA)
                                (return-from %fetch-row nil))
                              (%incf-block-position (1- current-row-nr))
                              (setf fetched-count (desc-rows-processed row-descriptor)
                                    last-operation :refresh)
                              1)|#
                        ;; Last set-cursor-position in some orientation does not fetch.
                        ((/= (%sql-fetch-scroll hstmt row-descriptor :relative
                                                :offset (1- current-row-nr))
                             SQL_NO_DATA)
                         (%incf-block-position (1- current-row-nr))
                         (setf fetched-count (desc-rows-processed row-descriptor))
                                ;last-operation :refresh)
                         1)
                        ((return-from %fetch-row nil)) )))
      (declare (fixnum row-nr))
      ;; Loop skipping rows with illegal statuses
      (loop (case (cond ((<= 1 row-nr fetched-count)
                         (setf current-row-status
                               (%decode-row-status (desc-row-status row-descriptor
                                                                    (1- row-nr)))))
                        ((/= (progn
                               (when (and (>= $ODBCVER 0300) (not retrieve-data))
                                 (%sql-set-attribute :hstmt hstmt SQL_ATTR_RETRIEVE_DATA
                                                     SQL_RD_ON)
                                 (setf retrieve-data t))
                               (%sql-fetch hstmt))
                             SQL_NO_DATA)
                         ;(let ((size (desc-array-size row-descriptor)));Access workaround
                         ;  (when (> size 1)
                         ;    (setf (desc-rows-processed row-descriptor)
                         ;          (min (desc-rows-processed row-descriptor) size))))
                         (%incf-block-position fetched-count)
                         (setf fetched-count (desc-rows-processed row-descriptor)
                               row-nr 1
                               current-row-status
                               (%decode-row-status (desc-row-status row-descriptor 0))))
                        (t                 		;  rowset  !
                         (setf block-position :end
                               ;last-operation :fetch
                               fetched-count 0 
                               current-row-nr 0
                               current-row-status :norow)
                         (return-from %fetch-row nil)))
              ((t :warn :updated) ;(#.SQL_ROW_SUCCESS[_WITH_INFO] #.SQL_ROW_UPDATED)
               (setf last-operation :fetch)
               (return (setf current-row-nr row-nr)))
              (:added ;#.SQL_ROW_ADDED
               ;; Oracle NB:  :add   - length-indicator-ptr,
               ;;	  SQL_COLUMN_IGNORE
               ;;	 SQL_ROW_UPDATES="Y", SQL_ROW_UPDATED  , 
               ;;	   by any user since row was last fetched
               (%sql-refresh-record hstmt row-nr))
              (otherwise ;(#.SQL_ROW_ERROR #.SQL_ROW_DELETED #.SQL_ROW_NOROW)
               (incf row-nr)))) )))

(defmethod db-fetch-rows ((cursor odbc-cursor) &key count flatp ignore-columns)
                                                   ;ignore-deleted
 ;;; Returts a list of count tuples or items if flatp is true.
  ;; Bookmark value is always returned as the first element of the tuple 
  ;; unless flatp is true or :bookmark occurs in the ignore-columns list.
  ;;  forward-only ,  array-size > 1,    long-.
  (when (active cursor)
    (let* ((handle (handle cursor))
           (row-descriptor (row-descriptor cursor))
           (columns-count (desc-count row-descriptor))
           (bookmark-record (unless (member :bookmark ignore-columns)
                              (bookmark-record row-descriptor)))
           (records (desc-records row-descriptor))
           (array-size (desc-array-size row-descriptor))
           (rows-processed 0)			; number of rows collected in this call
           row-nr)
      (declare (fixnum array-size))
      (when (member :last ignore-columns) (decf columns-count))
      (cond ((<= columns-count 0)
             nil)
            (flatp
             (when (> columns-count 1)
               (error 'sql-user-error :message
                      "If more than one column is to be fetched, flatp must be null."))
             (let ((record (aref records 0)))	    ; single column descriptor record
               (loop while (and (or (null count) (< rows-processed count))
                                  (setq row-nr (%fetch-row handle row-descriptor)))
                     collect
                           ;(case (desc-row-status row-descriptor (1- row-nr))
                           ;  ((#.SQL_ROW_SUCCESS #.SQL_ROW_SUCCESS_WITH_INFO
                           ;    #.SQL_ROW_UPDATED #.SQL_ROW_ADDED)
                      (cond ((desc-long-p record)
                             (when (> array-size 1)		; change pos in rowset
                               (%sql-position-to handle row-nr))
                             (%read-datum-in-chunks handle record))
                            ((%read-datum record (1- row-nr))))
                           ;  (#.SQL_ROW_ERROR "#<error>")
                           ;  (#.SQL_ROW_DELETED "#<deleted>") 
                           ;  (#.SQL_ROW_NOROW))
                     do (incf rows-processed))))
            (t
             (loop while (and (or (null count) (< rows-processed count))
                              (setq row-nr (%fetch-row handle row-descriptor)))
                   for tuple = (loop for record across records
                                     repeat columns-count
                                     collect
                                     (cond ((desc-long-p record)
                                            (when (> array-size 1) ; change pos in rowset
                                              (%sql-position-to handle row-nr))
                                            (%read-datum-in-chunks handle record))
                                           (t (%read-datum record (1- row-nr)))))
                   collect (if bookmark-record
                               (cons (%read-datum bookmark-record (1- row-nr)) tuple)
                               tuple)
                   do (incf rows-processed)))))))

(defmethod db-fetch-item ((cursor odbc-cursor) column &key current eof
                                                           (if-does-not-exist :error))
 ;;; Return one datum for the column, or raise error.
  ;; Args: if-does-not-exist  NIL = suppress "Column not found" error and just return nil
  ;; Value: = eof		If the cursor is closed or rowset is empty
  ;;	    = if-does-not-exist If the column required is missing
  (declare (optimize (speed 3) (space 0) #-debug (debug 0)))
  (if (active cursor)
      (let ((handle (handle cursor))
            (row-descriptor (row-descriptor cursor)))
        (if-bind (row-nr (if current
                             (current-row-nr row-descriptor)
                             (%fetch-row handle row-descriptor)))
          (if-bind (record (find-descriptor-record row-descriptor column
                                                   (eq if-does-not-exist :error)))
            (if (desc-long-p record)
                (progn
                  (when (> (desc-array-size row-descriptor) 1) ; change pos in rowset
                    (%sql-position-to handle row-nr))
                  (%read-datum-in-chunks handle record))
                (%read-datum record (1- row-nr)))
            if-does-not-exist)
          eof))
      eof))

(defmethod db-fetch-into ((cursor odbc-cursor) (list list)
                          &key columns (start-column 1) end-column current
                               (start 0) end)
 ;;; Store one row into the LIST.
  ;; Args: start,end  Determine subsequence positions in object.
  ;;       start-column  >0 - bookmarks can only be fetched via the columns argument!
  ;; Value: object if succeeded, nil if end of result set, or some error indicator.
  (when (active cursor)
    (let ((handle (handle cursor))
          (row-descriptor (row-descriptor cursor)))
      (when-bind (row-nr (if current
                             (current-row-nr row-descriptor)
                             (%fetch-row handle row-descriptor)))
        (let ((array-size (desc-array-size row-descriptor))
              (n (if end (- end start) most-positive-fixnum)))
          (declare (fixnum array-size n))
          (if columns
              ;; Explicit list of column specifiers
              (loop for column in columns
                    and rest = (nthcdr start list) then (rest rest)
                    while rest
                    repeat n
                    for record = (find-descriptor-record row-descriptor column)
                    if (desc-long-p record)
                      when (> array-size 1)		; change current pos in rowset
                        do (%sql-position-to handle row-nr)
                      end
                    and do (setf (car rest) (%read-datum-in-chunks handle record))
                    else do (setf (car rest) (%read-datum record (1- row-nr))))
              ;; Observe [start-column .. end-column] diapason
              (loop with records = (desc-records row-descriptor)
                    with end-nr fixnum = (or end-column (desc-count row-descriptor))
                    for rest = (nthcdr start list) then (rest rest)
                    while rest
                    repeat n
                    for column-nr fixnum upfrom start-column upto end-nr
                    for record = (aref records (1- column-nr))
                    if (desc-long-p record)
                      when (> array-size 1)		; change current pos in rowset
                        do (%sql-position-to handle row-nr)
                      end
                    and do (setf (car rest) (%read-datum-in-chunks handle record))
                    else do (setf (car rest) (%read-datum record (1- row-nr)))))
          list)))))

(defmethod db-fetch-into ((cursor odbc-cursor) (vector vector)
                          &key columns (start-column 1) end-column current
                               (start 0) end)
 ;;; Store one row into the LIST.
  ;; Args: start,end  Determine subsequence positions in object.
  ;;       start-column  >0 - bookmarks can only be fetched via the columns argument!
  ;; Value: object if succeeded, nil if end of result set, or some error indicator.
  (declare (optimize (speed 3) (space 0) #-debug (debug 0)
                     #+(and lispworks (not debug)) (hcl:fixnum-safety 0))
           (fixnum start-column start))
  (when (active cursor)
    (let ((handle (handle cursor))
          (row-descriptor (row-descriptor cursor)))
      (when-bind (row-nr (if current
                             (current-row-nr row-descriptor)
                             (%fetch-row handle row-descriptor)))
        (let ((array-size (desc-array-size row-descriptor))
              (end (or end (length vector))))
          (declare (fixnum array-size end))
          (if columns
              ;; Explicit list of column specifiers
              (loop for column in columns
                    and index fixnum upfrom start below end
                    for record = (find-descriptor-record row-descriptor column)
                    if (desc-long-p record)
                      when (> array-size 1)		; change current pos in rowset
                        do (%sql-position-to handle row-nr)
                      end
                    and do (setf (aref vector index) (%read-datum-in-chunks handle
                                                                            record))
                    else do (setf (aref vector index) (%read-datum record (1- row-nr))))
              ;; Observe [start-column .. end-column] diapason
              (loop with records = (desc-records row-descriptor)
                    with end-nr fixnum = (or end-column (desc-count row-descriptor))
                    for column-nr fixnum upfrom start-column upto end-nr
                    and index upfrom start below end
                    for record = (aref records (1- column-nr))
                    if (desc-long-p record)
                      when (> array-size 1)		; change current pos in rowset
                        do (%sql-position-to handle row-nr)
                      end
                    and do (setf (aref vector index) (%read-datum-in-chunks handle
                                                                            record))
                    else do (setf (aref vector index) (%read-datum record
                                                                   (1- row-nr)))))
          vector)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Updating data via cursors
;;; We assign SQL_COLUMN_IGNORE to the indicator value in order for
;;; SQLBulkOperations or SQLSetPos to ignore the value in the data buffer:
;;; - When updating a row of data, the column value is not changed.
;;; - When inserting a new row of data by a call to SQLBulkOperations, the column
;;;   value is set to its default or, if the column does not have a default, to NULL. 

(defun %write-row (cursor operation index columns values cv-alist)
 ;;; Args: 
  ;;  operation ::= :add | :update | function (lambda (hstmt) ...)
  ;;  index : Row index (starting from 0)
  ;;  columns : List of column names or numbers (starting from 1); 
  ;;		if null assumed all non-bookmark columns
  ;; NB: There is no possiblity to write to a bookmark column by means of %write-row
  ;; Oracle 8:
  ;;	We cannot update long columns by this method as Oracle has not SQL_GD_BOUND
  ;;	(prohibits binding such columns), but only SQL_GD_BLOCK.
  ;;	We can use parameterized positioned UPDATE instead.
  ;; Oracle 10:
  ;;	Driver Application Options > SQLGetData Extensions - turn on support of the
  ;;	SQL_GD_ANY_COLUMN, SQL_GD_ANY_ORDER and SQL_GD_BOUND along with SQL_GD_BLOCK.
  ;;	(It also forces Retrieval of Longs as a drawback.)
  (cond ((and cv-alist (null values) (null columns)))
        ((and (null cv-alist) values)        
         (when (atom values) (setq values (list values)))
         (when (and columns
                    (/= (if (atom columns)
                            (progn (setq columns (list columns)) 1)
                            (length columns))
                        (length values)))
           (error 'sql-user-error :message
  "The number of columns differs from the number of values for INSERT or UPDATE.")))
        (t
         (error 'sql-user-error :message
                "Bad or ambiguous keyword combination for INSERT or UPDATE.")))

  (when (active cursor)
    (let ((handle (handle cursor))
          (row-descriptor (row-descriptor cursor))
          (long-data ())	; long data items
          (long-records ())
          (ignored ()))	; ignored records and old value of desc-indicator-ptr
      (flet ((test (record column)
               (cond ((descriptor-record-p column)
                      (eq record column))
                     ((numberp column)
                      (eql (desc-record-number record) column))
                     (t (string-equal (desc-name record) (ordinary-argument column))))))
			;((stringp column)
			; (string-equal (desc-name record) column))
			;((typep column 'sql-ident)
			; (string-equal (desc-name record) (name column)))
			;((symbolp column)
			; (string-equal (desc-name record) (symbol-name column)))
        ;; Filter out long data items and write short ones into buffers.
	;; Variable rest caters for data when neither columns nor av-pairs are given
        (loop with records = (desc-records row-descriptor)
              and rest = values
              for record across records
              for position = (if cv-alist
                                 (assoc record cv-alist :test #'test)
                                 (position record columns :test #'test))
              for datum = (cond ((consp position)		; cv-alist pair
                                 (cdr position))
                                (position			; number
                                 (nth position values))
                                ((and rest (null columns))	; sequential elt
                                 (setq position t)
                                 (pop rest)))
              ;do (format t "~&record ~S datum ~S value position = ~S" 
              ;           record datum position)
              if position			; column descriptor record was found
                if (desc-long-p record)
                do (push record long-records)
                   (push datum long-data)
                else do (%write-datum datum record index)
              else do (setf ignored
                             (acons record (%get-integer (desc-indicator-ptr record)
                                                         index)
                                    ignored)
                            (%get-integer (desc-indicator-ptr record) index)
                             SQL_COLUMN_IGNORE)) )
      ;(format t "&ignored-records ~S" ignored)
      ;; Because long data columns typically are not bound, the application must
      ;;  - bind the column before calling SQLBulkOperations or SQLSetPos
      ;;  - and unbind it after calling SQLBulkOperations or SQLSetPos.
      ;;   32-   ,  parameter-data-ulong
      ;;  TargetValuePtr     !
      (when long-records
        (if (logtest (db-get-info (database cursor) SQL_GETDATA_EXTENSIONS)
                     SQL_GD_BOUND)
            (loop with sql-need-long-data-len = (db-get-info (database cursor)
                                                             SQL_NEED_LONG_DATA_LEN)
                  for record in long-records
                  and datum in long-data
                  do (setf (%get-integer (desc-indicator-ptr record) index)
                           (if sql-need-long-data-len
                               (%sql-len-data-at-exec (length datum))
                               SQL_DATA_AT_EXEC))
                  (%sql-bind-column handle
                                    (desc-record-number record)
                                    (desc-c-type record)
                                    (desc-data-ptr record)
                                    (desc-octet-length record)
                                    (desc-octet-length-ptr record)))
            (progn
              (cerror "Ignore new values of these columns."
                      'sql-simple-error :format-control
                      "ODBC driver cannot update via a cursor long data columns:~@{ ~S~}"
                      :format-arguments (mapcar #'desc-name long-records))
              (setq long-records nil))))

      (unwind-protect
          (when (= (if (keywordp operation)
                       (%sql-set-pos handle (1+ index) operation) ;translate to row-nr
                       (funcall operation handle))
                   SQL_NEED_DATA)
            (setf (active cursor) :need-data)
            (%write-data-at-exec long-data long-records handle row-descriptor)
            (setf (active cursor) :open))	; old-active?
        ;; OT1H: cancel-statement must be used cautiously because it leads to
        ;; unrecoverable [Driver Manager HY010] Function sequence error,
        ;; and nothing works except closing the cursor.
        ;; OTOH: not canceling can also lead to the same error, and even closing
        ;; will not work due to imposibility to cancel after we drop state :need-data.
        (when (eq (active cursor) :need-data)
          (cancel-statement cursor)
          (setf (active cursor) :open))
        (when (and (not (keywordp operation)) (> (desc-array-size row-descriptor) 1))
          (%sql-set-attribute :hstmt handle SQL_ATTR_ROW_ARRAY_SIZE
                              (desc-array-size row-descriptor)))
        ;; Restore the indicator-ptr values. Though %write- functions set it properly,
        ;; %sql-refresh-record invoked on a row with :added or :updated status
        ;; can refuse to refresh, e.g. Oracle driver.
        (loop for (record . value) in ignored
              do (setf (%get-integer (desc-indicator-ptr record) index) value))
        ;; Unbind long data columns
        (dolist (record long-records)			
          (%sql-unbind-column handle (desc-record-number record)))
) ) ) )

(defmethod db-delete-row ((cursor odbc-cursor) &key where)
  ;; Args: where Bookmark
 ;;; After the row has been deleted, the following are not valid for the row:
  ;; - positioned update and delete statements,
  ;; - calls to SQLGetData, 
  ;; - and calls to SQLSetPos with Operation set to anything except SQL_POSITION.
  ;; For drivers that support packing (like Access), the row is deleted
  ;; from the cursor when new data is retrieved (refetched) from the data source. 
  ;;   .   SQL_ROW_DELETED,   
  ;;       
  ;;	[ODBC state HY109] "Invalid cursor position; no keyset defined."
  ;;   ( status ptrs ,    )
  ;;	[ODBC state 01001] "Cursor operation conflict."
  ;;     SQL_ROW_ERROR.
  ;;    rowset   ,   
  (when (active cursor)
    (let ((row-descriptor (row-descriptor cursor))
          (handle (handle cursor)))
      (if where
          (cond ((bookmark-record row-descriptor)
                 (cond ((%sql-func-exists (database cursor) SQL_API_SQLBULKOPERATIONS)
                        (when (> (desc-array-size row-descriptor) 1)
                          (%sql-set-attribute :hstmt handle SQL_ATTR_ROW_ARRAY_SIZE 1))
                        ;; This is only needed what bind status ptrs!
                        (case (current-row-status cursor)
                          ((:deleted :error) ; :norow
                           (set-cursor-position cursor :first :retrieve-data nil)))
                           ;(set-cursor-position cursor :relative :offset 0
                           ;                     :retrieve-data nil)))
                           ;(%sql-fetch-scroll (handle cursor) row-descriptor :relative
                           ;                   :offset 0 :retrieve-data nil)))
                           ;(setf (current-row-status descriptor) nil)))
                        ;; write bookmark column of the first rowset row
                        (%write-datum where (bookmark-record row-descriptor) 0)
                        (%sql-bulk-operations handle :delete)
                        (when (> (desc-array-size row-descriptor) 1)
                          (%sql-set-attribute :hstmt handle SQL_ATTR_ROW_ARRAY_SIZE
                                              (desc-array-size row-descriptor)))
                        ;; undefine block cursor position
                        (setf (last-operation row-descriptor) :delete
                              (block-position row-descriptor) nil
                              (current-row-nr row-descriptor) 1
                              (current-row-status row-descriptor) nil)
                        t)
                       ((set-cursor-position cursor :bookmark
                                             :offset where
                                             :retrieve-data nil)
                        (%sql-delete-record handle 1)
                        (setf (current-row-status row-descriptor) :deleted))))
                ((set-cursor-position cursor :absolute
                                      :offset where
                                      :retrieve-data nil)
                 (%sql-delete-record handle 1)
                 (setf (current-row-status row-descriptor) :deleted)))
          (with-slots (fetched-count current-row-nr current-row-status) row-descriptor
            (cond ((<= 1 current-row-nr fetched-count)
                   (%sql-delete-record handle current-row-nr)
                   (setf current-row-status :deleted))
                  (t (error 'sql-database-error :database (database cursor)
                            :expression 'db-delete-row
                            :message "No current row to DELETE.")) )) ))))

(defmethod db-insert-row ((cursor odbc-cursor) &key columns values cv-alist)
  ;; Value: The record bookmark if the bookmark column is bound
  ;;	    or t if not bound or impossible to determine.
  ;;        -    SQL_ROW_ADDED.
  ;;      %sql-refresh-record.
  ;; NB:   %sql-set-pos   bookmark  
  (let ((row-descriptor (row-descriptor cursor)))
    (cond ((%sql-func-exists (database cursor) SQL_API_SQLBULKOPERATIONS)
           (when (> (desc-array-size row-descriptor) 1)
             ;; the attribute restored back in unwind-protected code of %write-row
             (%sql-set-attribute :hstmt (handle cursor) SQL_ATTR_ROW_ARRAY_SIZE 1))
           (%write-row cursor
                       (lambda (hstmt) (%sql-bulk-operations hstmt :add))
                       0 columns values cv-alist)
           (setf (last-operation row-descriptor) :add
                 (block-position row-descriptor) nil	; undefine block cursor position
                 (current-row-nr row-descriptor) 1
                 (current-row-status row-descriptor) nil)
           (if-bind (record (bookmark-record row-descriptor))
             (%read-datum record 0)
             t))
          (t 
           (%write-row cursor :add 0 columns values cv-alist)
           (when (= (current-row-nr row-descriptor) 1)
             (setf (current-row-status row-descriptor) :added))
           t))))

(defmethod db-update-row ((cursor odbc-cursor) &key columns values cv-alist where)
 ;;; Update current row or, if where is specified, the row pointed by it.
  ;; Args: where  Integer bookmark or absolute position in the entire result set
  (let ((row-descriptor (row-descriptor cursor)))
    (cond (where
           (cond ((bookmark-record row-descriptor)
                  (cond ((%sql-func-exists (database cursor) SQL_API_SQLBULKOPERATIONS)
                         (when (> (desc-array-size row-descriptor) 1)
                           ;; the attribute restored back in unwind-protected %write-row
                           (%sql-set-attribute :hstmt (handle cursor)
                                               SQL_ATTR_ROW_ARRAY_SIZE 1))
                         (case (current-row-status cursor)
                           ((:deleted :error) ; :norow)
                            (set-cursor-position cursor :first :retrieve-data nil)))
                         ;; write bookmark column of the first rowset row 
                         (%write-datum where (bookmark-record row-descriptor) 0)
                         (%write-row cursor
                                     (lambda (hstmt)(%sql-bulk-operations hstmt :update))
                                     0 columns values cv-alist)
                         ;; undefine block cursor position
                         (setf (last-operation row-descriptor) :update
                               (block-position row-descriptor) nil
                               (current-row-nr row-descriptor) 1
                               (current-row-status row-descriptor) nil)
                         t)
                        ((set-cursor-position cursor :bookmark
                                              :offset where
                                              :retrieve-data nil)
                         (%write-row cursor :update (1- (current-row-nr row-descriptor))
                                     columns values cv-alist)
                         (setf (current-row-status row-descriptor) :updated))))
                 ((set-cursor-position cursor :absolute
                                       :offset where
                                       :retrieve-data nil)
                  (%write-row cursor :update (1- (current-row-nr row-descriptor))
                              columns values cv-alist)
                  (setf (current-row-status row-descriptor) :updated))))
          ((<= 1 (current-row-nr row-descriptor) (fetched-count row-descriptor))
           ;; Rewrite last row fetched
           (%write-row cursor :update (1- (current-row-nr row-descriptor))
                       columns values cv-alist)
           (setf (current-row-status row-descriptor) :updated))
          (t (error 'sql-database-error :database (database cursor)
                    :expression 'db-update-row
                    :message "No current row to UPDATE.")) )))
