;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2005 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Type descriptors and conversions
;;; Rowset and parameter set descriptors
;;; Fetching result set

(in-package :ystok.odbc)

;;; Maximum character or binary buffer that we are allocating in Lisp
;;; in case this size cannot be determined from type descriptor.
;;; It must be as large as VARCHAR type max length.
;;; Otherwise the row descriptor record would be marked as long-p and not bound.
;;; Hence some catalog functions could fail (in Oracle) due to preceding
;;; of that allegedly long column to other ordinary bound columns.
;;; TODO: Set statement's SQL_ATTR_MAX_LENGTH attibute to the same value
;;;	  to reduce the size of long data retrieved.
(defparameter *default-max-length* 4000)

(defstruct (type-descriptor (:type vector)
                            (:conc-name type-desc-)
                            (:constructor nil)
                            (:copier nil))
  type-name		; Varchar not NULL.
  data-type		; Smallint not NULL.  SQL data type.
  column-size		; Integer (PRECISION in ODBC 2.0).
  literal-prefix	; Varchar. Character or characters used to prefix a literal
  literal-suffix	; Varchar. Character or characters used to terminate a literal
  create-params		; Varchar. A list of keywords, separated by commas.
  nullable		; Smallint not NULL. Whether the data type accepts a NULL value.
  case-sensitive	; Smallint not NULL. 
  searchable		; Smallint not NULL. How the data type is used in WHERE clause.
  unsigned-attribute	; Smallint. Whether the data type is unsigned.
  fixed-prec-scale	; Smallint not NULL. (MONEY in ODBC 2.0)
  auto-unique-value	; Smallint.  Whether autoincrementing.
  local-type-name	; Varchar. Localized version of the data sourcedependent name
  minimum-scale		; Smallint.  The minimum scale of the data type
  maximum-scale		; Smallint.  The maximum scale of the data type
  #+odbc3.0 sql-data-type	; Smallint NOT NULL.
  #+odbc3.0 sql-datetime-sub	; Smallint.
  #+odbc3.0 num-prec-radix	; Integer.
  #+odbc3.0 interval-precision)	; Smallint.

(defun make-type-descriptor-boa (&rest args)
  (loop with v = (make-array #+odbc3.0 19 #-odbc3.0 15 :initial-element nil)
        for arg in args
        and i upfrom 0 below (length v)
        do (setf (aref v i) arg)
        finally (return v)))

(defmacro %data-type-subscript (data-type)
  `(if (>= ,data-type 91)	;SQL_TYPE_DATE = 91
       (+ ,data-type 24 -91)	;SQL_VARCHAR - SQL_GUID = 23
       (+ ,data-type 11)))	;SQL_GUID = -11

(defun %initialize-type-info (database)
  (let ((cursor (obtain-statement database 'odbc-cursor))
        (type-info-vector (make-array 27 :initial-element nil)))
    (unwind-protect       
        (progn
          (setf (expression cursor) '%sql-get-type-info)
          (%sql-get-type-info (handle cursor) SQL_ALL_TYPES)
          (%initialize-rowset cursor)
          (loop for (type-name data-type . rest) in (db-fetch-rows cursor)
                for type-descriptor = (svref type-info-vector
                                             (%data-type-subscript data-type))
                unless type-descriptor
                do (setf type-descriptor
                          (apply #'make-type-descriptor-boa type-name data-type rest)
                         (svref type-info-vector (%data-type-subscript data-type))
                          type-descriptor)
                   ;(case (type-desc-data-type type-descriptor)
                   ;  ((#.SQL_CHAR #.SQL_BINARY)
                   ;   (setq *default-max-length* (min *default-max-length*
                   ;                         (type-desc-column-size type-descriptor)))))
           )
          (setf (type-info-vector database) type-info-vector))
     (db-close cursor)
   )))

(defun %get-type-descriptor (database type)
  (when-bind (type-info-vector (type-info-vector database))
    (cond ((numberp type)				; sql-type
           (svref type-info-vector (%data-type-subscript type))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 

(defclass descriptor (handled-object)
 ;;; Implementation or application descriptor for column-wise binding, i.e.
  ;; SQL_DESC_ALLOC_TYPE attribute value equals to SQL_BIND_BY_COLUMN.
 (;; SQL_DESC_COUNT [All]
  ;; This SQLSMALLINT header field specifies the 1-based index of the highest-numbered
  ;; record that contains data. When the driver sets the data structure for
  ;; the descriptor, it must also set the SQL_DESC_COUNT field to show how many
  ;; records are significant. When an application allocates an instance of this
  ;; data structure, it does not have to specify how many records to reserve room for.
  ;; As the application specifies the contents of the records, the driver takes
  ;; any required action to ensure that the descriptor handle refers to a data
  ;; structure of the adequate size
  ;; NB:   (fill-pointer (records desc))?
  (count :initarg :count :accessor desc-count :initform 0)

  ;; SQL_DESC_ARRAY_SIZE [Application descriptors]
  ;; ARD <=> SQL_ATTR_ROW_ARRAY_SIZE - the number of rows in the rowset. 
  ;; APD <=> SQL_ATTR_PARAMSET_SIZE - the number of values for each parameter. 
  (array-size :initarg :array-size :accessor desc-array-size :initform 1)

  ;; SQL_DESC_ARRAY_STATUS_PTR [All]
  ;; SQLUSMALLINT * header field points to an array of SQLUSMALLINT values.
  ;; IRD <=> SQL_ATTR_ROW_STATUS_PTR. 
  ;;  The array is initially populated by a call to SQLBulkOperations, SQLFetch,
  ;;  SQLFetchScroll, or SQLSetPos. If the call did not return SQL_SUCCESS or
  ;;  SQL_SUCCESS_WITH_INFO, the contents of the array pointed to by this field
  ;;  are undefined. Values:
  ;;  SQL_ROW_SUCCESS: The row was successfully fetched and has not changed since
  ;;	it was last fetched.
  ;;  SQL_ROW_SUCCESS_WITH_INFO: The row was successfully fetched and has not changed
  ;;	since it was last fetched. However, warning was returned about the row.
  ;;  SQL_ROW_ERROR: An error occurred while fetching the row.
  ;;  SQL_ROW_UPDATED: The row was successfully fetched and has been updated since
  ;;	it was last fetched. If the row is fetched again, its status is SQL_ROW_SUCCESS.
  ;;  SQL_ROW_DELETED: The row has been deleted since it was last fetched.
  ;;  SQL_ROW_ADDED: The row was inserted by SQLBulkOperations.
  ;;	If the row is fetched again, its status is SQL_ROW_SUCCESS.
  ;;  SQL_ROW_NOROW: The rowset overlapped the end of the result set, and no row was
  ;;	returned that corresponded to this element of the row status array. 
  ;;
  ;; IPD <=> SQL_ATTR_PARAM_STATUS_PTR
  ;;
  ;; ARD <=> SQL_ATTR_ROW_OPERATION_PTR 
  ;;  SQL_ROW_PROCEED: The row is included in the bulk operation using SQLSetPos.
  ;;	(This setting does not guarantee that the operation will occur on the row.
  ;;	If the row has the status SQL_ROW_ERROR in the IRD row status array,
  ;;	the driver might not be able to perform the operation in the row.)
  ;;  SQL_ROW_IGNORE: The row is excluded from the bulk operation using SQLSetPos. 
  ;;
  ;; APD <=> SQL_ATTR_PARAM_OPERATION_PTR
  (array-status-ptr :initarg :array-status-ptr :accessor desc-array-status-ptr
                    :initform nil)

  ;; SQL_DESC_ROWS_PROCESSED_PTR [Implementation descriptors]
  ;; In an IRD, this SQLUINTEGER * header field points to a buffer containing
  ;; the number of rows fetched after a call to SQLFetch or SQLFetchScroll,
  ;; or the number of rows affected in a bulk operation performed by a call to
  ;; SQLBulkOperations or SQLSetPos, including error rows.
  ;; In an IPD, this SQLUINTEGER * header field points to a buffer containing
  ;; the number of sets of parameters that have been processed, including error sets.
  ;; No number will be returned if this is a null pointer.
  ;; IRD?? ARD <=> SQL_ATTR_ROWS_FETCHED_PTR
  ;; IPD?? APD <=> SQL_ATTR_PARAMS_PROCESSED_PTR
  (rows-processed-ptr :initarg :rows-processed-ptr :accessor desc-rows-processed-ptr
                      :initform nil)

  (records :initarg :records :accessor desc-records
           :initform (make-array 0 :adjustable t :fill-pointer t))
 ))

(defmacro desc-rows-processed (descriptor)
  `(%get-uinteger (desc-rows-processed-ptr ,descriptor)))

(defmacro desc-row-status (descriptor index)
  `(%get-usmallint (desc-array-status-ptr ,descriptor) ,index))

(defclass row-descriptor (descriptor)
 ;;; Block cursor implementation
 ((last-operation :accessor last-operation :initform nil)
  ;; approximate absolute position, one of the following:
  ;;   nil 	 undefined, e.g. after SQLBulkOperations
  ;;   :start	 before start
  ;;   :end	 after end
  ;;   >=0	 number from start
  ;;   <0	 number frim end
  ;; Last value of the SQL_RETRIEVE_DATA cursor attibute. This would be better saved with
  ;; the cursor itself, but can stand here as we DON'T DISPOSE DESCRIPTORS any longer.
  (retrieve-data :accessor retrieve-data :initform t)
  (block-position :accessor block-position :initform :start)
  ;; number of rows processed by the last fetch
  (fetched-count :accessor fetched-count :initform 0)
  ;; row number within rowset that we have fetched or positioned on last.
  (current-row-nr :accessor current-row-nr :initform 0)
  ;; status of the last row fetched or operated on. Used as cache and a substitute 
  ;; for row status array when interacting with an ODBC 2.x driver.
  (current-row-status :accessor current-row-status :initform nil)
  ;; statement SQL_ATTR_FETCH_BOOKMARK_PTR attribute used during SQL_FETCH_BY_BOOKMARK
  (bookmark-record :accessor bookmark-record :initarg :bookmark :initform nil)
  (bookmark-ptr :accessor bookmark-ptr :initform nil)
  ;(bookmark-octet-length :accessor bookmark-octet-length)
 ))

;(defun bookmark-octet-length (row-descriptor)
;  (desc-octet-length (bookmark-record row-descriptor)))

(defstruct (descriptor-record (:conc-name desc-)
                              (:copier nil)
                              (:constructor nil))
 ;;; Descriptor records, representing columns or parameters.
  ;; NB: octet-length-ptr and indicator-ptr can point to the same array
  ;; LispWorks NB: most-positive-fixnum = 8388607

  ;; ODBC column number used for binding internally versus record position reflecting
  ;; the position of the selection in a cursor.
  ;; NB: Long- .       SQL_GD_ANY_ORDER!
  (record-number 0 :type fixnum)

  ;; SQL_DESC_NAME [Implementation descriptors]
  ;; Column alias or column name. 
  ;; An application can set the SQL_DESC_NAME field of an IPD to a parameter name
  ;; or alias to specify stored procedure parameters by name.
  ;; The SQL_DESC_NAME field of an IRD is a read-only field.
  (name nil :type string)		

  ;; SQL_DESC_BASE_COLUMN_NAME [IRDs]
  ;; This read-only SQLCHAR * record field contains the base column name for
  ;; the result set column. If a base column name does not exist (as in the case
  ;; of columns that are expressions), this variable contains an empty string.
  ;; SQL_DESC_BASE_TABLE_NAME [IRDs]
  ;; This read-only SQLCHAR * record field contains the base table name for
  ;; the result set column. If a base table name cannot be defined or is not
  ;; applicable, this variable contains an empty string.

  ;; SQL_DESC_CONCISE_TYPE [All]
  ;; This SQLSMALLINT field specifies the concise data type for all data types,
  ;; including the datetime and interval data types.
  ;; The values in the SQL_DESC_CONCISE_TYPE, SQL_DESC_TYPE, and 
  ;; SQL_DESC_DATETIME_INTERVAL_CODE fields are interdependent. Each time one of the
  ;; fields is set, the other must also be set. SQL_DESC_CONCISE_TYPE can be set by
  ;; a call to SQLBindCol or SQLBindParameter, or SQLSetDescField.
  ;; SQL_DESC_TYPE can be set by a call to SQLSetDescField or SQLSetDescRec. 
  ;;
  ;; If SQL_DESC_CONCISE_TYPE is set to a concise data type other than an interval
  ;; or datetime data type, the SQL_DESC_TYPE field is set to the same value and
  ;; the SQL_DESC_DATETIME_INTERVAL_CODE field is set to 0.
  ;;
  ;; If SQL_DESC_CONCISE_TYPE is set to the concise datetime or interval data type,
  ;; the SQL_DESC_TYPE field is set to the corresponding verbose type (SQL_DATETIME
  ;; or SQL_INTERVAL) and the SQL_DESC_DATETIME_INTERVAL_CODE field is set to
  ;; the appropriate subcode.

  ;; SQL_DESC_TYPE [All]
  ;; This SQLSMALLINT record field specifies the concise SQL or C data type for
  ;; all data types except datetime and interval data types. For the datetime and
  ;; interval data types, this field specifies the verbose data type, which
  ;; is SQL_DATETIME or SQL_INTERVAL
  (c-type 99 :type fixnum)		;SQL_C_DEFAULT
  (sql-type 99 :type fixnum)		;SQL_C_DEFAULT. SQL_UNKNOWN_TYPE?

  ;; SQL_DESC_DATA_PTR [Application descriptors and IPDs]
  ;; Points to a variable that will contain the parameter value (for APDs) 
  ;; or the column value (for ARDs). This field is a deferred field. It is not used
  ;; at the time it is set but is used at a later time by the driver to retrieve data.
  data-ptr			

  ;; SQL_DESC_LENGTH [All]
  ;; This SQLUINTEGER is either the maximum or actual length of a character string
  ;; in characters or a binary data type in bytes. 
  ;; It is the maximum length for a fixed-length datatype, or the actual length for
  ;; a variable-length data type. Its value always excludes the null-termination
  ;; character that ends the character string. 
  ;; For values whose type is SQL_TYPE_DATE, SQL_TYPE_TIME, SQL_TYPE_TIMESTAMP,
  ;; or one of the SQL interval data types, this field has the length in characters
  ;; of the character string representation of the datetime or interval value.
  (length 0 :type fixnum)
   
  ;; SQL_DESC_OCTET_LENGTH [All]
  ;; This SQLINTEGER record field contains the length, in bytes, of a character
  ;; string or binary data type. 
  ;; For fixed-length character or binary types, this is the actual length in bytes.
  ;; For variable-length character or binary types, this is the max length in bytes.
  ;; This value always excludes space for the null-termination character for
  ;; implementation descriptors and always includes space for the null-termination
  ;; character for application descriptors. For application data, this field
  ;; contains the size of the buffer. For APDs, this field is defined only
  ;; for output or input/output parameters.
  ;; NB: As for ARD and APD, we do count extra space for null-termination.
  (octet-length 0 :type fixnum)	

  ;; SQL_DESC_OCTET_LENGTH_PTR [Application descriptors]
  ;; SQLINTEGER * record field points to a variable that will contain the total
  ;; length in bytes of a dynamic argument (for parameter descriptors) or 
  ;; of a bound column value (for row descriptors). 
  ;;
  ;; For an APD, this value is ignored for all args except character string and binary;
  ;; if this field points to SQL_NTS, the dynamic argument must be null-terminated.
  ;; To indicate that a bound parameter will be a data-at-execution parameter, 
  ;; an application sets this field in the appropriate record of the APD to a variable
  ;; that, at execute time, will contain the value SQL_DATA_AT_EXEC or the result
  ;; of the SQL_LEN_DATA_AT_EXEC macro. If there is more than one such field,
  ;; SQL_DESC_DATA_PTR can be set to a value uniquely identifying the parameter
  ;; to help the application determine which parameter is being requested.
  ;;
  ;; If the OCTET_LENGTH_PTR field of an ARD is a null pointer, the driver does
  ;; not return length information for the column. If the SQL_DESC_OCTET_LENGTH_PTR
  ;; field of an APD is a null pointer, the driver assumes that character strings 
  ;; and binary values are null-terminated. (Binary values should not be 
  ;; null-terminated but should be given a length to avoid truncation.)
  ;; NB: We always allocate extra space for terminating #\Null in deferred input buffer
  ;;     when char data are expected or the type is unknown beforehand.
  octet-length-ptr

  ;; SQL_DESC_INDICATOR_PTR [Application descriptors]
  ;; In ARDs, this SQLINTEGER * record field points to the indicator variable. 
  ;; This variable contains SQL_NULL_DATA if the column value is a NULL.
  ;; For APDs, the indicator variable is set to SQL_NULL_DATA to specify NULL
  ;; dynamic arguments. Otherwise, the variable is zero (unless the values in
  ;; SQL_DESC_INDICATOR_PTR and SQL_DESC_OCTET_LENGTH_PTR are the same pointer).
  ;; 
  ;; If the SQL_DESC_INDICATOR_PTR field in an ARD is a null pointer, the driver
  ;; is prevented from returning information about whether the column is NULL or not.
  ;; If the column is NULL and SQL_DESC_INDICATOR_PTR is a null pointer, 
  ;; SQLSTATE "22002" (Indicator variable required but not supplied) is returned
  ;; when the driver attempts to populate the buffer after a call to SQLFetch
  ;; or SQLFetchScroll. If the call to SQLFetch or SQLFetchScroll did not return
  ;; SQL_SUCCESS or SQL_SUCCESS_WITH_INFO, the contents of the buffer are undefined.
  ;;
  ;; The SQL_DESC_INDICATOR_PTR field determines whether the field pointed to by
  ;; SQL_DESC_OCTET_LENGTH_PTR is set. If the data value for a column is NULL, 
  ;; the driver sets the indicator variable to SQL_NULL_DATA. The field pointed to
  ;; by SQL_DESC_OCTET_LENGTH_PTR is then not set. If a NULL value is not
  ;; encountered during the fetch, the buffer pointed to by SQL_DESC_INDICATOR_PTR
  ;; is set to zero and the buffer pointed to by SQL_DESC_OCTET_LENGTH_PTR is set
  ;; to the length of the data.
  ;indicator-ptr

  ;; SQL_DESC_PRECISION [All]
  ;; This SQLSMALLINT field contains
  ;; - the number of digits for an exact numeric type, 
  ;; - the number of bits in the mantissa (binary precision) for an approximate
  ;;   numeric type,
  ;; - or the numbers of digits in the fractional seconds component for
  ;;   the SQL_TYPE_TIME, SQL_TYPE_TIMESTAMP, or SQL_INTERVAL_SECOND data type.
  ;; This field is undefined for all other data types.
  (precision 0 :type fixnum)
   
  ;; SQL_DESC_SCALE [All]
  ;; SQLSMALLINT field contains the defined scale for decimal and numeric data types.
  (scale 0 :type fixnum)
   
  ;; SQL_DESC_NULLABLE [Implementation descriptors]
  ;; In IRDs, this read-only SQLSMALLINT record field is
  ;;  SQL_NULLABLE if the column can have NULL values, 

  ;;  SQL_NO_NULLS if the column does not have NULL values, or
  ;;  SQL_NULLABLE_UNKNOWN if it is not known whether the column accepts NULL values.
  ;; This field pertains to the result set column, not the base column.
  ;; In IPDs, this field is always set to SQL_NULLABLE because dynamic parameters
  ;; are always nullable and cannot be set by an application.
  (nullable SQL_NULLABLE_UNKNOWN :type fixnum)
  
  ;; SQL_DESC_DISPLAY_SIZE [IRDs]
  ;; This read-only SQLINTEGER record field contains the maximum number of characters
  ;; required to display the data from the column. 

  ;#define SQL_DESC_DATETIME_INTERVAL_CODE 1007
  ;#define SQL_DESC_UNNAMED                1012
  long-p	; if true, we don't bind the  column, but fetch via SQLGetData.
  lisp-type	; Lisp type or pseudo-type (keyword) specifier.
)

(defun desc-column-size (record)
  (case (desc-sql-type record)
    ((#.SQL_DOUBLE #.SQL_DECIMAL #.SQL_NUMERIC #.SQL_FLOAT #.SQL_REAL)
     (desc-precision record))
    (t (desc-length record))))

(defun desc-decimal-digits (record)
  (case (desc-sql-type record)
    ((#.SQL_DECIMAL #.SQL_NUMERIC)
     (desc-scale record))
    (t (desc-precision record))))

;; We don't use a separate inticator array
(defmacro desc-indicator-ptr (record) `(desc-octet-length-ptr ,record))

#|(defun desc-display-size (record)
    (case (desc-sql-type record)
      (otherwise(desc-length record))))|#

(defun display-size (sql-type &optional column-size decimal-digits type-descriptor)
  (case sql-type
    (#.SQL_INTEGER		; 11 if signed (sign and 10) or 10 if unsigned (10 digits)
     11)
     #+odbc3.0 (#.SQL_BIGINT	; 20 if signed (sign and 19) or 20 digits if unsigned
                20)
    ((#.SQL_FLOAT #.SQL_DOUBLE) ; sign, 15 digits, dec.point, letter E, sign, and 3 digits
     24)
    ((#+odbc3.0 #.SQL_TYPE_TIMESTAMP #.SQL_TIMESTAMP)
     ;; 19 (for a timestamp in the yyyy-mm-dd hh:mm:ss format) or
     ;; 20 + s (for a timestamp in the yyyy-mm-dd hh:mm:ss[.fff] format,
     ;; where s is the fractional seconds precision).
     (if (zerop (orf decimal-digits
                    (or (and type-descriptor (type-desc-maximum-scale type-descriptor))
                        0)))
         19
         (+ 20 decimal-digits)))
    (#.SQL_REAL			; sign, 7 digits, dec.point, letter E, sign, and 2 digits
     14)
    (#.SQL_SMALLINT		; 6 if signed (sign and 5) or 5 if unsigned (5 digits)
     6)
    (#.SQL_BIT			; 1 digit
     1)
    ((#.SQL_BINARY #.SQL_VARBINARY #.SQL_LONGVARBINARY) 
     ;; The defined or maximum (for variable types) length of the column times 2.
     ;; (Each binary byte is represented by a 2-digit hexadecimal number.)
     (if (orf column-size (and type-descriptor (type-desc-column-size type-descriptor)))
         (ash column-size 2)))
    (#.SQL_TINYINT
     ;; 4 if signed (sign and 3) or 3 if unsigned (3 digits).
     (if (and type-descriptor
              (eql (type-desc-unsigned-attribute type-descriptor) SQL_TRUE))
         3
         4))
    ((#+odbc3.0 #.SQL_TYPE_DATE	#.SQL_DATE)
     (or column-size
         10))			; 10 (a date in the format yyyy-mm-dd).    
    ((#+odbc3.0 #.SQL_TYPE_TIME #.SQL_TIME)
     ;; 8 (a time in the format hh:mm:ss) or 
     ;; 9 + s (a time in the format hh:mm:ss[.fff], where s is the fractional seconds)
     (or column-size
         (if (zerop (orf decimal-digits
                         (or (and type-descriptor(type-desc-maximum-scale type-descriptor))
                             0)))
             8
             (+ 9 decimal-digits))))
    ((#.SQL_NUMERIC #.SQL_DECIMAL)
     ;; The precision of the column plus 2 (a sign, precision digits, and a decimal point)
     (if (orf column-size (and type-descriptor (type-desc-column-size type-descriptor)))
         (+ column-size 2)))
    #+odbc3.0(#.SQL_GUID	; # of chars in aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee format
              36)              
    (otherwise
     (or column-size
         (and type-descriptor (type-desc-column-size type-descriptor))))
) )

(defstruct (column-desc-record (:include descriptor-record)
                               (:conc-name desc-)))

(defmacro bookmark-record-p (column-desc-record)
  `(zerop (desc-record-number ,column-desc-record)))

(defstruct (param-desc-record (:include descriptor-record)
                              (:conc-name desc-))
  ;(nullable SQL_NULLABLE :type fixnum)        ; redefine for different default 

  ;; SQL_DESC_PARAMETER_TYPE [IPDs]
  ;; This SQLSMALLINT record field is set to SQL_PARAM_INPUT for an input parameter, 
  ;; SQL_PARAM_INPUT_OUTPUT for an input/output parameter, or 
  ;; SQL_PARAM_OUTPUT for an output parameter. It is set to SQL_PARAM_INPUT by default.
  ;;
  ;; For an IPD, the field is set to SQL_PARAM_INPUT by default if the IPD 
  ;; is not automatically populated by the driver (the SQL_ATTR_ENABLE_AUTO_IPD
  ;; statement attribute is SQL_FALSE). An application should set this field
  ;; in the IPD for parameters that are not input parameters.
  (parameter-type SQL_PARAM_INPUT :type fixnum)
  ;parameter				; sql-parameter instance if any
)

(defun ordinary-argument (arg &optional default)
 ;;; Make a string comparable with descriptor record name
  ;; Args: default  "" or "%" could be of use.
  ;; NB: Maybe we should convert #\- to #\_ only for sql-ident instances?
  (cond ((stringp arg) arg)
        ((null arg) default)
        ((symbolp arg) (substitute #\_ #\- (symbol-name arg)))
        (t (ordinary-argument (name arg))) )) 	;(typep identifier 'sql-ident)

(defun find-descriptor-record (descriptor designator &optional (errorp t))
  (declare (optimize (speed 3) (space 0) #-debug (safety 0)))
  (let ((record (cond ((numberp designator)
                       (aref (desc-records descriptor) (1- designator)))
                      ((eq designator :bookmark)
                       (bookmark-record descriptor))
                      (t
                       (find (ordinary-argument designator) (desc-records descriptor)
                             :test #'string-equal :key #'desc-name)))))
                      ;((stringp item)
                      ; (find item (desc-records descriptor)
                      ;       :test #'string-equal :key #'desc-name))
                      ;((symbolp item)
                      ; (find (symbol-name item) (desc-records descriptor)
                      ;       :test #'string-equal :key #'desc-name))
                      ;(t
                      ; (find (string (name item)) (desc-records descriptor)
                      ;       :test #'string-equal :key #'desc-name)))))
    (cond (record)
          (errorp (error 'sql-user-error :message
                         (format nil "Wrong ODBC descriptor record specification ~s"
                                 designator))))))

;;; Type conversion control principle. Provide methods only on arrows pictured:
;;;	SQL type <-> C type <-> Lisp type
;;; No explicit method is provided for direct conversion:
;;;	SQL type <-/-> Lisp type

(defmethod db-sql-to-c-type (database sql-type column-size decimal-digits)
 ;;; Return nil if sql-type is an unknown type or unsure.
  ;; NB: For simplicity implement only SQL_C_SBIGINT and omit SQL_C_UBIGINT.
  (case sql-type
           ((#.SQL_INTEGER
 #-odbc3.0   #.SQL_BIGINT)	SQL_C_SLONG)		; or SQL_C_ULONG
           ((#.SQL_CHAR #.SQL_VARCHAR #.SQL_LONGVARCHAR)
     				SQL_C_CHAR)
 #+Unicode ((#.SQL_WCHAR #.SQL_WVARCHAR #.SQL_WLONGVARCHAR) 
				SQL_C_WCHAR)
           ((#.SQL_FLOAT #.SQL_DOUBLE) SQL_C_DOUBLE)
 #+odbc3.0 (#.SQL_TYPE_TIMESTAMP SQL_C_TYPE_TIMESTAMP)
	   (#.SQL_TIMESTAMP	SQL_C_TIMESTAMP)
;#+lispworks(#.SQL_REAL	SQL_C_DOUBLE)		; All floats are double in LWW and LWL
	   (#.SQL_REAL		SQL_C_FLOAT)
	   (#.SQL_SMALLINT	SQL_C_SSHORT)		; or SQL_C_USHORT
	   (#.SQL_BIT		SQL_C_BIT)
	   ((#.SQL_BINARY #.SQL_VARBINARY #.SQL_LONGVARBINARY) 
				SQL_C_BINARY)
	   (#.SQL_TINYINT   (if (eql (type-desc-unsigned-attribute
                                      (%get-type-descriptor database sql-type)) SQL_TRUE)
                       		SQL_C_UTINYINT
                       		SQL_C_STINYINT))
 #+odbc3.0 (#.SQL_TYPE_DATE	SQL_C_TYPE_DATE)
 #+odbc3.0 (#.SQL_TYPE_TIME	SQL_C_TYPE_TIME)
	   (#.SQL_DATE		SQL_C_DATE)
	   (#.SQL_TIME		SQL_C_TIME)
           ;; Converting to double seems to be performed via text, so takes current
           ;; DBMS client or computer locale settings into account.
           ;; LW's fli:set-locale affects Access dirver (or entire ODBC Driver Manager?).
	   ((#.SQL_NUMERIC #.SQL_DECIMAL)		; Corresponds Access Currency
            (cond ((/= decimal-digits 0)		; Unsupported by Orcale 8
 #+odbc3.0			SQL_C_NUMERIC		; scale /= 0
 #-odbc3.0	   		SQL_C_DOUBLE)
                  ((< column-size 5) SQL_C_SSHORT)	; scale = 0, precision < 5
                  (t 		SQL_C_SLONG)))		; scale = 0, precision >= 5
 #+odbc3.0 (#.SQL_BIGINT    ;(if (eql (type-desc-unsigned-attribute
                            ;          (%get-type-descriptor database sql-type)) SQL_TRUE)
                            ;   SQL_C_UBIGINT
                       		SQL_C_SBIGINT)
 #+odbc3.0 (#.SQL_GUID		SQL_C_CHAR)		; default conversion ODBC 3.5
    				; or SQL_C_GUID
  ))

(defmethod db-c-to-sql-type (database c-type &optional column-size decimal-digits)
 ;;; Return nil if sql-type is of an unknown type or unsure.
  ;; Used by: parameter descriptors as a rule.
 "Values: sql-type, column-size, decimal-digits."
  (let* (type-descriptor
         (sql-type
          (case c-type
            ((#.SQL_C_SLONG #.SQL_C_ULONG)
             SQL_INTEGER)
            (#.SQL_C_CHAR
 	     ;; Jet 3.5 does not allow SQL_VARCHAR as a parameter type!
             (cond ((setq type-descriptor (%get-type-descriptor database SQL_CHAR))
                    (if column-size
                        (if (<= column-size (type-desc-column-size type-descriptor))
                            SQL_CHAR
                            SQL_LONGVARCHAR)
                        (progn (setq column-size (type-desc-column-size type-descriptor))
                               SQL_CHAR)))
                   ((setq type-descriptor (%get-type-descriptor database SQL_VARCHAR))
                    (if column-size
                        (if (<= column-size (type-desc-column-size type-descriptor))
                            SQL_VARCHAR
                            SQL_LONGVARCHAR)
                        (progn (setq column-size (type-desc-column-size type-descriptor))
                               SQL_VARCHAR)))
                   (t (unless column-size (setq column-size *default-max-length*))
                      SQL_CHAR)))
            #+Unicode
            (SQL_C_WCHAR
             (cond ((setq type-descriptor (%get-type-descriptor database SQL_WCHAR))
                    (if column-size
                        (if (<= column-size (type-desc-column-size type-descriptor))
                            SQL_WCHAR
                            SQL_WLONGVARCHAR)
                        (progn (setq column-size (type-desc-column-size type-descriptor))
                          SQL_WCHAR)))
                   ((setq type-descriptor (%get-type-descriptor database SQL_WVARCHAR))
                    (if column-size
                        (if (<= column-size (type-desc-column-size type-descriptor))
                            SQL_WVARCHAR
                            SQL_WLONGVARCHAR)
                        (progn (setq column-size (type-desc-column-size type-descriptor))
                               SQL_WVARCHAR)))
                   (t (unless column-size (setq column-size *default-max-length*))
                      SQL_WCHAR)))
            (#.SQL_C_DOUBLE SQL_DOUBLE)    
  #+odbc3.0 (#.SQL_C_TYPE_TIMESTAMP SQL_TYPE_TIMESTAMP)
#+lispworks (#.SQL_C_FLOAT SQL_REAL)
            ((#.SQL_C_SSHORT #.SQL_C_USHORT) SQL_SMALLINT)
            (#.SQL_C_BIT SQL_BIT)	; Used to represent boolean values in Access
            ((#.SQL_C_UTINYINT #.SQL_C_STINYINT) SQL_TINYINT)
            (#.SQL_C_BINARY
             (cond ((setq type-descriptor (%get-type-descriptor database SQL_BINARY))
                    (if column-size
                        (if (<= column-size (type-desc-column-size type-descriptor))
                            SQL_BINARY
                            SQL_LONGVARBINARY)
                        (progn (setq column-size (type-desc-column-size type-descriptor))
                               SQL_BINARY)))
                   ((setq type-descriptor (%get-type-descriptor database SQL_VARBINARY))
                    (if column-size
                        (if (<= column-size (type-desc-column-size type-descriptor))
                            SQL_VARBINARY
                            SQL_LONGVARBINARY)
                        (progn (setq column-size (type-desc-column-size type-descriptor))
                               SQL_VARBINARY)))
                   (t (unless column-size (setq column-size *default-max-length*))
                      SQL_BINARY)))
  #+odbc3.0 (#.SQL_C_NUMERIC
             (or (find-if (lambda (data-type) (%get-type-descriptor database data-type))
                          '(#.SQL_NUMERIC #.SQL_DECIMAL))
                 SQL_VARCHAR))
  #+odbc3.0 (#.SQL_C_TYPE_DATE SQL_TYPE_DATE)
  #+odbc3.0 (#.SQL_C_TYPE_TIME SQL_TYPE_TIME)
            (#.SQL_C_TIMESTAMP SQL_TIMESTAMP)
            (#.SQL_C_DATE SQL_DATE)
            (#.SQL_C_TIME SQL_TIME)
  #+odbc3.0 ((#.SQL_C_SBIGINT #.SQL_C_UBIGINT) SQL_BIGINT)
        )))
    (values sql-type column-size decimal-digits)))

(defgeneric db-lisp-to-c-type (database lisp-type)
 ;;; Determine c-type when sql-type is unknown (usually due to missing SQLDescribeParam)
  ;; Values: 1> c-type, 2> column-size or nil if are determined from type descriptor later.
 (:method (database lisp-type)
  (declare (ignore database lisp-type))
  (values nil nil)))

(defun %describe-lisp-object (database datum 	;&rest args)
                              &optional lisp-type sql-type column-size decimal-digits
                              &aux odbc-column-size)
 "Find appropriate SQL- and C- types for representing Lisp object. Arguments:
 DATUM  ignored if null.
 LISP-TYPE if null, (type-of datum) is used instead;
   if a keyword, names an SQL type and ARGS are SQL arguments;
   otherwise symbol or list and ARGS are Lisp type arguments.
 Values: c-type, sql-type, column-size, decimal-digits, nullable."
  ;; As an input values, we usually get:
  ;; - either lisp-type, so we try to extract sizes from it;
  ;; - or sql-type, column-size, and decimal-digits returned by %sql-describe-...
  ;; NB: We should be consistent in whether args are Lisp type or SQL type parameters.
  ;; NB: Converting to SQL_C_CHAR of non-character column must be accompanied by
  ;;	 returning the display-size of the original sql-type as a column-size.
  ;; NB: Due to possible NLS-problem with numbers, we should always convert explicitly
  ;;	 the given SQL_NUMERIC to float or rational via SQL_C_NUMERIC.
  ;; NB: SQL_C_WCHAR is only returned if sql-type is known as SQL_W...CHAR.
  ;;	 Are there other cases?
  (unless lisp-type
    (setq lisp-type (type-of datum)))
  (cond ((and (consp lisp-type) (numberp (second lisp-type)))
         (shiftf odbc-column-size column-size (second lisp-type)))
        (column-size)				; column-size comes as an argument
         ;(setq column-size-from-odbc t))
        ((vectorp datum)
         (shiftf odbc-column-size column-size (length datum))))
  (when (and (consp lisp-type) (numberp (third lisp-type)))
    (setq decimal-digits (third lisp-type)))
  (let ((type-descriptor nil))
   (flet ((numeric (&optional (seq #1='(#.SQL_NUMERIC #.SQL_DECIMAL)) default)
           ;; Provide a numeric c-type corresponding to sql-type. When no sql-type 
           ;; is given explicitly, choose it automatically among the supported by DBMS.
           (case sql-type
             ((#.SQL_NUMERIC #.SQL_DECIMAL)
 #+odbc3.0    SQL_C_NUMERIC
 #-odbc3.0    (setq colum-size (display-size sql-type column-size decimal-digits
                                             (setq type-descriptor
                                                   (%get-type-descriptor database type))))
 #-odbc3.0    SQL_C_CHAR)
             ((nil)
              (dolist (type seq)
                (when (setq type-descriptor (%get-type-descriptor database type))
                  (setq sql-type type)
                  (return)))
 #+odbc3.0    SQL_C_NUMERIC
 #-odbc3.0    (when sql-type
                (setq colum-size (display-size sql-type column-size decimal-digits
                                               type-descriptor)))
 #-odbc3.0    SQL_C_CHAR)
             (otherwise
              (when default
 #+odbc3.0      SQL_C_NUMERIC)))))
   (let* (;(sql-type nil)
          (c-type 
           (case (first-or-self lisp-type)
            ((string simple-string simple-base-string	; either Lisp args (length)
              character base-char standard-char
 #+lispworks  lw:text-string #+lispworks lw:simple-text-string #+lispworks lw:simple-char
              :string :clob :blob)			; or SQL args (length)
             (case sql-type
               ((#.SQL_CHAR #.SQL_VARCHAR #.SQL_LONGVARCHAR)
                SQL_C_CHAR)
 #+Unicode     ((#.SQL_WCHAR #.SQL_WVARCHAR #.SQL_WLONGVARCHAR) 
                SQL_C_WCHAR)
               ((#.SQL_BINARY #.SQL_VARBINARY #.SQL_LONGVARBINARY) ; to hex string
                ;; We are going to convert it later with %get-binary
                SQL_C_BINARY)
               (otherwise				
                (when sql-type				; need a buffer of display-size
                  (setq column-size (display-size sql-type column-size decimal-digits
                                        (setq type-descriptor
                                              (%get-type-descriptor database sql-type)))))
                SQL_C_CHAR)))
            ((integer fixnum bignum)			; Lisp args (low high)
             (cond ((or (eql sql-type SQL_VARBINARY)	; Bookmark
                        (eql sql-type SQL_BINARY))	; Bookmark (MsOra driver)
                    SQL_C_VARBOOKMARK)			; =SQL_C_BINARY
                   ((eql sql-type SQL_BIGINT)		; Use bigint only if supported
                    SQL_C_SBIGINT)
                   ((when-bind (result					
                     (cond ((and column-size decimal-digits)	  ; (low high) or actual
                            (cond ((> column-size decimal-digits) ; comes along with
                                   (cond ((<= column-size 5)	  ; sql-type
                                          SQL_C_SSHORT)
                                         ((<= column-size 10)
                                          SQL_C_SLONG)
                                         ((eq lisp-type 'fixnum)  ; DI 2008-Nov-27 ??
                                          SQL_C_SLONG)))
                                  ((<= -32768 column-size decimal-digits 32767)
                                   SQL_C_SSHORT)
                                  ((<= -2147483648 column-size decimal-digits 2147483647)
                                   SQL_C_SLONG)))
                           (datum
                            (let ((integer-length (integer-length datum)))
                              (cond ((< integer-length 16)
                                     SQL_C_SSHORT)
                                    ((< integer-length 32)
                                     SQL_C_SLONG)
                                    ((%get-type-descriptor database SQL_BIGINT)
                                     SQL_C_SBIGINT))))
                           ((eq lisp-type 'fixnum)
                            SQL_C_SLONG) ))
                      (nilf column-size decimal-digits)	; not to mess with SQL's sizes and 
                      result))))			; get them finally from type-desc
            (:integer ;:counter				; SQL args (precision)
             (if (or (and column-size (<= column-size 4))
                     (and datum (< (integer-length datum) 16)))
                 SQL_C_SSHORT
                 SQL_C_SLONG))				; Opt for SQL_C_SBIGINT?
            (:smallint
             SQL_C_SSHORT)
            (bit ;boolean :boolean
             SQL_C_BIT)
            ((unsigned-byte :unsigned-byte)		; Lisp args (bit-length)
             (cond ((null column-size) SQL_C_UTINYINT)
                   ((<= column-size 8)
                    (nilf column-size) SQL_C_UTINYINT)
                   ((<= column-size 16)
                    (nilf column-size) SQL_C_USHORT)
                   (t (nilf column-size) SQL_C_ULONG)))
            ((signed-byte :signed-byte)			; Lisp args (bit-length)
             (cond ((null column-size) SQL_C_STINYINT)
                   ((<= column-size 8)
                    (nilf column-size) SQL_C_STINYINT)
                   ((<= column-size 16)
                    (nilf column-size) SQL_C_SSHORT)
                   (t (nilf column-size) SQL_C_SLONG)))
            ((double-float float :double :float)
             ;(or (numeric) SQL_C_DOUBLE))
            ;(:float					; SQL args (binary precision)
             (cond ;((numeric))
                   ((and column-size (> column-size 53))
                    (unless sql-type (setq sql-type SQL_FLOAT))
                    (setq column-size (display-size sql-type column-size decimal-digits
                                        (setq type-descriptor
                                              (%get-type-descriptor database sql-type))))
                    SQL_C_CHAR)	
                   ((and (floatp datum) (> (float-digits datum) 53))
                    (unless sql-type (setq sql-type SQL_FLOAT))
                    (setq column-size (display-size sql-type column-size decimal-digits
                                        (setq type-descriptor
                                              (%get-type-descriptor database sql-type))))
                    SQL_C_CHAR)
                   (t
                    SQL_C_DOUBLE)))
            ((:timestamp sql.sys:universal-time
 #+local-time local-time:local-time
 #+local-time local-time:utc-time)
             ;; MSDN also mentions SQL_TYPE_UTCDATETIME and SQL_TYPE_UTCTIME, but
             ;; - there is no corresponding C-type and C structure described,
             ;; - who supports them?
             (if (>= $ODBCVER 0300) SQL_C_TYPE_TIMESTAMP SQL_C_TIMESTAMP))
            ;; The precision of a NUMERIC(p,s) is decimal precision exactly equal to p. 
            ((ratio rational :numeric)			; SQL args (precision[,scale])
             (numeric #1# t))
            ;; The precision of a DECIMAL(p,s) is an implementation-defined decimal
            ;; precision that is no less than p.
            (:decimal					; SQL args (precision[,scale])
             (numeric '(#.SQL_DECIMAL #.SQL_NUMERIC) t))
	    ;; Convert bit to octet length only if it comes from lisp-type or datum
            (bit-vector				; make sure it is a bit-length
             (when (and column-size odbc-column-size)
               (setf column-size (ceiling column-size 8)))
             SQL_C_BINARY)
	    ;; For "fixed-length type", always revert to ODBC column-size if possible
            (simple-bit-vector
             (cond (odbc-column-size (setf column-size odbc-column-size))
                   (column-size (setf column-size (ceiling column-size 8))))
             SQL_C_BINARY)
            ((array simple-array vector octet-vector)
             (let ((type (cond ((consp lisp-type) (second lisp-type))
                               (datum (upgraded-array-element-type
                                       (array-element-type datum)))))
                   (size (cond ((consp lisp-type)
                                (if (consp (caddr lisp-type))
                                    (caaddr lisp-type)
                                    (caddr lisp-type)))
                               (datum (array-total-size datum)))))
               (setf decimal-digits 0)
               (case (if (consp type) (first type) type)
                 ((unsigned-byte signed-byte :unsigned-byte :signed-byte)
                  (when size				; We've got more pricise value
                    (setf column-size 
                           (if-bind (bit-count (and (consp type) (second type)))
                             (* size (ceiling bit-count 8))
                             size)))
                  SQL_C_BINARY))))
            ((single-float :real)
             SQL_C_FLOAT) ;(or (numeric) SQL_C_FLOAT))
            (:bookmark SQL_C_VARBOOKMARK)		; =SQL_C_BINARY
            (:bigint
             SQL_C_SBIGINT)
            (long-float
             (cond ((numeric))
                   (t
                    (unless sql-type (setq sql-type SQL_DOUBLE))
                    (setq column-size (display-size sql-type column-size decimal-digits
                                        (setq type-descriptor
                                              (%get-type-descriptor database sql-type))))
                    SQL_C_CHAR)))
#+local-time((local-time:date :date)
             (if (>= $ODBCVER 0300) SQL_C_TYPE_DATE SQL_C_DATE))	; DI 2007-Apr-11
                 ;(if (setq type-descriptor (%get-type-descriptor database SQL_TYPE_DATE))
                 ;    SQL_C_TYPE_DATE SQL_C_TYPE_TIMESTAMP)
                 ;(if (setq type-descriptor (%get-type-descriptor database SQL_DATE))
                 ;    SQL_C_DATE SQL_C_TIMESTAMP)))
         )))
     (or c-type
         ;; If we got here, first, give priority to conversion from lisp-type, 
         ;; then, try the default conversion from sql-type as a last resort.
         (multiple-value-bind (type size) (db-lisp-to-c-type database lisp-type)
           (prog1 (setq c-type type)
             (when size (setq column-size size))))
         (and sql-type
              (multiple-value-bind (type size) 
                  (db-sql-to-c-type database sql-type column-size decimal-digits)
                (prog1 (setq c-type type)
                  (when size (setq column-size size)))))
         (progn (cerror "Convert Lisp object to a character data type."
                        'sql-type-coercion-error :database database :data-type lisp-type)
           (setq c-type SQL_C_CHAR)))
     (unless sql-type
       (multiple-value-setq (sql-type column-size decimal-digits)
           (db-c-to-sql-type database c-type column-size decimal-digits)))
     (if (orf type-descriptor (%get-type-descriptor database sql-type))
         (values c-type sql-type 
                 (or column-size (type-desc-column-size type-descriptor))
                 (or decimal-digits (type-desc-maximum-scale type-descriptor) 0)
                 (type-desc-nullable type-descriptor))
         (values c-type sql-type 
                 (or column-size 0)
                 (or decimal-digits 0)
                 SQL_NULLABLE)) ))))

;;; Descriptors and descriptor records

(defun %make-desc-record (database lisp-type array-size record-number name
                          c-type sql-type column-size decimal-digits
                          &optional nullable parameter-type)
 ;;; Create new descriptor record and allocate low-level buffers.
  ;; Args: array-size  Rowset or paramset size
  ;;       column-size =0 if cannot be determined
  ;; NB: For long data, only a single item from one row of the rowset or paramset
  ;;     is fetched or deposited at once.
  (unless c-type
    (setq c-type (db-sql-to-c-type database sql-type column-size decimal-digits)))
  (let* ((precision (or column-size 0))
         (scale (or decimal-digits 0))
         (length column-size)
         (octet-length length)
         (long-p nil)
         (data-ptr
          (case c-type					; order by "most often used first"
            (#.SQL_C_SLONG
             (%new-ptr 'sql-integer array-size))	;:long
            (#.SQL_C_CHAR
             (let* ((type-descriptor
                     (if (member sql-type '(#.SQL_CHAR #.SQL_VARCHAR #.SQL_LONGVARCHAR))
                         (%get-type-descriptor database sql-type)
                         nil))
                    (max-length  ; can be null during first %initialize-type-info call
                     (if type-descriptor 
                         (min (type-desc-column-size type-descriptor) *default-max-length*)
                         *default-max-length*)))
               (cond ((or (zerop column-size) (> column-size max-length))
                      (setq octet-length (1+ max-length)		
                            long-p t)
                      ;; For long data, only a single item from one row of the rowset
                      ;; or paramset is fetched or deposited at once.
                      (%new-cstring octet-length 'SQL-CHAR))
                     (t	
                      (setq octet-length (1+ column-size))
                      (%new-cstring (* octet-length array-size) 'SQL-CHAR)) )))
             ;(when (> (* octet-length array-size) 32767)
             ; (error "~&Buffer size ~d bigger than 32767. Reduce rowset-size (now ~d)."
             ;         (* row-fetch-count octet-length) row-fetch-count))
  #+Unicode (#.SQL_C_WCHAR
             (let* ((type-descriptor
                     (if (member sql-type '(#.SQL_WCHAR #.SQL_WVARCHAR #.SQL_WLONGVARCHAR))
                         (%get-type-descriptor database sql-type)
                         nil))
                    (max-length  ; can be null during first %initialize-type-info call
                     (if type-descriptor
                         (min (type-desc-column-size type-descriptor) *default-max-length*)
                         *default-max-length*)))
               (cond ((or (zerop column-size) (> column-size max-length))
                      (setq octet-length (* 2 (1+ max-length))
                            long-p t)
                      (%new-cstring (1+ max-length) 'SQL-WCHAR))	;:wchar-t))
                     (t
                      (setq octet-length (* 2 (1+ column-size)))
                      (%new-cstring (* (1+ column-size) array-size) 'SQL-WCHAR)) )))
            (#.SQL_C_DOUBLE
             (%new-ptr 'sql-double array-size))		;:double
            ((#.SQL_C_TIMESTAMP
  #+odbc3.0   #.SQL_C_TYPE_TIMESTAMP)
             (setq precision decimal-digits
                   scale 0)
             (%make-struct sql-c-timestamp array-size))
            (#.SQL_C_SSHORT
             (%new-ptr 'sql-smallint array-size))	;:short
            ((#.SQL_C_BIT #.SQL_C_STINYINT)
             ;(setq octet-length 1  length 1  precision 1  scale 0)
             (%new-ptr 'sql-byte array-size))
            (#.SQL_C_UTINYINT
             (%new-ptr 'sql-octet array-size))		;:unsigned-byte
            (#.SQL_C_BINARY ;=SQL_C_VARBOOKMARK 
             (let ((max-length
                    (if-bind (type-descriptor (%get-type-descriptor database sql-type))
                      (min (type-desc-column-size type-descriptor) *default-max-length*)
                      *default-max-length*)))
               (cond ((or (zerop column-size) (> column-size max-length))
                      (setq octet-length max-length
                            long-p t)
                      (%new-binary octet-length))
                     (t
                      (%new-binary (* octet-length array-size))) )))
            ((#.SQL_C_DATE
  #+odbc3.0   #.SQL_C_TYPE_DATE)
             (setq precision 0
                   scale 0)
             (%make-struct sql-c-date array-size))
            ((#.SQL_C_TIME
  #+odbc3.0   #.SQL_C_TYPE_TIME)
             (setq precision decimal-digits
                   scale 0)
             (%make-struct sql-c-time array-size))
            (#.SQL_C_FLOAT
             (setq precision column-size)
             (%new-ptr 'sql-real array-size))		;:float
            (#.SQL_C_ULONG				; seldom used
             (%new-ptr 'sql-uinteger array-size))	;:unsigned-long
            (#.SQL_C_USHORT
             (%new-ptr 'sql-usmallint array-size))	;:unsigned-short
  #+odbc3.0 (#.SQL_C_NUMERIC
             (%make-struct sql-c-numeric array-size))
  #+odbc3.0 ((#.SQL_C_SBIGINT #.SQL_C_UBIGINT)
             (%make-struct sql-c-sbigint array-size))	; unused sql-c-ubigint
            (otherwise
             (cerror "Convert data to a character string."
                     'sql-type-coercion-error :database database :data-type sql-type)
             (setq c-type SQL_C_CHAR)
             (let ((max-length 
                    (if-bind (type-descriptor (%get-type-descriptor database SQL_VARCHAR))
                      (min (type-desc-column-size type-descriptor) *default-max-length*)
                      *default-max-length*)))
               (cond ((or (zerop column-size) (> column-size max-length))
                      (setq octet-length (1+ max-length)		
                            long-p t)
                      (%new-cstring octet-length 'SQL-CHAR))
                     (t
                      (setq octet-length (1+ column-size))
                      (%new-cstring (* octet-length array-size) 'SQL-WCHAR)) )))
            ))
         (octet-length-ptr (%new-ptr 'sql-integer array-size)))
    (if parameter-type
        (make-param-desc-record
          :record-number record-number
          :name name
          :c-type c-type :sql-type sql-type :lisp-type lisp-type
          :length length :octet-length octet-length
          :data-ptr data-ptr
          :octet-length-ptr octet-length-ptr
          :nullable (or nullable SQL_NULLABLE)
          :precision precision :scale scale :long-p long-p         
          :parameter-type parameter-type)
        (make-column-desc-record
          :record-number record-number
          :name name
          :c-type c-type :sql-type sql-type :lisp-type lisp-type
          :length length :octet-length octet-length
          :data-ptr data-ptr
          :octet-length-ptr octet-length-ptr
          ;:indicator-ptr octet-length-ptr		; use the same as octet
          :nullable (or nullable SQL_NULLABLE)
          :precision precision :scale scale :long-p long-p)
      )))

(defun %make-column-desc-record (database hstmt array-size column-nr &optional lisp-type)
 ;;; Arsg: array-size	Number of rows to fetch at once.
  ;;	   column-nr	Becomes record-number.
  (let ((c-type nil))
    (multiple-value-bind (name sql-type column-size decimal-digits nullable)
        (%sql-describe-column hstmt column-nr) ;max-name-length
      (when lisp-type
        (let (dummy)        		; don't change original sql-type
          (declare (ignore dummy))
          (multiple-value-setq (c-type dummy column-size decimal-digits)
              (%describe-lisp-object database nil lisp-type
                                     sql-type column-size decimal-digits))))
      (%make-desc-record database lisp-type array-size column-nr name
                         c-type sql-type column-size decimal-digits nullable))))

#|(defun %renumber-records (records)
 ;;; Sort record numbers by assigning larger numbers to long data records.
  ;; Doesn't move any record inside the records array.
  (loop with long-index = (count-if-not #'desc-long-p records)
        and short-index = 0
        for record across records
        do (setf (desc-record-number record) (if (desc-long-p record)
                                                 (incf long-index)
                                                 (incf short-index)))))|#

(defun ensure-hdesc (descriptor hstmt &optional (type SQL_ATTR_APP_ROW_DESC))
  (declare (fixnum type))
  (or (handle descriptor)
      (setf (handle descriptor) (%sql-get-attribute :hstmt hstmt type))))

(defmethod initialize-instance :after ((self descriptor) &rest initargs
                                       &key array-size &allow-other-keys)
  (declare (ignore initargs))
  (when array-size
    (setf (desc-array-status-ptr self) (%new-ptr 'sql-usmallint array-size)))
  (setf (desc-rows-processed-ptr self) (%new-ptr 'sql-uinteger)))

(defmethod reinitialize-instance :after ((self descriptor) &rest initargs
                                         &key array-size &allow-other-keys)
 ;;; NB:    ,  shared-initialize   array-size
  ;;       1,      
  (declare (ignore initargs))
  (when (and array-size (not (eql (desc-array-size self) array-size)))
    (when-bind (ptr (desc-array-status-ptr self))
      (%dispose-ptr ptr))
    (setf (desc-array-status-ptr self) (%new-ptr 'sql-usmallint array-size))))

(defmethod shared-initialize :after ((self row-descriptor) slot-names
                                     &rest initargs
                                     &key cursor array-size result-types
                                          start-column end-column column-numbers
                                     &allow-other-keys)
 ;;; Works when the cursor is specified explicitly after the cursor type is set.
  (declare (ignore slot-names initargs))
  (when cursor
   (with-slots (database handle expression) cursor
    (ensure-hdesc self handle)
    (let ((records (desc-records self))
          (result-types (or result-types (collect-types expression)))
          (bookmark-record nil)
          (max-short nil)	; maximum record-number of short data if any
          (min-long nil)	; minimum record-number of long data if any
          lisp-type record)
      (setf (fill-pointer records) 0)
      (if column-numbers
	  ;; used for predefined queries like catalog info, unlikely includes long data
          (dolist (column-nr column-numbers)
            (declare (fixnum column-nr))
            (setq lisp-type (if (zerop column-nr)
                                :bookmark		; bignum ;(vector unsigned-byte)
                                (nth (1- column-nr) result-types))
                  record (%make-column-desc-record database handle array-size
                                                   column-nr lisp-type))
            (if (desc-long-p record)
                (setq min-long (min (or min-long most-positive-fixnum) (desc-record-number
                                                                        record)))
                (setq max-short (max (or max-short 0) (desc-record-number record))))
            (if (zerop column-nr)				; bookmark
                (setq bookmark-record record)
                (vector-push-extend record records)))
          (do ((upto (if end-column
                         (min end-column (%sql-columns-count handle))
                         (%sql-columns-count handle)))
               (column-nr (or start-column 1) (1+ column-nr)))	; default start-column
              ((> column-nr upto))
            (declare (fixnum column-nr upto))
            (setq lisp-type (if (zerop column-nr)
                                :bookmark		; bignum (vector unsigned-byte)
                                (nth (1- column-nr) result-types))
                  record (%make-column-desc-record database handle array-size
                                                   column-nr lisp-type))
                ;when (zerop column-nr)				; bookmark
                ;do (setf (desc-lisp-type record) 'vector)
            (if (desc-long-p record)
                (unless min-long (setq min-long (desc-record-number record)))
                (setq max-short (desc-record-number record)))
            (if (zerop column-nr)				; bookmark
                (setq bookmark-record record)
                (vector-push-extend record records)) ))

      ;; Verify if long columns go at the end.
      (when (and min-long max-short (< min-long max-short)
                 (not (logtest (db-get-info database SQL_GETDATA_EXTENSIONS)
                               SQL_GD_ANY_ORDER)))
        (error 'sql-database-data-error :database database :expression expression
               :message "Long type columns must follow short ones in a query."))
        ;;    ,       ,
        ;;     !
        ;(%renumber-records records)
      (when (and min-long (> array-size 1) (eq (cursor-type cursor) :forward-only))
        (cerror "Set the rowset size to 1."
                'sql-database-data-error :database database :expression expression
                :message "Forward-only cursor cannot fetch long data in block.")
        (setq array-size 1))
      ;(loop record accross records
      ;      maximize (* (desc-octet-length record) array-size))
      ; (setf array-size (min (floor 32767 (1+ max-size)) *max-row-fetch-count*))
      (setf (desc-count self) (fill-pointer records)
            (desc-array-size self) array-size
            (bookmark-record self) bookmark-record
  ))) )
  self)

(defmethod %dispose-descriptor ((descriptor descriptor) &optional drop-p)
  (when drop-p
    (when-bind (ptr (desc-array-status-ptr descriptor))
      (%dispose-ptr ptr)
      (nilf (desc-array-status-ptr descriptor)))
    (when-bind (ptr (desc-rows-processed-ptr descriptor))
      (%dispose-ptr ptr)
      (nilf (desc-rows-processed-ptr descriptor)))
    (nilf (handle descriptor)))
  (loop for record across (desc-records descriptor)
        ;for data-ptr = (desc-data-ptr record)
        ;and octet-length-ptr = (desc-octet-length-ptr record)
        when (desc-data-ptr record)
          do (%dispose-ptr (desc-data-ptr record))
             (nilf (desc-data-ptr record))
        when (desc-octet-length-ptr record)
          do (%dispose-ptr (desc-octet-length-ptr record))
             (nilf (desc-octet-length-ptr record)) )
        ;when (desc-indicator-ptr record)
  (setf (fill-pointer (desc-records descriptor)) 0))

(defmethod %dispose-descriptor :after ((descriptor row-descriptor) &optional drop-p)
  (declare (ignore drop-p))
  (when-bind (record (bookmark-record descriptor))
    (when-bind (ptr (desc-data-ptr record))
      (%dispose-ptr ptr)
      (nilf (desc-data-ptr record)))
    (when-bind (ptr (desc-octet-length-ptr record))
      (%dispose-ptr ptr)
      (nilf (desc-octet-length-ptr record)))
    (nilf (bookmark-record descriptor)))
  (when-bind (ptr (bookmark-ptr descriptor))
    (%dispose-ptr ptr)
    (nilf (bookmark-ptr descriptor))))

#|(defun (setf rowset-size) (size cursor)
  (when (>= $ODBCVER 0300)
    (with-slots (handle row-descriptor) cursor
      (when row-descriptor
        (%sql-set-attribute :hstmt handle SQL_ATTR_ROW_ARRAY_SIZE size)
        (setf (desc-array-size row-descriptor) size)))))|#

(defun %initialize-rowset (cursor &key (size SQL_ROWSET_SIZE_DEFAULT) result-types
                                       (start-column 1) end-column column-numbers)
 ;;; Initialize fetching results after SQLExecDirect, SQLExecute, or catalog function.
  (unless (>= $ODBCVER 0300) (setq size 1))
  (with-slots (handle row-descriptor) cursor
    (if row-descriptor
        (reinitialize-instance row-descriptor
                               :cursor cursor
                               :array-size size
                               :result-types result-types
                               :start-column start-column 
                               :end-column end-column
                               :column-numbers column-numbers)
        (setf row-descriptor (make-instance 'row-descriptor ;(%make-row-descriptor cursor
                                            :cursor cursor
                                            :array-size size
                                            :result-types result-types
                                            :start-column start-column 
                                            :end-column end-column
                                            :column-numbers column-numbers)))
    (%shared-initialize-rowset cursor)
    #|(cond ((>= $ODBCVER 0300)
           (%sql-set-attribute :hstmt handle SQL_ATTR_ROW_ARRAY_SIZE
                               (setq size (desc-array-size row-descriptor)))
           (when (or (> size 1) (eq cursor-type :keyset-driven))
             (%sql-set-attribute :hstmt handle SQL_ATTR_ROW_STATUS_PTR
                                 (desc-array-status-ptr row-descriptor)))
           (when (> size 1)
             (%sql-set-attribute :hstmt handle SQL_ATTR_ROWS_FETCHED_PTR
                                 (desc-rows-processed-ptr row-descriptor))))
    (when (= size 1) ;(or (> size 1) (eq cursor-type :keyset-driven))
      (setf (desc-row-status row-descriptor 0) SQL_ROW_SUCCESS)
    ;(when (= size 1)
      (setf (desc-rows-processed row-descriptor) 1))
          (t ;; Emulate success
           (setf (desc-row-status row-descriptor 0) SQL_ROW_SUCCESS
                 (desc-rows-processed row-descriptor) 1)))
    (setf (block-position row-descriptor) :start)	; shared-initialize?
    (allf 0 (fetched-count row-descriptor)
            (current-row-nr row-descriptor))
    (nilf (current-row-status row-descriptor)
          (last-operation row-descriptor))
    (with-slots (bookmark-ptr) row-descriptor
      (when bookmark-ptr
        (%dispose-ptr bookmark-ptr)
        (nilf bookmark-ptr)))|#

    (when-bind (record (bookmark-record row-descriptor))	; bind bookmark if any
      (%sql-bind-column handle
                        (desc-record-number record)
                        (desc-c-type record)
                        (desc-data-ptr record)
                        (desc-octet-length record)
                        (desc-octet-length-ptr record)))

    (loop for record across (desc-records row-descriptor)
          unless (desc-long-p record)   ; if true, we fetch in chunks with %sql-get-data
          do (%sql-bind-column handle
                               (desc-record-number record)
                               (desc-c-type record)
                               (desc-data-ptr record)
                               (desc-octet-length record)
                               (desc-octet-length-ptr record))
          and when (eql (desc-c-type record) SQL_C_NUMERIC)
                do ;(%sql-set-desc-field (handle self) (desc-record-number record)
                   ;                      SQL_DESC_CONCISE_TYPE SQL_C_NUMERIC)
                   (%sql-set-desc-field (ensure-hdesc row-descriptor handle)
                                        (desc-record-number record)
                                        SQL_DESC_SCALE (desc-scale record)) )))

(defun %shared-initialize-rowset (cursor)
 ;;; Initializing or simply reinitializing when the result set columns stay the same.
  ;; NB: Setting SQL_ATTR_ROW_STATUS_PTR     
  ;;	   . ,   - bookmark-  
  ;;      Fetch,     .    , 
  ;;	 add,  .
  ;- NB: Setting SQL_ATTR_ROW_STATUS_PTR for non-block cursor (size=1)
  ;-	 seems undesirable. This undermines Driver Manager. Both Access and Oracle
  ;-	 drivers stop working: crash, provide wrong out-len-indicators,
  ;-     especially the in catalog function SQLColumns().
  ;- MSDN inconsistent with this:
  ;-	 "If using block cursors, scrollable cursors, or both, the application: ...
  ;-     - Calls SQLSetStmtAttr to set SQL_ATTR_ROW_STATUS_PTR to point to an array
  ;-	   of status records.
  ;-	 - Calls SQLSetStmtAttr to set SQL_ATTR_ROWS_FETCHED_PTR to point to an
  ;-	   SQLINTEGER..."
  (with-slots (handle row-descriptor cursor-type) cursor
    ;(%sql-set-attribute :hstmt handle SQL_ATTR_ROW_BIND_TYPE SQL_BIND_BY_COLUMN)
    (let* ((size (desc-array-size row-descriptor))
           (bind-ptrs (or (> size 1)
                          (eq cursor-type :keyset-driven))))
      (when (>= $ODBCVER 0300)
        (%sql-set-attribute :hstmt handle SQL_ATTR_ROW_ARRAY_SIZE size)
        (%sql-set-attribute :hstmt handle SQL_ATTR_ROW_STATUS_PTR
                            (if bind-ptrs
                                (desc-array-status-ptr row-descriptor)
                                *null-pointer*))
        (%sql-set-attribute :hstmt handle SQL_ATTR_ROWS_FETCHED_PTR
                            (if bind-ptrs
                                (desc-rows-processed-ptr row-descriptor)
                                *null-pointer*)))
      (unless (and (>= $ODBCVER 0300) bind-ptrs)  ; emulate success and one resulting row
        (setf (desc-row-status row-descriptor 0) SQL_ROW_SUCCESS
              (desc-rows-processed row-descriptor) 1)))
    ;; Don't touch retrieve-data as it reflects the cursor attribute
    (setf (block-position row-descriptor) :start)
    (allf 0 (fetched-count row-descriptor)
            (current-row-nr row-descriptor))
    (nilf (current-row-status row-descriptor)
          (last-operation row-descriptor)) ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Low-level reading and writing various data type
;;; Args: ptr  Usually of a correct FLI-type

(defun timestamp-wo-zone-to-universal-time (ptr)
  (values
   (encode-universal-time 
    (%foreign-slot ptr sql-c-timestamp second)
    (%foreign-slot ptr sql-c-timestamp minute)
    (%foreign-slot ptr sql-c-timestamp hour)
    (%foreign-slot ptr sql-c-timestamp day)
    (%foreign-slot ptr sql-c-timestamp month)
    (%foreign-slot ptr sql-c-timestamp year))
   (%foreign-slot ptr sql-c-timestamp fraction)))

;(defun universal-time-to-timestamp (ptr time &optional (fraction 0))
(defgeneric time-to-timestamp-wo-zone (ptr time)
  (:method (ptr (universal-time integer) &aux (fraction 0))
   (multiple-value-bind (second minute hour day month year)
       (decode-universal-time universal-time)
     (setf (%foreign-slot ptr sql-c-timestamp second) second
           (%foreign-slot ptr sql-c-timestamp minute) minute
           (%foreign-slot ptr sql-c-timestamp hour) hour
           (%foreign-slot ptr sql-c-timestamp day) day
           (%foreign-slot ptr sql-c-timestamp month) month
           (%foreign-slot ptr sql-c-timestamp year) year
           (%foreign-slot ptr sql-c-timestamp fraction) fraction))))

(defun date-to-universal-time (ptr)
  (encode-universal-time
   0 0 0
   (%foreign-slot ptr sql-c-date day)
   (%foreign-slot ptr sql-c-date month)
   (%foreign-slot ptr sql-c-date year)))

(defgeneric time-to-date (ptr time)
 (:method (ptr (universal-time integer))
   (multiple-value-bind (second minute hour day month year)
       (decode-universal-time universal-time)
     (declare (ignore second minute hour))
     (setf (%foreign-slot ptr sql-c-date day) day
           (%foreign-slot ptr sql-c-date month) month
           (%foreign-slot ptr sql-c-date year) year))))

(defgeneric time-to-time-wo-zone (ptr time)
 (:method (ptr (sec integer))
  ;; Seconds (i.e. local-time-sec) -> SQL_TYPE_TIME
  ;; Q: Shall we treat integer argument as universal-time?
   (multiple-value-bind (minute second) (floor sec 60)
    (declare (fixnum minute second))
    (multiple-value-bind (hour minute) (floor minute 60)
      (declare (fixnum hour minute))
      (setf (%foreign-slot ptr sql-c-time second) second
            (%foreign-slot ptr sql-c-time minute) minute
            (%foreign-slot ptr sql-c-time hour) hour)))) )

(defun time-wo-zone-to-universal-time (ptr)
  (values
   (encode-universal-time 
    (%foreign-slot ptr sql-c-time second)
    (%foreign-slot ptr sql-c-time minute)
    (%foreign-slot ptr sql-c-time hour)
    #-pcl 0 #+pcl 1
    #-pcl 0 #+pcl 1
    #-pcl 0 #+pcl 1900)))
   ;(%foreign-slot ptr sql-c-timestamp fraction)  ;DIV 2002-Aug-05 Does not apply.

;;; BIGINT

#+odbc3.0
(declaim (inline %get-sbigint %put-sbigint) 			;%get-ubigint %put-ubigint
         (ftype (function (t) integer) %get-sbigint)		;%get-ubigint
         (ftype (function (t integer) (values)) %put-sbigint))	;%put-ubigint
#+odbc3.0
(defun %get-sbigint (ptr)
  (declare (optimize (speed 3) (debug 0) (space 0) #+lispworks (safety 3)))
  (%make-int64 (%foreign-slot ptr sql-c-sbigint high)
               (%foreign-slot ptr sql-c-sbigint low)))
#+odbc3.0
(defun %put-sbigint (ptr datum)
  (declare (integer datum)
           (optimize (speed 3) (debug 0) (space 0) (safety 0)))
  (setf (%foreign-slot ptr sql-c-sbigint high) (%int64-high datum)
        (%foreign-slot ptr sql-c-sbigint low) (%int64-low datum))
  (values))
#|
(defun %get-ubigint (ptr)
  (declare (optimize (speed 3) (debug 0) (space 0) (safety 0)))
  (%make-int64 (%foreign-slot ptr sql-c-ubigint high)
               (%foreign-slot ptr sql-c-ubigint low)))
(defun %put-ubigint (ptr datum)
  (declare (integer datum)
           (optimize (speed 3) (debug 0) (space 0) (safety 0)))
  (setf (%foreign-slot ptr sql-c-ubigint high) (%int64-high datum)
        (%foreign-slot ptr sql-c-ubigint low) (%int64-low datum))
  (values))
|#
;;; A number is stored in the val field of the SQL_NUMERIC_STRUCT structure is
;;; a scaled integer, in little endian mode. For example, 
;;; the number 10.001 base 10, with a scale of 4, is scaled to an integer of 100010.
;;; Because this is 186AA in hexadecimal format, the value in SQL_NUMERIC_STRUCT would be
;;; "AA 86 01 00 00  00", with the number of bytes defined by SQL_MAX_NUMERIC_LEN.
;;;
;;; The precision and scale fields of the SQL_C_NUMERIC data type are never used for
;;; input from an application, only for output from the driver to the application.
;;; When the driver writes a numeric value into the SQL_NUMERIC_STRUCT, it will use
;;; its own driver-specific default as the value for the precision field, and it will
;;; use the value in the SQL_DESC_SCALE field of the application descriptor (which
;;; defaults to 0) for the scale field. An application can provide its own values
;;; for precision and scale by setting the SQL_DESC_PRECISION and SQL_DESC_SCALE fields
;;; of the application descriptor.
;;;
;;; Access:
;;;  1)	Precision defaults to 28
;;;  2) Precision jumps to 38 if we specify it (and scale) explicitly.
;;;  3) Specifying scale <4 while binding could warn during fetch:
;;;	[ODBC state 01S07 info 13] Fractional truncation on column number XX
;;;  4) Currency range is [-922,337,203,685,477.5808 .. 922,337,203,685,477.5807].
;;;	Otherwise signals
;;;		[ODBC state 22003 error 34] Numeric value out of range (null)
;;; 	and the cursor position is lost.
;;;  5) We can actually set the scale field of the SQL_NUMERIC_STRUCT structure to 0
;;;	for simplifying integer output from application to driver!

(declaim (inline %get-numeric %put-numeric)
         (ftype (function (t t t) (values)) %put-numeric))
(defun %get-numeric (ptr lisp-type)
 ;;; Args: lisp-type  Defaults to ratio.
  ;; LW: SYSTEM:LISP-MEMORY-COPY-BIGNUM?
  ;; LW CAUTION: Some parts work only when compiled with (safety 3)!
  ;;		 Otherwise, #<pointer out of bounds> even without sys::alloc-bignum$fixnum
  ;;		  , -    SYSTEM:NEGATE.
  (declare (optimize (speed 3) (debug 0) (space 0) (safety 0)))
  (let* ((precision (%foreign-slot ptr sql-c-numeric precision))
         (scale (%foreign-slot ptr sql-c-numeric scale))
         (sign (%foreign-slot ptr sql-c-numeric sign)) ; 1 if positive, 0 if negative
         (value-ptr (%foreign-slot-pointer ptr nil value :type SQL-CHAR))
         (octet-length (cond ((<= precision 19) 8)
                             ((<= precision 28) 12)
                             (16)))
         (big #+lispworks (sys::alloc-bignum$fixnum (ash octet-length -2))
              #-lispworks 0))
    (declare (fixnum octet-length scale) (type octet precision sign) (type bignum big))
    ;(format t "~%;%get-numeric: p=~s scale=~s sign=~s ~{~2,'0x~}" precision scale sign
    ;        (loop for i upfrom 0 below 16 collect (ffc::%svref value-ptr i)))
    (locally (declare (optimize (safety 0) #+lispworks (hcl:fixnum-safety 0)))
      (loop for pos fixnum downfrom (ash (1- octet-length) 3) by 8
            and  i  fixnum downfrom (1- octet-length) to 0 
            do (setf (ldb (byte 8 pos) big) (the (unsigned-byte 8)
                                                 (%get-octet value-ptr i)))))
                      ;(%svref value-ptr i :type '(unsigned-byte 8))))))
    (case lisp-type
      ((float double-float #+lispworks single-float)	; no separate single-float in LW
       (locally (declare (optimize (safety 3)))
         (let ((value (if (< 0 scale)
                          (* big (expt 0.1d0 (the fixnum scale)))
                          (* big (expt 10d0 (the fixnum (- (the fixnum scale))))))))
           (if (zerop sign) (- value) value))))
 #-lispworks
      (single-float					; untested
       (locally (declare (optimize (safety 3)))
         (let ((value (if (< 0 scale)
                          (* big (expt 0.1s0 (the fixnum scale)))
                          (* big (expt 10s0 (the fixnum (- (the fixnum scale))))))))
           (if (zerop sign) (- value) value))))
      (string
       (format nil "~,v,vf" (max scale 0) (- scale) ;precision
               (if (zerop sign) (- big) big)))
      (t
       (let ((value (the rational (/ big (the rational (expt 10 (the fixnum scale)))))))
         (if (zerop sign)
             (locally (declare (optimize (safety 3)))
               (- value))
             value))) )))
;(format nil "~,v,vf" 0 -4 123456789)

(defun %put-numeric (ptr datum record)
 ;;; Should we test for breaching (desc-precision record) ourselves?
  (declare (optimize (speed 3) (debug 0)))
  (if (realp datum)
      (let* ((value-ptr (%foreign-slot-pointer ptr nil value :type SQL-CHAR))
             (scale 0)	;(the fixnum (desc-scale record)))
             (big (if (integerp datum)
                      datum
                      (the integer (abs (round (* datum (the integer (expt 10
                                    (the fixnum (setq scale (desc-scale record)))))))))))
             (octet-length (the fixnum (min (ceiling (integer-length big) 8) 16))))
        (declare (fixnum scale octet-length) (integer big))
        ;(format t "~%big ~s" big)
        (locally (declare (optimize (safety 0) #+lispworks (hcl:fixnum-safety 0)))
          (%fill-vector value-ptr :end 16)			;SQL_MAX_NUMERIC_LEN
          (loop for  i  fixnum upfrom 0 below octet-length
                and pos fixnum upfrom 0 by 8
                do ;(setf (%svref value-ptr i :type '(unsigned-byte 8))
                   (%put-octet value-ptr (ldb (byte 8 pos) big) i))
          ;(loop for i fixnum upfrom octet-length below 16	;SQL_MAX_NUMERIC_LEN
          ;      do (setf (%svref value-ptr i :type '(unsigned-byte 8)) 0))
          ;; For parameter, the structure can be non-initialized - set all slots!
          (setf (%foreign-slot ptr sql-c-numeric precision) (desc-precision record)
                (%foreign-slot ptr sql-c-numeric scale) scale
                (%foreign-slot ptr sql-c-numeric sign) (if (minusp datum) 0 1))))
      (error 'sql-user-error :message (format nil
             "While writing to numeric column ~s, the datum is not a real number: ~s"
             (desc-name record) datum)))
  (values))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Reading and writing single piece of data from/to buffers

(declaim (ftype (function (t &optional fixnum) t) %read-datum)
         (ftype (function (t t &optional fixnum) t) %write-datum))
(defun %read-datum (record &optional (index 0))
 ;;;  Convert rowset or paramset buffer pointed by data-ptr to a single lisp object
  ;; Args: record   column- or param-desc-record
  ;;       index    Row index in the rowset or parameter index in paramset
  ;; NB: Regularly in here, we only check for boolean lisp type that can be represented
  ;;	 by one of: SQL_CHAR or SQL_BIT.
  ;;	 In no other cases, the lisp-type should be checked (except the Oracle driver
  ;;	 bug with SQL_C_NUMERIC).
  ;; LW NB: We can simply replace several case clauses by this one:
  ;;		(otherwise (fli:dereference data-ptr :index index))
  ;;	    due to type information wired into the data pointer itself.
  (declare (type fixnum index)
           (optimize (speed 3) (space 0) #-debug (debug 0) #-debug (safety 0)
                     #+(and lispworks (not debug)) (hcl:fixnum-safety 0)))
  (let ((out-len-ind (%get-integer (desc-octet-length-ptr record) index)))
    (if (= out-len-ind SQL_NULL_DATA)	; the right way is to check for null
        ;; NB:  Unicode-   SQL_NULL_DATA 
        ;;      -65536,  (desc-c-type record) = SQL_C_WCHAR!
        *null*				; (%get-integer (desc-indicator-ptr record))
        (with-accessors ((data-ptr desc-data-ptr)) record
          (case (desc-c-type record)
 #-lispworks (#.SQL_C_SLONG
              (%get-slong data-ptr index))
             (#.SQL_C_CHAR
              ;; When truncation occurs, out-len-ind hold the length of available
              ;; data in bytes, which can be larger than the buffer size.
              (let* ((desc-octet-length (desc-octet-length record))
                     (length (min out-len-ind (1- desc-octet-length)))
                     (string (if (zerop index)
                                 (%get-string data-ptr length)
                                 (let ((start (* index desc-octet-length)))
                                   (%get-string data-ptr (+ start length) start)))))
                (declare (fixnum desc-octet-length length))
            #1= (case (desc-sql-type record)
                  ((#.SQL_DECIMAL #.SQL_NUMERIC)
                   ;; Oracle WORKAROUND - backtrack to what db-sql-to-c-type did
                   ;; May read incorrectly if the driver incorrectly convert to string:
                   ;; - after fli:set-locale
                   ;; - after setting session NLS,
                   ;;   e.g. Oracle "123,45" -read-from-string-> 123
                   ;; Oracle 10: Choose US numeric settings required by ODBC standard
                   ;; - either DSN Configuration > Application Options > Numeric Settings 
                   ;; - or connection string keyword NUM=US to driver-connect.
                   (let ((lisp-type (desc-lisp-type record)))
                     (if (and lisp-type (not (keywordp lisp-type))
                              (subtypep lisp-type 'string))
                         string
                         (let ((*read-base* 10))
                           (read-from-string string)))))
                  (otherwise
                   (case (desc-lisp-type record)
                     ((boolean :boolean)		; shall we return :true or :false?
                      (and (eql (length string) 1)
                           (let ((char (schar string 0)))
                             (or (char-equal char #\T)
                                 (char-equal char #\Y)
                                 (char-equal char #\1))))) ; PosgresSQL when Bools as Chars
                     (character
                      (if (zerop (length string)) #\Space (schar string 0)))
                     (otherwise
                      string))))))
 #+Unicode   (#.SQL_C_WCHAR
              (let* ((desc-octet-length (desc-octet-length record))
                     (length (min out-len-ind (- desc-octet-length 2)))
                     (string (if (zerop index)
                                 (%get-wstring data-ptr length)
                                 (let ((start (* index desc-octet-length)))
                                   (%get-wstring data-ptr (+ start length) start)))))
                (declare (fixnum desc-octet-length length))
                #1#))
 #-lispworks (#.SQL_C_DOUBLE
              (%get-double data-ptr index))
             ((#.SQL_C_TIMESTAMP
 #+odbc3.0     #.SQL_C_TYPE_TIMESTAMP)
              (with-dynamic-pointer (ptr :index index) data-ptr
                #-local-time (multiple-value-bind (universal-time frac)
                                 (timestamp-wo-zone-to-universal-time ptr)
                               (funcall *time-encode-function* universal-time frac))
                #+local-time (timestamp-wo-zone-to-local-time ptr)))
 #-lispworks (#.SQL_C_SSHORT 
              (%get-sshort data-ptr index))
             (#.SQL_C_BIT
              (let ((bit (%get-bit data-ptr index)))
                (case (desc-lisp-type record)
                  ((boolean :boolean)		; shall we return :true or :false?
                   (not (zerop bit)))
                  (otherwise
                   bit))))
 #-lispworks (#.SQL_C_UTINYINT
              (%get-octet data-ptr index))
             (#.SQL_C_BINARY ;=SQL_C_VARBOOKMARK
              (let ((desc-octet-length (desc-octet-length record)))
                (%get-binary data-ptr
                             (min out-len-ind desc-octet-length)
                             (or (first-or-self (desc-lisp-type record))
                                 *default-read-binary-type*)
                             (* index desc-octet-length))))
 #-lispworks (#.SQL_C_STINYINT
              (%get-byte data-ptr index))	
 #-lispworks (#.SQL_C_FLOAT
              (%get-single data-ptr index))
 #-lispworks (#.SQL_C_ULONG
              (%get-ulong data-ptr index))
 #-lispworks (#.SQL_C_USHORT
              (%get-ushort data-ptr index))	
             ((#.SQL_C_DATE
 #+odbc3.0     #.SQL_C_TYPE_DATE)
              (with-dynamic-pointer (ptr :index index) data-ptr
                #-local-time (funcall *time-encode-function* (date-to-universal-time ptr))
                #+local-time (date-to-local-time ptr)))
             ((#.SQL_C_TIME
 #+odbc3.0     #.SQL_C_TYPE_TIME)
              (with-dynamic-pointer (ptr :index index) data-ptr
                ;; DIV 2002-Aug-05 No fraction part in SQL_C_TIME
                #-local-time (funcall *time-encode-function*
                                      (time-wo-zone-to-universal-time ptr))
                #+local-time (time-wo-zone-to-local-time ptr)))
             (#.SQL_C_NUMERIC
              (with-dynamic-pointer (ptr :index index) data-ptr
                (%get-numeric ptr (or (desc-lisp-type record)
                                      *default-read-numeric-type*))))
 #+odbc3.0   ((#.SQL_C_SBIGINT #.SQL_C_UBIGINT)
              (with-dynamic-pointer (ptr :index index) data-ptr
                (%get-sbigint ptr)))			; unused %get-ubigint
             (otherwise
 #+lispworks  (fli:dereference data-ptr :index index)	; colapse several clauses
 #-lispworks  "#<unreadable>"))))))

(defun %write-datum (datum record &optional (index 0))
 ;;;  Convert lisp object to buffer contents pointed by data-ptr
  ;; Args: datum   Lisp object
  ;;	   record  column- or param-desc-record
  ;;       index   Row index in the rowset or parameter index in paramset
  ;; NB: We must set length-idicator to something non-negative. We assign 0 for the
  ;;     most of the types because the driver should checks this length only if 
  ;;     the column is character or binary.
  ;;     Nevertheless for SQL_C_DOUBLE, we assign 8 since the Access driver treats
  ;;     0 as null when updating or adding via SQLSetPos() or SQLBulkOperations().
  (declare (type fixnum index))
  (with-accessors ((data-ptr desc-data-ptr)
                   (len-ind-ptr desc-octet-length-ptr)) record
    (if (eql datum *null*)
        (setf (%get-integer len-ind-ptr index) SQL_NULL_DATA)
        (case (desc-c-type record)
 #-lispworks (#.SQL_C_SLONG
              (setf (%get-integer len-ind-ptr index) 0)
              (%put-slong data-ptr datum index))
 	     (#.SQL_C_CHAR
          #1= (unless (stringp datum)
                (setq datum
                      (case (desc-sql-type record) ; Backtrack to what db-sql-to-c-type did
                        ((#.SQL_DECIMAL #.SQL_NUMERIC)
                         ;; Oracle: May write incorrectly as when the driver 
                         ;;	    numeric setting is wrong, e.g. "123,45"
                         (write-to-string datum :base 10))
                        (otherwise
                         (case (desc-lisp-type record)
                           ((boolean :boolean)
                            (if (boolean-true-p datum) "T" "F"))
                           (otherwise ;character
                            (princ-to-string datum)))))))
              (setf (%get-integer len-ind-ptr index)	; returns size in bytes 
                    (%put-string data-ptr datum (desc-octet-length record)
                                 :index (* index (desc-octet-length record)))))
 #+Unicode   (#.SQL_C_WCHAR
              #1#
              (setf (%get-integer len-ind-ptr index)	; returns size in characters 
                    (%put-wstring data-ptr datum (desc-octet-length record)
                                  :index (* index (desc-octet-length record)))))
 	     (#.SQL_C_DOUBLE
	      ;; Due to Access driver bug, assign non-zero indicator
              (setf (%get-integer len-ind-ptr index)
                    #+lispworks #.(fli:size-of 'SQL-DOUBLE) #-lispworks 8)
              (%put-double data-ptr datum index))
             ((#.SQL_C_TIMESTAMP
 #+odbc3.0     #.SQL_C_TYPE_TIMESTAMP)
              (setf (%get-integer len-ind-ptr index) 0)
              (with-dynamic-pointer (ptr :index index) data-ptr
                (time-to-timestamp-wo-zone ptr datum)))
 #-lispworks (#.SQL_C_SSHORT 
              (setf (%get-integer len-ind-ptr index) 0)
              (%put-sshort data-ptr datum index))	
             (#.SQL_C_BIT 
              (setf (%get-integer len-ind-ptr index) 0)
              (%put-bit data-ptr ;datum
                        (if (boolean-true-p datum) 1 0)
                        index))
 #-lispworks (#.SQL_C_UTINYINT
              (setf (%get-integer len-ind-ptr index) 0)
              (%put-octet data-ptr datum index))
             (#.SQL_C_BINARY ;=SQL_C_VARBOOKMARK
              (setf (%get-integer len-ind-ptr index)	; returns size in bytes
                    (%put-binary data-ptr datum (desc-octet-length record) :index index)))
 #-lispworks (#.SQL_C_STINYINT
              (setf (%get-integer len-ind-ptr index) 0)
              (%put-byte data-ptr datum index))
 #-lispworks (#.SQL_C_FLOAT
              (setf (%get-integer len-ind-ptr index) 0)
              (%put-single data-ptr datum index))
	     ((#.SQL_C_DATE
 #+odbc3.0     #.SQL_C_TYPE_DATE)
              (setf (%get-integer len-ind-ptr index) 0)
              (with-dynamic-pointer (ptr :index index) data-ptr
                (time-to-date ptr datum)))
	     ;((#.SQL_C_TIME #+odbc3.0 #.SQL_C_TYPE_TIME)
             ; ;; DIV 2002-Aug-05 No fraction part in SQL_C_TIME
 #+odbc3.0   (#.SQL_C_NUMERIC
              (setf (%get-integer len-ind-ptr index) 0)
              (with-dynamic-pointer (ptr :index index) data-ptr
                (%put-numeric ptr datum record)))
 #+odbc3.0   ((#.SQL_C_SBIGINT #.SQL_C_UBIGINT)
              (setf (%get-integer len-ind-ptr index) 0)
              (with-dynamic-pointer (ptr :index index) data-ptr
                (%put-sbigint ptr datum)))		; unused %put-ubigint
             (otherwise
 #+lispworks  (setf (%get-integer len-ind-ptr index) 0
                    (fli:dereference data-ptr :index index) datum)
             ;(setf len-ind-ptr *null-pointer*))
		;(#.SQL_C_ULONG
		; (%put-ulong data-ptr datum index))
		;(#.SQL_C_USHORT
                ; (%put-ushort data-ptr datum index))	
	       ;(multiple-value-bind (universal-time frac)
               ;   (time-wo-zone-to-universal-time data-ptr)
               ; (funcall *time-decode-function* universal-time frac)))
               ; (funcall *time-conversion-function*
               ;          (time-wo-zone-to-universal-time data-ptr)))
  ) )   )    )
  datum)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Reading and writing in chunks

(defun %read-datum-in-chunks (hstmt record)
 ;;; Fetch one Lisp object by multiple SQLGetData calls.
  ;; NB: SQLGetData should never be called for a forward-only cursor when
  ;;     the rowset size is greater than 1.
  (let* ((column-nr (desc-record-number record))
         (c-type (desc-c-type record))
         (data-ptr (desc-data-ptr record))
         (octet-length (desc-octet-length record))	; total buffer length
         (octet-len-ind-ptr (desc-octet-length-ptr record))
         type						; foreign buffer type
         max-valid-octet-length				; part of buffer with data
         lisp-type
         vector
         (res (%sql-get-data hstmt column-nr c-type 	; SQL_ARD_TYPE
                             data-ptr octet-length octet-len-ind-ptr))
         (out-len-ind (%get-integer octet-len-ind-ptr)))
    (when (= out-len-ind SQL_NULL_DATA)
      (return-from %read-datum-in-chunks *null*))

    ;; Differentiate over the return code of the first call.
    (case c-type
      (#.SQL_C_CHAR
       (setq type 'SQL-CHAR	;:char
             lisp-type #+lispworks (lw:external-format-type *foreign-ef*)
                       #-lispworks 'character
             max-valid-octet-length (1- octet-length)
             vector (if (= out-len-ind #.SQL_NO_TOTAL) ; don't know how long it will be
                        (make-array 0 :adjustable t :element-type lisp-type)
                        (make-string out-len-ind :element-type lisp-type))))
      #+Unicode
      (#.SQL_C_WCHAR
       (setq type 'SQL-WCHAR	;:wchar-t
             lisp-type #+lispworks 'lw:simple-char
                       #-lispworks 'character
             max-valid-octet-length (- octet-length 2)
             vector (if (= out-len-ind #.SQL_NO_TOTAL) ; don't know how long it will be
                        (make-array 0 :adjustable t :element-type lisp-type)
                        (make-string (truncate out-len-ind 2)
                                     :element-type lisp-type))))
      (#.SQL_C_BINARY
       (setq type 'SQL-OCTET
             lisp-type '(unsigned-byte 8)
             max-valid-octet-length octet-length
             vector (if (= out-len-ind #.SQL_NO_TOTAL) ; don't know how long it will be
                        (make-array 0 :adjustable t :element-type lisp-type)
                        (make-array out-len-ind :element-type lisp-type))))
      (otherwise
       (error 'sql-type-coercion-error :data-type (desc-sql-type record))))

    ;; Main loop
    (loop with offset fixnum = 0 	; lisp vector or string start
          until (= res SQL_NO_DATA)
          do (setf offset (%buffer-into-vector data-ptr type
                                               (if (= out-len-ind SQL_NO_TOTAL)
                                                   max-valid-octet-length
                                                   (min out-len-ind
                                                        max-valid-octet-length))
                                               vector offset))
          ;while (and (= res SQL_SUCCESS_WITH_INFO)
          ;           (equal (%sql-state +SQL-NULL-HANDLE+ +SQL-NULL-HANDLE+ hstmt)
          ;                  "01004"))
          do (setf res (%sql-get-data hstmt column-nr c-type ; SQL_ARD_TYPE
                                      data-ptr octet-length octet-len-ind-ptr)
                   out-len-ind (%get-integer octet-len-ind-ptr)))
    (case c-type
      ((#.SQL_C_CHAR #+Unicode #.SQL_C_WCHAR)
       (if (stringp vector) vector (coerce vector 'string)))
      (otherwise vector))))

(defun %write-datum-in-chunks (datum hstmt record) ; &optional (array-size 1)
 ;;; Feed datum for data-at-execution column or parameter
  ;; Value: Total octets transferred
  ;; NB: In case of error, SQLPutData transits to the state preceding the call of
  ;;     SQLSetPos or SQLBulkOperations.
  (let ((data-ptr (desc-data-ptr record))
        (octet-length (desc-octet-length record)))	; one item buffer length
    (if datum
        (case (desc-c-type record)
          (#.SQL_C_CHAR
           (unless (stringp datum) (setq datum (princ-to-string datum)))
           (loop with size = (length datum)
                 for start upfrom 0 below size by octet-length
                 for byte-count fixnum = (%put-string data-ptr datum octet-length
                                                      :start start)
                 unless (%sql-succeeded (%sql-put-data hstmt data-ptr byte-count))
                 do (loop-finish)
                 sum byte-count))
          #+Unicode
          (#.SQL_C_WCHAR
           (unless (stringp datum) (setq datum (princ-to-string datum)))
           (loop with size = (* 2 (length datum))
                 for start upfrom 0 below size by octet-length
                 for byte-count fixnum = (%put-wstring data-ptr datum octet-length
                                                       :start start)
                 unless (%sql-succeeded (%sql-put-data hstmt data-ptr byte-count))
                 do (loop-finish)
                 sum byte-count))
          (#.SQL_C_BINARY
           (loop with length = (length datum)
                 for start upfrom 0 below length by octet-length
                 for byte-count fixnum = (%put-binary data-ptr datum octet-length
                                                      :start start)
                 unless (%sql-succeeded (%sql-put-data hstmt data-ptr byte-count))
                 do (loop-finish)
                 sum byte-count)))
        (progn (%sql-put-data hstmt data-ptr SQL_NULL_DATA)
               0))))

(defun row-data-address (hstmt)
 ;;; Returns an address a 32-bit integer value associated with the data-at-exec column.
  ;; When array-size > 1, the value is not data-ptr address, but actually
  ;;	Bound Address + Binding Offset + ((Row Number  1) * Element Size)
  (with-temporary-allocation ((ptr sql-uinteger))
    (when (= (%sql-param-data hstmt ptr) SQL_NEED_DATA)
      (%get-ulong ptr))))

(defun %write-data-at-exec (data records hstmt descriptor)
                           ;&optional (count (desc-array-size descriptor)))
 ;;; Writes long data for unbound columns. Arguments:
  ;;  data	 List of long column values.
  ;;              (each element is either datum or sequence of length array-size.)
  ;;  records	 List of long data records from (desc-records descriptor)
  ;;  descriptor Row descriptor
  ;; Value: Number of octets written.
  (loop with last fixnum = (1- (desc-array-size descriptor)) ; index of last element
        for address = (row-data-address hstmt)
        while address
        sum      	; Search for corresponding record and keep track of its position
        (loop for record in records	
              and position fixnum upfrom 0
              for start = (%pointer-to-address (desc-data-ptr record))
              when (<= start address (+ start (* last (desc-octet-length record))))
              return (%write-datum-in-chunks (nth position data) hstmt record)
              finally (return 0))))	; return zero if no record found - very strange!


#||
(defun %reinitialize-rowset (cursor &key (size SQL_ROWSET_SIZE_DEFAULT))
 ;;; Reinitializing simply when the result set columns stay the same.
  ;(unless (>= $ODBCVER 0300) (setq size 1))
  (with-slots (handle row-descriptor cursor-type) cursor 
    (when (and (>= $ODBCVER 0300)
               (not (eql (desc-array-size row-descriptor) size)))
      (%sql-set-attribute :hstmt handle SQL_ATTR_ROW_ARRAY_SIZE
                          (setf (desc-array-size row-descriptor) size))
      (when (or (> size 1) (eq cursor-type :keyset-driven))
        (%sql-set-attribute :hstmt handle SQL_ATTR_ROW_STATUS_PTR
                            (desc-array-status-ptr row-descriptor)))
      (when (> size 1)
        (%sql-set-attribute :hstmt handle SQL_ATTR_ROWS_FETCHED_PTR
                            (desc-rows-processed-ptr row-descriptor))))
    (unless (or (> size 1) (eq cursor-type :keyset-driven))
      (setf (desc-row-status row-descriptor 0) SQL_ROW_SUCCESS))
    (when (= size 1)
      (setf (desc-rows-processed row-descriptor) 1))
    (setf (block-position row-descriptor) :start)	; shared-initialize?
    (allf 0 (fetched-count row-descriptor)
            (current-row-nr row-descriptor))
    (nilf (current-row-status row-descriptor)
          (last-operation row-descriptor))
) )
||#

