;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Object-oriented interface to SQL - Metaobject and variable definitions
;;; Copyright (c) 2002-2005 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Local time to YstokSQL/ODBC interface

(in-package :ystok.odbc)

;;; SQL Expression specifics

(defmethod output-sql ((expr local-time:date) (database odbc-database))
  (declare (ignore database))
  (write-string "{d '" *sql-stream*)
  (local-time:format-timestring *sql-stream* expr nil nil 4 0)
  (write-string "'}" *sql-stream*))

(defmethod output-sql ((expr local-time:local-time) (database odbc-database))
 ;;; Take the number of digits to the right of the decimal point in the seconds part
  ;; (fractional seconds) into account
  ;; As some drivers do not like #\, as a decimal separator (or pass it to back-ends
  ;; without translating), we use dot to conform to SQL standard.
  (write-string "{ts '" *sql-stream*)
  (let* ((type-descriptor (%get-type-descriptor database SQL_TYPE_TIMESTAMP))
         (maximum-scale (if type-descriptor (type-desc-maximum-scale type-descriptor)))
         (*decimal-separator* #\.))
    (local-time:format-timestring *sql-stream* expr nil nil 4
                                  (if (and maximum-scale (plusp maximum-scale)) 4 3)
                                  #\- #\: #\Space))
  (write-string "'}" *sql-stream*))


;; The value of the fraction field is the number of billionths of a second 
;; and ranges from 0 through 999,999,999 (1 less than 1 billion).
;; For example, the value of the fraction field for a half-second is 500,000,000, 
;; for a thousandth of a second (one millisecond) is 1,000,000,
;; for a millionth of a second (one microsecond) is 1,000, 
;; and for a billionth of a second (one nanosecond) is 1.
;;
;; Fraction is valuable only if 
;;	(>= (type-desc-minimum-scale (%get-type-descriptor a SQL_TYPE_TIMESTAMP)) 0)

(defparameter *local-time-decode-function* 'local-time:decode-as-local 
                                           ;'local-time:decode-as-utc
  "Function to decode a local-time instance when writing to the ODBC buffer
 of type TIME(STAMP) WITHOUT TIMEZONE")

;;; ODBC SQL_TYPE_TIMESTAMP
(defmacro timestamp-wo-zone-to-local-time (ptr)
  `(local-time:encode-local-time 
    (floor (%foreign-slot ,ptr sql-c-timestamp fraction) 1000000)
    (%foreign-slot ,ptr sql-c-timestamp second)
    (%foreign-slot ,ptr sql-c-timestamp minute)
    (%foreign-slot ,ptr sql-c-timestamp hour)
    (%foreign-slot ,ptr sql-c-timestamp day)
    (%foreign-slot ,ptr sql-c-timestamp month)
    (%foreign-slot ,ptr sql-c-timestamp year)
    local-time:*session-timezone* local-time:*default-encode-type* nil))

(defmethod time-to-timestamp-wo-zone (ptr (time local-time:date))
  (multiple-value-bind (year month day) (local-time:decode-date time)
    (setf (%foreign-slot ptr sql-c-timestamp fraction) 0
          (%foreign-slot ptr sql-c-timestamp second) 0
          (%foreign-slot ptr sql-c-timestamp minute) 0
          (%foreign-slot ptr sql-c-timestamp hour) 0
          (%foreign-slot ptr sql-c-timestamp day) day
          (%foreign-slot ptr sql-c-timestamp month) month
          (%foreign-slot ptr sql-c-timestamp year) year)))

(defmethod time-to-timestamp-wo-zone (ptr (time local-time:local-time))
  (multiple-value-bind (msec second minute hour day month year) ;(decode-as-local time)?
      (funcall *local-time-decode-function* time)
    (setf (%foreign-slot ptr sql-c-timestamp fraction) (* msec 1000000)
          (%foreign-slot ptr sql-c-timestamp second) second
          (%foreign-slot ptr sql-c-timestamp minute) minute
          (%foreign-slot ptr sql-c-timestamp hour) hour
          (%foreign-slot ptr sql-c-timestamp day) day
          (%foreign-slot ptr sql-c-timestamp month) month
          (%foreign-slot ptr sql-c-timestamp year) year)))

;;; ODBC SQL_TYPE_DATE
(defmacro date-to-local-time (ptr)
  `(local-time:encode-date
    (%foreign-slot ,ptr sql-c-timestamp year)
    (%foreign-slot ,ptr sql-c-timestamp month)
    (%foreign-slot ,ptr sql-c-timestamp day)
    nil))

(defmethod time-to-date (ptr (time local-time:date))
 ;;; lt:date -> SQL_TYPE_DATE
  (multiple-value-bind (year month day) (local-time:decode-date time)
    (setf (%foreign-slot ptr sql-c-date day) day
          (%foreign-slot ptr sql-c-date month) month
          (%foreign-slot ptr sql-c-date year) year)))

(defmethod time-to-date (ptr (time local-time:local-time))
 ;;; lt:local-time -> SQL_TYPE_DATE
  (multiple-value-bind (msec second minute hour day month year)
      (funcall *local-time-decode-function* time)
    (declare (ignore msec second minute hour))
    (setf (%foreign-slot ptr sql-c-date day) day
          (%foreign-slot ptr sql-c-date month) month
          (%foreign-slot ptr sql-c-date year) year)))

;;; ODBC SQL_TYPE_TIME
(defmacro time-wo-zone-to-local-time (ptr)
 ;;; Q: Shall not we just return local-time-sec?
  `(local-time:encode-local-time
    0 ;(floor (%foreign-slot ,ptr sql-c-timestamp fraction) 1000000)
    (%foreign-slot ,ptr sql-c-time second)
    (%foreign-slot ,ptr sql-c-time minute)
    (%foreign-slot ,ptr sql-c-time hour)
    0 0 0	; @00-1-11-30! Do we need to introduce time-day, time-month and time-year?
    local-time:*session-timezone* 'local-time:local-time nil))

(defmethod time-to-time-wo-zone (ptr (time local-time:local-time))
  ;; lt:local-time -> SQL_TYPE_TIME
  (multiple-value-bind (msec second minute hour) ;(lt:decode-as-local time)?
      (funcall *local-time-decode-function* time)
    (declare (ignore msec day month year))
    (setf (%foreign-slot ptr sql-c-time second) second
          (%foreign-slot ptr sql-c-time minute) minute
          (%foreign-slot ptr sql-c-time hour) hour)))
