;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2005 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Parameterized execution and prepared statements
;;;
;;; NB1: You cannot set the concurrency level on a prepared statement. 
;;; NB2: If a statement has been prepared and you cannot _unprepare_ it,
;;;	 you need SQLCancel().

(in-package :ystok.odbc)

(defun %make-param-desc-record (database hstmt array-size pos parameter value)
 ;;; Args: array-size  Paramset size - number of parameter tuples to set or retrieve.
  ;;       parameter   sql-parameter marker or something else?
  (let* ((plist (and parameter (plist parameter)))
         (lisp-type (getf plist :type))
         sql-type c-type column-size decimal-digits nullable)
    (cond ((or lisp-type value)
           (multiple-value-setq (c-type sql-type column-size decimal-digits nullable)
               (%describe-lisp-object database value lisp-type)))
          ((%sql-func-exists database SQL_API_SQLDESCRIBEPARAM)	; does not in Access
           ;; Oracle always describes parameters generated by an SQL expression
           ;; as of type SQL_VARCHAR.
           (multiple-value-setq (sql-type column-size decimal-digits nullable)
               (%sql-describe-param hstmt (1+ pos))))
          (t
           (multiple-value-setq (c-type sql-type column-size decimal-digits nullable)
               (%describe-lisp-object database value 'string))))
    (%make-desc-record database lisp-type array-size (1+ pos) (getf plist :name)
                       c-type sql-type column-size decimal-digits nullable
                       ;; Jet 4.0: Only input parameters
                       (ecase (getf plist :direction :input)
                         (:input  SQL_PARAM_INPUT)
                         (:output SQL_PARAM_OUTPUT)
                         (:io	  SQL_PARAM_INPUT_OUTPUT))) ))

(defun %make-param-descriptor (stmt &optional (array-size 1) parameters values)
  (with-slots (handle database prepared) stmt
    (let ((records (make-array 0 :adjustable t :fill-pointer t)))
      (dotimes (pos (if prepared
                        (%sql-params-count handle)
                        (length (or parameters values))))
        (vector-push-extend (%make-param-desc-record database handle array-size pos 
                                                     (pop parameters) (pop values))
                            records))
      ;(loop record accross records
      ;      maximize (* (desc-octet-length record) array-size))
      ; (setf array-size (min (floor 32767 (1+ max-size)) *max-row-fetch-count*))
      (make-instance 'descriptor
           :count (fill-pointer records)
           :array-size array-size
           :array-status-ptr (%new-ptr 'sql-usmallint array-size)
           :rows-processed-ptr (%new-ptr 'sql-uinteger)
           :records records))))

(defun %initialize-paramset (stmt parameters values &key (size 1))
 ;;; Initialize descriptor records before (prepared = nil) or after preparing stmt.
  (unless (>= $ODBCVER 0300) (setq size 1))
  (with-slots (handle param-descriptor) stmt
    (setf param-descriptor (%make-param-descriptor stmt size
                              (or parameters (collect-parameters (expression stmt)))
                              values))
    (when (> size 1)
      ;; Works fine with Oracle 8.1.7. Is not supported by Jet 3.5 or 4.0.
      (%sql-set-attribute :hstmt handle SQL_ATTR_PARAMSET_SIZE size)
      ;; In Oracle due to SQL_PARC_NO_BATCH, parameter status array is not filled.
      (%sql-set-attribute :hstmt handle SQL_ATTR_PARAM_STATUS_PTR
                                        (desc-array-status-ptr param-descriptor))
      ;(%put-ushort (desc-array-status-ptr param-descriptor) SQL_PARAM_PROCEED)
      (%sql-set-attribute :hstmt handle SQL_ATTR_PARAMS_PROCESSED_PTR
                                        (desc-rows-processed-ptr param-descriptor)))
    ;; Bind input parameters in all cases. However for long data, we will set 
    ;; the contents of indicator-ptr (or octet-length-ptr) to SQL_LEN_DATA_AT_EXEC
    ;; before executing in order to feed in chunks with %sql-put-data.
    ;; NB1: Jet 3.5 does not allow SQL_VARCHAR as a parameter type!
    ;; NB2: As reported by Roland Sevin (upsis<at>club<tiret>internet<point>fr),
    ;;	    column-size is not necessary for all drivers, e.g. MySQL can work without it,
    ;;	    but MsAccess or DBase cannot.
    (loop for record across (desc-records param-descriptor)
          do (%sql-bind-parameter handle
                                  (desc-record-number record)
                                  (desc-parameter-type record)
                                  (desc-c-type record)
                                  (desc-sql-type record)
                                  ;(if (desc-long-p record) 0 
                                  (desc-column-size record)
                                  (desc-decimal-digits record)
                                  (desc-data-ptr record)
                                  (desc-octet-length record)
                                  (desc-octet-length-ptr record))) ))

(defun %execute-parameterized (stmt expr parameters values &optional paramset-size timeout)
 ;;; Set parameter values and execute
  ;; Args: expr  If null, do not bind params - execute the statement with new values.
  ;; Value: number of tuples of parameters processed.
  (with-slots (handle database expression param-descriptor active) stmt
    ;(when (not active) (error "Statement ~s is inactive." stmt))
    (cond ((and expr (not (prepared stmt)))		; get ready for direct execution
           (setf expression expr)			; i.e. not calling %sql-prepare
           (%initialize-paramset stmt parameters values	; fill in param-descriptor
                                 :size (orf paramset-size 1)))
          ((null paramset-size)
           (setq paramset-size (desc-array-size param-descriptor))))
    (when param-descriptor
      ;; Fill in length/indicator and feed data for "short" parameters
      (loop with sql-need-long-data-len = (db-get-info database SQL_NEED_LONG_DATA_LEN)
            for record across (desc-records param-descriptor)
            and value in values
            if (> paramset-size 1)
            ;; Is this processing order pertinent to column-wise parameter binding
            do (loop for index upfrom 0 below paramset-size
                     for item = (elt value index)
                     if (desc-long-p record)
                     do (setf (%get-integer (desc-indicator-ptr record) index)
                               (if sql-need-long-data-len
                                   (%sql-len-data-at-exec (length item))
                                   SQL_DATA_AT_EXEC))
                     else do (%write-datum item record index))
            else if (desc-long-p record)
            do (setf (%get-integer (desc-indicator-ptr record))
                      (if sql-need-long-data-len
                          (%sql-len-data-at-exec (length value))
                          SQL_DATA_AT_EXEC))
                                 ;(if (consp parameter)
                                 ;    (cadr parameter) ;; size of stream
                                 ;    (min (print (length parameter)) +max-precision+)
            else do (%write-datum value record))
      (when timeout
        (%sql-set-attribute :hstmt handle SQL_ATTR_QUERY_TIMEOUT timeout))
      (when (eql (if (prepared stmt)
                     (%sql-execute handle)
                     (%sql-exec-direct handle (write-sql-to-string expr database)))
                 SQL_NEED_DATA)
        ;; Feed data for data-at-execution parameters.
        ;; Cancel the statement if something wrong happens.
        (setf active :need-data)
        (unwind-protect
            (loop with records = (desc-records param-descriptor)
                  with pointer = (parameter-data-ptr handle)
                  while pointer
                  for position = (position pointer records
                                       :test #'ffc:%ptr-eql  ;(and (fli:pointerp ar1))
                                       :key #'desc-data-ptr)
                  ; (assert position (position) "Parameter descriptor record not found")
                  for record = (aref records position)
                  if (> paramset-size 1)		; loop over the sequence value
                  do (loop with value = (nth position values)
                           with ptr = nil
                           for index upfrom 0 below paramset-size
                           do (%write-datum-in-chunks (elt value index) handle record)
                           while (and (setf ptr (parameter-data-ptr handle))
                                      (ffc:%ptr-eql ptr pointer))
                           finally (setf pointer ptr))
                  else do (%write-datum-in-chunks (nth position values) handle record)
                  and do (setf pointer (parameter-data-ptr handle))
                  finally (setf active t))
          ;; CAUTION!
          ;(when (eq active :need-data) (cancel-statement stmt))
          (setf active :need-data)
         ))
      (if (> paramset-size 1)
          (%get-uinteger (desc-rows-processed-ptr param-descriptor))
          1))))

(defun parameter-data-ptr (hstmt)
 ;;; Returns pointer to data used as a 32-bit value associated with the data-at-exec
  ;; parameter. 
  ;; When paramset-size > 1, this functions works as follows.
  ;;  Oracle: Returns data-ptr value paramset-size times one after the other for the 
  ;;          first parameter then switches to data-ptr of the second and so on.
  ;;  Other DBMS: ?
  (with-temporary-allocation ((ptr :ptr))
    (when (= (%sql-param-data hstmt ptr) SQL_NEED_DATA)
      (%get-ptr ptr))))

;;; Links to functional SQL interface

(defmethod db-prepare-statement ((database odbc-database) expr
                                 &key parameters values (paramset-size 1) timeout)
  (db-prepare-statement (obtain-statement database
                          (typecase expr
                            (sql-query-exp 'odbc-param-cursor)
                            (string (if (and (>= (length expr) 6)
                                             (equal (subseq expr 0 6) "SELECT"))
                                        'odbc-param-cursor
                                        'odbc-param-statement))
                            (t 'odbc-param-statement)))
                        expr
                        :parameters parameters :values values
                        :paramset-size paramset-size
                        :timeout timeout))

(defmethod db-prepare-statement ((stmt odbc-param-statement) expr
                                 &key parameters values (paramset-size 1) timeout
                                 &allow-other-keys)
  (with-slots (handle expression prepared) stmt
    (setf expression expr)
    (when timeout
      (%sql-set-attribute :hstmt handle SQL_ATTR_QUERY_TIMEOUT timeout))
    (%sql-prepare handle (write-sql-to-string expr (database stmt)))
    (setf prepared t))
  (%initialize-paramset stmt parameters values :size paramset-size)
  stmt)

(defmethod db-prepare-statement ((cursor odbc-param-cursor) query-expression
                                 &key ;parameters values (paramset-size 1)
                                      (rowset-size 1)
                                      (start-column 1) end-column column-numbers
                                 &allow-other-keys)
  (call-next-method)
  (%initialize-rowset cursor :size rowset-size
                             :start-column start-column :end-column end-column
                             :column-numbers column-numbers)
  cursor)

(defmethod db-close :before ((stmt odbc-param-statement) &key &allow-other-keys)
  (with-slots (handle param-descriptor prepared) stmt
    (when handle (%sql-free-statement handle :reset))
    (when param-descriptor
      (%dispose-descriptor param-descriptor t)
      (nilf param-descriptor))
    (nilf prepared)))

;(defmethod db-cancel :after ((stmt odbc-param-statement))
;  (with-slots (prepared) stmt
;    (setf prepared nil)))

(defmethod db-execute-command ((stmt odbc-param-statement) expr
                               &key parameters values paramset-size timeout)
  (let ((result (%execute-parameterized stmt expr parameters values paramset-size timeout)))
    (values result (%sql-rows-count (handle stmt)))))

(defmethod db-open-cursor ((cursor odbc-param-cursor) query-expression
                           &key ;(cursor-type :forward-only)
                                ;concurrency
                                (rowset-size 1)
                                start-column end-column column-numbers
                                parameters values (paramset-size 1)
                                timeout
                           &allow-other-keys)
  (%execute-parameterized cursor query-expression parameters values paramset-size timeout)
  (unless (prepared cursor)
    (%initialize-rowset cursor :size rowset-size
                               :start-column start-column :end-column end-column
                               :column-numbers column-numbers))
  cursor)

(defmethod db-map-parameters ((database odbc-database) result-type expression
                              &rest sequences)
 ;;; Executes a parameterized statement over values taken from each sequence in turn.
  ;; Arguments:
  ;; RESULT-TYPE : Performs a type coersion of the value returned like in MAP.
  ;; EXPRESSION : Arbitrary statement expression with as many parameters markers
  ;;              as there are sequences.
  ;; SEQUENCES : Parameter values forming a paramset with the size equal to the
  ;;             length of the shortest.
  (with-prepared-statement (stmt expression :database database)
    (apply #'db-map-parameters stmt result-type nil sequences)))

(defmethod db-map-parameters ((stmt odbc-param-statement) result-type expression
                              &rest sequences)
  (apply #'map result-type
         (lambda (&rest values) (db-execute-command stmt nil :values values))
         sequences))

(defmethod db-map-parameters ((cursor odbc-param-cursor) result-type expression
                              &rest sequences)
  (apply #'map result-type
         (lambda (&rest values)
           (prog2
               (db-open-cursor cursor nil :values values)
               (db-fetch-rows cursor)
             (%sql-free-statement (handle cursor) :close))) ; close cursor
         sequences))

