;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2005 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; SQL to ODBC interface

(in-package :ystok.odbc)

(defmethod db-type-initialize ((db-type (eql :odbc)) &optional version)
  ;#+:mk (mk:load-system :odbc)  
  ;#+(and (not :mk) :lispworks) (lw:load-system :odbc)
  (make-instance 'odbc-environment :db-type db-type))

(defmethod initialize-instance :after ((self odbc-environment)
                                       &key (version #+odbc3.0 SQL_OV_ODBC3
                                                     #-odbc3.0 SQL_OV_ODBC2)
                                       &allow-other-keys)
 ;;; When the call fails, should we decrease SQL_ATTR_ODBC_VERSION to SQL_OV_ODBC2?
  (unless (henv self)                               
    (let ((henv (%sql-alloc-handle :henv)))
      (%sql-set-attribute :henv henv SQL_ATTR_ODBC_VERSION version)
      (setf (henv self) henv))))

(defmethod initialize-instance :after ((db odbc-database) &key db-type
                                                          &allow-other-keys)
  (declare (ignore db-type))
  (unless (handle db)
    (setf (handle db) (%sql-alloc-handle :hdbc (henv db)))))

(eval-when (:compile-toplevel :load-toplevel :execute)
  (defconstant +connection-spec-template+ '(&optional dsn &key user password)
    "Default connection specification template"))

(defmethod db-name-from-spec ((database-type (eql :odbc)) (connection-spec list))
  (check-connection-spec connection-spec database-type #.+connection-spec-template+)
  (destructuring-bind #.+connection-spec-template+ connection-spec
    (declare (ignore user password))
    dsn))

(defmethod db-user-name-from-spec ((database-type (eql :odbc)) (connection-spec list))
  (check-connection-spec connection-spec database-type #.+connection-spec-template+)
  (destructuring-bind #.+connection-spec-template+ connection-spec
    (declare (ignore dsn password))
    user))

(defmethod sql.sys::db-password-from-spec ((database-type (eql :odbc)) (connection-spec list))
  (check-connection-spec connection-spec database-type #.+connection-spec-template+)
  (destructuring-bind #.+connection-spec-template+ connection-spec
    (declare (ignore dsn user))
   password))

(defmethod db-connect ((db-type (eql :odbc)) (connection-spec string) 
                       &rest args &key &allow-other-keys)
 ;;; Convert a standard CommonSQL specification string to the 'right' one.
  ;(declare (ignore args))
  (apply #'db-connect
         (find-environment db-type)
         (list (db-name-from-spec db-type connection-spec)
               :user (db-user-name-from-spec db-type connection-spec)
               :password (sql.sys::db-password-from-spec db-type connection-spec))
         args))

(defmethod db-connect ((db-type (eql :odbc)) (connection-spec list) &rest args &key)
  (apply #'db-connect (find-environment db-type) connection-spec args))

(defmethod db-connect ((db-type (eql :odbc)) (connection-spec pathname) &rest args &key)
  (apply #'db-connect (find-environment db-type) connection-spec args))

(defmethod db-connect ((environment odbc-environment) (connection-spec pathname)
                       &rest args &key)
  (apply #'db-connect environment (list connection-spec) args))

(defmethod db-connect ((environment odbc-environment) connection-spec &rest args &key)
 ;;; NB: As user and password can be supplied with both connection-spec and args, 
  ;;	 merge these credentials while giving preference to args.
  (check-connection-spec connection-spec environment #.+connection-spec-template+)
  (destructuring-bind #.+connection-spec-template+ connection-spec
    (let* ((uid (or (getf args :user) user))
           (pwd (or (getf args :password) password))
           (db (make-instance 'odbc-database ;:db-type :odbc
                              :environment environment
                              :spec connection-spec :name dsn
                              :user uid :password pwd)))
      ;; If this is the first connection opened, find out exact Driver Manager version
      (unless (sql.sys::connections environment)
        (when-bind (version (read-from-string (remove #\.	; e.g. "03.52.0000"
                                               (%sql-get-info (handle db) SQL_ODBC_VER))
                                              nil nil :end 4))
          (setq $ODBCVER version)))

      ;; If connection cannot be established, we are thrown out of here.
      (apply #'db-connect db dsn :user uid :password pwd
             (remove-properties args '(:user :password))))))


(defmethod db-connect ((db odbc-database) dsn &key user password
                                                   autocommit
                                                   (unicode *unicode*)
                                                   driver-av-pairs database-type
                                                   timeout)
 ;;; Establish database class eventually; can also be used to reconnect
  ;; Args:
  ;;  autocommit  Oracle recommends manual commit or rollback operations using SQLEndTran
  ;;		  for improving concurrency and performance.
  ;;  driver-av-paris  nil, string or pair list with driver-specific arguments.
  ;;  password	  Any will fit when the empty password was set for the user-name in
  ;;		  Access system .mdw-database (where groups and user accounts are stored).
  ;;		  This works similarly both via SQLConnect and SQLDriverConnect.
  ;;  timeout	  Number of seconds to wait for a login request to complete before
  ;;		  returning to the application;
  ;;		  0 = timeout is disabled and a connection attempt will wait indefinitely.
  (when timeout
    (%sql-set-attribute :hdbc (handle db) SQL_ATTR_LOGIN_TIMEOUT timeout))
  (when (cond ((or driver-av-pairs
                   (null dsn)
                   (pathnamep dsn))
                   ;(uri:uri-p dsn))
               (setf (connected db) (driver-connect db database-type driver-av-pairs
                                                    :unicode unicode
                                                    :dsn dsn :uid user :pwd password)))
              (t
               (if (and #-Unicode nil unicode (>= $ODBCVER 0350))
                   (%sql-connectw (handle db) dsn user password)
                   (%sql-connect (handle db) dsn user password))
               ;; If we got this far without erring out, the database was 
               ;; successfully connected.
               (setf (connected db) t)))
    (let ((server-name (db-get-info db SQL_SERVER_NAME))
          (dbms-name (db-get-info db SQL_DBMS_NAME))
          (dbms-version (db-get-info db SQL_DBMS_VER)))
      ;; need SERVER-NAME and DBMS-NAME because many drivers mix this up
      (flet ((dbms-type-p (db-name)
               (or (search db-name server-name :test #'char-equal)
                   (search db-name dbms-name :test #'char-equal))))
        (cond ((dbms-type-p "Access")
               (change-class db (if (search "03.50" dbms-version)
                                    (or (find-class 'access-3.5-database nil)
                                        'access-odbc-database)
                                    'access-odbc-database)))
              ((dbms-type-p "Oracle")
               (change-class db 'oracle-odbc-database))
              ((dbms-type-p "PostgreSQL")
               (change-class db 'postgresql-odbc-database))
              ((dbms-type-p "SQLite")
               (change-class db 'sqlite3-odbc-database))
              ((dbms-type-p "Excel")
               (change-class db 'excel-odbc-database))
              ((dbms-type-p "MSSQL")
               (change-class db 'mssql-database))
              ((dbms-type-p "MySQL")
               (change-class db 'mysql-odbc-database))))
      (when (/= (db-get-info db SQL_TXN_CAPABLE) SQL_TC_NONE)
      	;; has transaction support
        (%sql-set-attribute :hdbc (handle db) SQL_ATTR_AUTOCOMMIT 
                            (if autocommit SQL_AUTOCOMMIT_ON SQL_AUTOCOMMIT_OFF))))
    (%initialize-type-info db)
    db))

(defgeneric driver-connect (db be-type driver-av-pairs
                            &key hwnd completion-mode dsn uid pwd &allow-other-keys)
 (:documentation "Use SQLDriverConnect and invoke an interactive dialog.
Arguments:
 DRIVER-AV-PAIRS : driver attributes as list ((attr value)...) or string \"attr=value;...\".
 BE-TYPE : specific back-end type, e.g. :access (useful for specializing on).
 COMPLETION-MODE : one of :complete, :required, :prompt, :noprompt."))

(defmethod driver-connect ((db odbc-database) be-type (driver-av-pairs string) &rest args
                           &key)
  (apply #'driver-connect db be-type (parse-driver-attributes driver-av-pairs) args))

(defmethod driver-connect ((db odbc-database) be-type (driver-av-pairs list)
                           &key (hwnd t) (completion-mode :complete)
                                dsn uid pwd (unicode *unicode*))
 ;;; Args: driver-av-pairs List of (attribute value) pairs
  (declare (ignore be-type))
  (when-bind (completion (funcall (if (and #-Unicode nil unicode (>= $ODBCVER 0350))
                                      #'%sql-driver-connectw
                                      #'%sql-driver-connect)
                                  (hdbc db) (%ensure-hwnd hwnd)
                                  (make-driver-attributes-string
                                   (merge-av-pairs driver-av-pairs
                                                   :dsn dsn :uid uid :pwd pwd))
                                  completion-mode))
    ;; Refresh database object slots to match completed connection attributes
    (let ((av-pairs (parse-driver-attributes completion)))
      (with-accessors ((name database-name)
                       (user-name user-name)
                       (password password)) db
        (flet ((non-string-null (x) (and x (imp (stringp x) (plusp (length x))))))
          (unless (non-string-null name)
            (setf name (second (or (assoc :dsn av-pairs :test #'string-equal)
                                   (assoc :dbq av-pairs :test #'string-equal)))))
          (unless (non-string-null user-name)
            (setf user-name (second (assoc :uid av-pairs :test #'string-equal))))
          (unless (non-string-null password)
            (setf password (second (assoc :pwd av-pairs :test #'string-equal)))) )))
    t))


(defmethod db-disconnect ((database odbc-database))
  (dolist (stmt (statements database))
    (db-close stmt :drop-p t))
    ;(when (active stmt)
    ;  (with-slots (handle) stmt
    ;    (when hstmt 
    ;      (%sql-free-handle :hstmt handle) ;(%free-statement hstmt :drop)
    ;      (setf handle nil)))))
  (with-slots (handle) database
    (%sql-disconnect handle)
    (%sql-free-handle :hdbc handle)
    (setf handle nil)))

;;; Transaction-related

(defparameter *access-mode-alist* '((:read-write . #.SQL_MODE_READ_WRITE)
                                    (:read-only  . #. SQL_MODE_READ_ONLY)))

(defparameter *isolation-level-alist* '((:uncommitted  . #.SQL_TXN_READ_UNCOMMITTED)
                                        (:committed    . #.SQL_TXN_READ_COMMITTED)
                                        (:repeatable   . #.SQL_TXN_REPEATABLE_READ)
                                        (:serializable . #.SQL_TXN_SERIALIZABLE)))

(defmethod db-start-transaction ((database odbc-database) &optional access isolation)
 ;;; Access NB: If any cursor is open, signals
  ;; 		 [ODBC state HY011 error 72].
  ;;		 [Microsoft][ODBC Microsoft Access Driver]Attribute cannot be set now.
  ;;  In what way does Access itself manage to combine bound forms with transactions?
  ;;  Use another connection?
  (declare (ignore mode isolation))
  (unless (= (db-get-info database SQL_TXN_CAPABLE) SQL_TC_NONE)
    (with-slots (handle) database
    (let ((autocommit (= (%sql-get-attribute :hdbc handle SQL_ATTR_AUTOCOMMIT)
                         SQL_AUTOCOMMIT_ON))
          (old-access nil)
          (old-isolation nil))
      (when autocommit
        (%sql-set-attribute :hdbc handle SQL_ATTR_AUTOCOMMIT SQL_AUTOCOMMIT_OFF))
      (when access
        (setq old-access (car (rassoc (%sql-get-attribute :hdbc handle
                                                          SQL_ATTR_ACCESS_MODE)
                                      *access-mode-alist*)))
        (%sql-set-attribute :hdbc handle SQL_ATTR_ACCESS_MODE
                            (cdr-assoc access *access-mode-alist*)))
      (when isolation
        (setq old-isolation (car (rassoc (%sql-get-attribute :hdbc handle
                                                             SQL_ATTR_TXN_ISOLATION)
                                         *isolation-level-alist*)))
        (%sql-set-attribute :hdbc handle SQL_ATTR_TXN_ISOLATION
                            (cdr-assoc isolation *isolation-level-alist*)))
      (values autocommit old-access old-isolation)))))

(defmethod db-end-transaction ((database odbc-database) completion-type
                               &optional autocommit access isolation)
  (%sql-end-transaction :hdbc (handle database) completion-type))

(defmethod db-end-transaction :after ((database odbc-database) completion-type
                                      &optional autocommit access isolation)
  (with-slots (handle) database
    (when autocommit
      (%sql-set-attribute :hdbc handle SQL_ATTR_AUTOCOMMIT SQL_AUTOCOMMIT_ON))
    (when access
      (%sql-set-attribute :hdbc handle SQL_ATTR_ACCESS_MODE
                          (cdr-assoc access *access-mode-alist*)))
    (when isolation
      (%sql-set-attribute :hdbc handle SQL_ATTR_TXN_ISOLATION
                          (cdr-assoc isolation *isolation-level-alist*))) ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Statements

(defmethod initialize-instance :after ((stmt odbc-statement) &key &allow-other-keys)
  (with-slots (handle) stmt
    (unless handle	;     shared-initialize?
      (setf handle (%sql-alloc-handle :hstmt (handle (database stmt)))))))

(defmethod reinitialize-instance :after ((stmt odbc-statement) &key &allow-other-keys)
  (with-slots (handle) stmt
    (unless handle
      (setf handle (%sql-alloc-handle :hstmt (handle (database stmt)))))))

(defmethod obtain-statement (database statement-class
                             &key #+lispworks (predicate 'lw:true))
  "Create new or find a passive statement object and reuse it by switching to active.
 This makes the functions db-execute-command and db-query thread safe."
  ;; To Do: Does keeping forward-only and keyset-driven cursors separately
  ;;	    instead of altering attributes give any performance benefit?
  (or (without-scheduling 			; not context switch allowed here 
        (when-bind (passive-stmt (find-if (lambda (stmt)
                                            (and (not (active stmt))
                                                 (eq (type-of stmt) statement-class)
                                                 #+lispworks (funcall predicate stmt)))
                                          (statements database)))
          (setf (active passive-stmt) t)
          ;(reinitialize-instance passive-stmt :active t)
          passive-stmt))
      (make-instance statement-class :database database
                                     :active t)))

(defmethod db-execute-command ((database odbc-database) expression
                               &key parameters values timeout)
  (if (or parameters values)
      (db-execute-command (obtain-statement database 'odbc-param-statement)
                          expression
                          :parameters parameters :values values  :timeout timeout)
      (db-execute-command (obtain-statement database 'odbc-statement)
                          expression
                          :timeout timeout)))

(defmethod db-execute-command ((stmt odbc-statement) expr
                               &key timeout &allow-other-keys)
  (with-slots (handle database expression) stmt
      ;(unless hstmt (setf hstmt (%sql-alloc-handle :hstmt hdbc)))
    (unwind-protect
        (progn
          (when expr
            (setf expression expr))
          (when-bind (string (write-sql-to-string expression database))
            (when timeout
              (%sql-set-attribute :hstmt handle SQL_ATTR_QUERY_TIMEOUT timeout))
            (let ((result (%sql-exec-direct handle string)))
              (values result (%sql-rows-count handle)))))
      (db-close stmt))))

(defmethod db-close ((stmt odbc-statement) &key drop-p &allow-other-keys)
  (with-slots (handle active) stmt
    (case active
      ((:need-data :still-executing)
       (db-cancel stmt)))
    (when (and handle drop-p)
      (%sql-free-handle :hstmt handle) ;(%free-statement hstmt :drop)
      (setf handle nil)
      (setf (statements (database stmt)) (delete stmt (statements (database stmt)))))
    (unless (eq active :closed)		; allow relinquishing if state is :unbound or t
      (setf active nil)))
  stmt) 

;;; PM: I am not quite sure how useful this is. I have tried to cancel a query from a
;;;	different thread, but this does not seem to work. On the other hand, Command-.
;;;	(in MCL) does cancel a statement (and obviously does not use cancel-query).
;;;	How does this work? Idea: look into SQLSetConnectOption.

(defmethod db-cancel ((stmt odbc-statement))
 ;;; Cancel the following types of processing on a statement: 
  ;; - A function running asynchronously on the statement.
  ;; - A function on a statement that needs data.
  ;; - A function running on the statement on another thread. 
  ;; Can signal SQL_ERROR if some function was executing.
  ;; Value:  SQL_SUCCESS If the cancel request was processed.
  ;;		The application must continue to call the original function
  ;;		until the return code is not SQL_STILL_EXECUTING.
  ;;		If that function was successfully canceled, the return code is
  ;;		SQL_ERROR and SQLSTATE HY008 (Operation canceled).
  (with-slots (handle active) stmt
    (when (%sql-succeeded (%sql-cancel handle))
      (setf active t))))

;;; Cursor-involving operations

(defmethod db-open-cursor ((database odbc-database) query-expression &rest args
                           &key ;(start-column 1) end-column column-numbers
                                parameters values
                           &allow-other-keys)
  ;(declare (ignore start-column end-column column-numbers))
  (if (or parameters values)
      (apply #'db-open-cursor (obtain-statement database 'odbc-param-cursor)
             query-expression args)
      (apply #'db-open-cursor (obtain-statement database 'odbc-cursor)
             query-expression args)))

(defmethod db-query ((database odbc-database) query-expression
                     &key flatp result-types parameters values timeout)
  (let ((cursor (obtain-statement database (if (or parameters values)
                                               'odbc-param-cursor
                                               'odbc-cursor))))
    (unwind-protect       
        (db-query cursor query-expression :flatp flatp  :result-types result-types
                  :parameters parameters :values values :timeout timeout)
      (db-close cursor))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  Catalog/metadata inquiry functions

(defmacro with-cursor-handle-results ((handle database &rest rowset-args)
                                      &body odbc-call)
 ;;; Returns the entire result set of the pseudo-query performed by odbc-call
  ;; Usually there is no point to return column names when rowset-args are given.
  ;; Set cursor attibutes to forward-only in order for Oracle not to complain
  ;; when we obtain an old keyset-driven statement:
  ;;	[ODBC state 01S02] [Oracle][ODBC]Option value changed.
  (with-gensyms (cursor)
    `(let* ((,cursor (obtain-statement ,database 'odbc-cursor))
            (,handle (handle ,cursor)))
       (unwind-protect       
           (progn
             (setf (expression ,cursor) ',(caar odbc-call))
             (%initialize-cursor ,cursor :forward-only)
             ,@odbc-call
             (%initialize-rowset ,cursor ,@rowset-args)
             ,(if rowset-args
                  `(db-fetch-rows ,cursor)
                  `(values (db-fetch-rows ,cursor)
                           (map 'list #'desc-name
                                (desc-records (row-descriptor ,cursor))))))
         (db-close ,cursor)))))

(defmethod db-column-type ((database odbc-database) column table schema &key catalog)
 ;;; Values: 1> Native SQL type specifier, 2> column size, 3> decimal digits, 4> nullable
  ;; NB: With some drivers, SQLColumns adds something in parentheses to TYPE_NAME, e.g.
  ;;      Oracle : "TIMESTAMP(6)" 
  ;;      SQLite3 : "VARCHAR(30)"
  ;;	 The first value is returned with this kind of stuff stripped off.
  (when-bind (record (first	; 6 TYPE_NAME, 7 COLUMN_SIZE, 9 DECIMAL_DIGITS, 11 NULLABLE
                      (with-cursor-handle-results (handle database
                                                   :column-numbers '(6 7 9 11))
                        (%sql-columns handle
                                      (name catalog)
                                      (if (eq schema :all)
                                          SQL_ALL_SCHEMAS
                                          (ordinary-argument schema))
                                      (ordinary-argument table)
                                      (ordinary-argument column)))))
    (destructuring-bind (type-name column-size decimal-digits nullable) record
      (values (let* ((position (position #\( type-name :test #'char=))
                     (keyword (intern (nstring-upcase (if position
                                                          (subseq type-name 0 position)
                                                          type-name))
                                      :keyword)))
                (cond ((memq keyword '(:char :varchar :binary :varbinary))
                       (list keyword column-size))
                      ((memq keyword '(:decimal :numeric :number))
                       (list* keyword column-size
                              (and decimal-digits (list decimal-digits))))
                      ((and (eq keyword :timestamp) decimal-digits)
                       (list keyword decimal-digits))
                      (t keyword)))
              column-size decimal-digits nullable))))

(defmethod db-list ((environment odbc-environment) connection-spec &key (type :all))
 "Returns a list of (data-source-name driver-name) pairs.
Args: type ::= :all | :user | :system."
   (with-slots (handle) environment
     (loop for direction = (or type :first) then :next
           for data = (multiple-value-list (%sql-data-sources handle
                                                              :direction direction))
           while (car data)
           collect data)))

(defmethod db-list ((db-type (eql :odbc)) connection-spec &key type)
  (db-list (find-environment db-type) connection-spec :type type))

(defmethod db-drivers ((db-type (eql :odbc)))
 "Returns a list of the form (driver-name . av-pairs)."
   (with-slots (handle) (find-environment db-type)
     (loop for direction = :first then :next
           for (driver-name driver-av-pairs) = (multiple-value-list (%sql-drivers handle
                                                                 :direction direction))
           while driver-name
           collect (cons driver-name
                         (parse-driver-attributes driver-av-pairs #\Null)))))

(defmethod db-describe-columns ((database odbc-database) table
                                &key column catalog schema)
  (with-cursor-handle-results (handle database)
    (%sql-columns handle (name catalog) (ordinary-argument schema)
                  (ordinary-argument table "%")
                  (ordinary-argument column "%"))))

(defmethod db-describe-columns :around ((database odbc-database) (expr sql-query-exp)
                                        &key column catalog schema)
 ;;; As SQLColumns() works only for tables and views, detect expr like SELECT * FROM view
  (with-slots (sql.sys::selections sql.sys::from) expr
    (unless (rest sql.sys::selections)
      (let ((identifier (first sql.sys::selections)))
        (when (and (eq (type-of identifier) 'sql-ident)
                   (eq (name identifier) '*)
                   (eq (type-of sql.sys::from) 'sql-ident-table))
          (call-next-method database sql.sys::from
                            :column column :catalog catalog :schema schema))))))

(defmethod db-list-columns ((database odbc-database) table schema &key catalog)
  (let ((cursor (obtain-statement database 'odbc-cursor)))
    (unwind-protect       
        (progn
          (setf (expression cursor) '%sql-columns)
          (%initialize-cursor cursor :forward-only)
          (%sql-columns (handle cursor) (name catalog)
                        (if (eq schema :all) SQL_ALL_SCHEMAS (ordinary-argument schema))
                        (ordinary-argument table)
                        "%")
          (%initialize-rowset cursor :column-numbers '(4)) ; COLUMN_NAME 4
          (db-fetch-rows cursor :flatp t))
      (db-close cursor))))

(defmethod db-describe-foreign-keys ((database odbc-database) pk-table fk-table
                                     &key pk-catalog pk-schema
                                          fk-catalog fk-schema)
 ;;; Values: (PKTABLE_CAT PKTABLE_SCHEM PKTABLE_NAME PKCOLUMN_NAME
  ;;          FKTABLE_CAT FKTABLE_SCHEM FKTABLE_NAME FKCOLUMN_NAME
  ;;          KEY_SEQ UPDATE_RULE DELETE_RULE FK_NAME PK_NAME DEFERRABILITY)
  (with-cursor-handle-results (handle database)
    (%sql-foreign-keys handle
       (name pk-catalog) (ordinary-argument pk-schema) (ordinary-argument pk-table)
       (name fk-catalog) (ordinary-argument fk-schema) (ordinary-argument fk-table))))

(defmethod db-describe-primary-keys ((database odbc-database) table &key catalog schema)
  ;; Values: (TABLE_CAT TABLE_SCHEM TABLE_NAME COLUMN_NAME KEY_SEQ PK_NAME)
  ;;         ordered by TABLE_CAT, TABLE_SCHEM, TABLE_NAME, and KEY_SEQ.
  (with-cursor-handle-results (handle database)
    (%sql-primary-keys handle (name catalog) (ordinary-argument schema)
                       (ordinary-argument table))))

(defmethod db-describe-special-columns ((database odbc-database) table
                                        &key (type SQL_BEST_ROWID)
                                             (scope SQL_SCOPE_CURROW)
                                             (nullable t)
                                             catalog schema)
  ;; Value: (SCOPE COLUMN_NAME DATA_TYPE TYPE_NAME COLUMN_SIZE BUFFER_LENGTH
  ;;         DECIMAL_DIGITS PSEUDO_COLUMN) order by actual SCOPE
  (with-cursor-handle-results (handle database)
    (%sql-special-columns handle type (name catalog) (ordinary-argument schema)
                          (ordinary-argument table) scope
                          (if nullable SQL_NULLABLE SQL_NO_NULLS))))

(defmethod db-describe-statistics ((database odbc-database) table
                                   &key unique (reserved 0)  ;=SQL_QUICK (or SQL_ENSURE)
                                   catalog schema)
  ;; Args: table Explicit table, "%" does not work.
  ;; Value: (TABLE_CAT TABLE_SCHEM TABLE_NAME NON_UNIQUE INDEX_QUALIFIER INDEX_NAME
  ;;         TYPE ORDINAL_POSITION COLUMN_NAME ASC_OR_DESC CARDINALITY PAGES
  ;;         FILTER_CONDITION)
  ;; Access NB:       .
  (with-cursor-handle-results (handle database)
    (%sql-statistics handle (name catalog) (ordinary-argument schema)
                     (ordinary-argument table)
                     (if unique SQL_INDEX_UNIQUE SQL_INDEX_ALL)
                     reserved)))

(defmethod db-list-indexes ((database odbc-database) table &key catalog schema)
  (do ((acc ())
       (rest (db-describe-statistics database table :catalog catalog :schema schema)
             (rest rest))
       tuple)
      ((null rest) (nreverse acc))
    (unless (eql (seventh (setq tuple (first rest))) SQL_TABLE_STAT)	; filter out
      (setq acc (cons (sixth tuple) acc)))))				; table statistics

(defmethod db-list-indexes ((database odbc-database) (table (eql :all)) 
                            &key catalog schema)
 ;;; Access NB: Index names can be duplicated.
  (mappend (lambda (table) (db-list-indexes database table
                                            :catalog catalog :schema schema))
           (db-list-tables database schema :catalog catalog)))


(defmethod db-describe-tables ((database odbc-database) table
                               &key (table-type "TABLE") catalog schema)
  ;; Args:
  ;;	schema-name ::= :all or null
  ;; 	table-type ::= Value List (comma-separated) of
  ;;			"" (all)
  ;;                    | "TABLE" (ordinary tables)
  ;;			| "SYSTEM TABLE" (system tables)
  ;;			| "SYNONYM" (attached tables) 
  ;;			| "VIEW" (row-returning queries)
  ;;              E.g. "TABLE,VIEW"
  ;; Value: (("F:\\Lisp\\Test\\TehnoTab" NIL "Table1" "TABLE" NIL)...)
  (with-cursor-handle-results (handle database)
    (%sql-tables handle (name catalog)
                 (if (eq schema :all) SQL_ALL_SCHEMAS (ordinary-argument schema))
                 (ordinary-argument table) table-type)))

(defmethod db-list-tables ((database odbc-database) schema &key table catalog)
  (let ((cursor (obtain-statement database 'odbc-cursor)))
    (unwind-protect       
        (progn
          (setf (expression cursor) '%sql-tables)
          (%sql-tables (handle cursor)
                       (name catalog)
                       (if (eq schema :all) SQL_ALL_SCHEMAS (ordinary-argument schema))
                       (ordinary-argument table "%")
                       "TABLE")
          (%initialize-rowset cursor :column-numbers '(3)) ; TABLE_NAME 3
          (db-fetch-rows cursor :flatp t))
      (db-close cursor))))

(defun %drop-schema-object (kind identifier database if-does-not-exist
                            &optional plist (sql-state "42S02"))
 ;;; Helper for hanlding exceptions during droping tables, views or the like
 ;;; Args: sql-state Corresponding the error when the object does not exitst;
  ;;		     the ODBC default "42S02" reads "Base table or view not found".
  ;;	   plist ::= (({:CASCADE | :RESTRICT | "PURGE" (Oracle)} t) ...)
  (handler-bind ((odbc-error (lambda (c)  		; Base table or view not found
                               (when (equal (sql-state c) sql-state)
                                 (ecase if-does-not-exist
                                   (:ignore (return-from %drop-schema-object nil))
                                   (:error t))))))
    (execute-command (sql.sys::write-drop-exp-to-string kind identifier database plist)
                     :database database)))

(defmethod db-drop-table ((database odbc-database) table &rest args
                          &key (if-does-not-exist :error))
  (%drop-schema-object :table table database if-does-not-exist
                       (remove-properties args '(:if-does-not-exist))))
                 ;:suffix (if (equal (subseq (db-get-info database SQL_DBMS_VER) 0 2) "10")
                 ;             "PURGE" nil)		; Move to oracle-odbc.lisp?
  #|(handler-bind ((odbc-error (lambda (c)  		; Base table or view not found
                               (when (equal (sql-state c) "42S02")
                                 (ecase if-does-not-exist
                                   (:ignore (return-from db-drop-table nil))
                                   (:error t))))))
    (execute-command (string-append "DROP TABLE "
                                    (write-sql-identifier-to-string table database)
                     :database database)))|#

(defmethod db-list-views ((database odbc-database) schema &key view catalog)
  (let ((cursor (obtain-statement database 'odbc-cursor)))
    (unwind-protect       
        (progn
          (setf (expression cursor) '%sql-tables)
          (%sql-tables (handle cursor)
                       (name catalog)
                       (if (eq schema :all) SQL_ALL_SCHEMAS (ordinary-argument schema))
                       (ordinary-argument view "%")
                       "VIEW")
          (%initialize-rowset cursor :column-numbers '(3)) ; TABLE_NAME 3
          (db-fetch-rows cursor :flatp t))
      (db-close cursor))))

(defmethod db-drop-view ((database odbc-database) view &rest args
                         &key (if-does-not-exist :error))
  (%drop-schema-object :view view database if-does-not-exist
                       (remove-properties args '(:if-does-not-exist))))
  #|(handler-bind ((odbc-error (lambda (c)  		; Base table or view not found
                               (when (equal (sql-state c) "42S02")
                                 (ecase if-does-not-exist
                                   (:ignore (return-from db-drop-view nil))
                                   (:error t))))))
    (execute-command (string-append "DROP VIEW "
                                    (write-sql-identifier-to-string view database))
                     :database database)))|#

(defmethod db-describe-type ((database odbc-database) (sql-type integer))
  (with-cursor-handle-results (handle database)
    (%sql-get-type-info handle sql-type)))

(defmethod db-get-info ((database odbc-database) info-type)
  (cond ((numberp info-type)
         (with-slots (info) database
           (let ((value (gethash info-type info :unbound)))
             (if (eq value :unbound)
                 (setf (gethash info-type info) (%sql-get-info (handle database)
                                                               info-type))
                 value))))))
        ;((keywordp info-type) ; does this make it more readable?

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Expressions

(defmethod db-identifier-case ((database odbc-database))
  (let ((case (db-get-info database SQL_IDENTIFIER_CASE)))
    (cond ((eql case SQL_IC_UPPER)	:upcase)
          ((eql case SQL_IC_LOWER)	:downcase)
          ((eql case SQL_IC_SENSITIVE)	:sensitive)
          ((eql case SQL_IC_MIXED)	:mixed))))

(defmethod db-identifier-quote-char ((database odbc-database))
  (schar (db-get-info database SQL_IDENTIFIER_QUOTE_CHAR) 0))

(defmethod db-special-character-p ((database odbc-database) char)
  (find char (db-get-info database SQL_SPECIAL_CHARACTERS)))

(defmethod output-sql ((expr vector) (database odbc-database))
  (let ((type (upgraded-array-element-type (array-element-type expr))))
    (case (if (consp type) (first type) type)
      ((unsigned-byte signed-byte)
       (if-bind (type-descriptor (or (%get-type-descriptor database SQL_VARBINARY)
                                     (%get-type-descriptor database SQL_BINARY)))
         (progn
           (when-bind (string (type-desc-literal-prefix type-descriptor))
             (write-string string *sql-stream*))
           (write-string (object-hex-string expr (and (consp type) (second type)))
                         *sql-stream*)
           (when-bind (string (type-desc-literal-suffix type-descriptor))
             (write-string string *sql-stream*)))
         (call-next-method)))
      (otherwise
       (call-next-method)))))

(defmethod output-sql ((expr bit-vector) (database odbc-database))
  (if-bind (type-descriptor (or (%get-type-descriptor database SQL_VARBINARY)
                                (%get-type-descriptor database SQL_BINARY)))
    (progn (when-bind (string (type-desc-literal-prefix type-descriptor))
             (write-string string *sql-stream*))
           (write-string (object-hex-string expr) *sql-stream*)
           (when-bind (string (type-desc-literal-suffix type-descriptor))
             (write-string string *sql-stream*)))
    (call-next-method)))

(defmethod output-sql ((expr sql-ident-table) (database odbc-database))
 ;;; Output [catalog.][schema.]table[ alias] or [schema.]table[@catalog][ alias]
  ;; Accept a pathname as catalog, check if aliases are enabled.
  (declare (ignore database))
  (with-slots (name alias plist) expr
    (let ((catalog (getf plist :catalog))
          (schema (getf plist :schema))
          info)
      (when catalog
        (unless (db-get-info database SQL_CATALOG_NAME)
          (error 'sql-database-error :database database
                 :message "Catalog names are not supported."))
        (when (eql (setq info (db-get-info database SQL_CATALOG_LOCATION))
                   SQL_CL_START)
          (write-sql-identifier (name catalog) *sql-stream* database)
          (write-string (db-get-info database SQL_CATALOG_NAME_SEPARATOR) *sql-stream*)))
      (when (and schema (not (zerop (db-get-info database SQL_SCHEMA_USAGE))))
        (write-sql-identifier schema *sql-stream* database)
        (write-char #\. *sql-stream*))
      (write-sql-identifier name *sql-stream* database)
      (when (and catalog (eql info SQL_CL_END))
        (write-string (db-get-info database SQL_CATALOG_NAME_SEPARATOR) *sql-stream*)
        (write-sql-identifier (name catalog) *sql-stream* database))
      (when (and alias ystok.sql.sys::*in-subselect*)
        (write-char #\Space *sql-stream*)
        (write-sql-identifier alias *sql-stream* database)))))

(defmethod output-sql ((expr sql-concat-exp) (database odbc-database))
 ;;; Only two arguments are accepted - introduce internal calls to CONCAT
  ;; Leave intact if scalar function escapes are not supported at all
  ;; (e.g. within SQLite driver).
  (if (logtest (db-get-info database SQL_STRING_FUNCTIONS) #x00000001) ; SQL_FN_STR_CONCAT
      (with-slots (components) expr
        (labels ((output (args)
                   (let ((first (first args))
                         (rest (rest args)))
                     (if rest
                         (progn
                           (write-string "{fn CONCAT(" *sql-stream*)
                           (output-sql first database)
                           (write-char #\, *sql-stream*)
                           (output rest)
                           (write-string ")}" *sql-stream*))
                         (output-sql first database)))))
          (output components)))
      (call-next-method)))

(defmethod output-sql ((expr sql-cast-exp) (database odbc-database))
 ;;; Transform CAST(v AS type) into {fn CONVERT(v, SQL_type)} if needed
  ;; NB: Though CAST is supported in the following back-ends,
  ;;  - SQLite3: SQL_SQL_CONFORMANCE signals.
  ;;  - Oracle: SQL_CONVERT_FUNCTIONS and SQL_SQL92_VALUE_EXPRESSIONS do not mention _CAST.
  ;(if (or (> (db-get-info database SQL_SQL_CONFORMANCE) 1)	;SQL_SC_SQL92_ENTRY
  ;        (logtest (db-get-info database SQL_SQL92_VALUE_EXPRESSIONS) 2)) ;SQL_SVE_CAST 
  (if (or (and (>= $ODBCVER 0300)				   ; SQL_FN_CVT_...
               (logtest (db-get-info database SQL_CONVERT_FUNCTIONS) #x00000002)) ;_CAST
          (not (logtest (db-get-info database SQL_CONVERT_FUNCTIONS) #x00000001)));_CONVERT
      (call-next-method)				; CAST is supported or CONVERT not
      (let ((modifier (slot-value expr 'modifier)))
        (write-string "{fn CONVERT(" *sql-stream*)
        (output-sql (slot-value expr 'components) database)
        (write-string ",SQL_" *sql-stream*)
        (write-string (apply #'db-sql-type-specifier database
                             (first-or-self modifier)
                             (and (consp modifier) (rest modifier)))
                      *sql-stream*)
        (write-string ")}" *sql-stream*))))

(eval-when (:compile-toplevel :load-toplevel :execute)
  (defconstant +scalar-functions+
    '(
      ;; CURRENT_DATE (ODBC 3.0)  Returns the current date. 
      (current-date	#.SQL_TIMEDATE_FUNCTIONS . #x00020000)	; SQL_FN_TD_CURRENT_DATE
      ;; CURRENT_TIME[(time-precision)] (ODBC 3.0)
      ;;  Returns the current local time. The time-precision argument
      ;;  determines the seconds precision of the returned value. 
      (current-time	#.SQL_TIMEDATE_FUNCTIONS . #x00040000)	; SQL_FN_TD_CURRENT_TIME
      ;; CURRENT_TIMESTAMP[(timestamp-precision)] (ODBC 3.0) 
      ;;  Returns the current local date and local time as a timestamp value. 
      ;;  The timestamp-precision argument determines the seconds precision
      ;;  of the returned timestamp 
      (current-timestamp #.SQL_TIMEDATE_FUNCTIONS . #x00080000)	; SQL_FN_TD_CURRENT_TIMESTAMP

      (ABS	#.SQL_NUMERIC_FUNCTIONS . #x00000001)	; SQL_FN_NUM_ABS
      (ACOS	#.SQL_NUMERIC_FUNCTIONS . #x00000002)	; SQL_FN_NUM_ACOS
      (ASIN	#.SQL_NUMERIC_FUNCTIONS . #x00000004)	; SQL_FN_NUM_ASIN
      (ATAN	#.SQL_NUMERIC_FUNCTIONS . #x00000008)	; SQL_FN_NUM_ATAN
                 ;ATAN2	#x00000010L
      (CEILING	#.SQL_NUMERIC_FUNCTIONS . #x00000020)	; SQL_FN_NUM_CEILING
      (COS	#.SQL_NUMERIC_FUNCTIONS . #x00000040)	; SQL_FN_NUM_COS
                 ;COT	#x00000080L
      (EXP	#.SQL_NUMERIC_FUNCTIONS . #x00000100)	; SQL_FN_NUM_EXP
      (FLOOR	#.SQL_NUMERIC_FUNCTIONS . #x00000200)	; SQL_FN_NUM_FLOOR
      (LOG	#.SQL_NUMERIC_FUNCTIONS . #x00000400)	; SQL_FN_NUM_LOG
      (MOD	#.SQL_NUMERIC_FUNCTIONS . #x00000800)	; SQL_FN_NUM_MOD
      (PI	#.SQL_NUMERIC_FUNCTIONS . #x00010000)	; SQL_FN_NUM_PI
                 ;SIGN 	#x00001000L
      (SIN	#.SQL_NUMERIC_FUNCTIONS . #x00002000)	; SQL_FN_NUM_SIN
      (SQRT	#.SQL_NUMERIC_FUNCTIONS . #x00004000)	; SQL_FN_NUM_SQRT
      (TAN	#.SQL_NUMERIC_FUNCTIONS . #x00008000)	; SQL_FN_NUM_TAN
                 ;RAND 	#x00020000L
                 ;DEGREES	#x00040000L
                 ;LOG10	#x00080000L)
      (POWER	#.SQL_NUMERIC_FUNCTIONS . #x00100000)	; SQL_FN_NUM_POWER
                 ;RADIANS	#x00200000L
      ;; Jet NB: ROUND always computes a la Access CInt  lng -
      ;; when the fractional part is exactly 0.5, it is rounded to the nearest even number;
      ;; for example, 0.5 rounds to 0, and 1.5 rounds to 2.
      (ROUND	#.SQL_NUMERIC_FUNCTIONS . #x00400000)	; SQL_FN_NUM_ROUND
      (TRUNCATE	#.SQL_NUMERIC_FUNCTIONS . #x00800000)	; SQL_FN_NUM_TRUNCATE

      (SUBSTRING	#.SQL_STRING_FUNCTIONS . #x00000800)	; SQL_FN_STR_SUBSTRING
  )  )

  (export (mapcar #'car +scalar-functions+))
)

(sql:defsql-function () . #.(mapcar #'car +scalar-functions+))

(defmethod output-sql ((expr sql-function-exp) (database odbc-database))
 ;;; Adorn the call of a function known to ODBC by interoperable {fn ...} notation.
  ;; Leave intact the call if the backend-specific function is unknown or 
  ;; if scalar function escapes are not supported at all (e.g. within SQLite driver).
  (let ((tuple (assoc (name expr) +scalar-functions+ :test #'string=)))
         ;(and (not (zerop (db-get-info database SQL_NUMERIC_FUNCTIONS))) ; 0 for Postgres!
         ;      (not (zerop (db-get-info database SQL_STRING_FUNCTIONS)))
         ;      (member (name expr) +scalar-function-symbols+ :test #'string=))
    (cond ((and tuple (logtest (db-get-info database (cadr tuple)) (cddr tuple)))
           (write-string "{fn " *sql-stream*)
           (call-next-method)
           (write-char #\} *sql-stream*))
          (t
           (call-next-method)))))

(defmethod output-sql ((expr sql-call-procedure-exp) (database odbc-database))
  (with-slots (name sql.sys::result args) expr
    (write-char #\{ *sql-stream*)
    (when sql.sys::result
      (output-sql sql.sys::result database)
      (write-char #\= *sql-stream*))
    (write-string "call " *sql-stream*)
    (output-sql name database)
    (when args (output-sql args database))	; embrace arguments only when given
    (write-char #\} *sql-stream*)))

;;; Joins: Use {oj...} escapes or, for obsolete DBMSs, specialize by converting to WHERE.
#|(defmethod output-sql ((expr sql-join-exp) (database odbc-database))
  (with-slots (operator components on sql.sys::using) expr
    (write-string "{oj " *sql-stream*)
    (sql.sys::%output-join-exp operator components on sql.sys::using database
                               nil) ; no parentheses
    (write-char #\} *sql-stream*)))|#

#|;;;;;;;;;;; Unused stuff ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; one for odbc-database is missing
;(defmethod terminate ((cursor odbc-cursor))
;  ;;(format tb::*local-output* "~%*** terminated: ~s" cursor)
;  (with-slots (hstmt) cursor
;    (when hstmt
;      ;(%free-statement hstmt :drop)
;    (%dispose-column-ptrs cursor)))

 |#
