;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2005 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Condition classes

(in-package :ystok.sql.sys)

(define-condition sql-condition () ())

(define-condition sql-error (error sql-condition) ())

(define-condition sql-simple-error (simple-condition sql-error) ())

(define-condition sql-user-error (sql-error)
 ((message :reader message :initarg :message :initform nil))
 (:report (lambda (c stream)
           (format stream "An SQL lisp code error occurred~:[ #<unspecified>~;:~%~:*~A~]."
                   (message c))))
 (:documentation "Used to signal errors in Lisp code."))

(define-condition sql-warning (warning sql-condition) ())

(define-condition sql-simple-warning (simple-condition sql-warning) ())

(define-condition sql-undefined-function-warning (sql-warning)
 ((name :reader name :initarg :name)
  (args :initarg :args))
 (:report (lambda (c stream)
            (format stream "Undefined SQL function or expansion of user-defined procedure in call ~S.~%Default SQL function expression syntax applied."
                    (cons (name c) (slot-value c 'args))))))


(define-condition sql-reader-error (sql-user-error simple-error) () ;sql-syntax-error
 (:report (lambda (c stream)
            (format stream "An SQL lisp reader error:~%~:[~?~;~:*~A~]."
                    (message c)
                    (simple-condition-format-control c)
                    (simple-condition-format-arguments c))))
 (:documentation "Generated by the SQL bracket syntax interface."))

(define-condition sql-type-check-error (sql-user-error) ()
 (:documentation "Generated when type checking in the SQL interface causes an error."))

(define-condition sql-invalid-spec-error (sql-user-error)
  ((connection-spec :initarg :connection-spec :reader connection-spec)
   (db-type :initarg :database-type :reader db-type)
   (template :initarg :template :reader template))
  (:report (lambda (c stream)
	     (format stream "The connection specification ~A
 is invalid for database type ~A.
 The connection specification must conform to~%~A"
		     (connection-spec c) (db-type c) (template c)))))

(defmacro check-connection-spec (connection-spec database-type template)
 ;;; Check list-like connection specification against the template,
  ;; and signal an sql-invalid-spec-error if they don't match.
  `(handler-case
       (destructuring-bind ,template ,connection-spec
	 (declare (ignore ,@(set-difference template lambda-list-keywords)))
	 t)
     (error () (error 'sql-invalid-spec-error
		      :connection-spec ,connection-spec
		      :database-type ,database-type
		      :template (quote ,template)))))

(define-condition sql-exists-condition (sql-condition)
   ((old-db :initarg :old-db :reader old-db)
    (new-db :initarg :new-db :reader new-db :initform nil))
   (:report (lambda (c stream)
	      (format stream "In call to ~S:~%" 'connect)
	      (cond
		((null (new-db c))
		 (format stream "There is an existing connection ~A to database ~A."
			 (old-db c) (database-name (old-db c))))
		((eq (new-db c) (old-db c))
		 (format stream "Using existing connection ~A to database ~A."
			 (old-db c) (database-name (old-db c))))
		(t
		 (format stream "Created new connection ~A to database ~A~% ~
although there is an existing connection (~A)."
			 (new-db c)
			 (database-name (new-db c))
			 (old-db c)))))))

(define-condition sql-exists-warning (sql-exists-condition sql-warning) ())

(define-condition sql-exists-error (sql-exists-condition sql-user-error) ())

(define-condition sql-closed-error (sql-user-error)
 ((database :initarg :database :reader database))
 (:report (lambda (c stream)
            (format stream "The database ~A has already been closed." (database c)))))

(defun signal-closed-database-error (database)
  (cerror "Ignore this error and return nil." 'sql-closed-error
	  :database database))

(define-condition sql-nodb-error (sql-user-error) ()
 (:default-initargs
  :message "No database is open. Call sql:connect first."))

(defun signal-no-database-error ()
  (cerror "Ignore this error and return nil." 'sql-nodb-error))


(define-condition sql-database-condition (sql-condition)
 ((database   :reader database   :initarg :database   :initform nil)
  (expression :reader expression :initarg :expression :initform nil)
  (message    :reader message    :initarg :message    :initform nil)
  (sql-state  :reader sql-state	 :initarg :sql-state  :initform nil ; primary error id
              :reader error-id   :initarg :error-id)		    ; CommonSQL compliant
  (secondary-error-id :reader secondary-error-id :initarg :secondary-error-id
	      :initform nil)))					    ; DB return code , e.g.

(define-condition sql-database-error (sql-database-condition sql-error) ()
 (:report (lambda (c stream)
           (format stream "[Error~@[ ~A~]~@[/~A~]] using database ~:[#<unknown>~;~:*~A~].
Message: ~A.~@[~%The SQL statement: ~S~]"
                   (sql-state c) (secondary-error-id c)
                   (and (database c) (database-name (database c)))
                   (message c) (expression c)) ))
 (:documentation "Signaled when the underlying DBMS generates an error."))

(define-condition sql-database-data-error (sql-database-error) ()
 (:documentation "Signaled with the SQL data passed to a database."))

(define-condition sql-type-coercion-error (sql-database-data-error) 
 ((data-type :initarg :data-type))		; DBMS type code or name; ODBC: sql-type
 (:report (lambda (c stream)
            (format stream "Error while using database ~:[#<unknown>;~:*~A~]:
 ~@[The SQL statement: ~S~%~] Undefined conversion from DBMS type ~A to a Lisp type.~@[~%Message: ~A.~]~@[~% (SQLSTATE is ~S)~]"
                    (and (database c) (database-name (database c)))
                    (expression c) (slot-value c 'data-type) (message c)
                    (sql-state c))
            ))
 (:documentation
  "Signaled when type coercion of the data from the underlying DBMS causes an error."))


(define-condition sql-connection-error (sql-database-error)
 ;;; TODO: Provide mask-password for connection-spec of any kind.
  ((connection-spec :initarg :connection-spec :reader connection-spec)
   (db-type :initarg :db-type :initarg :database-type :reader db-type))
  (:report (lambda (c stream)
	     (format stream "[Error~@[ ~S~]~@[/~A~]] trying to connect to database ~A
of database-type ~A:~%Message: ~A."
		     (sql-state c) (secondary-error-id c)
                     (connection-spec c) (db-type c) (message c)))))

(define-condition sql-timeout-error (sql-connection-error)
 ((timeout :initarg :timeout :reader timeout :initform nil)) ; timeout value in seconds
 (:documentation "Signaled when the database times out while processing some operation."))

(define-condition sql-fatal-error (sql-connection-error) ()
 (:documentation "Signaled when the database connection is no longer usable."))

;; Means the code can work without change, once the other users stop using the database.
(define-condition sql-temporary-error (sql-database-error) ()
 (:documentation "Signaled when other users are working with
 the same database, for example, a table lock, running out of resources,
 or when DBMS is still executing another command."))


;;; Condition subclasses to print in a user-friendly way within the client application
;;; Q: Add the constraint name slot in hope that some DBMS returns it?

(define-condition sql-check-error (sql-database-data-error)
 ((table :reader table :initarg :table))	; the name of table
 (:documentation
  "Signaled when the underlying DBMS detects a table-level CHECK constraint violation."))

(define-condition sql-foreign-key-error (sql-database-data-error)
 ((table :reader table  :initarg :table))	; the name of referencing table
 (:documentation
  "Signaled when the underlying DBMS detects a FOREIGN KEY constraint violation."))

(define-condition sql-not-nullable-error (sql-database-data-error)
 ((table  :reader table  :initarg :table)	; the name of table
  (column :reader column :initarg :column))	; the name of column
 (:documentation
  "Signaled when the underlying DBMS detects a NOT NULL constraint violation."))

(define-condition sql-unique-error (sql-database-data-error)
 ((table   :reader table   :initarg :table)	; the name of table
  (columns :reader columns :initarg :columns))	; the (names of) columns
 (:documentation
  "Signaled when the underlying DBMS detects a UNIQUE KEY or INDEX constraint violation."))
