;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2006 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Classes and methods constructing SQL expression and statements

(in-package :ystok.sql.sys)

;(eval-when (:compile-toplevel :load-toplevel :execute)
;(defvar *default-schema* "YstokSQL")
;(defun in-schema (schemaname)
;  (setf *default-schema* schemaname))
;(defvar *object-schemas* (make-hash-table :test #'equal)
;  "Hash of schema name to class constituent lists.")
;)

(mapc (lambda (symbol) (setf (get symbol 'sql-keywordp) t))
      '(:all :as :asc
        :check :check-option :constraint	;:cascade
        :default :desc :distinct
        :false :first :foreign-key
        :ignore-null :last
        :not-null :nulls
        :on-delete :on-update
        :pivot :primary-key
        :transform :true
        :unique :unknown :using
        :with))

(defun sql-keywordp (symbol)  
 ;;; Symbol name coincides with on of the SQL_KEYWORDS but does not name a data type.
  (and (keywordp symbol)
       (get symbol 'sql-keywordp)))
  ;(member symbol '(:distinct :all :as :asc :desc :nulls :first :last)
  ;        :test #'eq))

(defun sql (&rest args)
  (with-sql-string-stream ()
    (loop for (arg . rest) on args
          do (output-sql arg *default-database*)  ;:toplevel-p t
          when rest do (write-char #\Space *sql-stream*))))

(defun write-sql-to-string (expr database &optional terminate)
 "Create full-fledged SQL string from expr that is passed to the database.
 If terminate, concatenate semicolon at the end."
 ;;; Value: string when success,
  ;;	 or NIL or another value throwed to write-sql-to-string tag during output-sql
  (if (stringp expr)
      (if terminate
          (let ((length (length expr)))
            (if (> length  0)
                (if (char= (char expr (1- length)) #\;)
                    expr
                    (string-append expr ";"))
                ";"))
          expr)
      (catch 'write-sql-to-string
        (with-sql-string-stream ()
          (output-sql expr database)  ;:toplevel-p t
          (when terminate
            (write-char #\; *sql-stream*))))))

(defgeneric output-sql (expr database)
 (:documentation "Writes SQL expression expr into *sql-stream* for database.
   Is not obliged to return the resulting string."))

(defmethod output-sql ((expr (eql t)) database)
 ;;; Depends on database and datatype; T is an Oracle hack
  (declare (ignore database))
  (write-string "'T'" *sql-stream*))

(defmethod output-sql ((expr symbol) database)
  (declare (ignore database))
  (if (eq expr *null*)
      (write-string "NULL" *sql-stream*)
      (let ((symbol-name (symbol-name expr)))
        (if (keywordp expr)
            (if (sql-keywordp expr)			        ; treat as a keyword
                (write-string (substitute #\Space #\- symbol-name) *sql-stream*)
                (output-sql symbol-name *sql-stream*))		; treat as a string
                ;(write-sql-string symbol-name *sql-stream*)
            (write-string symbol-name *sql-stream*)))))		; operator

(defmethod output-sql ((expr integer) database)
  (declare (ignore database))
  (princ expr *sql-stream*))

(defmethod output-sql ((expr ratio) database)
  (declare (ignore database))
  (write-string (float-to-sql-string (float expr)) *sql-stream*))

(defmethod output-sql ((expr float) database)
  (declare (ignore database))
  (write-string (float-to-sql-string expr) *sql-stream*))

(defmethod output-sql ((expr string) database)
 ;;; Construct character string literal
  (write-char #\' *sql-stream*)
  (loop for char across expr
        if (char= char #\') do (write-char #\' *sql-stream*)
        do (write-char char *sql-stream*))
  (write-char #\' *sql-stream*))

(defmethod output-sql ((expr character) database)
  (output-sql (string expr) database))

(defmethod output-sql ((expr cons) database)
  ;(format nil "(~{~A~^,~})" )
  (write-char #\( *sql-stream*)
  (loop for (arg . rest) on expr
        do (output-sql arg database)
        when rest do (write-char #\, *sql-stream*))
  (write-char #\) *sql-stream*))

(defmethod output-sql ((expr vector) database)
  (let ((type (upgraded-array-element-type (array-element-type expr))))
    (case (if (consp type) (first type) type)
      ((unsigned-byte signed-byte)
       (write-char #\' *sql-stream*)
       (write-string (object-hex-string expr (and (consp type) (second type)))
                     *sql-stream*)
       (write-char #\' *sql-stream*))
      (otherwise      
       (loop with n = (length expr)
             for i upfrom 0 below n
             do (output-sql (aref expr i) database)
             when (< i (1- n))
               do (write-char #\, *sql-stream*))))))

(defmethod output-sql ((expr bit-vector) database)
  (write-char #\' *sql-stream*)
  (write-string (object-hex-string expr) *sql-stream*)
  (write-char #\' *sql-stream*))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; SubExpression Classes

(defclass sql-expression ()
  ())

(defmethod make-load-form ((self sql-expression) &optional environment)
  (make-load-form-saving-slots self :environment environment))

(defgeneric collect-table-refs (expr)
  (:documentation "Return a list of sql-ident-table instances without duplicates.")
  (:method (expr) (declare (ignore expr)) ()))

;;; In accordance with the SQL-92 specification, parameter markers cannot be placed
;;; in the following locations.
;;; - In a SELECT list
;;; - As both expressions in a comparison-predicate
;;; - As both operands of a binary operator
;;; - As both the first and second operands of a BETWEEN operation
;;; - As both the first and third operands of a BETWEEN operation
;;; - As both the expression and the first value of an IN operation
;;; - As the operand of a unary + or  operation
;;; - As the argument of a set-function-reference
(defgeneric collect-parameters (expr)
  (:documentation "Return a list of sql-parameter instances.")
  (:method (expr) (declare (ignore expr)) ())
  (:method ((expr cons))
   (nconc (collect-parameters (first expr))
          (collect-parameters (rest expr))))
  (:method ((expr vector))
   (loop for arg across expr
         nconc (collect-parameters arg)))
 )

(defmethod print-object ((self sql-expression) stream)
  (print-unreadable-object (self stream :type nil)
    (format stream "~A \"~A\"" (type-of self) (sql self)))
  self)

;;; The following definition for straight ups could be avoided by simply using strings.
(defclass sql-verbatim (sql-expression)
  ((text :initarg :string :initarg :text :initform ""))
  (:documentation "A literal SQL expression."))

(defmethod output-sql ((expr sql-verbatim) database)
  (declare (ignore database))
  (write-string (slot-value expr 'text) *sql-stream*))

(defgeneric name (object)
 (:method (object) (declare (ignore object)) nil)
 (:method ((object symbol)) object)
  ;(when object (sysmbol-name object)))
 (:method ((object string)) object)
 (:method ((object pathname)) (namestring object))
 (:method ((object sql-expression)) nil))

(defgeneric plist (object)
 (:method (object) (declare (ignore object)) nil))

(defclass sql-ident (sql-expression)
  ((name :initarg :name :initform nil :accessor name))
  (:documentation "A generic SQL identifer."))

(defmethod output-sql ((expr sql-ident) database)
 ;;; Used to represent unqualified identifiers including [*]
  (write-sql-identifier (name expr) *sql-stream* database))

(defmethod write-sql-identifier ((identifier sql-ident)
                                 &optional stream database (name-only nil) (quote *quote*))
  (if name-only
      (write-sql-identifier (name identifier) stream database name-only quote)
      (let ((*sql-stream* stream)
            (*quote* quote))
        (output-sql identifier database))))

(defun output-sql-identifiers (list database &optional (name-only t) (paren t))
 ;;; Output non-empty identifier list with no qualifiers (table or schema) by default
  (when paren (write-char #\( *sql-stream*))
  (loop for (arg . rest) on (listify list)
        do (output-sql-identifier arg database name-only)
        when rest do (write-char #\, *sql-stream*))
  (when paren (write-char #\) *sql-stream*)))

(defclass sql-column (sql-ident)
  ((qualifier :initarg :qualifier :initform nil)	  ; name or sql-ident-table
   (plist :initarg :plist :accessor plist :initform nil)) ; :type, :asc, :desc, :as
  (:documentation "An SQL column reference."))

(defmethod collect-table-refs ((expr sql-column))
  (when-bind (qualifier (slot-value expr 'qualifier))
    (when (symbolp qualifier)
      (list (make-instance 'sql-ident-table :name qualifier)))))

(defmethod output-sql ((expr sql-column) database)
 ;;; [qualifier.]name[ type]
  (declare (ignore database))
  (with-slots (qualifier name plist) expr
    ;(when-bind (catalog (getf plist :catalog) ...)
    (when qualifier
      (write-sql-identifier (alias qualifier) *sql-stream* database)
      (write-char #\. *sql-stream*))
    (write-sql-identifier name *sql-stream* database)
    (loop for (key value) on plist by #'cddr
          ;for key = (first rest)
          do (case key
               ((:asc :desc)
                (write-char #\Space *sql-stream*)
                (write-string (string key) *sql-stream*))
               (:as
                (write-char #\Space *sql-stream*)
                (write-string (string key) *sql-stream*)
                (write-char #\Space *sql-stream*)
                (write-sql-identifier value *sql-stream* database))))))

(defgeneric sql-column (object &optional table)
 (:method ((column symbol) &optional qualifier)
  (make-instance 'sql-column :name column :qualifier qualifier)) )

(defclass sql-ident-table (sql-ident)
  ((alias :initarg :alias :initform nil)	; correlation name
   (plist :initarg :plist :accessor plist :initform nil))	; :schema, :catalog
  (:documentation "An SQL table factor."))

(defgeneric alias (object)
 (:method (object) (name object))
 (:method ((table sql-ident-table)) (or (slot-value table 'alias) (name table))) )

(defmethod output-sql ((expr sql-ident-table) database)
 ;;; Output [catalog.][schema.]table [alias]
  ;; NB: alias is being output only if *in-subselect* is true!
  (declare (ignore database))
  (with-slots (name alias plist) expr
    (when-bind (catalog (getf plist :catalog))
      (write-sql-identifier catalog *sql-stream* database)
      (write-char #\. *sql-stream*))
    (when-bind (schema (getf plist :schema))
      (write-sql-identifier schema *sql-stream* database)
      (write-char #\. *sql-stream*))     
    (write-sql-identifier name *sql-stream* database)
    (when (and alias *in-subselect*)
      (write-char #\Space *sql-stream*)
      (write-sql-identifier alias *sql-stream* database))))

;;; Parameter markers may include the following properties in plist:
;;;  :name	External name of a stored procedure parameter
;;;		(SQL Server: dynamic parameter, which does not have a default value.)
;;;  :key	Keyword used to initialize it when invoked via functional interface
;;;  :type	Lisp object type that could help binding the parameter
;;;  :direction :input, :output, :io
(defclass sql-parameter (sql-expression)
  ((plist :initarg :plist :accessor plist :initform nil))
  (:documentation "An SQL parameter marker."))

(defmethod print-object ((self sql-parameter) stream)
  (if *print-readably*
      ;(signal (make-condition 'print-not-readable :object object)))
      (format stream "[?~{ ~S~}]" (plist self))
      (print-unreadable-object (self stream :type t)
        (format stream "[?~{ ~S~}]" (plist self))))
  self)	  ; return the object to avoid lots of trouble (advised by C.Beck)

(defmethod collect-parameters ((expr sql-parameter))
  (list expr))

(defmethod output-sql ((expr sql-parameter) database)
  (declare (ignore database))
  (write-char #\? *sql-stream*))

;;; ToDo: Arity checking in def-sql?

(defclass sql-relational-exp (sql-expression)
  ((operator :initarg :operator :initform nil)               ; symbol as a rule
   (components :initarg :components :initform nil))
  (:documentation "An SQL relational expression (expr1 op expr2...)."))

(defmethod collect-table-refs ((expr sql-relational-exp))
  (remove-duplicates (mapcan #'collect-table-refs (slot-value expr 'components))
                     :test #'equal
                     :key 'name))

(defmethod collect-parameters ((expr sql-relational-exp))
  (mapcan #'collect-parameters (slot-value expr 'components)))

(defun %output-relational-exp (operator components database &optional plist)
 ;;; Helper also used by specializations on output-sql, e.g.
  ;;	#<method output-sql (sql-concat-exp desktop-mixin)>
  (write-char #\( *sql-stream*)
  (loop for (arg . rest) on (if (consp (car components))
                                (car components)
                                components)
        do (output-sql arg database)
        when rest
        do (write-char #\Space *sql-stream*)
        (write-string (string operator) *sql-stream*) ;(output-sql operator database) 
        (write-char #\Space *sql-stream*))
  (%output-plist plist database)
  (write-char #\) *sql-stream*))

(defmethod output-sql ((expr sql-relational-exp) database)
  (with-slots (operator components) expr
    (%output-relational-exp operator components database)))

;;; The following seems to be of value only for Postgress
;;; that does not provide locale-specific comparison in WHERE clause.
(defclass sql-comparison-exp (sql-relational-exp)
  ((case :initarg :case :initform nil))		; ::= :downcase | :upcase | :ignore
  (:documentation "Binary predicate with possible case conversions on strings."))


(defclass sql-like-exp (sql-relational-exp)
  ((plist :accessor plist :initarg :plist :initform nil))
  (:documentation "An SQL 'like' or variants."))

(defmethod output-sql ((expr sql-like-exp) database)
  (with-slots (operator components plist) expr
    (%output-relational-exp operator components database
                            (remove-properties plist '(:case)))))
 
#|(defclass sql-upcase-like (sql-relational-exp)
  ()
  (:documentation "An SQL 'like' that upcases its arguments."))

(defmethod output-sql ((expr sql-upcase-like) database)
 ;;; Similiarly to [like [upper arg1] [upper arg2]]
  (with-slots (components) expr
    (write-string "(UPPER(" *sql-stream*)
    (output-sql (first components) database)
    (write-string ") LIKE UPPER(" *sql-stream*)
    (output-sql (second components) database)
    (write-string "))" *sql-stream*)))|#

(defclass sql-assignment-exp (sql-relational-exp)
  ()
  (:documentation "An SQL Assignment expression expr1 = expr2."))

(defmethod output-sql ((expr sql-assignment-exp) database)
 ;;; Output <column-name> = value
  ;; SQL-2003: (1) No qualifier allowed in column (allowed in Access),
  ;;	(2) value can be a scalar subquery (not supported by Access).
  (with-slots (operator components) expr
    ;(output-sql (first components) database)
    (write-sql-identifier (first components) *sql-stream* database t)
    (write-char #\Space *sql-stream*)
    (output-sql operator database)
    (write-char #\Space *sql-stream*)
    (let ((*in-subselect* t))
      (output-sql (second components) database))))

(defclass sql-between-exp (sql-relational-exp)
  ()
  (:documentation "An SQL (x between y and z) predicate."))

(defmethod output-sql ((expr sql-between-exp) database)
  (with-slots (components) expr
    (write-char #\( *sql-stream*)
    (output-sql (first components) database)
    (write-string " BETWEEN " *sql-stream*)
    (output-sql (second components) database)
    (write-string " AND " *sql-stream*)
    (output-sql (third components) database)
    (write-char #\) *sql-stream*)))

(defclass sql-concat-exp (sql-relational-exp)
  ((operator :initform "||"))
  (:documentation
    "An SQL string concatenation. Redefine output-sql for non-standard back-ends!"))

(defclass sql-value-exp (sql-expression)
  ((modifier :initarg :modifier :initform nil)
   (components :initarg :components :initform nil))
  (:documentation "An SQL term with unary operator, e.g. (not expr), (exists ())... "))

(defmethod collect-table-refs ((expr sql-value-exp))
  (when (listp (slot-value expr 'components))
    (remove-duplicates (mapcan #'collect-table-refs (slot-value expr 'components))
                       :test #'equal
                       :key 'name)))

(defmethod collect-parameters ((expr sql-value-exp))
  (when (listp (slot-value expr 'components))
    (mapcan #'collect-parameters (slot-value expr 'components))))

(defmethod output-sql ((expr sql-value-exp) database)
  (with-slots (modifier components) expr
    (if modifier
        (let ((listp (not (memq modifier '(ANY ALL)))))	; parens are not allowed
          (when listp (write-char #\( *sql-stream*))
          (output-sql modifier database)
          (write-char #\Space *sql-stream*)
          (output-sql (first components) database)	;components?-introduce extra parens
          (when listp (write-char #\) *sql-stream*)))
        (output-sql components database))))  	; output components surrounded by parens

(defclass sql-cast-exp (sql-value-exp)
  ()
  (:documentation "An SQL-92,99 cast expression CAST(value AS type)."))

(defmethod collect-table-refs ((sql sql-cast-exp))
  (when-bind (components (slot-value sql 'components))
    (collect-table-refs components)))

(defmethod collect-parameters ((sql sql-cast-exp))
  (collect-parameters (slot-value sql 'components)))

(defun %output-cast-exp (expr database)
  (with-slots (modifier components) expr
    (if modifier
        (progn
          (write-string "CAST(" *sql-stream*)
          (output-sql components database)
          (write-string " AS " *sql-stream*)
          ;(output-sql modifier database)
          (write-string (apply #'db-sql-type-specifier database (first-or-self modifier)
                                (and (consp modifier) (rest modifier)))
                        *sql-stream*)
          (write-char #\) *sql-stream*))
        (output-sql components database))))

(defmethod output-sql ((expr sql-cast-exp) database)
  (%output-cast-exp expr database))

(defclass sql-function-exp (sql-expression)
  ((name :initarg :name :initform nil :reader name)
   (args :initarg :args :initform nil))
  (:documentation "An SQL function expression f(expr1,expr2...)."))

(defmethod collect-table-refs ((expr sql-function-exp))
  (remove-duplicates (mapcan #'collect-table-refs (slot-value expr 'args))
                     :test #'equal
                     :key 'name))

(defmethod collect-parameters ((expr sql-function-exp))
  (mapcan #'collect-parameters (slot-value expr 'args)))

(defun %output-function-exp (name args database)
  (write-sql-identifier name *sql-stream* database)   ;(output-sql name database)
  (write-char #\( *sql-stream*)
  (loop for (arg . rest) on args
        do (output-sql arg database)
        when rest do (write-char #\, *sql-stream*))
  (write-char #\) *sql-stream*))

(defmethod output-sql ((expr sql-function-exp) database)
  (with-slots (name args) expr
    (%output-function-exp name args database)))

(defclass sql-aggregate-function-exp (sql-function-exp)
  ((modifier :initarg :modifier :initform nil))
  (:documentation "An SQL aggreage function expression, e.g. COUNT(DISTINCT expr1)."))

(defmethod output-sql ((expr sql-aggregate-function-exp) database)
  (with-slots (name modifier args) expr
    (output-sql name database)
    (write-char #\( *sql-stream*)
    (when modifier
      (output-sql modifier database)
      (write-char #\Space *sql-stream*))
    (loop for (arg . rest) on args
          do (output-sql arg database)
          when rest do (write-char #\, *sql-stream*))
    (write-char #\) *sql-stream*) ))

;;; SQL-2003
;;;	CASE <operand> WHEN expr1 THEN result1 WHEN expr2 THEN result2 ... ELSE result END
;;;	CASE WHEN cond1 THEN result1 WHEN cond2 THEN result2 ... ELSE result END
;;; SQL-2003:	COALESCE(expr1,expr2[,expr3]...)
;;; ODBC 1.0:	IFNULL(expr1, expr2) 
;;; MS Jet:	Nz(expr[,value]) - Inside Access only and returns string!
;;; Oracle:	NVL(expr1,expr2), NVL2(expr1,expr2,expr3)
(defclass sql-case-exp (sql-function-exp) ())

(defmethod output-sql ((expr sql-case-exp) database )
  (flet ((out-args (args)
           (loop for (arg result) in args
                 if (eq arg t)
                 do (write-string " ELSE " *sql-stream*)
                    (output-sql result database)
                 else do (write-string " WHEN " *sql-stream*)
                         (output-sql arg database)
                         (write-string " THEN " *sql-stream*)
                         (output-sql result database)
                 finally (write-string " END" *sql-stream*))))
    (with-slots (name args) expr
      (case name
        (COND
         (write-string "CASE" *sql-stream*)
         (out-args args))
        (CASE
         (write-string "CASE " *sql-stream*)
         (output-sql (first args) database)
         (out-args (rest args)))
        (otherwise				; COALESCE
         (call-next-method))))))

(defclass sql-call-procedure-exp (sql-function-exp)
  ((result :initarg :result :initform nil))	; e.g. {?=call proc-name(args)}
  (:documentation "User-defined stored procedure or action query call expression,
 e.g. {call set_order_status(?,'SHIPPED')}."))

;(defclass sql-backend-function-exp (sql-function-exp)
;  ()
;  (:documentation "A backend-specific function expression, e.g. CEIL(expr)."))
;

(defmethod output-sql ((expr sql-call-procedure-exp) database)
  (warn 'sql-undefined-function-warning
        :name (if-bind (result (slot-value expr 'result))
                (cons (name expr) result)
                (name expr))
        :args (slot-value expr 'args))
  #|(warn 'call-procedure-warning
       :format-control "Undefined expansion of user-defined procedure call for database ~S:
  ~@[~S=~]~A ~S, .
  Default SQL function expression syntax applied."
       :format-arguments (list database (slot-value expr 'result) (name expr)
                               (slot-value expr 'args)))|#
  (call-next-method))

(defclass sql-join-exp (sql-relational-exp)
  ((on :initarg :on :initform nil)
   (using :initarg :using :initform nil)))	; can be :natural

(defun %output-join-exp (operator components on using database &optional (paren t))
  (when paren (write-char #\( *sql-stream*))
  (output-sql (first components) database)
  (write-char #\Space *sql-stream*)
  (when (eq using :natural)
    (write-string "NATURAL " *sql-stream*))
  (write-string (symbol-name operator) *sql-stream*)
  (write-char #\Space *sql-stream*)
  (output-sql (second components) database)
  (cond (on
         (write-string " ON " *sql-stream*)
         (output-sql on database))
        ((and using (neq using :natural))
         (write-string " USING " *sql-stream*)
         (output-sql-identifiers (listify using) database)))
  (when paren (write-char #\) *sql-stream*)))

(defmethod output-sql ((expr sql-join-exp) database)
  (with-slots (operator components on using) expr
    (%output-join-exp operator components on using database)))
  ;(error 'sql-reader-error 
  ;       :format-control "No JOIN expansion defined for generic database."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Query Statements

(defclass sql-query-exp (sql-expression)
  ((selections :initarg :selections :initform nil)
   (all :initarg :all :initform nil)
   (flatp :initarg :flatp :initform nil)
   (set-operation :initarg :set-operation :initform nil) ; UNION, INTERSECTION, EXCEPT
   (distinct :initarg :distinct :initform nil)
   (from :initarg :from :initform nil)
   (where :initarg :where :initform nil)
   (group-by :initarg :group-by :initform nil)
   (having :initarg :having :initform nil)
   (order-by :initarg :order-by :initform nil)
   (pivot :initarg :pivot :initform nil)		; for Access only
   (limit :initarg :limit :initform nil)
  )
  (:documentation "An SQL SELECT query."))

(defmethod collect-table-refs ((expr sql-query-exp))
  (remove-duplicates (collect-table-refs (slot-value expr 'where))
                     :test #'equal 
                     :key 'name))
                     ;(collect-table-refs (slot-value expr 'set-operation))

(defmethod collect-parameters ((expr sql-query-exp))
  ;(remove-duplicates ??
  (nconc ;(collect-parameters (slot-value expr 'selections))
    (collect-parameters (slot-value expr 'where))
    (collect-parameters (slot-value expr 'group-by))
    (collect-parameters (slot-value expr 'having))
    (collect-parameters (slot-value expr 'set-operation))
    (collect-parameters (slot-value expr 'order-by))
    (collect-parameters (slot-value expr 'pivot))))

(defparameter *select-arguments*
  '(:all :database :distinct :flatp :from :where :group-by
    :having :order-by :set-operation :pivot :limit :refresh))	; but not :transform!
    ;:offset

(defun query-arg-p (symbol)
  (member symbol *select-arguments*))

(defun query-get-selections (select-args)
  "Return two values: the list of select-args up to the first keyword,
 uninclusive, and the args from that keyword to the end."
  (if-bind (first-key-arg (position-if #'query-arg-p select-args))
    (values (subseq select-args 0 first-key-arg)
            (subseq select-args first-key-arg))
    select-args))

(defun sql-query-exp (&rest args)
  (multiple-value-bind (selections key-args) (query-get-selections args)
    (destructuring-bind (&key all flatp set-operation distinct
                              from where group-by having order-by pivot limit
                         &allow-other-keys)
        key-args
      (unless selections
        (error 'sql-reader-error
               :format-control "No target columns supplied to select statement."))
      (unless from
        (error 'sql-reader-error
               :format-control "No source tables supplied to select statement."))
      (make-instance 'sql-query-exp
                     :selections selections
                     :all all :flatp flatp :set-operation set-operation
                     :distinct distinct :from from :where where
                     :group-by group-by :having having :order-by order-by
                     :pivot pivot :limit limit))))

(defun collect-types (expr)
 ;;; Return a list of lisp-types for selection arguments.
  ;; Is not invoked if the :result-types argument is given explicitly.
  (typecase expr
    (sql-column
     (getf (plist expr) :type))
    (sql-parameter
     (getf (plist expr) :type))
    (sql-cast-exp
     (slot-value expr 'modifier))
    (sql-query-exp			; beware of :as and *
     (with-slots (selections) expr
       (do* ((result ())
             (rest selections (if (eq arg :as) (cddr rest) (cdr rest)))
             (arg (first rest) (first rest)))
           ((or (null rest) (eq (name arg) '*))
            (nreverse result))
         (if (eq arg :as)		; try to exctact type from the correlation name
             (unless (first result)	; if none was available from the previous arg
               (setf (first result) (collect-types (second rest))))
             (push (collect-types arg) result))) ))
    (number (type-of expr)) ))

(defmethod output-sql ((expr sql-query-exp) database)
  (with-slots (distinct selections from where group-by having set-operation order-by limit)
      expr
    (when *in-subselect*
      (write-char #\( *sql-stream*))
    (write-string "SELECT " *sql-stream*)
    (when distinct
      (write-string "DISTINCT " *sql-stream*)
      (unless (eql distinct t)
        (write-string "ON " *sql-stream*)
        (output-sql distinct database)
        (write-char #\Space *sql-stream*)))
    (let ((*in-subselect* t))			; with aliases and inner "()" enable
      ;; A la (output-sql (apply #'vector selections)), but hanlde AS keyword.
      (loop for (arg . rest) on selections
            unless (eq arg :as)
            do (output-sql arg database)
            and when rest
                  if (eq (first rest) :as)
                  do (write-string " AS " *sql-stream*)
                  else do (write-char #\, *sql-stream*))
      (write-string " FROM " *sql-stream*)
      (if (listp from)
          (output-sql (apply #'vector from) database)
          (output-sql from database))
      (when where
        (write-string " WHERE " *sql-stream*)
        (output-sql where database))
    )
    (when group-by
      (write-string " GROUP BY " *sql-stream*)
      (if (listp group-by)
          (output-sql (apply #'vector group-by) database)
          (output-sql group-by database)))
    (when having
      (write-string " HAVING " *sql-stream*)
      (output-sql having database))
    (when set-operation
      (output-sql set-operation database))
    (when order-by
      (write-string " ORDER BY " *sql-stream*)
      (if (listp order-by)
          ;; A la (output-sql (apply #'vector order-by) database), but handle ASC/DESC
          (loop with prev = nil
                for arg in order-by
                if (sql-keywordp arg) do (write-char #\Space *sql-stream*)
                else if prev do (write-char #\, *sql-stream*)
                do (output-sql arg database)
                   (setq prev t))
          ;; ASC or DESC can be embedded into sql-column as :asc or :desc
          (output-sql order-by database)))
    (when limit
      (write-string " LIMIT " *sql-stream*)
      (output-sql limit database))
    ;(when offset
    ;  (write-string " OFFSET " *sql-stream*)
    ;  (output-sql offset database))
    (when *in-subselect*
      (write-char #\) *sql-stream*))))

(defclass sql-set-query-exp (sql-relational-exp)
  ((modifier :initarg :modifier :initform nil)))	; all or distinct

(defmethod output-sql ((expr sql-set-query-exp) database)
  (with-slots (operator components modifier) expr
    (let ((*in-subselect* nil))
      (dolist (arg components)
        (write-char #\Space *sql-stream*)
        (write-string (string operator) *sql-stream*)
        (when modifier
          (write-char #\Space *sql-stream*)
          (write-string (string modifier) *sql-stream*))
        (write-char #\Space *sql-stream*)
        (output-sql arg database)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; DML Statements

(defclass sql-insert-exp (sql-expression)
  ((into :reader table :initarg :into :initform nil)
   (columns :initarg :columns :initform nil)
   (values :initarg :values :initform nil)
   (query :initarg :query :initform nil))
  (:documentation "An SQL INSERT statement."))

(defun sql-insert-exp (&key into columns values cv-alist av-pairs query)
  (unless into
    (error 'sql-reader-error :format-control ":into keyword not supplied for INSERT."))
  (cond ((and cv-alist (null values) (null columns) (null query))
         (setq columns (mapcar #'car cv-alist)
               values (mapcar #'cdr cv-alist)))
        ((and av-pairs (null values) (null columns) (null query))
         (setq columns (mapcar #'first av-pairs)
               values (mapcar #'second av-pairs)))
        ((and (null cv-alist) (null av-pairs) (xor values query))
         (when (and columns values
                    (/= (if (atom columns) 1 (length columns))
                        (if (atom values) 1 (length values))))
           (error 'sql-reader-error :format-control
             "The number of columns differs from the number of values for INSERT.")))
        (t
         (error 'sql-reader-error
                 :format-control "Bad or ambiguous keyword combination for INSERT.")))
  (make-instance 'sql-insert-exp :into into
                                 :values values :columns columns :query query))
    
(defmethod collect-parameters ((expr sql-insert-exp))
  (nconc (collect-parameters (slot-value expr 'values))
         (collect-parameters (slot-value expr 'query))))

(defmethod output-sql ((expr sql-insert-exp) database)
 ;;; Treat symbol and string INTO specially.
  (with-slots (into columns values query) expr
    (write-string "INSERT INTO " *sql-stream*)
    (typecase into
      (symbol (write-sql-identifier into *sql-stream* database))
      (string (write-sql-identifier into *sql-stream* database))
      (t (output-sql into database)))
    (when columns
      (write-char #\Space *sql-stream*)
      (output-sql-identifiers columns database))
      ;(output-sql (listify columns) database))
    (when values
      (write-string " VALUES " *sql-stream*)
      (output-sql (listify values) database))
    (when query
      (write-char #\Space *sql-stream*)
      (output-sql query database))))


(defclass sql-update-exp (sql-expression)
  ((table   :reader table   :initarg :table :initform nil)
   (columns :reader columns :initarg :columns :initform nil)
   (values :initarg :values :initform nil)
   (where :initarg :where :initform nil))
  (:documentation "An SQL UPDATE statement."))

(defun sql-update-exp (table &key columns values cv-alist av-pairs where)
  (unless table
    (error 'sql-reader-error :format-control "Table not supplied for UPDATE."))
  (cond ((and columns values (null cv-alist) (null av-pairs))
         (unless (= (if (atom columns) 1 (length columns))
                    (if (atom values) 1 (length values)))
           (error 'sql-reader-error :format-control
             "The number of columns differs from the number of values for UPDATE.")))
        ((and cv-alist (null columns) (null values))
         (setq columns (mapcar #'car cv-alist)
	       values (mapcar #'cdr cv-alist)))
        ((and av-pairs (null columns) (null values))
         (setq columns (mapcar #'first av-pairs)
	       values (mapcar #'second av-pairs)))
        (t
         (error 'sql-reader-error
                 :format-control "Bad or ambiguous keyword combination for UPDATE.")))
  (make-instance 'sql-update-exp :table table
                                 :columns columns :values values
			         :where where))

(defmethod collect-parameters ((expr sql-update-exp))
  (nconc (collect-parameters (slot-value expr 'values))
         (collect-parameters (slot-value expr 'where))))

(defmethod output-sql ((expr sql-update-exp) database)
 ;;; Access NB: Scalar subqueries are not allowed in values
  (with-slots (table where columns values) expr
    (write-string "UPDATE " *sql-stream*)
    (typecase table
      (symbol (write-sql-identifier table *sql-stream* database))
      (string (write-sql-identifier table *sql-stream* database))
      (t (let ((*in-subselect* t))			; enable aliases
           (output-sql table database))))
    (write-string " SET " *sql-stream*)
    (let ((*in-subselect* t))				; can mention scalar subqueries
      (output-sql (coerce (mapcar (lambda (column value) (sql-operation '== column value))
                                  (listify columns) (listify values))
                          'vector)
                  database))
    (when where
      (write-string " WHERE " *sql-stream*)
      (output-sql where database))))


(defclass sql-delete-exp (sql-expression)
  ((from :initarg :from :initform nil)
   (where :initarg :where :initform nil))
  (:documentation "An SQL DELETE statement."))

(defun sql-delete-exp (&key from where)
  (unless from
    (error 'sql-reader-error :format-control "FROM not supplied for DELETE."))
  (make-instance 'sql-delete-exp :from from :where where))

(defmethod collect-parameters ((expr sql-delete-exp))
  (collect-parameters (slot-value expr 'where)))

(defmethod output-sql ((expr sql-delete-exp) database)
  (with-slots (from where) expr
    (write-string "DELETE FROM " *sql-stream*)
    (typecase from
      (symbol (write-sql-identifier from *sql-stream* database))
      (string (write-sql-identifier from *sql-stream* database))
      (t (let ((*in-subselect* t))				; enable aliases
           (output-sql from database))))
    (when where
      (write-string " WHERE " *sql-stream*)
      (output-sql where database))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; DDL Statements
;;; Defining a separate class for every drop statement is reasonable only for
;;; an application parsing SQL scripts.

(defun %output-drop-exp (kind name database &optional plist)
 ;;; Args: plist  May include ON for kind :index as well as :restrict or :cascade
  (when (and (getf plist :restrict) (getf plist :cascase))
    (error 'sql-reader-error
           :format-control "Ambiguous keyword combination for DROP: ~s"
           :format-arguments (list plist)))
  (write-string "DROP " *sql-stream*)
  (princ kind *sql-stream*)
  (write-char #\Space *sql-stream*)
  (output-sql-identifier name database)
  (%output-plist plist database))

(defun write-drop-exp-to-string (&rest args)
  (with-sql-string-stream ()
    (apply #'%output-drop-exp args)))

(defclass sql-ddl-expression (sql-expression)
  ((name :initarg :name :accessor name)))

;;; Table

(defclass sql-create-table-exp (sql-ddl-expression)
  ((columns :initarg :columns)
   ;(description :initarg :description)
   (constraints :initarg :constraints :accessor constraints :initform nil)
   (plist :initarg :plist :initform nil))	; with, tablespace etc.
  (:documentation "An SQL CREATE TABLE statement."))

(defun sql-create-table-exp (name description &rest args
                             &key constraints &allow-other-keys)
  (let ((columns (loop for column in description
                       if (and (consp column) (keywordp (first column)))
                       do (push (if (eq (first column) :constraint) ;trim the keyword
                                    (rest column)
                                    column)
                                constraints)
                       else collect column)))
    (make-instance 'sql-create-table-exp :name name :columns columns
 					 :constraints constraints
                                         :plist (remove-properties args '(:constraints)))))


(defun %output-create-table-exp (name columns constraints database &optional plist)
  (write-string "CREATE TABLE " *sql-stream*)	; {GLOBAL | LOCAL | TEMPORARY} TABLE
  (output-sql-identifier name database)
  (write-string " (" *sql-stream*)
  (loop for (arg . rest) on columns
        when (and (consp arg) (keywordp (first arg)))
        do (output-table-constraint arg database)
        else do (output-column arg database)
        when rest
        do (write-string ", " *sql-stream*))
  (loop for constraint in constraints
        do ;(write-string ", " *sql-stream*)
           (output-table-constraint constraint database t))
  (write-char #\) *sql-stream*)
  (%output-plist plist database))

(defmethod output-sql ((expr sql-create-table-exp) database)
  (with-slots (name columns constraints plist) expr
    (%output-create-table-exp name columns constraints database plist)))

(defun output-column (column database &optional name-only)
  (cond ((consp column)
         (destructuring-bind (identifier &optional type &rest constraints) column
           (output-sql-identifier identifier database t)
           (write-char #\Space *sql-stream*)
           (cond ((and (null type) (typep identifier 'sql-column))
                  (setq type (getf (plist identifier) :type)))
                 ((and (sql-keywordp type)	     ; skip :unique, :not-null, or the like
                       (typep identifier 'sql-column))
                  (setq constraints (cons type constraints)
                        type (getf (plist identifier) :type))))
           (unless type
             #1=(error 'sql-reader-error
                       :format-control "Type is unspecified for column ~S."
                       :format-arguments (list column)))
           (write-string (apply #'db-sql-type-specifier database
                                (first-or-self type)
                                (and (consp type) (rest type)))
                         *sql-stream*)
           (when (eq (first constraints) :default)
             (write-string " DEFAULT " *sql-stream*)
             (output-sql (second constraints) database)
             (setq constraints (cddr constraints)))
           (output-column-constraints constraints database)))
        (name-only
         (output-sql-identifier column database t)) ; only name when in ALTER TABLE DROP
        ((typep column 'sql-column)
         (output-sql-identifier column database t)
         (write-char #\Space *sql-stream*)
         (if-bind (type (getf (plist column) :type))
           (write-string (apply #'db-sql-type-specifier database
                                (if (consp type) (first type) type)
                                (and (consp type) (rest type)))
                         *sql-stream*)
           #1#))
         (t (error 'sql-reader-error
                   :format-control "Wrong column description ~S."
                   :format-arguments (list column)))))
  
(defmethod output-column-constraints (constraints database)
  (when (eq (first constraints) :constraint)		; provide constraint name
    (write-string " CONSTRAINT " *sql-stream*)
    (output-sql-identifier (second constraints) database)
    (setq constraints (cddr constraints)))
  (loop with constraints = constraints
        while constraints
        do (write-char #\Space *sql-stream*)
           (case (first constraints)
             ((:not-null :primary-key :unique)
              (output-sql (first constraints) database)
              (setq constraints (cdr constraints)))
             (:references
              (write-string "REFERENCES " *sql-stream*)
              (output-sql-identifier (second constraints) database)
              (output-sql-identifiers (third constraints) database)
              ;(output-sql (listify (third constraints)) database)
              (setq constraints (cdddr constraints)))
             ((:on-update :on-delete)
		;; SQL-99: <referential action> ::= CASCADE | SET NULL | SET DEFAULT
		;;				 | RESTRICT | NO ACTION (default)
              (output-sql (first constraints) database)
              (case (second constraints)
                (:cascade  (write-string " CASCADE" *sql-stream*))
                (:set-null (write-string " SET NULL" *sql-stream*))
                (otherwise (error 'sql-reader-error
                            :format-control "Keyword :cascade or :set-null expected.")))
              (setq constraints (cddr constraints)))
             (:check
              (write-string "CHECK" *sql-stream*)
 	      ;; relational-exp is always enclosed in () - do we need listify?
              (output-sql (listify (second constraints)) database)
              (setq constraints (cddr constraints)))
             (otherwise
              (error 'sql-reader-error
                     :format-control "Unexpected column constraint ~S."
                     :format-arguments constraints)))))

(defgeneric output-table-constraint (constraint database &optional separator)
 ;;; Args: separator If true, output comma before
  ;; Value: True = success,
  ;;        NIL  = constraint is not supported by back-end and was not actually output.
  ;;		   (The specialization should not signal in that case.)
 (:method (constraint database &optional separator)
  (when separator
    (write-char #\, *sql-stream*))
  (when (eq (first constraint) :constraint)
    (setq constraint (rest constraint)))
  (unless (keywordp (first constraint))
    (write-string "CONSTRAINT " *sql-stream*)
    (output-sql-identifier (first constraint) database)	; provide constraint name
    (write-char #\Space *sql-stream*)
    (setq constraint (rest constraint)))
  (when constraint		; only name when used in ALTER TABLE DROP CONSTRAINT
    (case (first constraint)
      (:foreign-key
       (destructuring-bind (cols ref-keyword parent-table referenced-cols
                            &key on-update on-delete)
           (rest constraint)
         (unless (eq ref-keyword :references)
           (error 'sql-reader-error :format-control "Keyword :references expected."))
         (write-string "FOREIGN KEY" *sql-stream*)
         (output-sql-identifiers cols database)
         ;(output-sql (listify cols) database)
         (write-string " REFERENCES " *sql-stream*)
         (output-sql-identifier parent-table database)
         (output-sql-identifiers referenced-cols database)
         ;(output-sql (listify referenced-cols) database)
         (when-bind (on (or on-update on-delete))
           (write-string (if on-update "ON UPDATE" "ON DELETE") *sql-stream*)
           (case on
             (:cascade  (write-string " CASCADE" *sql-stream*))
             (:set-null (write-string " SET NULL" *sql-stream*))
             (otherwise (error 'sql-reader-error
                         :format-control "Keyword :cascade or :set-null expected."))))))
      ((:primary-key :unique :not-null)
       (output-sql (first constraint) database)
       (output-sql-identifiers (second constraint) database))
      (:check
       (output-sql (first constraint) database)
       (output-sql (listify (second constraint)) database))
      (otherwise
       (error 'sql-reader-error :format-control "Unexpected table constraint ~S."
           	  	        :format-arguments (list constraint)))))
  t)
 (:method ((constraint string) database &optional separator)
  (when separator
    (write-char #\, *sql-stream*))
  (write-string constraint *sql-stream*)) )

(defclass sql-alter-table-exp (sql-ddl-expression)
  ((operator :initarg :operator)
   (args :initarg :args))
   ;(add :initform nil)
   ;(modify :initarg :modify :initform nil)
   ;(drop :initarg :drop :initform nil))
  (:documentation "An SQL ALTER TABLE statement."))

;;; Index

(defclass sql-create-index-exp (sql-ddl-expression)
  ((on :initarg :on)
   (columns :initarg :columns)
   ;; Options that should be processed specially:
   ;;  :unique t
   ;;  :type :== { :bitmap | :btree | :hash ...} - customize in specializations.
   ;; The other are output as an usual plist.
   (plist :initarg :plist :initform nil))
  (:documentation "An SQL CREATE INDEX statement."))

(defmethod output-sql ((expr sql-create-index-exp) database)
  (with-slots (name on columns plist) expr
    (write-string "CREATE " *sql-stream*)
    (when (getf plist :unique) (output-sql :unique database))
    ;(when-bind (type (getf plist :type)) (output-sql type database))	; oracle
    (write-string " INDEX " *sql-stream*)
    (output-sql-identifier name database)
    (write-string " ON " *sql-stream*)
    (output-sql-identifier on database)
    ;(when-bind (type (getf plist :using)) (output-sql-identifier type database t))
    ;(output-sql-identifiers columns database)
    (output-sql (listify columns) database)	; to index on expressions (PostgreSQL)
    (%output-plist (remove-properties plist '(:unique :type)) database)))

  
(defclass sql-create-view-exp (sql-ddl-expression)
  ((columns :initarg :columns)
   (query :initarg :query :initform nil)
   (plist :initarg :plist :initform nil))	; with, force or other options
  (:documentation "An SQL CREATE VIEW statement."))

(defmethod output-sql ((expr sql-create-view-exp) database)
  (with-slots (name columns query plist) expr
    (write-string "CREATE " *sql-stream*)
    (case (getf plist :if-exists)
      (:replace (write-string "OR REPLACE " *sql-stream*)))
    (when (getf plist :force) (write-string "FORCE " *sql-stream*))
    (write-string "VIEW " *sql-stream*)
    (output-sql-identifier name database)
    (when columns				; does not work for Jet 3.5
      (output-sql-identifiers columns database))
      ;(output-sql (listify columns) database))
    (write-string " AS " *sql-stream*)
    (output-sql query database)
    (loop for (arg value) on plist by #'cddr
          do (case arg
               ((:if-exists :force))
               (otherwise
                 (write-char #\Space *sql-stream*)
                 (output-sql arg database)
                 (write-char #\Space *sql-stream*)
                 (output-sql value database))))))

(defun %output-create-view-no-columns (expr database) ; keywords)
 ;;; Columns and other properties may not works for some back-ends,
  ;; e.g. Jet 3/4, SQLite3.
  ;; Args: keywords ::= T means all allowed, or list of keywords allowed?
  ;; CAUTION: As we are altering the expression instances, they can stop working
  ;;          when shared with other database types!
  (with-slots (name columns query plist) expr
    (write-string "CREATE VIEW " *sql-stream*)
    (output-sql-identifier name database)
    (when columns
      (if (typep query 'sql-query-exp)
          (setf (slot-value query 'selections)
                (do ((acc ())
                     (columns columns (rest columns))
                     (selections (slot-value query 'selections) (rest selections)))
                    ((null (and columns selections))
                     (nreconc acc selections))
                  (let ((column (first columns))
                        (arg (first selections)))
                    (setq acc (if (equal (name column) (name arg))	; name coincides -
                                  (cons arg acc)			; avoid circularity
                                  (list* column :as arg acc)))))
                columns nil)
          (error 'sql-user-error
             :message "Column list is not supported in CREATE VIEW.")))
    (write-string " AS " *sql-stream*)
    (output-sql query database)
    (when plist
      (warn 'sql-simple-warning
            :format-control "CREATE VIEW has ignored properties: ~s."
            :format-arguments plist))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Juxtaposition of Lisp and SQL types
;;; Rule of thumb: keyword denotes a native SQL type, non-keyword symbol - a Lisp type

(defmethod db-sql-type-specifier (database lisp-type &rest args)
 ;;; Default specifier - use lisp-type as is.
  (declare (ignore database))
  (format nil "~A~@[(~{~A~^,~})~]" lisp-type args))

(defmethod db-sql-type-specifier (database (lisp-type (eql 'character)) &rest args)
 ;;; Fixed-length character type. For PostgreSQL if length=1, use "char"
  (declare (ignore database lisp-type))
  (format nil "CHAR(~A)" (or (first args) 1)))
#+lispworks
(defmethod db-sql-type-specifier (database (lisp-type (eql 'lw:simple-char)) &rest args)
  (declare (ignore database lisp-type))
  (apply #'db-sql-type-specifier database 'character args))

(defmethod db-sql-type-specifier (database (lisp-type (eql 'simple-string)) &rest args)
  (declare (ignore database lisp-type))
  (if args (format nil "CHAR(~A)" (first args)) "CHAR"))
(defmethod db-sql-type-specifier (database (lisp-type (eql 'simple-base-string))
                                  &rest args)
  (declare (ignore lisp-type))
  (apply #'db-sql-type-specifier database 'simple-string args))
#+lispworks
(defmethod db-sql-type-specifier (database (lisp-type (eql 'lw:simple-text-string))
                                  &rest args)
  (declare (ignore lisp-type))
  (apply #'db-sql-type-specifier database 'simple-string args))

(defmethod db-sql-type-specifier (database (lisp-type (eql 'string)) &rest args)
  (declare (ignore database lisp-type))
  (if args (format nil "VARCHAR(~A)" (first args)) "VARCHAR"))
(defmethod db-sql-type-specifier (database (lisp-type (eql 'base-string)) &rest args)
  (declare (ignore lisp-type))
  (apply #'db-sql-type-specifier database 'string args))
#+lispworks
(defmethod db-sql-type-specifier (database (lisp-type (eql 'lw:text-string)) &rest args)
  (declare (ignore lisp-type))
  (apply #'db-sql-type-specifier database 'string args))

(defmethod db-sql-type-specifier (database (lisp-type (eql 'integer)) &rest args)
  (declare (ignore database lisp-type))
  (if args
      (destructuring-bind (low high) args
        (cond ((<= -32768 low high 32767) "SMALLINT")
              ;((<= 0 low high 255) "BYTE")
              (t "INTEGER")))
      "INTEGER"))

(defmethod db-sql-type-specifier (database (lisp-type (eql 'bignum)) &rest args)
  (declare (ignore database lisp-type args))
  "BIGINT")

(defmethod db-sql-type-specifier (database (lisp-type (eql 'signed-byte)) &rest args)
  (declare (ignore database lisp-type))
  (if (or (null args) (<= (first args) 16)) "SMALLINT" "INTEGER"))

(defmethod db-sql-type-specifier (database (lisp-type (eql 'unsigned-byte)) &rest args)
  (declare (ignore database lisp-type))
  (cond ((or (null args) (<= (first args) 8)) "BYTE")
        ((<= (first args) 16) "SMALLINT")
        (t "INTEGER")))

(defmethod db-sql-type-specifier (database (lisp-type (eql 'float)) &rest args)
 ;; To convert from binary to decimal precision, multiply by 0.30103.
 ;; To convert from decimal to binary precision, multiply by 3.32193.
  (declare (ignore database lisp-type))
  (if args
      (destructuring-bind (low high) args
        (format nil "FLOAT(~A)" (max (float-precision low) (float-precision high))))
      "FLOAT"))

(defmethod db-sql-type-specifier (database (lisp-type (eql 'double-float)) &rest args)
  (declare (ignore database lisp-type args))
  "DOUBLE PRECISION")

(defmethod db-sql-type-specifier (database (lisp-type (eql 'single-float)) &rest args)
  (declare (ignore database lisp-type args))
  "REAL")

(defmethod db-sql-type-specifier (database (lisp-type (eql 'simple-bit-vector)) &rest args)
  (declare (ignore database lisp-type))
  (if args (format nil "BINARY(~A)" (ceiling (first args) 8)) "BINARY"))

(defmethod db-sql-type-specifier (database (lisp-type (eql 'bit-vector)) &rest args)
  (declare (ignore database lisp-type))
  (if args (format nil "VARBINARY(~A)" (ceiling (first args) 8)) "VARBINARY"))

(defmethod db-sql-type-specifier (database (lisp-type (eql 'octet-vector)) &rest args)
  (declare (ignore database lisp-type))
  (if args (format nil "VARBINARY(~A)" (first args)) "VARBINARY"))

(defmethod db-sql-type-specifier (database (lisp-type (eql 'vector)) &rest args)
  (declare (ignore database lisp-type))
  (if args
      (destructuring-bind (type length) args
        (ecase (if (consp type) (first type) type)
          ((signed-byte unsigned-byte)
           (when (and (consp type) (second type))
             (setq length (* length (ceiling (second type) 8))))
           (format nil "VARBINARY(~A)" length))))
      "VARBINARY"))

(defmethod db-sql-type-specifier (database (lisp-type (eql 'universal-time))
                                  &rest args)
  (declare (ignore database lisp-type args))
  "TIMESTAMP")

;;; Local time to SQL-99

#+local-time
(defmethod db-sql-type-specifier (database (lisp-type (eql 'local-time:date))
                                  &rest args)
  (declare (ignore database lisp-type args))
  "DATE")

#+local-time
(defmethod db-sql-type-specifier (database (lisp-type (eql 'local-time:local-time))
                                  &rest args)
  (declare (ignore database lisp-type args))
  "TIMESTAMP")	;"WITHOUT TIME ZONE"

#+local-time
(defmethod db-sql-type-specifier (database (lisp-type (eql 'local-time:utc-time))
                                  &rest args)
  (declare (ignore database lisp-type args))
  "TIMESTAMP WITH TIME ZONE")


#||
(defun %output-drop-index-exp (name on database &optional plist)
  (write-string "DROP INDEX " *sql-stream*)
  (output-sql-identifier name database)
  (when on 
    (write-string " ON " *sql-stream*)
    (output-sql-identifier on database))
  (%output-plist plist database))

(defclass sql-drop-index-exp (sql-ddl-expression)
  ((on :initarg :on)
   ;(if-does-not-exist :initarg :if-does-not-exist :initform :error))
   (plist :initarg :plist :initform nil))	; :if-does-not-exist, :purge, :cascade
  (:documentation "An SQL DROP INDEX statement."))

(defmethod output-sql ((expr sql-drop-index-exp) database)
  (with-slots (name on) expr
    (write-string "DROP INDEX " *sql-stream*)
    (output-sql-identifier name database)
    (when on 
      (write-string " ON " *sql-stream*)
      (output-sql-identifier on database))))
||#
