;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2005 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Functional interface to SQL

(in-package :ystok.sql.sys)

;(defparameter *trace-sql* nil
;  "If set to a true value, the sql commands sent to the database are additionally
;written to *TRACE-OUTPUT*, or, in the case *TRACE-SQL* is bound to an output stream,
;to that stream. Useful for debugging.")

(defvar *initialized-database-types* ()
  "Contains a list of database types which have been initialized by calls
to initialize-database-type.")

(defun initialize-database-type (&key (database-type *default-database-type*))
  "Initialize the given database-type, if it is not already initialized, 
 as indicated by *initialized-database-types*.
 Value: Environment object if success."
  (unless (member database-type *initialized-database-types*)
    (when-bind (environment (db-type-initialize database-type))
      (setf (get database-type 'environment) environment)
      (unless *default-environment*
        (setq *default-environment* environment))
      (push database-type *initialized-database-types*)
      environment)))

(defun find-environment (db-type &key (errorp t))
  (cond ((typep db-type 'environment) db-type)
        ((get db-type 'environment))
        (errorp (cerror "Try to initialize it now."
                        'sql-user-error :message (format nil 
                        "Database environment ~S is not initialized." db-type))
                (initialize-database-type :database-type db-type))))

(defmacro with-database ((database) &body body)
 "Bind *default-database* to the database given in the scope of BODY"
  `(let ((*default-database* ,database))
     ,@body))

(declaim (ftype (function (cons) cons) %pair-to-dotted) (inline %pair-to-dotted))
(defun %pair-to-dotted (pair)
  (declare (type cons pair) (optimize (debug 0)))
  (cons (first pair) (second pair)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Connecting

(defvar *connect-if-exists* :warn-old
  "Default value for the if-exists parameter of the connect function.
 One of :new :warn-new :error :warn-old :old")

;; should this be something different from a list?
(defvar *connected-databases* ()
  "A list of all connected databases in all environments.")

(defun connected-databases (&optional environment)
 "Return the list of active database objects."
  (if environment
      (remove-if (complement #'connected)
                 (connections (find-environment environment)))
      *connected-databases*))

(defun find-database (database &key (errorp t) database-type user)
  (typecase database
    (database        ; Return the database object itself
     database)
    ((or string pathname)
     (or (find-if (lambda (db)
                    (and (equal (database-name db) database)
                         (imp user (equal (user-name db) user))
                         (imp database-type (equal (db-type db) database-type))))
                  (connected-databases))
	 (when errorp
	   (cerror "Return nil."
		   'sql-simple-error
		   :format-control "There exists no database called ~A."
		   :format-arguments (list database)))))))

;;; Standard parameters from connection string.
(defmethod db-name-from-spec (db-type (connection-spec string))
  (declare (ignore db-type))
  (first (split-seq #\/ connection-spec :count 1 :keep-empty-subseqs t)))

(defmethod db-user-name-from-spec (db-type (connection-spec string))
  (declare (ignore db-type))
  (let ((user-name (second (split-seq #\/ connection-spec :count 2
                                      :keep-empty-subseqs t))))
    user-name)) ;(if (equal user-name "") nil user-name)))

(defmethod db-password-from-spec (db-type (connection-spec string))
  (declare (ignore db-type))
  (third (split-seq #\/ connection-spec)))


(defmethod db-name-from-spec (db-type (connection-spec pathname))
  (declare (ignore db-type))
  connection-spec)


(defun connect (connection-spec &rest args
                                &key (if-exists *connect-if-exists*) 
                                     (database-type *default-database-type*)
                                     (make-default t)
                                &allow-other-keys)
 "Connects to an SQL database.
Arguments:
 CONNECTION-SPEC : Either \"database-name/user-name/password\", where
     database-name : Name of database to be used in connection string.
     user :  Username, default is empty string \"\".
     password : Default is NIL.
  or arbitrary flat lamblda-list, e.g. (my-database :user \"admin\")
 IF-EXISTS : Default: *connect-if-exists*. One of 
  :NEW      : A new connection is created regardless of prior connections existing. 
  :WARN-NEW : Like :new, but a warning is issued. 
  :OLD      : If an old connection exists, it is used, otherwise a new one is created. 
  :WARN-OLD : Like :OLD, but a warning is issued if an old connection is used. 
  :ERROR    : If a connection already exists, an error is signalled.
 DATABASE-TYPE : A database type (e.g., :odbc). Default: *default-database-type*.
 AUTOCOMMIT : Whether to open connection in autocommit mode. Default: nil.
 MAKE-DEFAULT : If non-nil, *default-database* is set to the database object.
 DRIVER-AV-PARIS : nil, string or pair list with driver-specific arguments
     passed to driver-connect.
 TIMEOUT    : Number of seconds to wait for a login request to complete;
              0 disables timeout and a connection attempt will wait indefinitely.
Value: the database object or NIL if cannot be connected."
  ;; NB: args may also include :user and/or :password - these overwrites those
  ;;	 specified in connection-spec
  (let ((old-db (find-database (db-name-from-spec database-type connection-spec)
                   :database-type database-type 
                   :user (or (getf args :user)
                             (db-user-name-from-spec database-type connection-spec))
                   :errorp nil))
        (result nil))
    (without-properties (args args :if-exists :database-type :make-default) 
      (if old-db
          (ecase if-exists
            (:new
             (setq result (apply 'db-connect database-type connection-spec args)))
            (:warn-new
             (setq result (apply 'db-connect database-type connection-spec args))
             (warn 'sql-exists-warning :old-db old-db :new-db result))
            (:error
             (restart-case (error 'sql-exists-error :old-db old-db)
               (create-new ()
                 :report "Create a new connection."
                 (setq result (apply 'db-connect database-type connection-spec args)))
               (use-old ()
                 :report "Use the existing connection."
                 (setq result old-db))))
            (:warn-old
             (setq result old-db)
             (warn 'sql-exists-warning :old-db old-db :new-db old-db))
            (:old
             (setq result old-db)))
          (setq result (apply 'db-connect database-type connection-spec args))))
    (when (and result (connected result))
      (pushnew result *connected-databases*)
      (pushnew result (connections (environment result)))
      (when make-default (setq *default-database* result))
      result)))

;(defmethod db-disconnect ((database t))
;  (signal-no-database-error))

(defun disconnect (&key (database *default-database*) (rollback t))
  "Close the connection to database.
 If rollback is t (the default), rolls back uncommited transactions."
 ;;; NB: We actually must have retained information whether the database was made default
  ;;	 on connecting and restore *default-database* only in that case.
  (cond ((null database)
         (signal-no-database-error))
        ((not (connected database))
         (signal-closed-database-error database))
        (t (when rollback (db-end-transaction database :rollback))
           (db-disconnect database)
           (setf *connected-databases* (delete database *connected-databases*)
                 (connected database) nil
                 (connections (environment database))
                  (delete database (connections (environment database))) )
           (when (eq database *default-database*)
             (setq *default-database* (first *connected-databases*))))))

(defun disconnect-all-databases (&optional environment)
  (dolist (database (connected-databases environment))
    (disconnect :database database)))

#+mcl
(pushnew #'disconnect-all-databases *lisp-cleanup-functions*)

#+lispworks
(lw:define-action "When quitting image" "Disconnect all databases"
                  'disconnect-all-databases
                  :once) ; because "When quitting image" is actually run twice

(defmacro with-connection ((database connection-spec &rest args) &body body)
 "Establish a database connection, execute the body, and guarante closing the connection.
Arguments:
 DATABASE : symbol the new database connection object is bound to in the body of the macro.
 CONNECTION-SPEC and other ARGS : same arguments as in CONNECT."
 `(let ((,database (connect ,connection-spec :if-exists :new ,@args)))
    (unwind-protect (progn ,@body)
      (disconnect :database ,database :rollback nil))))

#|(defmacro with-connection ((database connection-spec &rest args
                            &key (database-type '*default-database-type*)
                                 (bind-as-default t) &allow-other-keys)
                            &body body)
 ;BIND-AS-DEFAULT: If true, *DEFAULT-DATABASE* is bound to the database object
 ;in the scope of the body.
  (with-gensyms (successp)
    `(let* ((,database (connect ,connection-spec :if-exists :new
                                :database-type ,database-type
                                :make-default nil
                                ,@(remove-properties args
                                                     '(:database-type :bind-as-default))))
            (,successp nil)
            ,@(when bind-as-default
                `((*default-database* ,database))))
       (unwind-protect
         (multiple-value-prog1
           (progn ,@body)
           (setf ,successp t))
         (disconnect :database ,database :rollback (not ,successp)))))|#

(defun ensure-connected (database)
  (cond ((null database)
         (signal-no-database-error))
        ;((typep database database))) ; around environment
        ((not (connected database))
         (signal-closed-database-error database))))

;;; Statements

(defun execute-command (stmt-expression &key database parameters values timeout)
 "Execute an SQL command (text or expression) or prepared statement."
  ;; TODO: Use generics only to ease tracing/recording?
  (cond ((typep stmt-expression 'statement)
         (db-execute-command stmt-expression nil
                             :parameters parameters :values values  :timeout timeout))
        (t
         (ensure-connected (or database (setq database *default-database*)))
         ;(record-sql-command stmt-expression database)
         (db-execute-command database stmt-expression
                             :parameters parameters :values values  :timeout timeout)
         ;(record-sql-result (call-next-method) database))
) )     )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  DQL  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun query (query-expression &key database flatp result-types parameters values timeout)
 "Execute an SQL query (text or expression) or prepared cursor.
Value: A list of lists of values from the result set of that expression
 or list of values if flatp is true."
  (cond ((typep query-expression 'statement)
         (db-query query-expression nil :flatp flatp  :result-types result-types
                   :parameters parameters  :values values  :timeout timeout))
        (t
         (ensure-connected (or database (setq database *default-database*)))
         ;(record-sql-command query-expression database)
         (db-query database query-expression :flatp flatp  :result-types result-types
                   :parameters parameters  :values values  :timeout timeout)
         ;(record-sql-result (call-next-method) database))
       )))


;;; Generic printing helpers %print-tuple and %print-tuples
;;; Notes:
;;; - Sizes do not count one extra space used as a column separator.
;;; - If format-control looks like "~v...", we pass the size to the format function
;;;   directly, otherwise pass only datum.
;;; - If format-control is a function, we always pass the size to it (?).
;;; TODO1: Move them to Ystok-Library?
;;; TODO2: Let tuple be a vector?

(defun %print-tuple (tuple stream formats sizes)
  (terpri stream)
  (loop for arg in tuple
        and fmts = (if (consp formats) formats ()) then (rest fmts)
        and szs = sizes then (rest szs)
        for size = (first szs)
        for format-control = (or (first fmts)
                                 (and (atom formats) formats)
                                 (if size "~vA" "~A"))
        do (if arg
               (cond (size
                      (cond ((functionp format-control)
                             (funcall format-control stream size arg))
                            ((char-equal (char format-control 1) #\v)
                             (format stream format-control size arg))
                            ((let* ((string (format nil format-control arg))
                                    (length (length string)))
                               (write-string string stream)
                               (when (< length size)		; enhance with blank
                                 (format stream "~v@T" (- size length)))))))
                     ((functionp format-control)
                      (funcall format-control stream arg))
                     (t (format stream format-control arg)))
               (when size
                 (format stream "~v@T" size)))
           (write-char #\Space stream)))			; output separator

(defun %print-tuples (tuples &key titles formats sizes (divider-p t)
                                  (stream *standard-output*)
                             &aux divider)
 ;;; Args: tuples ::= ((datum11 datum12 ...) (datum21 datum22 ...) ...)
  ;;       divider-p  Output line of "-" after titles if any
  (unless sizes
    ;; Compute sizes automatically as the maximum width of print representaton of data.
    (setq sizes (if titles
                    (mapcar 'length titles)
                    (make-list (length (first tuples)) :initial-element 3)))
    (dolist (tuple tuples)
      (do ((rest tuple (rest rest))
           (fmts (if (consp formats) formats ()) (rest fmts))
           (szs sizes (rest szs)))
          ((null rest))
        (let* ((arg (first rest))
               (format-control (or (first fmts) (and (atom formats) formats)))
               (new (length (cond ((functionp format-control)
                                   (funcall format-control nil nil arg))
                                  (format-control
                                   (if (char-equal (char format-control 1) #\v)
                                       (format nil format-control nil arg)
                                       (format nil format-control arg)))
                                  ((stringp arg)
                                   arg)
                                  (t (princ-to-string arg))))))
          (when (< (first szs) new)
            (setf (first szs) new))))))
  (when titles
    (%print-tuple titles stream nil sizes)
    (when divider-p
      (%print-tuple (setq divider (mapcar (lambda (size)
                                            (make-string size :initial-element #\-))
                                          sizes))
                    stream nil sizes)))
  (dolist (tuple tuples)
    (%print-tuple tuple stream formats sizes))
  (when (and tuples divider)
    (%print-tuple divider stream nil sizes)))

(defun print-query (query-expression &key titles formats (sizes #\Tab)
                                     (stream *standard-output*)
                                     (database sql:*default-database*) result-types)
 ;;; Args: formats  List of format controllers, i.e. string or functions, e.g. (yl::~f).
  ;;		    If null for a column, the default "~A" format string assumed.
  ;;		    If not list but a controller, it is used for all columns.
  ;;       sizes    List of widths passed to format controllers.
  ;;		    If a character, use it as separator (useful for exporting data).
  ;;		    If =T, minimal sizes are computed - rather ineffective as
  ;;                requires the entire recordset to be loaded into memory.
  (cond ((characterp sizes)
         (when titles
           (format stream "~&~:{~S~:^~C~}"
                   (mapcar (lambda (title) (list (string title) sizes)) titles)))
         (map-query nil
                    (lambda (&rest tuple)
                      (terpri stream)
                      (loop for (arg . rest) on tuple
                            and fmts = (if (consp formats) formats ())
                                       then (rest fmts)
                            for format-control = (or (first fmts)
                                                     (and (atom formats) formats)
                                                     "~A")
                            when arg
                            do (cond ((functionp format-control)
                                      (funcall format-control stream arg))
                                     ((char-equal (char format-control 1) #\v)
                                      (format stream format-control nil arg))
                                     (t (format stream format-control arg)))
                            when rest
                            do (write-char sizes stream)))
                    query-expression
                    :database database :result-types result-types))
        ((eq sizes t)
         (multiple-value-bind (tuples columns)
             (sql:query query-expression :database database :result-types result-types)
           (%print-tuples tuples :stream stream
                          :titles (or titles columns) :formats formats)))  ; :sizes nil!
        (t
         ;; Otherwise explicit list or no sizes - try to output in fixed-width format
         (let (divider)
           (when titles
             (%print-tuple titles stream nil sizes)			; :formats nil!
             (%print-tuple (setq divider (mapcar (lambda (size)
                                                   (make-string size :initial-element #\-))
                                                 (or sizes (mapcar 'length titles))))
                           stream nil sizes))
           (map-query nil
                      (lambda (&rest tuple) (%print-tuple tuple stream formats sizes))
                      query-expression
                      :database database :result-types result-types)
           (when titles (%print-tuple divider stream nil sizes)))) )
  (values))


#+mcl
(defmethod initialize-instance ((cursor cursor) &key &allow-other-keys)
  (call-next-method)
  (terminate-when-unreachable cursor))

(defun open-cursor (query-expression &rest args &key (database *default-database*)
                                     &allow-other-keys) ; result-types parameters values)
 "Create a new cursor and executes EXPRESSION, without closing open statements.
 Returns the cursor object."
  (without-properties (args args :database)
    (apply #'db-open-cursor database query-expression args)))
                  ;:parameters parameters :values values))

(defun close-statement (statement &rest args)
 "Close cursor (dump result-set), unbind columns, drop descriptor memory."
  (apply #'db-close statement args))

(defun cancel-statement (statement)
 "Cancels arbitrary statement object."
  (db-cancel statement))

(defmacro with-cursor ((cursor query-expression &rest args
                       &key (database '*default-database*) &allow-other-keys)
                       &body body)
  `(let ((,cursor (db-open-cursor ,database ,query-expression
                                  ,@(remove-properties args '(:database)))))
     (unwind-protect (locally ,@body)
       (db-close ,cursor))))		; set cursor inactive or dispose memory 

(defun fetch-into (cursor object &rest args
                                 &key columns start-column end-column current
                                 &allow-other-keys)
 "Fill in the OBJECT with column data from the next or most recenlty fetched row.
Arguments:
 COLUMNS : List of names (strings or symbols, :bookmark is allowed) or numbers.
 START-COLUMN : First column number in the rowset row; default is 1.
 END-COLUMN : Last inclusive column number; default is the last column number.
 CURRENT : If true, do not fetch the next row but acquire data from the current row.
Value: The OBJECT itself."
  (declare (ignore columns start-column end-column current))
  (apply #'db-fetch-into cursor object args))

(defun fetch-item (cursor column &key current)
 "Return one datum from the next or most recenlty fetched row.
Argumens:
 COLUMN : Column specifier as a name (string or symbol, :bookmark is allowed) or number.
 CURRENT : If true, do not fetch the next row but acquire an item from the current row."
  (db-fetch-item cursor column :current current))

(defun fetch-rows (cursor &rest args &key count flatp &allow-other-keys)
 "Fetch the first COUNT rows of the query result set; (all rows, if COUNT is nil)."
  (declare (ignore count flatp))
  (apply #'db-fetch-rows cursor args))

(defun select (&rest args)
 "Execute a select statement and returns a list of fetched result rows.
Signature: select &rest selections &key all flatp set-operation distinct from
           where group-by having order-by database.
Value: List of tuples. When flatp is null (the default), each tuple is
  - either list of column values
  - or list of instances, all selections are view classes."
 ;;; NB1: Object selections is like (class-name1 ... class-nameN [selection1]...)
  ;;	  where ordinary selections from tables/views are allowed after classes.
  ;; NB2: The :from argument is allowed in object-oriended select and should mention
  ;;      "ordinary" tables/views that could be theta-joined with the ones extracted
  ;;      from the view class selections via the :where condition.
  (multiple-value-bind (selections key-args) (query-get-selections args)
    (when selections
      (let* ((rest (member-if-not #'symbolp selections))
             (names (ldiff selections rest)))
        (if (and names (every (lambda (arg) (find-class arg nil)) names))
            (object-select names rest key-args)
            (destructuring-bind (&key (database *default-database*) flatp result-types
                                      timeout &allow-other-keys)
                key-args
              (query (apply #'sql-query-exp args) :database database
                     :flatp flatp  :result-types result-types  :timeout timeout)))))))

(defun map-query (result-type function query-expression
                  &key (database *default-database*) result-types
                       parameters values timeout)
 "Execute a query and map a function over the result rows.
Arguments:
 RESULT-TYPE : Performs a type coersion of the value returned like in cl:map.
 FUNCTION : A function accepting either
   - as many arguments as there are result columns if flatp is nil (as called via apply),
   - or one argument if flatp is true (as called via funcall).
   Flatp can only be specified via QUERY-EXPRESSION [select ...], otherwise assumed nil.
 QUERY-EXPRESSION : A select expression."
  (let ((flatp (and (typep query-expression 'sql-query-exp)
                    (slot-value query-expression 'flatp)))
        (cursor (db-open-cursor database query-expression
                                :result-types result-types
                                :parameters parameters :values values  :timeout timeout)))
    (unwind-protect
        (ecase (if (atom result-type) result-type (first result-type))
          (list 
           (if flatp
               (do ((acc ()) datum)
                   ((eq (setq datum (db-fetch-item cursor 1 :eof *unbound-value*))
                        *unbound-value*)
                    (nreverse acc))
                 (setq acc (cons (funcall function datum) acc)))
               (do ((acc ())
                    (row (make-list (db-columns-count cursor))))
                   ((null (db-fetch-into cursor row)) (nreverse acc))
                 (setq acc (cons (apply function row) acc)))))
          ((nil)
           (if flatp
               (do (datum)
                   ((eq (setq datum (db-fetch-item cursor 1 :eof *unbound-value*))
                        *unbound-value*))
                 (funcall function datum))
               (do ((row (make-list (db-columns-count cursor))))
                   ((null (db-fetch-into cursor row)))
                 (apply function row))))
           #|(loop with row = (make-list (db-columns-count cursor))
                 ;while (db-fetch-rows cursor :count 1 :into row)
                 while (db-fetch-into cursor row)
                 do (if flatp (funcall function row) (apply function row))))|#
          ((vector string array bit-vector
            simple-vector simple-string simple-array simple-bit-vector
            base-string simple-base-string
            #+lispworks (lw:text-string) #+lispworks (lw:simple-text-string))
	   (do* ((result-set (db-fetch-rows cursor :flatp flatp) (rest result-set))
                 (result (make-sequence result-type (length result-set)))
                 (i 0 (1+ i)))
                ((null result-set) result)
             (declare (fixnum i))
             (let ((row (first result-set)))
               (setf (aref result i)
                     (if flatp (funcall function row) (apply function row)))))) )
      (db-close cursor))))

(defmacro do-query (((&rest args) query-expression
                     &key (database '*default-database*) result-types
                          parameters values timeout)
                    &body body)
 "Execute a query; for each result row, the body is executed, while the column 
 values are bound to symbols accessible in the body.
Arguments:
 ARGS : A list of symbols to bind the result columns to. May contain a &rest argument. 
  The last argument may be a stream argument of the form (:stream <var>); then, the
  symbol <var> is bound to a SQL-STREAM for input which is opened before execution 
  of the body and closed afterwards.
 QUERY-EXPRESSION : A select expression.
 BODY : body to execute for each row."
  (let* ((last-arg (car (last args)))
         (last-stream-p (and (consp last-arg) (eq :stream (car last-arg))))
         (simple-args (when last-stream-p (butlast args)))
         (stream-arg (when last-stream-p (cadr last-arg)))
         (arglen (let ((rest-pos (position '&rest args)))
                   (cond ((not rest-pos)
                          (length args))
                         ((= rest-pos (- (length args) 2))
                          (if last-stream-p
                              (error "Stream and &rest arguments are not allowed together.")
                              nil))
                         (t
                          (error "Malformed argument list ~a." args)))))
         (cursor (gensym "q"))
	 (row (gensym "row")))
    `(with-cursor (,cursor ,query-expression :database ,database
                                             :result-types ,result-types
                                             :parameters ,parameters :values ,values
                                             :timeout ,timeout)
       ,(if last-stream-p
            `(let ((,stream-arg (make-sql-stream ,cursor :column ,(1- arglen))))
               (loop with data and data-found-p
                     do (multiple-value-setq (data data-found-p)
                            (db-fetch-rows ,cursor :count 1
                                           :ignore-columns '(:last)))
                     while data-found-p
                     do (open-sql-stream ,stream-arg)
                     (destructuring-bind ,simple-args data ,@body))
               (close ,stream-arg))
            `(loop with ,row = (make-list (db-columns-count ,cursor))
                   while (db-fetch-into ,cursor ,row)
                   do (destructuring-bind ,args ,row ,@body))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  DML  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun insert-records (&key into columns attributes values cv-alist av-pairs query 
			    (database *default-database*)
                            auto-unique-valued)
 "Insert a record or records into a table.
Arguments:
 INTO : Table or cursor;
        views are not supported by some DMBSs.
 AUTO-UNIQUE-VALUED : Column designator or list designators; allowed only for cursor.
Values: Those returned by execute-command or
        the value of the auto-unique-valued column or a list of such values."
 ;; Q: Shall we open a cursor "SELECT * FROM into [FOR INSERT]" atomatically when
 ;;    auto-unique-valued is specified? Maybe specialize via db-execute-command method?
  (if (typep into 'cursor)
      (let ((result (db-insert-row into
                     :columns (or columns attributes) :values values
                     :cv-alist (or cv-alist (mapcar #'%pair-to-dotted av-pairs)))))
        (if (and auto-unique-valued (integerp result))			; bookmark
            (progn (set-cursor-position into :bookmark :offset result)
              (if (consp auto-unique-valued)
                  (fetch-into into (make-list (length auto-unique-valued)) :current t
                              :columns auto-unique-valued)
                  (fetch-item into auto-unique-valued :current t)))
            result))
      (execute-command (sql-insert-exp :into into 
                                       :columns (or columns attributes) :values values 
                                       :cv-alist cv-alist :av-pairs av-pairs
                                       :query query)
                       :database database)))
  
(defun update-records (table &key columns attributes values cv-alist av-pairs where 
				  (database *default-database*)) ;&allow-other-keys
 "Update a record or records."
  (if (typep table 'cursor)
      (db-update-row table :columns (or columns attributes) :values values
                           :cv-alist (or cv-alist (mapcar #'%pair-to-dotted av-pairs))
                           :where where)		; where could be a bookmark
      (execute-command (sql-update-exp table
                                       :columns (or columns attributes) :values values
                                       :cv-alist cv-alist :av-pairs av-pairs
                                       :where where)
                       :database database)))

(defun delete-records (&key from where
			    (database *default-database*))
 "Delete a record or records."
 ;;; Value: 1> return code (e.g. SQL_NO_DATA = 100 if nothing was deleted)
  ;;	    2> number of records deleted.
  (if (typep from 'cursor)
      (db-delete-row from :where where)			; where could be a bookmark
      (execute-command (sql-delete-exp :from from :where where)
                       :database database)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  DDL  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Table

(defun create-table (name description &rest args
                     &key constraints (database *default-database*) &allow-other-keys)
 "Creates a database table.
Arguments:
 NAME : table name as symbol or expression.
 DESCRIPTION : A list of either column descriptors of the form:
  (column-name [type] [:default value]
        [:constraint constraint-name]
	[:primary-key | :unique | :not-null]
	[:references parent-table referenced-column]
	[:check condition]).
   COLUMN-NAME is a attibute expression or symbol (for the latter, keep-symbol-names!)
   TYPE is a symbol or a list of column type, precision, and scale.
  or constraint clauses of the form:
  ([:constraint [constraint-name]] {:primary-key | :unique} column-or-list)
  ([:constraint [constraint-name]] :foreign-key column-or-list
      :references parent-table referenced-column-or-list
      [{:on-update | :on-delete} {:cascade | :set-null}] )
  ([:constraint [constraint-name]] :check condition).
 CONSTRAINTS : A list of constraint clauses with an optional :constraint keyword:
  ([[:constraint] constraint-name] ...<as above>...)."
 (declare (ignore constraints))
 (execute-command (apply #'sql-create-table-exp name description
                         (remove-properties args '(:database))) ;:constraints constraints)
                  :database database))
  ;(pushnew name (gethash *default-schema* *table-schemas*) :test #'equal)))

(defun alter-table (name operator args &key (database *default-database*)
                         &allow-other-keys)
 "Modifies the design of a table.
Arguments:
 NAME : table name as symbol or expression.
 OPERATOR : :add, :alter, or :drop
 ARGS : list of either column descriptors of the form:
  column-name
  (column-name [type [:default value]
        [:constraint constraint-name]
	[:primary-key | :unique | :not-null]
	[:references parent-table referenced-column]
	[:check condition]]), where
   COLUMN-NAME is an attribute expression or symbol,
   TYPE is a symbol or a list of column type, precision, and scale;
  or constraint clauses of the form:
  (:constraint constraint-name)
  (:constraint constraint-name {:primary-key | :unique} column-or-list)
  (:constraint constraint-name :foreign-key column-or-list
      :references parent-table referenced-column-or-list
      [{:on-update | :on-delete} {:cascade | :set-null}] )
  (:constraint constraint-name :check condition)."
  (execute-command (make-instance 'sql-alter-table-exp :name name
                                  :operator operator :args args)
                                  ;:add add :modify modify :drop drop)
                   :database database))


(defun drop-table (table &rest args &key (database *default-database*) 
                                         (if-does-not-exist :error) &allow-other-keys)
 "Drops the TABLE from the DATABASE.
Arguments:
 IF-DOES-NOT-EXIST : Determine the behavior when the table does not exist:
   :ignore - return nil,  :error - signal an error.
 ARGS : Specifies drop behavior as a plist (with indicators :cascade, :restrict,
   and/or back-end specific)."
  (apply #'db-drop-table database table :if-does-not-exist if-does-not-exist
         (remove-properties args '(:database :if-does-not-exist))))

(defmethod db-drop-table (database table &rest args &key (if-does-not-exist :error))
  (when (ecase if-does-not-exist
          (:ignore (table-exists-p table :database database))
          (:error t))
    (execute-command (write-drop-exp-to-string :table table database
                        (remove-properties args '(:if-does-not-exist)))
             ;(string-append "DROP TABLE " (write-sql-identifier-to-string table database))
                     :database database)))

(defun list-tables (&key (database *default-database*) schema owner catalog)
 "List all tables in the database specified.
Arguments:
 SCHEMA (or OWNER for backward compatibility)
  nil  - only user tables are considered,
  :all - all tables are considered,
  string - only tables in the schema specified are considered.
 Value:  Table names as a list of strings."
  (db-list-tables database (or schema owner) :catalog catalog))

(defun table-exists-p (table &key (database *default-database*) schema owner catalog)
 "Value: boolean"
  (if (db-list-tables database (or schema owner) :table table :catalog catalog) t nil))

;;; View

(defun create-view (name &rest args &key columns attributes column-list
                                         as with with-check-option ;if-exists
                                         (database *default-database*) &allow-other-keys)
  (declare (ignore with))
  (execute-command (make-instance 'sql-create-view-exp
      :name name
      :columns (or columns attributes column-list)	; CommonSQL compatibility argument
      :query as
      :plist (if with-check-option			; CommonSQL compatibility argument
                 (list* :with :check-option
                   (remove-properties args '(:columns :attributes :as :with-check-option
                                             :database)))
                 (remove-properties args '(:columns :attributes :as :database))))
                   :database database))

(defun drop-view (view &rest args &key (database *default-database*)
                                       (if-does-not-exist :error) &allow-other-keys)
 "Drops the VIEW from the DATABASE.
Arguments:
 IF-DOES-NOT-EXIST : Determine the behavior when the view does not exist:
   :ignore - return nil, :error - signal an error.
 ARGS : Specifies drop behavior as a plist (with indicators :cascade, :restrict,
   and/or back-end specific)."
  (apply #'db-drop-view database view :if-does-not-exist if-does-not-exist
         (remove-properties args '(:database :if-does-not-exist))))

(defmethod db-drop-view (database view &rest args &key (if-does-not-exist :error))
  (when (ecase if-does-not-exist
          (:ignore (view-exists-p view :database database))
          (:error t))
    (execute-command (write-drop-exp-to-string :view view database
                       (remove-properties args '(:if-does-not-exist)))
             ;(string-append "DROP VIEW " (write-sql-identifier-to-string view database))
                     :database database)))

(defun list-views (&key (database *default-database*) schema owner catalog)
 "List all views in the database specified.
Arguments:
 SCHEMA (or OWNER for backward compatibility)
  nil  - only user tables are considered,
  :all - all tables are considered,
  string - only tables in the schema specified are considered.
 Value:  View names as a list of strings."
  (db-list-views database (or schema owner) :catalog catalog))

(defun view-exists-p (view &key (database *default-database*) schema owner catalog)
 "Value: boolean"
  (if (db-list-views database (or schema owner) :view view :catalog catalog) t nil))

;;; Index

(defun create-index (name &rest args &key on columns attributes unique type
                                         (database *default-database*)
                                     &allow-other-keys)
 "Arguments: type ::= :bitmap, :btree, or the like."
  (declare (ignore unique type))
  (execute-command (make-instance 'sql-create-index-exp 
                                  :name name :on on :columns (or columns attributes)
                                  :plist (remove-properties args
                                           '(:on :columns :attributes :database)))
                   :database database))

(defun drop-index (name &rest args
                        &key on (database *default-database*) (if-does-not-exist :error))
 "Drop index NAME from.
Arguments:
 ON : table name needed for some DBMSs, for example Access."
  (apply #'db-drop-index database name  :on on  :if-does-not-exist if-does-not-exist
         (remove-properties args '(:database :on :if-does-not-exist))))

(defmethod db-drop-index (database index &rest args &key on (if-does-not-exist :error))
 ;;; According SQL standard, the default method does not output ON in DROP INDEX
  (when (ecase if-does-not-exist
          (:ignore (index-exists-p index :database database :on on))
          (:error  t))
    (execute-command
     (write-drop-exp-to-string :index index database
                               (remove-properties args '(:on :if-does-not-exist)))
     :database database)))
   ;(with-sql-string-stream ()
   ;  (%output-drop-index-exp index on database
   ;                          (remove-properties args '(:on :if-does-not-exist))))
   ;(make-instance 'sql-drop-index-exp :name index :on on
   ;               :plist (remove-properties args'(:on :if-does-not-exist)))

(defun list-indexes (&key (database *default-database*) on schema owner catalog)
 "Return a list of strings representing index names in the DATABASE.
Arguments:
 SCHEMA : If a string denoting a user name, only indexes owned by the user are listed;
    if eql to :all then all indexes are listed.
 OWNER : Synonym of SCHEMA.
 ON : Limits the results to indexes on the specified tables.
   If equals to nil (the default), all tables are considered,
   Other meaningful values are a string, symbol or SQL expression
   representing a table name or a list of such table identifiers."
  (let ((schema (or schema owner)))
    (if (and on (neq on :all))
        (mappend (lambda (table) (db-list-indexes database table
                                                  :catalog catalog :schema schema))
                 (listify on))
        (db-list-indexes database :all :catalog catalog :schema schema))))

(defun index-exists-p (index &key (database *default-database*) on schema owner catalog)
 "Test for existence of an index called INDEX in the DATABASE.
Arguments:
 SCHEMA : If a string denoting a user name, only indexes owned by the user are checked;
    if eql to :all then all indexes are checked.
 OWNER : Synonym of SCHEMA.
 ON : Limits the results to indexes on the specified tables, see list-indexes.
    Passing nil (the default) as the ON value can be pointless
    because several tables can have an index with the same name imposed.
Value: Boolean"
  (if (member (write-sql-identifier-to-string index database t nil)
              (list-indexes :database database :on on
                            :schema (or schema owner) :catalog catalog)
              :test #'string-equal)
      t nil))

;;; Sequence

(defun create-sequence (name &rest args &key (database *default-database*)
                                        &allow-other-keys)
 "Creates a sequence generator called NAME in the DATABASE."
  (apply #'db-create-sequence database name (remove-properties args '(database))))

(defun drop-sequence (name &key (if-does-not-exist :error) (database *default-database*))
  "Drops the sequence generator NAME from the DATABASE.
 Arguments:
  IF-DOES-NOT-EXIST : Determine the behavior when the table does not exist:
    :ignore - returns nil, :error - signal an error."
  (ecase if-does-not-exist
    (:ignore
     (unless (sequence-exists-p name :database database)
       (return-from drop-sequence)))
    (:error t))
  (db-drop-sequence database name))

(defun alter-sequence (name start &rest args &key (database *default-database*)
                                             &allow-other-keys)
 "Set the non-issued base value of the sequence generator NAME to START.
After that, the next call of sequence-next will return (START + increment).
 Value: START returned."
  (apply #'db-alter-sequence database name start (remove-properties args '(database))))

(defun list-sequences (&key (database *default-database*) schema owner catalog)
 "Return a list of strings representing sequence names in the DATABASE.
Arguments:
 SCHEMA (or OWNER for backward compatibility)
   nil  - only user sequences are considered,
   string -  denoting a user name, only sequences owned by the user are listed;
   :all  -  all sequences are listed."
  (db-list-sequences database :catalog catalog :schema (or schema owner)))

(defun sequence-exists-p (name &key (database *default-database*) schema owner catalog)
  "Tests for the existence of a sequence called NAME in the DATABASE."
  (if (member (write-sql-identifier-to-string name database t nil)
              (list-sequences :database database
                              :schema (or schema owner) :catalog catalog)
              :test #'string-equal)
      t nil))
  
(defun sequence-next (name &key (database *default-database*))
 "Increment and return the next value in the sequence generator NAME."
  (db-sequence-next database name))

(defun sequence-last (name &key (database *default-database*))
 "Return the last issued value of the sequence generator NAME."
  ;; SQL-2003 BUG:
  ;;  The current base value is defined in SQL-2003 as the lowest/highest
  ;;  non-issued(??) value for the generator.
  (db-sequence-last database name))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  PREPARED STATEMENTS  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Works for insert and select statements

(defun prepare-statement (expression &key (database *default-database*)
                                          parameters values (paramset-size 1) timeout)
 "Create and prepare a statement for repeated execution.
Arguments:
 EXPRESSION : SQL expression to be prepared with parameter markers [?] embedded.
 PARAMETERS : List of the sql-parameter instances; only necessary if not all 
  expression is a verbatim SQL string. (Otherwise it is possible to deduce them
  including types and names from the expression.)
 VALUES : List of arbitrary Lisp objects; only used to supply type information
  if its not available in parameters; those values are not actually assigned.
Value: A prepared statement object."
  (ensure-connected database)
  (db-prepare-statement database expression :parameters parameters :values values
                        :paramset-size paramset-size  :timeout timeout))

(defmacro with-prepared-statement ((variable expression &rest args)
                                   &body body)
 "Prepares a (parameterized) statement for repeated execution.
 A new statement or cursor object is created and bound to the symbol VARIABLE. 
 EXECUTE-COMMAND or QUERY (and its variants) calls can be used inside this macro,
 with :values argument supplying parameter values."
  `(let ((,variable (prepare-statement ,expression ,@args)))
     (unwind-protect
         (progn ,@body)
       (close-statement ,variable))))

(defun map-parameters (result-type expression sequences
                       &key (database *default-database*))
 "Execute parameterized EXPRESSION over each of parameter tuples given by SEQUENCES."
  (apply #'db-map-parameters database result-type expression sequences))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Database configuration operations and inquiries about catalog and metadata (dictionary)
;;; The database-type can be one of the following:
;;; - db-type, i.e. environment type,
;;; - a back-end type, e.g. :access or :sqlite3,
;;; - a database class name, e.g. access-odbc-database.
;;; - a database connection object, e.g. posgtresql-odbc-database,
;;;   which supports CREATE/DROP DATABASE statement.

#+deprecated
(defun data-sources (&optional (db-type *default-database-type*))
 "Value: a list of (data-source description) - pairs"
   (db-data-sources db-type))

(defun create-database (connection-spec &rest args &key database-type &allow-other-keys)
 "Create a database in the DBMS."
  ;(when (stringp connection-spec)
  ;  (setq connection-spec (string-to-list-connection-spec connection-spec)))
  (apply 'db-create database-type connection-spec
         (remove-properties args '(:database-type))))

(defun destroy-database (connection-spec &rest args &key database-type &allow-other-keys)
 "Destroy a database in the DBMS."
  (apply 'db-destroy database-type connection-spec
         (remove-properties args '(:database-type))))

(defun list-databases (connection-spec &rest args &key database-type &allow-other-keys)
 "Value: a list of databases existing in the DBMS specified by DATABASE-TYPE and matching CONNECTION-SPEC."
  (apply 'db-list database-type connection-spec
         (remove-properties args '(:database-type))))

(defun status (&optional full)
 "Print information about the currently connected databases to *STANDARD-OUTPUT*.
 Args: FULL If true, more detailed information about each database is printed."
  (format t "~&CONNECTION STATUS: ~A" (timestring))
  (%print-tuples 
     (do* ((tuples ())
           (rest (connected-databases) (rest rest))
           (db (first rest) (first rest)))
          ((null rest)
           (nreverse tuples))
       (push (list* (if (eq db *default-database*) "*" +null-string+)
                    (database-name db)
                    (string-downcase (string (db-type db)))
                    (if (command-recording-stream db)
                        (if (result-recording-stream db) "Both" "Commands")
                        (if (result-recording-stream db) "Results" "nil"))
                    (when full
                      (list ;(if (conn-pool db) "t" "nil")
                            (length (db-list-tables db :all))
                            (length (db-describe-tables db "%" :table-type "VIEW")))))
             tuples))
      :titles `(,+null-string+ "DATABASE" "TYPE" "RECORDING" 
                ,@(when full '(#|POOLED|# "TABLES" "VIEWS")))
  )
  (values))


(defun list-attributes (table &key (database *default-database*) schema owner catalog)
  "List the columns of TABLE in the database specified.
 Value:  Columns as a list of strings."
  (db-list-columns database table (or schema owner) :catalog catalog))

(defun attribute-type (column table &key (database *default-database*) schema owner catalog)
 "Return the data type of the COLUMN in TABLE.
Arguments: DATABASE specifies the database to query, defaults to *default-database*.
Value: Keyword, which has the symbol name uppercased,
  or a list (data-type-keyword column-size decimal-digits)"
  (db-column-type database column table (or schema owner
                                            (when (typep column 'sql-column)
                                              (getf (plist column) :schema)))
                  :catalog catalog))


#||;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; EVALUATION ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;(defun describe-tables (&key (database *default-database*)
;                             catalog schema (table-name "%") (table-type "%"))
;  (db-describe-tables database catalog-name schema-name table-name table-type))

;; Drop non-generally accepted defaults for catalog (qualifier) or schema (owner) owner.
;; Please leave them for a database-specific db-describe-columns method.
(defun describe-columns (&key catalog	;(table-qualifier "")
                              schema	;(table-owner (user-name *default-database*))
                              (table-name "%")
                              (column-name "%")
                              (database *default-database*))
  (db-describe-columns database
                       (if catalog (string-upcase catalog))
                       (if schema (string-upcase schema))
                       (string-upcase table-name)
                       (string-upcase column-name)))
||#
