;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2005 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Database variables, classes, and generic functions prototyping back-end API

(in-package :ystok.sql.sys)

(defvar *default-database-type* :odbc
  "Specifies the default type of database - only :ODBC is supported for now")

(defvar *default-environment* nil "Default environment object.")
(defvar *default-database* nil "Default database object.")

(defparameter *null* nil
  "Lisp representation of SQL Null value, default = nil.
 May be locally bound to something else if a certain type is necessary.")

(defun boolean-true-p (value)
  "Test for Lisp value representing boolean truth: everything except 0, :false, or nil
 is true; the :true keyword is recommended."
 (cond ((numberp value) (if (zerop value) 0 1))
       ((eq value :false) nil)
       (value)))

(defparameter *default-read-binary-type* 'octet-vector
 "Lisp type for converting binary data to when no type is specified within selection.
 One of bit-vector, string (hexadecimals), or octet-vector
 (equivalent to (vector (unsigned-byte 8))")

(defparameter *default-read-numeric-type* 'ratio
 "Lisp type for converting decimal data with scale >0 to when no type is specified
 within selection. One of double-float, ratio, or string")

;;; LMH Add hook to allow application to determine the time conversion.
;;; Circumvent fractional seconds problem as encode-universal-time need not accept them.
#-local-time
(defparameter *time-encode-function* (lambda (time &optional fraction)
                                       (declare (ignore fraction))
                                       time)
 "Bound to a function that converts from a Lisp universal time (and a fractional
 as possible second argument) to the desired representation of date/time/timestamp.")

(defclass db-object () ()
 (:documentation "Top-level class in the impelementational database machinery;
 has nothing in common with standard-db-object."))

(defclass environment (db-object)
 ((db-type :initarg :db-type :reader db-type)
  (connection-pooling)  ; SQL_CP_OFF, SQL_CP_ONE_PER_DRIVER, or SQL_CP_ONE_PER_HENV
  (connections :initarg :connections :accessor connections :initform nil
       :documentation "All database objects ever created, connected or disconnected.")
 ) 
 (:documentation "A kind of connection pool associated with each database type."))

(defclass database (db-object) ;odcl::transactable-store
 ((name :initarg :name :accessor database-name)		; CommonSQL standardized reader
  (spec :initarg :spec)					; original connection specificat.
  (user-name :initarg :user     :accessor user-name :initform "") ; was user-id in 0.2
  (password  :initarg :password :accessor password  :initform "")
  (environment :initarg :environment :reader environment :initform nil)
  (connected :accessor connected :initform nil)
  ;; resource pool of (active and inactive) statement objects
  (statements :accessor statements :initform ())
  (view-classes :accessor view-classes :initform ())

  (command-recording-stream :accessor command-recording-stream :initform nil)
  (result-recording-stream  :accessor result-recording-stream  :initform nil)
  ;(schema :accessor database-schema :initform nil)
  ;(transaction-level :accessor transaction-level :initform 0)
 )
 (:documentation "Class used to store connection information. 
 Several database objects can coexist, and are listed in *connected-databases*.
 The API packages (ODBC, DTF, or others) should subclass this class.
 There is a *default-database*, so you do not always need to mention the
 database you want to work with."))

(defmethod print-object ((self database) stream)
  (print-unreadable-object (self stream :identity t)
    (format stream "~a \"~a@~a\"~:[ disconnected~;~]" (class-name (class-of self)) 
                                  (user-name self) (database-name self)
                                  (connected self))))

(defmethod db-type ((db database)) (db-type (environment db)))

;;; Mixin classes for specific databases not necessarily accessed via odbc.
(defclass db-mixin () ())
(defclass sequenceless-mixin () ())			; Missing native sequence support
(defclass desktop-mixin (db-mixin) ())			; MS Desktop Drivers
(defclass access-mixin (desktop-mixin sequenceless-mixin) ())
;(defclass excel-mixin (desktop-mixin) ())
(defclass mysql-mixin (db-mixin) ())
(defclass oracle-mixin (db-mixin) ())
(defclass postgresql-mixin (db-mixin) ())
(defclass sqlite3-mixin (db-mixin sequenceless-mixin) ()) ; SQLite3

(defclass statement (db-object)
 ((expression :initform nil :initarg :expression :accessor expression
              :documentation "Statement string or expression object")
  (database :initarg :database :reader database
            :documentation "Database object the statement is to be run against")
  (active :initform nil :initarg :active :accessor active
          :documentation "One of: t (busy), :need-data, :still-executing"))
 (:documentation
  "An SQL statement object; stores expression and database to run against."))

(defmethod initialize-instance :after ((stmt statement) &key &allow-other-keys)
  (when (slot-boundp stmt 'database)
    (push stmt (statements (database stmt)))))

;;; "Result set" could be a better name, but looks a bit long.
(defclass cursor (statement)
 ((cursor-type :initarg :cursor-type :accessor cursor-type :initform :forward-only))
 (:documentation "A statement with result set."))

(defclass param-statement (statement)
 ((prepared :initarg :prepared :accessor prepared :initform nil)))
  ;(parameters) ;   sql-parameter  param-desc-record?

(defclass param-cursor (param-statement cursor) ())

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Expression support

(defgeneric table (object)
 (:method (object) nil))

(defgeneric columns (object)
 (:method (object) ()))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Support for database types is added by defining methods for the following
;;; generic functions. There is no need to change the sql module.

(defgeneric db-type-initialize (db-type &optional version)
  (:documentation
   "The internal generic implementation of initialize-database-type."))

(defgeneric db-type-load-foreign (db-type)
  (:documentation
   "Load associated external libraries for the given database type."))

(defgeneric db-name-from-spec (database-type connection-spec)
 (:documentation "Returns the name of the database that would be created if connect
 was called with the connection-spec."))

(defgeneric db-user-name-from-spec (database-type connection-spec)
 (:method (database-type connection-spec) nil))
(defgeneric db-password-from-spec (database-type connection-spec)
 (:method (database-type connection-spec) nil))

(defgeneric db-connect (db-or-type connection-spec &key autocommit &allow-other-keys))
(defgeneric db-disconnect (database))

(defgeneric db-start-transaction (database &optional access isolation)
 ;;; Values: Old database attributes: 1) autocommit 2) access-mode 3) isolation-level
  (:method ((database database) &optional access isolation)   ; dummy for odbc or oracle
    (declare (ignore database mode isolation))))

;; Optional arguments - old values to be restored
(defgeneric db-end-transaction (db-object completion-type
                                &optional autocommit access isolation))

(defgeneric db-execute-command (db-object expression &key &allow-other-keys))
(defgeneric db-cancel (stmt))
 ;(:method :after ((stmt statement))
 ;(setf (active stmt) nil)))

(defgeneric db-query (database query-expression &key flatp &allow-other-keys))
(defgeneric db-open-cursor (database query-expression &key cursor-type
                                                      &allow-other-keys))
(defgeneric db-close (db-object &key drop-p unbind &allow-other-keys))

(defgeneric db-columns-count (cursor))
;(defgeneric db-column-names (cursor)) ; returns a vector
(defgeneric db-fetch-into (cursor object &key start-column end-column current
                                         &allow-other-keys))
(defgeneric db-fetch-item (cursor column &key current &allow-other-keys))
(defgeneric db-fetch-rows (cursor &key count flatp &allow-other-keys))
(defgeneric db-delete-row (cursor &key &allow-other-keys))
(defgeneric db-insert-row (cursor &key columns values cv-alist &allow-other-keys))
(defgeneric db-update-row (cursor &key columns values cv-alist &allow-other-keys))

(defgeneric current-row-number (cursor))
(defgeneric current-row-status (cursor)
 (:documentation
  "One of t, :warn, :error, :added, :deleted, :updated, :norow, or nil (unavailable)."))

(defgeneric set-cursor-position (cursor orientation &key offset retrieve-data errorp))
 ;;; Args: cursor  Opened cursor, scrollable for orientation other than :next.
  ;;	   orientation ::= :first | :last | :next | :prior | :relative | :absolute
  ;;       offset  Integer, perhaps a bookmark (shall we provide vector bookmarks?)
  ;;	   retrieve-data  If nil, populates only status array and bookmark column
  ;; Values: non-nil (new row number) - success.
  ;;         nil - failure; after failure rowset could:
  ;;		 - either gets status :norow (actually implemented in ODBC);
  ;;		 - or preserves its previous status.

(defgeneric db-map-parameters (db-object result-type expression &rest sequences))
(defgeneric db-prepare-statement (database expression &key parameters values
                                                      &allow-other-keys))
;(defgeneric db-map-bind-query (query type function &rest parameters))

;;; SQL syntax specializers
(defgeneric db-sql-type-specifier (database lisp-type &rest args)
 (:documentation "Return the type SQL type specifier as a string, 
 for the given lisp type and argumentss."))

(defgeneric db-identifier-case (database)
 ;;; Value ::= :upcase | :downcase | :sensitive | :mixed
  ;;	   Analogous to readtable-case, exept that :preseve is divided into the two:
  ;;	    :sensitive	Identifiers in SQL are case-sensitive and are stored 
  ;;			in mixed case in system catalog (never within SQL-92compliant)
  ;;	    :mixed	Identifiers in SQL are not case-sensitive
  ;;		 	and are stored in mixed case in system catalog.
  ;; NB: Quoted identifier case is always assumed :sensitive (as if SQL-92compliant)
 (:method (database)
  (declare (ignore database))
  :upcase))

(defgeneric db-identifier-quote-char (database)
 (:method (database)
  (declare (ignore database))
  #\"))

(defgeneric db-special-character-p (database char)
 (:documentation "Whether the CHAR is allowed inside schema object
 or column name but only within quoting.")
 (:method (database char)
  (declare (ignore database))
  (find char "#$^")))

;;; Catalog/metadata configuration and inquiries
;;; db-probe - can be implemented via db-list
(defgeneric db-create (db-type connection-spec &key &allow-other-keys))
(defgeneric db-destroy (db-type connection-spec &key &allow-other-keys))
(defgeneric db-list (db-type connection-spec &key &allow-other-keys))

(defgeneric db-describe-tables (database table &key table-type catalog schema))
(defgeneric db-list-tables (database schema &key table catalog)
 (:documentation "List all tables or matching TABLE in the given DATABASE."))
(defgeneric db-drop-table (database name &key if-does-not-exist &allow-other-keys))

(defgeneric db-list-views (database schema &key view catalog)
 (:documentation "List all views or matching VIEW in the given DATABASE."))
(defgeneric db-drop-view (database name &key if-does-not-exist &allow-other-keys))

(defgeneric db-list-columns (database table schema &key catalog)
 (:documentation "List all columns in table."))
(defgeneric db-column-type (database column table schema &key catalog)
 (:documentation "Return the type of column in table."))

(defgeneric db-list-indexes (database table &key catalog schema))
(defgeneric db-drop-index (database name &key on if-does-not-exist &allow-other-keys))

;(defgeneric db-table-exists-p (database table schema)
;  (:documentation "Returns true, if TABLE in SCHEMA of DATABASE."))

;;; Sequences

(defgeneric db-create-sequence (database name &rest args)
 (:documentation "Create a sequence in the DATABASE."))
(defgeneric db-drop-sequence (database name)
 (:documentation "Drop the sequence from the DATABASE."))
(defgeneric db-alter-sequence (database name start &rest args)
  (:documentation "Set sequence position and/or change other options."))
(defgeneric db-sequence-last (database name)
 (:documentation "Return the last issued value of the generator."))
(defgeneric db-sequence-next (database name)
 (:documentation "Increment sequence generator base value and return it."))
(defgeneric db-list-sequences (database &key catalog schema)
  (:documentation "List all sequences in the DATABASE."))


#||
#+unused
(defgeneric database-add-column (table column database)
  (:documentation "Add the column to the table."))

(defgeneric database-rename-column (table oldatt newname database)
  (:documentation "Rename the column in the table to NEWNAME."))

(defgeneric oid (object)
  (:documentation "Return the unique ID of a database object."))
||#
