;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2006 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Low-level expression functions, identifiers (schema object names)
;;; and general purpose utils

(in-package :ystok.sql.sys)

(defvar *sql-stream* nil)	; Stream accumulating SQL output
(defvar *in-subselect* nil)	; If true, deliniate with () and print table aliases
(defvar *quote* :default)	; T=quote unconditionally, NIL=do not, :default=analyse

(defmacro with-sql-string-stream ((&optional string
                                   &key (element-type 'yl:*default-char-type*))
                                  &body body)
 ;;; Convenience macro binding *sql-stream* to a string output stream.
 `(with-output-to-string (*sql-stream* ,string :element-type ,element-type)
    ,@body))

(defgeneric write-sql-identifier (identifier &optional stream database name-only quote)
 ;;; Used in places where we know that an identifier must occur in the expression.
  ;; Allows strings instead of symbols or sql-ident instances to be used
  ;; as table or column names.
  ;; Args: identifier Names schema objects and their parts, qualified or unqualified.
  ;;	   name-only  If true, output unqualified name only, e.g. for sql-column instance.
  ;;	   quote      T = unconditionally quote, NIL = do not quote, :default = analyse
 (:method ((identifier string)
           &optional stream (database *default-database*) name-only (quote t))
  ;; Args: The string usually represents a quoted identifier.
  (declare (ignore name-only))
  (let ((identifier-quote-char (if quote (db-identifier-quote-char database) nil)))
    (when identifier-quote-char (write-char identifier-quote-char stream))
    (write-string identifier stream)
    (when identifier-quote-char (write-char identifier-quote-char stream))))

 (:method ((identifier symbol)
           &optional stream (database *default-database*) name-only (quote *quote*))
  ;; Args: The symbol can represent either nonquoted or quoted identifier.
  ;; NB:
  ;; 1) We replace lisp-style hyphen with SQL-style underscore.
  ;; 2) We delimit by SQL_IDENTIFIER_QUOTE_CHARs if contains SQL_SPECIAL_CHARACTERS.
  ;; TODO: Delimit if coincides with on of the SQL_KEYWORDS?
  ;; 
  ;; What we are passing to SQL
  ;; +---------+---------+------------+------------+------------+
  ;; | Symbol  |Actual id|   Oracle   | PostgreSQL |   Access   |
  ;; |         |  in DB  | (:upcase)  |(:downcase) |  (:mixed)  | <- db-identifier-case
  ;; +---------+---------+------------+------------+------------+
  ;; |  MYTAB  |  MYTAB  |    MYTAB   |    MYTAB   |   MYTAB    |
  ;; | |mytab| |  mytab  |   "mytab"  |    mytab   |   mytab    |
  ;; | |MyTab| |  MyTab  |   "MyTab"  |   "MyTab"  |   MyTab    |
  ;; +---------+---------+------------+------------+------------+
  (declare (ignore name-only))
  (let ((symbol-name (substitute #\_ #\- (symbol-name identifier)))
        (identifier-quote-char nil))
    (declare (type simple-string symbol))
    (unless (string= symbol-name "*")
      (flet ((set-quote ()
               (or identifier-quote-char
                   (null quote)					; prohibit quoting
                   (setq identifier-quote-char (db-identifier-quote-char database)))))
        (do ((db-identifier-case (db-identifier-case database))
             (case nil)						; symbol name case
             (length (length symbol-name))
             (i 0 (1+ i))
             char)
            ((>= i length)
             ;; Assume that normal symbol has all letters upcase and do not quote it.
             (case case
               (:mixed
                (when (memq db-identifier-case '(:upcase :downcase)) (set-quote)))
               (:downcase
                (when (eq db-identifier-case :upcase) (set-quote)))))
          (declare (fixnum length i))
          (cond ((alphanumericp (setq char (schar symbol-name i)))
                 (cond ((eq case :mixed))
                       ((upper-case-p char)
                        (setq case (if (eq case :downcase) :mixed :upcase)))
                       ((lower-case-p char)
                        (setq case (if (eq case :upcase) :mixed :downcase))) ))
                ((char= char #\_))
                ((db-special-character-p database char)
                 (set-quote))
                (t (error 'sql-reader-error
                          :format-control "Inacceptable character '~A' in identifier ~S."
                          :format-arguments (list char identifier))) ))))
    (when identifier-quote-char (write-char identifier-quote-char stream))
    (write-string symbol-name stream)
    (when identifier-quote-char (write-char identifier-quote-char stream)))) )

(defun write-sql-identifier-to-string (identifier
                                       &optional (database *default-database*)
                                                 (name-only nil) (quote *quote*))
  (with-sql-string-stream ()
    (write-sql-identifier identifier *sql-stream* database name-only quote)))

(defmacro output-sql-identifier (expr database &optional (name-only nil)
                                                         (quote :default specified-p))
  ;; Args: name-only  If true, output unqualified name only, e.g. for sql-column instance
  `(write-sql-identifier ,expr *sql-stream* ,database ,name-only
                         ,@(when specified-p `(,quote))))
  #|(if (typep expr 'sql-ident)
      (if name-only
          (write-sql-identifier (name expr) *sql-stream* database)
          (output-sql expr database))
      (write-sql-identifier expr *sql-stream* database)))|#

(defun %output-plist (plist database)
 ;;; If the value of a property is T, output only the property/option name
  (loop for (arg value) on plist by #'cddr
        when value
        do (write-char #\Space *sql-stream*)
           (princ arg *sql-stream*)	;(output-sql arg database)?
           (unless (eq value t)
             (write-char #\Space *sql-stream*)
             (output-sql value database))))

;;; Other helpers

(defun command-output (command &rest args)
 ;;; Execute a file or operation system command
  ;; Args: command  Use something like "cmd /c command" to invoke Windows command.
  ;;	   args     Key arguments including :current-directory, :wait, or :prefix.
  ;;		    (See system:call-system-showing-output for details.) 
  ;;                Undocumented :extra-environ works on Windows and seems ot accept
  ;;		    an alist, e.g. '(("TESTVAR" . "testval"))
  ;; Values: (1) status (2) output string.
  ;; TODO: Move to Ystok-Library?
  #+lispworks
  (let* ((status -1)
         (message
          (with-output-to-string (stream nil :element-type yl:*default-char-type*)
            (setq status (apply #'system:call-system-showing-output command
                                :output-stream stream  :show-cmd nil
                                 ;:shell-type "/bin/sh"
                                args)))))
    (values status message)))

(defun float-to-sql-string (number)
 ;;; Convert exponent character to match SQL <approximate numeric literal>
  #+lispworks4 (princ-to-string number)	; LispWorks4 always seems to use E as exponent char
  #-lispworks4 (nsubstitute-if #\E
                               (lambda (char) (find char "fFdDlLsS"))
                               (princ-to-string number)
                               :count 1))

(defmethod object-hex-string ((vector bit-vector) &optional (bit-count (length vector)))
  (let* ((byte-count (ceiling bit-count 8))
         (string (make-string (* 2 byte-count) :element-type 'base-char)))
    (dotimes (i byte-count)
      (let ((byte1 0)
            (byte0 0))
        (loop for j from 0 to 3
              for index = (+ (ash i 3) j)
              while (< index bit-count)
              do (setf byte1 (logior byte1 (ash (bit vector (+ (ash i 3) j)) (- 3 j)))))
        (loop for j from 4 to 7
              for index = (+ (ash i 3) j)
              while (< index bit-count)
              do (setf byte0 (logior byte0 (ash (bit vector (+ (ash i 3) j)) (- 7 j)))))
        (setf (char string (* i 2)) (hex-digit-char byte1)
              (char string (+ (* i 2) 1)) (hex-digit-char byte0))))
    string))

(defun timestring (&optional (time (get-universal-time)))
 ;;; Weak replacement of local-time utilities
  (multiple-value-bind (seconds minutes hours day month year)
      (decode-universal-time time)
    (format nil "~4D-~2,'0D-~2,'0D ~2,'0D:~2,'0D:~2,'0D"
            year month day hours minutes seconds)))


#+deprecated
(defun write-sql-string (string &optional stream)
 ;;; Construct character string literal
  (write-char #\' stream)
  (loop for char across string
        if (char= char #\') do (write-char #\' stream)
        else if (char= char #\\) do (write-char #\\ stream) ; PostgreSQL extension?
        do (write-char char stream))
  (write-char #\' stream))

#+unused
(defun byte-vector-to-hex-string (vector &optional (bit-count 8))
  (case bit-count
    ((8 nil)
     (loop with length = (length vector)
           with string = (make-string (* length 2) :element-type 'base-char)
           for i upfrom 0 below length
           and j upfrom 0 by 2
           for byte = (aref vector i)
           do (setf (schar string j) (code-hex-digit (ash byte -4))
                    (schar string (1+ j)) (code-hex-digit (logand byte #x0F)))
           finally (return string)))))

#||
(defun getenv (var)
  "Return the value of the environment variable."
  #+allegro (sys::getenv (string var))
  #+clisp (ext:getenv (string var))
  #+(or cmu scl)
  (cdr (assoc (string var) ext:*environment-list* :test #'equalp
              :key #'string))
  #+lispworks (lw:environment-variable (string var))
  #+mcl (ccl::getenv var)
  #+sbcl (sb-ext:posix-getenv var))

(eval-when (:compile-toplevel :load-toplevel :execute)
  (when (char= #\a (schar (symbol-name '#:a) 0))
    (pushnew :clsql-lowercase-reader *features*)))

(defun convert-to-db-default-case (str database)
  (if database
      (case (db-type-default-case (database-underlying-type database))
	(:upper (string-upcase str))
	(:lower (string-downcase str))
	(t str))
    ;; Default CommonSQL behavior is to upcase strings
    (string-upcase str)))
	    
(defun ensure-keyword (name)
  "Returns keyword for a name"
  (etypecase name
    (keyword name)
    (string (nth-value 0 (intern (symbol-name-default-case name) :keyword)))
    (symbol (nth-value 0 (intern (symbol-name name) :keyword)))))
||#