;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2005 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Functional interface to SQL - LOOP extension
;;;
;;; Extends loop to provide a clause for iterating over query results.
;;; Signature: loop {for | as} var [[of-type] type-spec] 
;;;                 being {the | each} {records | record | tuples | tuple}
;;;                 {in | of} query-expression
;;;		    [using (database db)] 
;;;                 [using (parameters pars)] [using (values vals)]
;;; Arguments:  
;;;	VAR	   A record variable represented by a list or vector.
;;;     TYPE-SPEC  If eql to vector, VAR is bound to simple vector.
;;;	QUERY-EXPRESSION  An SQL query sentence; :flatp is ignored if given.
;;;	DB, PARS, VALS	  YSQL extensions
;;; Values: A loop return value.
;;; 
;;; Each iteration of the loop assigns the next record of the result set
;;; to the variable VAR. Destructuring can be used in VAR to bind variables 
;;; to specific columns of the records resulting from QUERY-EXPRESSION.
;;; In conjunction with the panoply of existing clauses available
;;; from the loop macro, the new iteration clause provides an
;;; integrated report generation facility.
;;;
;;; Implementation Note:
;;;	As we are allocating the single object for all records before
;;;     iteration and then modifying it on each iteration destructively, 
;;;	do not forget to make a copy when collecting!

(in-package :ystok.sql.sys)

#+lispworks
(defun loop-record-iteration-path (name variable data-type prep-phrases
                                   inc all-preps data)
 ;;; Args: name ::= loop::record | loop::records | loop::tuple | loop::tuples
  ;;       prep-phrases  Alist ((loop::in . expr) ...)
  ;;       data = specific-data of lw:define-loop-method
  (declare (ignore name data inc all-preps))
  (when (rest prep-phrases)
    (error 'conditions:loop-error
           :format-control "Duplicate IN or OF prepositional phrase: ~S." 
           :format-arguments (rest prep-phrases)))
  (let* ((query-expression (rest (loop::get-prep '(loop::of loop::in) prep-phrases)))
         (cursor (gensym "loop-cursor"))
         (database (if (assoc 'loop::database loop::*using-variables*)
                       (lw:named-variable 'database)))
         (parameters (if (assoc 'loop::parameters loop::*using-variables*)
                         (lw:named-variable 'parameters)))
         (values (if (assoc 'loop::values loop::*using-variables*)
                     (lw:named-variable 'values)))
         (heading `((with-cursor (,cursor ,query-expression 
                                          ,@(if database `(:database ,database))
                                          ,@(if parameters `(:parameters ,parameters))
                                          ,@(if values `(:values ,values))))))
       )
    ;(print loop::*using-variables*)
    (if (consp variable)
	(let ((record (gensym "record")))
          (values nil		; if t, test exit condition after the first body pass
                  heading					; outer form header
                  `((,record nil) ,@(mapcar #'list variable))	; var bindings
                  `((setq ,record (make-list (db-columns-count ,cursor)))) ; prologue
                  `((null (db-fetch-into ,cursor ,record)))
                  `(,variable ,record)				; destructing bind 1
                  ;`(a b))					; until (or a b)
                  ;`(pattern subject)				; destructing bind 2
        ) )
        (values nil heading
                `((,variable nil ,data-type))			; var bindings
                `((setq ,variable 				; prologue
                        ,(cond ((subtypep data-type 'list) 	; covers (null data-type)
                                `(make-list (db-columns-count ,cursor)))
                               ((eq data-type 'vector) 		;simple-vector
                                `(make-array (db-columns-count ,cursor)))
                               ((error 'conditions:loop-error
                                  :format-control "Invalid loop record datatype: ~S."
                                  :format-arguments (list data-type)))) ))
                `((null (db-fetch-into ,cursor ,variable)))
) ) ) )

#+lispworks
(lw:define-loop-method (loop::record loop::records loop::tuple loop::tuples)
                       loop-record-iteration-path
                       (:of :in))
                       ;loop::db-value) ; passad as data

#|
(loop initially (print "initially")
      for r of-type vector
      being each record of q using (:database db) using (:parameters pars)
      do (print r) (prin1 r)
      collect r)
(loop for (c1 c2 c3) ;of-type (integer integer integer)
      being each record of q using (:database db) 
        using (:parameters pars) using (:values '(1 2))
      do (print c1) (prin1 c2))
 |#

#|;;; MIT-LOOP extension 
;;; The follwing is simply repackaging MaiSQL code and does need adjusting.
;;; the [from db] loop clause is used instead of [using (database db)]
#+(or cmu scl)
(defun loop-record-iteration-path (variable data-type prep-phrases)
  (let ((in-phrase nil)
	(from-phrase nil))
    (loop for (prep . rest) in prep-phrases
	  do
	  (case prep
	    ((:in :of)
	     (when in-phrase
	       (ansi-loop::loop-error
		"Duplicate OF or IN iteration path: ~S." (cons prep rest)))
	     (setq in-phrase rest))
	    ((:from)
	     (when from-phrase
	       (ansi-loop::loop-error
		"Duplicate FROM iteration path: ~S." (cons prep rest)))
	     (setq from-phrase rest))
	    (t
	     (ansi-loop::loop-error
	      "Unknown preposition: ~S." prep))))
    (unless in-phrase
      (ansi-loop::loop-error "Missing OF or IN iteration path."))
    (unless from-phrase
      (setq from-phrase '(*default-database*)))
    (cond
      ((consp variable)
       (let ((query-var (ansi-loop::loop-gentemp 'loop-record-))
	     (db-var (ansi-loop::loop-gentemp 'loop-record-database-))
	     (result-set-var (ansi-loop::loop-gentemp
			      'loop-record-result-set-))
	     (step-var (ansi-loop::loop-gentemp 'loop-record-step-)))
	 (push `(when ,result-set-var
		 (database-dump-result-set ,result-set-var ,db-var))
	       ansi-loop::*loop-epilogue*)
	 `(((,variable nil ,data-type) (,query-var ,(first in-phrase))
	    (,db-var ,(first from-phrase))
	    (,result-set-var nil)
	    (,step-var nil))
	   ((multiple-value-bind (%rs %cols)
		(database-query-result-set ,query-var ,db-var)
	      (setq ,result-set-var %rs ,step-var (make-list %cols))))
	   ()
	   ()
	   (not (database-store-next-row ,result-set-var ,db-var ,step-var))
	   (,variable ,step-var)
	   (not ,result-set-var)
	   ()
	   (not (database-store-next-row ,result-set-var ,db-var ,step-var))
	   (,variable ,step-var))))	; destructuring-bind
      (t
       (let ((query-var (ansi-loop::loop-gentemp 'loop-record-))
	     (db-var (ansi-loop::loop-gentemp 'loop-record-database-))
	     (result-set-var (ansi-loop::loop-gentemp
			      'loop-record-result-set-)))
	 (push `(when ,result-set-var
		 (database-dump-result-set ,result-set-var ,db-var))
	       ansi-loop::*loop-epilogue*)
	 `(((,variable nil ,data-type) (,query-var ,(first in-phrase))
	    (,db-var ,(first from-phrase))
	    (,result-set-var nil))
	   ((multiple-value-bind (%rs %cols)
		(database-query-result-set ,query-var ,db-var)
	      (setq ,result-set-var %rs ,variable (make-list %cols))))
	   ()
	   ()
	   (not (database-store-next-row ,result-set-var ,db-var ,variable))
	   ()
	   (not ,result-set-var)
	   ()
	   (not (database-store-next-row ,result-set-var ,db-var ,variable))
	   ()))))))

#+(or cmu scl)
(ansi-loop::add-loop-path '(record records tuple tuples)
			  'loop-record-iteration-path
			  ansi-loop::*loop-ansi-universe*
			  :preposition-groups '((:of :in) (:from))
			  :inclusive-permitted nil)
 |#