;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2007 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Object-oriented interface to SQL - Metaobject and variable definitions
;;;
;;; Notes:
;;;  1) :home-key slot-name
;;;	The value may be a non-null atom or a list of elements, where elements can
;;;	be symbols (denoting column names), nil, :not-null, strings, integers or floats.
;;;     EXTENSION to CommonSQL: (:home-key symbol) is allowed for a standard slot.
;;;
;;;  2)	In CommonSQL, the :foreign-key keywords badly correlates with SQL keywords
;;;	used in constaint definition, cf.
;;;	  CREATE/ALTER TABLE table
;;;	   CONSTRAINT FOREIGN KEY fk-columns REFERENCES parent-table referenced-columns
;;;	To exclude misinterpretation, we use :join-key instead but retain
;;;	the :foreign-key synonym for compatibility.
;;;
;;;   :join-key slot-name
;;;	This is the name of the slot of the :join-classs class to be a subject for
;;;	the join. The argument slot-name may be an element or a list of elements, where
;;;	elements can be symbols, nil, strings, integers or floats.
;;;	(On the contrary to home keys, one can hardly imagine the application where
;;;	something other than non-null symbols are useful for join keys.)
;;;     EXTENSION to CommonSQL: (:join-key symbol) is allowed for a standard slot.
;;;
;;;   :join-class class-name
;;;	This can correspond to
;;;	- either the parent PRIMARY table in a constraint if :set t,
;;;	- or the SECONDARY table in a constraint if :set nil.
;;;	The :self keyword specified as the class-name means self-join,
;;;	i.e. against the same base table (useful for abstract mixins).
;;;
;;;   :target slot-name
;;;	For now, YSQL supports only deferred target slot though allows the slot-name
;;;	to refer not only to a join slot but also to a base or standard slot.
;;;     TODO: For a non-set "base-targeted" join slot, retrieving it in immediate manner
;;;	      could potentially lead to a more effective query as that would not fetch
;;;	      the whole joined instance but only the corresponding target column.
;;;
;;;   :order-by slot-name-or-column-or-list
;;;	For :set t, addition convenience clause for sorting lazyloaded join instances.
;;;
;;;  3) In CommonSQL the :virtual keyword is badly correlates with the virtual slot
;;;	example (see CLOS chapter in LW User Guide). We use :db-kind :standard instead
;;;	but retain the :virtual synonym for compatibility.
;;;
;;;  4) For db-kinds other than :join, we collect into db-info the following
;;;	:constraints
;;;	:type
;;;	:auto-unique-valued (generally, no update is possible if used instead of :default)
;;;
;;;  5) Instead of refined :join, one can
;;;	- declare the db-kind of slot :standard,
;;;	- define a method on slot-unbound,
;;;	- set/load-on-demand the value of the slot there.
;;;	The :standard can be used to implement component mapping (in ORM terminology).
;;;	Just list the columns corresponding to the slots of a component in the
;;;	:home-key option.
;;;
;;;  6) Constraint inheritance
;;;	As the whole :db-info slot option is taken from the most specific superclass,
;;;	all the constraints defined there are also inherited along with the slot.
;;;	The inheritance of class-level constraines is not implemented.

(in-package :ystok.sql.sys)

(deftype universal-time ()
 "An integer, positive or negative, as returned by ENCODE-UNIVERSAL-TIME."
  'integer)

(declaim (inline base-slot-p))
(defun base-slot-p (sd)
  (memq (db-kind sd) '(:base :key)))

(defclass standard-db-class (standard-class)
 ((table       :accessor table)		; base-table, supply sql-ident-table for alias
  (constraints :accessor constraints)	; list of lists a la create/alter-table
  (base-slots  :reader base-slots :initform ())	; sublist of effective slot definitions
  (key-slots   :reader key-slots :initform ())	; sublist of base-slots
  ;; CLSQL/USQ: :List ((name1 type1 db-kind1 column1) ...), db-kind ::= nil | :base | :key
  ;(descriptors :accessor descriptors  :initarg :descriptors  :initform ())
  ;; a la alias in sql-ident-table to avoid collisions in auto generated select query
  ;(qualifier  :accessor qualifier  :initarg :qualifier  :initform nil)
 )
 (:documentation "Metaclass for all view classes."))

;;; Allow mentioning standard classes in the list of superclasses
;;; CAUTION: Not tested.

(defmethod clos:validate-superclass ((class standard-db-class) (superclass standard-class))
  t)

;;; MOP standard: Unknown class options "quote" their values automatically
;;; Lispworks allows special processing

(defmethod clos:process-a-class-option ((class standard-db-class) name value)
 ;;; Arts: value  The tail of the defclass option form.
  (case name
    ((:table :base-table)
     (let ((first (first value)))
       (if (symbolp first)
           `(,name ',value)	; auto quote as per MOP
           `(,name ,first))))	; evaluate option to produce sql-ident-table instance
    ;(:constraints)		; auto quote due to MOP
    (otherwise
     (call-next-method))))

;;; (call-next-method) may invoke (finalize-inheritance)
;;; Prohibitions:
;;;  - A standard-class defined with defclass cannot refer to a standard-db-class
;;;    among its direct-superclasses,
;;;  - and v.v. (for LW only?).
;;; Optimization hint:
;;;     In the base-slots list, we move :auto-unique-valued slot descriptions to
;;;     the start. So they usually come to the head of the list. In this manner,
;;;     we try to "fix" their position within cursor row descriptors.
(defmethod initialize-instance :around ((class standard-db-class) &rest initargs
					&key name direct-superclasses	;direct-slots
                                             base-table table		; are synonyms
                                             constraints
                                        &allow-other-keys)
  (let ((root-class (find-class 'standard-db-object nil))
        (table (or base-table table)))
    ;(setf (qualifier class) (first qualifier))
    (setf (table class) (keep-symbol-names (if table (first-or-self table) name) class)
          (constraints class) (keep-symbol-names constraints class t))
    (if (or (null root-class)
            ;(member (find-class 'standard-db-class) direct-superclasses
            ;        :key 'class-of :test 'eq))	; as inherits from a standard-db-object,
            (member-if (lambda (class) (typep class 'standard-db-class))
                       direct-superclasses))
        (call-next-method)			; let its method proceed.
        (apply #'call-next-method class		; invoke standard-class's method
               :direct-superclasses (cons root-class direct-superclasses)
               (remove-properties initargs
                                  '(:direct-superclasses :base-table :table))))
    (setf (slot-value class 'base-slots)
           (stable-sort (remove-if-not #'base-slot-p
                                       (clos:class-slots class))	;-> compute-slots
                        (lambda (x y) (and (getf x :auto-unique-valued)
                                           (not (getf y :auto-unique-valued))))
                        :key 'db-info)
          (slot-value class 'key-slots) (remove :key (base-slots class)
                                                :key 'db-kind  :test-not 'eq)) ))
    ;(register-metaclass class direct-slots)))

(defmethod reinitialize-instance :around ((class standard-db-class) &rest initargs
                                          &key direct-superclasses ;direct-slots
                                               base-table table		; are synonyms
                                               constraints
                                          &allow-other-keys)
  (let ((root-class (find-class 'standard-db-object nil))
        (table (or base-table table)))
    ;(setf (qualifier class) (first qualifier))
    (setf (table class) (keep-symbol-names (if table
                                               (first-or-self table)
                                               (class-name class))
                                           class)
          (constraints class) (keep-symbol-names constraints class t))
    (if (or (null root-class)
            (eq class root-class) 	; no need to reinitialize in a non-standard way
            ;(memq (find-class 'standard-db-class) direct-superclasses
            ;      :key 'class-of))	; as inherits from a standard-db-object,
            (member-if (lambda (class) (typep class 'standard-db-class))
                       direct-superclasses))
        (call-next-method)			; let its method proceed.
        (apply #'call-next-method class		; invoke standard-class's method
               :direct-superclasses (cons root-class direct-superclasses)
               (remove-properties initargs
                                  '(:direct-superclasses :base-table :table))))
    ;; Very subtle bugs are possible due to class-slots returning old esd list.
    (clos:finalize-inheritance class)
    (setf (slot-value class 'base-slots)
           (stable-sort (remove-if-not #'base-slot-p
                                       (clos:class-slots class))  	;-> compute-slots
                        (lambda (x y) (and (getf x :auto-unique-valued)
                                           (not (getf y :auto-unique-valued))))
                        :key 'db-info)
          (slot-value class 'key-slots) (remove :key (base-slots class)
                                                :key 'db-kind  :test-not 'eq)) ))
    ;(register-metaclass class direct-slots)))

(defmethod clos:process-a-slot-option ((class standard-db-class) name value
                                       already-processed-options slot)
 ;;; Value: A plist of initargs.
  ;; Assumption: :db-kind must preceed :db-info in the option list.
  (case name
    (:column
     `(,name ,(if (symbolp value)
                  `',value	; retain quote for CommonSQL compatibility
                  value)	; evaluate the option to produce sql-ident instance
             ,.already-processed-options))
    (:db-kind
     `(,name ,(if (eq value :virtual) :standard value) ,.already-processed-options))
    (:db-info		; already-processed-options ::= ((quote :db-kind) (quote :join))
     `(,name ,(parse-db-info (getf already-processed-options :db-kind :base)
              ;(cadadr (member :db-kind already-processed-options :key 'second :test 'eq))
                             value)
             ,.already-processed-options))
    (otherwise
     (call-next-method))))

(defun parse-db-info (db-kind plist)
 ;;; Value: 'plist or (list :k1 'value1 ...) - acceptable for further parsing.
  ;; TODO: 1) Return (list ...) on occurences of (load-time-value ---)
  ;;	   2) Accept :constraint, :type etc.
  (when (oddp (length plist))
    (error 'sql-user-error :message (format nil
           "Wrong :db-info slot option - odd length of the property list:~%~S" plist)))
  (do* ((acc ())
        (rest plist (cddr rest))
        (key (first rest) (first rest))
        (value (second rest) (second rest)))
       ((null rest)
        (dolist (key (case db-kind			; verify for obligations
                       (:join '(:home-key :join-class :join-key))))
          (unless (getf acc key)
            (error 'sql-user-error :message (format nil
         "For :db-kind ~S, the property ~S is obligatory in the :db-info slot option:~%~S"
             db-kind key plist))))
        (case db-kind				        ; provide defaults
          (:join (when (eq (getf acc :set *unbound-value*) *unbound-value*)
                   (lw:nconcf acc (list :set t)))))
                 ; the :retrieval :deferred default is provided by routines themselves
        (list 'quote acc))
    (cond ((not (memq key (case db-kind
                            (:join '(:home-key :join-class :join-key :foreign-key
                                     :target-slot :retrieval :set :order-by))
                            ((:key :base) '(:auto-unique-valued :constraints))
                            (:standard '(:home-key :join-key)))))	; YSQL extension
           (cerror #1="Ignore wrong or duplicated property."
                   'sql-user-error :message (format nil
                   "For :db-kind ~S, illegal property ~S in the :db-info slot option:~%~S"
                   db-kind key plist)))
          ((eq (getf acc (if (eq key :foreign-key)
                             (setq key :join-key)	; CommonSQL compatibility
                             key)
                     *unbound-value*)
               *unbound-value*)
           (lw:nconcf acc (list key value))		; grab the property
           (case key
             (:retrieval (when (eq value :immediate)
                           (lw:nconcf acc (list :set nil))))))
          ((neq key :set)	; ignore :set possibly forced by preceding :retrieval
           (cerror #1# 'sql-user-error :message (format nil
                   "Duplicated property ~S in :db-info slot option:~%~S" key plist)))
) ) )

(defclass db-class-sd-mixin ()
 ((column :accessor column  :initarg :column
	  :documentation "The name of the SQL column or instance of sql-expr-ident
	 this slot is stored in. Defaults to the slot name.")
  (db-kind :accessor db-kind  :initarg :db-kind  :initform :base  :type keyword
	   :documentation "The kind of DB mapping which is performed for this slot:
	:base (default) - maps to an ordinary column of the DB view.
	:key - corresponds to part of unique keys for this view,
	:join - corresponds to a set of objects or a single instance
		of another view class,
	:standard - indicates an ordinary CLOS slot.")
  (db-info :accessor db-info  :initarg :db-info  :initform ()
	   :documentation "The list of database-related properties describing join,
	column type or constraints. Inherited from the most specific superclass.")
))
#|(db-type :accessor db-type :initform nil
           :documentation "Original (or generated onestablishing connection?) type
	specification, maybe with keyword specifiers.")
  (db-reader :accessor view-class-slot-db-reader :initarg :db-reader
    :initform nil :documentation
    "If a string, then when reading values from the DB, the string
will be used for a format string, with the only value being the value
from the database.  The resulting string will be used as the slot
value.  If a function then it will take one argument, the value from
the database, and return the value that should be put into the slot.")
   (db-writer :accessor view-class-slot-db-writer :initarg :db-writer :initform nil
    :documentation
    "If a string, then when reading values from the slot for the DB,
the string will be used for a format string, with the only value being
the value of the slot.  The resulting string will be used as the
column value in the DB.  If a function then it will take one argument,
the value of the slot, and return the value that should be put into
the database.")
   (db-constraints :accessor view-class-slot-db-constraints :initarg :db-constraints
    :initform nil :documentation
  "A keyword symbol representing a single SQL column constraint or list of such symbols.")
   (void-value :accessor view-class-slot-void-value :initarg :void-value :initform nil
    :documentation "Value to store if the SQL value is NULL. Default is NIL.")
   (specified-type :accessor specified-type :initform nil
    :documentation "KMR: Internal slot storing the :type specified by user.")))|#

(defmethod column  ((sd clos:standard-slot-definition)) nil)
(defmethod db-info ((sd clos:standard-slot-definition)) nil)
(defmethod db-kind ((sd clos:standard-slot-definition)) :standard)
(defmethod constraints ((sd clos:standard-slot-definition)) nil)

(defun join-slots (class &optional (retrieval :immediate))
 ;;; Value: list of join slot definitions with retrieval specified for the class
  (remove-if-not (lambda (sd) (and (eq (db-kind sd) :join)
                                   (or (eq retrieval t)
                                       (eq (getf (db-info sd) :retrieval :deferred)
                                           retrieval))))
                 (clos:class-slots class)))

(defun join-class (sd)
  (getf (db-info sd) :join-class))
(defmethod auto-unique-valued ((sd db-class-sd-mixin))
  (getf (db-info sd) :auto-unique-valued))
(defmethod constraints ((sd db-class-sd-mixin))
  (getf (db-info sd) :constraints))

(defun home-slot-names (slot &optional class)
 ;;; Get slot names to update (or retrieve?) corresponding to the SLOT given
  ;; Args: slot  Name or definition of a (changed) slot, possibly join (or key -?)
  ;;	   class Is only needed if slot is a name.
  ;; Values: List of base slot names.
  ;; Used by: Correct mapping: changed pane/column -> join-slot -> home-key slots
  ;; NB: We always cons a fresh list as the result can be nconc'ed by means of mapcan.
  (when-bind (sd (if (symbolp slot) (clos:find-slot-definition slot class) slot))
    (if (memq (db-kind sd) '(:join :standard))
        (mapcan (lambda (hk) (if (symbolp hk) (list hk)))
                (listify (getf (db-info sd) :home-key)))
        (list (if (symbolp slot) slot (clos:slot-definition-name slot))))))

(defmethod db-type ((sd db-class-sd-mixin))
  (let ((column (column sd)))
    (if (typep column 'sql-column)
        (getf (plist column) :type)
        (clos:slot-definition-type sd))))

(defclass db-class-dsd (db-class-sd-mixin clos:standard-direct-slot-definition) ())
(defclass db-class-esd (db-class-sd-mixin clos:standard-effective-slot-definition) ())

(defmethod clos:direct-slot-definition-class ((class standard-db-class) &rest initargs)
  (declare (ignore initargs))
  (find-class 'db-class-dsd))

(defmethod clos:effective-slot-definition-class ((class standard-db-class) &rest initargs)
  (declare (ignore initargs))
  (find-class 'db-class-esd))


;; Some MOPs, like openmcl 0.14.2, cons attribute values in a list.
;(declaim (inline delistify-dsd))
;(defun delistify-dsd (list) (if (and (listp list) (null (rest list))) (first list) list))

(defun make-sql-column (sd class
                        &optional (type (when-let (type (clos:slot-definition-type sd))
                                          (unless (or (eq type t)) ; (keywordp type))
                                            type))))
  (sql-expression :column (keep-symbol-names
                           (or (and (slot-boundp sd 'column) (name (column sd)))
                               (clos:slot-definition-name sd))
                           class)
                  :table (if class (alias (table class)))	; symbol-name already kept
                  :type type))

(defmethod clos:compute-effective-slot-definition ((class standard-db-class)
                                                   slot-name direct-slots)
 ;;; Args: direct-slots  Ordered list of dsds from class and its superclasses.
  ;; CAUTION: We must not mutate dsd as it can be used by subclasses!
  ;; Called by: finalize-inheritance -> compute-slots.
  (let* ((dsd (first direct-slots))
         (db-class-sd-mixin (typep dsd 'db-class-sd-mixin))
         type)
    ;; Store the user-specified type and then compute real Lisp type and store it
    ;; Avoid error during finalization of class: <type> is an illegal type specifier.
    ;; User should put exotic type into sql-column instance.
    ;; Q: Enhance type as (or ,type null) for nullable columns?
    (and db-class-sd-mixin ;(null (db-type dsd))
          ;(setf (db-type dsd) type) ;(compute-lisp-type-from-slot-specification dsd type)
         (setq type (clos:slot-definition-type dsd))
         (or (keywordp (first-or-self type))
             (memq type '(#+local-time lt:local-time)))
	 ;; replace dangerous non-lisp type specification by T before calling next method
         (setf (clos:slot-definition-type dsd) t))
      
    (let ((esd (call-next-method)))
      (with-slots (column db-kind db-info) esd
        (when db-class-sd-mixin
                ;; Defaults :db-kind to :base if not supplied
          (setf db-kind (if (slot-boundp dsd 'db-kind) (db-kind dsd) :base)
                (clos:slot-definition-type dsd) type	; restore original type of dsd
                ;db-type (db-type dsd) ;(delistify-dsd (db-type dsd))
	 
                ;; Ensure column only for base slots passing the type specified
                ;; into a newly created (or extant sql-column - dangerous?!) instance
                ;; Extract type information, first, from column, second, from dsd,
                ;; finally, from esd inherited from a superclass.
                column (let ((sql-ident (if (slot-boundp dsd 'column) (column dsd))))
                         (when (or sql-ident
                                   (memq db-kind '(:base :key)))
                           (make-sql-column dsd class
                            (cond ((and (typep sql-ident 'sql-column)	 ; explicit db type
                                        (getf (plist sql-ident) :type))) ; in column option
                                  ((and type (neq type t))		 ; try dsd
                                   type)
                                  ((and (setq type (clos:slot-definition-type esd))
                                        (neq type t))			 ; try esd
                                   type)))))
                         #|(if (and type (neq type t))			; useful type info
                             (if (typep sql-ident 'sql-column)
                                 (progn (unless (getf (plist sql-ident) :type)
                                          (setf (getf (plist sql-ident) :type) type))
                                     #1=(unless (slot-value sql-ident 'qualifier)
                                          (setf (slot-value sql-ident 'qualifier)
                                                (alias (table class))))
                                     sql-ident)
                                 (make-sql-column dsd class type))
                             (if (typep sql-ident 'sql-column)	; no type info to add
                                 (progn #1# sql-ident)
                                 (make-sql-column dsd class nil))))|#
          ;(setf column (make-sql-column dsd class)
          ;(setf column (if (slot-boundp dsd 'column) (column dsd) slot-name)
                ;(column-name-from-arg
                ; (if (slot-boundp dsd 'column)
                ;     (delistify-dsd (column dsd))
                ;     (column-name-from-arg (sql-escape slot-name))))
	 
                ;(slot-value esd 'db-writer)
                ;(when (slot-boundp dsd 'db-writer)
                ;  (delistify-dsd (view-class-slot-db-writer dsd)))

                ;(slot-value esd 'db-constraints)
                ;(when (slot-boundp dsd 'db-constraints)
                ;  (delistify-dsd (view-class-slot-db-constraints dsd)))
	 
                ;; Inherit from the most specific non-empty :db-info slot option and
                ;; replace the :self join class by the actual name of this class
                db-info (let* ((dsd (find-if (lambda (dsd) (and (slot-boundp dsd 'db-info)
                                                                (db-info dsd)))
                                             direct-slots))
                               (plist (if dsd (db-info dsd))))
                          (when (eq (getf plist :join-class) :self)
                            (setf plist (copy-list plist)
                                  (getf plist :join-class) (class-name class)))
                          plist)
                ;(specified-type esd) (delistify-dsd (specified-type dsd))
	 )
         ;; For all other slots, esd is not of db-class-esd yet. Do we need to convert?
	 #|(let (#+openmcl (type-predicate (slot-value esd 'ccl::type-predicate)))
           #-clisp (change-class esd 'db-class-esd
                                 #+allegro :name #+allegro (slot-definition-name dsd))
           #+openmcl (setf (slot-value esd 'ccl::type-predicate) type-predicate)
           (setf column (make-sql-column dsd class)
                 db-info nil
                 db-kind :standard)
                 ;(specified-type esd) (slot-definition-type dsd)
         )|#)
      esd))))
  
(let ((lw:*handle-warn-on-redefinition* :warn))
 (defmethod CLOS:FIND-SLOT-DEFINITION ((slots list) class &optional effectivep)
  ;; Exend the generic by the method on list of slots
  (do ((result ()) (rest slots (rest rest)))
      ((null rest) (nreverse result))
    (if-bind (sd (clos:find-slot-definition (first rest) class effectivep))
      (setq result (cons sd result))
      ;(slot-missing class (first objects) slot-name 'setf)
      (error 'sql-user-error :message (format nil "Slot ~A is missing from class ~A"
                                              (first rest) (class-name class)))))) )


#||;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  SPARE  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(setq l (stable-sort bs (lambda (x y) (and (getf (db-info x) :auto-unique-valued)
                                           (not (getf (db-info y) :auto-unique-valued))))))

;;; A non-nil value means that creating View Class instances or
;;; setting their slots automatically creates/updates the
;;; corresponding records in the underlying database.
;;; DI 2007-Apr-11:
;;;  This is rather exotic and can be used in application-specific manner for specific 
;;;  db classes. Such classes require the (:optimize-slot-access nil) option.
  
(defvar *db-auto-sync* nil)
(defvar %db-initializing% nil)	; T when we are already in initialize-instance

(defmethod (setf closs:slot-value-using-class) (new-value
                                          (class standard-db-class) instance slot-name)
  (declare (ignore new-value))
  (prog1 (call-next-method)
    (when (and *db-auto-sync* (not %db-initializing%) (not %db-reading%))
      (let ((sd #+lispworks (clos:find-slot-definition slot-name class)
                #-lispworks (prog1 slot-name
                              (setq slot-name (slot-definition-name slot-name)))))
        (unless (eq (db-kind sd) :standard)
          (update-record-from-slot instance slot-name))))))

(defmethod initialize-instance ((instance standard-db-object) &rest initargs &key)
  (declare (ignore initargs))
  (let ((%db-initializing% t))
    (call-next-method)))
    (when (and *db-auto-sync* (not %db-reading%))
      (update-records-from-instance instance))))

;;; OLD

(defclass %slot-order-test-class () ((a) (b)))
(clos:finalize-inheritance (find-class '%slot-order-test-class))
(let ((slots (clos:class-slots (find-class '%slot-order-test-class))))
    (ecase (slot-definition-name (first slots))
      (a)
      (b (pushnew :mop-slot-order-reversed cl:*features*)))))
       
(defun ordered-class-slots (class)
  #+mop-slot-order-reversed (reverse (class-slots class))
  #-mop-slot-order-reversed (class-slots class))	 ; lispworks!

(%make-base-slot-info :a 2)

(defun get-keywords (keys plist)
  (mapcar (lambda (key) (getf key plist)) keys))

(defstruct (base-slot-descriptor (:constructor %make-base-slot-descriptor))
  name
  type
  db-kind
  column)

;(defvar +extra-class-options+ '(:base-table))
;(defvar +extra-slot-options+
; '(:column :db-kind :db-type :db-constraints :db-info));:db-reader :db-writer :void-value
			       
(defun register-metaclass (class direct-slots)
  (setf #|(descriptors class)	; := ((name1 type1 db-kind1 column1) ...)
         (stable-sort (remove-if-not
                       (lambda (base-slot-descriptor)
                         (member (base-slot-descriptor-db-kind base-slot-descriptor)
                                 '(nil :base :key)))
                       (mapcar (lambda (slot)
                                 (apply '%make-base-slot-descriptor
                                        (retain-properties slot
                                                       '(:name :type :db-kind :column))))
                               direct-slots))
                      #'string<  :key #'base-slot-descriptor-name)|#
       (key-slots class) (remove :key (clos:class-slots class)
                                 :key 'db-kind  :test-not 'eq)))


(defmethod clos:compute-slots ((class standard-db-class))
 ;;; Need to sort order of class slots so they are the same across implementations.
  (let ((slots (call-next-method))
	desired-sequence
	output-slots)
    (dolist (c (clos:compute-class-precedence-list class))
      (dolist (s (clos:class-direct-slots c))
	(let ((name (clos:slot-definition-name s)))
	  (unless (find name desired-sequence)
	    (push name desired-sequence)))))
    (dolist (desired desired-sequence)
      (let ((slot (find desired slots :key #'clos:slot-definition-name)))
	(assert slot)
	(push slot output-slots)))
    output-slots))

(defun register-metaclass (class direct-slots)
  (setf (object-definition class)	; := ((name1 type1 db-kind1 column1) ...)
         (stable-sort (remove-if-not
                       '(lambda (quad) (member (third quad) '(nil :base :key)))
                       (mapcar (lambda (slot)
                                 (get-keywords '(:name :type :db-kind :column) slot))
                               direct-slots))
                      #'string<  :key #'car)
        (key-slots class)
        (remove-if-not (lambda (slot) (eql (slot-value slot 'db-kind) :key))
                       (clos:class-slots class))))

#+unused
(defparameter *db-info-lambda-list*
  '(&key join-class home-key foreign-key (key-join nil) (target-slot nil)
         (retrieval :immmediate) (set nil)))

(defmethod clos:slot-value-using-class ((class standard-db-class) instance slot-name)
 ;; This one does not work in a delivered image
  (declare (optimize (speed 3)))
  (unless %db-reading%
    (let ((sd #+lispworks (clos:find-slot-definition slot-name class)
              #-lispworks (prog1 slot-name
                            (setq slot-name (slot-definition-name slot-name)))))
      (when (and (eq (db-kind sd) :join)
                 (not (slot-boundp instance slot-name))
                 (database instance))
        (let ((%db-reading% t))
          (setf (slot-value instance slot-name) (join-slot-value-fault instance sd))))))
  (call-next-method))
||#
