;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2006 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Object-oriented interface to SQL - User's API

(in-package :ystok.sql.sys)

(defclass standard-db-object ()
 ((database :accessor database  :initarg :database  :initform nil  :db-kind :standard
            :documentation "Associated database where the instance is stored."))
 (:metaclass standard-db-class)
 (:documentation "Superclass for all View Classes."))


;;; Helpers

(defun check-slot-type (sd value)
  (let* ((slot-type (clos:slot-definition-type sd)) ; (specified-type sd)
         (basetype (first-or-self slot-type)))
    (when (and slot-type value (neq slot-type t))
      (unless (typep value basetype)
        (error 'sql-user-error
	       :message (format nil "Invalid value ~A in slot ~A, not of type ~A."
                                value (clos:slot-definition-name sd) slot-type))))))

;;; Database-dependent instance-independent converters

(defgeneric db-sql-value (database value type)
 (:documentation "Value-to-datum converter returning a Lisp object being used
 while generating SQL and passed to low-level cursor operations.
Arguments: 
 TYPE : Lisp or database-specific type specifier.")
 (:method (database value type)
  (declare (ignore type sd))
  value)
 (:method (database value (type (eql 'boolean)))
  (declare (ignore database type))
  (cond ((numberp value) value)
          ;((column-nullable column) (case value ((:false nil) value) (otherwise :true)))
        (value :true)
        (:false)))
 (:method (database (value integer) (type (eql 'universal-time)))
  ;; Used without local-time
  (declare (ignore database type))
  (timestring value)) )

(defgeneric db-datum-value (database datum type)
 (:documentation "Datum-to-value converter returning a Lisp object being stored in a slot.
Arguments: 
 TYPE : Lisp or database-specific type specifier.")
 (:method (database datum type)
  (declare (ignore database type))
  datum)
 (:method (database (datum number) (type (eql 'boolean)))
  (declare (ignore database type))
  (/= datum 0)) )


;;; User specialized generics - instance-dependent converters.
;;; NB: We cannot specialize on both database and instance in the same method.
;;; Q: Do we need SD?

(defgeneric slot-value-to-sql (value instance slot-name type &optional database sd)
 ;;; High-level manipulating via SQL.
 (:method (value instance slot-name type &optional database sd)
  (declare (ignore sd))
  (db-sql-value (or database (database instance)) value type)) )
  ;;(check-slot-type sd value)
  ;(value-to-sql (slot-value instance slot-name) type sd)))

(defgeneric slot-value-to-datum (value instance slot-name type &optional database sd)
 ;;; Different low-level updating via cursors and parameters
  ;; The default method returns the same as for SQL.
 (:method (value instance slot-name type &optional database sd)
  (slot-value-to-sql value instance slot-name database sd))
 (:method (value instance slot-name (type (eql 'boolean)) &optional database sd)
  (declare (ignore instance slot-name type database sd))
  (cond ((numberp value) value)
        ;((column-nullable column) (case value (:false 0) ((nil) nil) (otherwise 1)))
        (value 1)
        (0))) )

(defgeneric set-slot-value-from-datum (datum instance slot-name type &optional database sd)
 ;;; Args: datum  Datum returned by sql:select or low-level %read-datum.
  ;; Called by: db-fetch-into, set-slot-values-from-db
 (:method (datum instance slot-name type &optional database sd)
  (declare (ignore sd))
  (setf (slot-value instance slot-name)
        (db-datum-value (or database (database instance)) datum type))) )
  ;(setf (slot-value instance slot-name) datum))
 #|(:method (datum instance slot-name (type (eql 'boolean)) &optional sd)
  (declare (ignore type sd))
  (setf (slot-value instance slot-name) (if (numberp datum) (/= datum 0) datum)))|#

;;; Internal convenience functions for base slots

(defun %sd-value-sql (instance sd &optional database
                                            (slot-name (clos:slot-definition-name sd))
                                            (value (slot-value instance slot-name))
                                            (type (or (db-type sd) (type-of value))))
  (slot-value-to-sql value instance slot-name type sd database))

#+unused
(defun %sd-value-datum (instance sd &optional database
                                              (slot-name (clos:slot-definition-name sd))
                                              (value (slot-value instance slot-name))
                                              (type (or (db-type sd) (type-of value))))
  (slot-value-to-datum value instance slot-name type sd database))

#+unused
(defun (setf %sd-value-datum) (datum instance sd &optional database
                                                 (slot-name (clos:slot-definition-name sd))
                                                 (type (db-type sd)))
  (set-slot-value-from-datum datum instance slot-name type sd database)
  datum)


(defun make-pk-where (database instance)
  ;(declare (ignore database))
  (do* ((class (class-of instance))
        ;(table (table class))
        (sd)
        (conjuncts ())
        (sds (key-slots class) (rest sds)))
      ((null sds)
       (apply (sql:sql-operator 'and) conjuncts))
    (setq sd (first sds)
          conjuncts (cons (sql-operation '=
                            (sql-column sd class)
                            (%sd-value-sql instance sd database))
                            ;(slot-value-to-sql instance (clos:slot-definition-name sd)
                            ;                   (clos:slot-definition-type sd) sd))
                          conjuncts))))

(defmethod sql-ident-table ((class standard-db-class))
  (let ((table (table class)))
    (if (eq (type-of table) 'sql-ident-table)
        table						; cached value
        (setf (table class) (sql-expression :table (or (name table)
                                                       (class-name class)))))))

(defmethod sql-column ((sd db-class-sd-mixin) &optional class)
  (let ((column (column sd)))
    (case (type-of column)
      (sql-column
       column)						; cached value
      (sql-ident
       ;; Beware of rewriting coalesced SQL identifiers unintentionally.
       (let ((qualifier (if class (alias (table class))))
             (type (clos:slot-definition-type sd)))
         (if (or (if (eq type t) (setq type nil) type))
                 qualifier)
             (setf (column sd) (make-instance 'sql-column :name (name column)
                                              :qualifier qualifier
                                              :plist (if type `(:type ,type) ())))
             (change-class column 'sql-column)))
      (otherwise
       (when (base-slot-p sd)
         (setf (column sd) (make-sql-column sd class)))))))

(defun collect-base-sc-alist (class &optional (slots t) qualifier)
 ;;; Args: slots     T means all base slots or list of slot names
  ;;       qualifier Alternative table alias to be referenced by columns.
  ;; Value: ((sd . sql-column) ...) - base slots only
  ;; Used by: fetch-objects, update-instance-from-records, find-all
  (do ((alist ())
       (sds (base-slots class) (rest sds)))
      ((null sds)
       (if alist
           (nreverse alist)
           (error 'sql-user-error :message
                  (format nil "No slots of kind :base in view class ~A"
                          (class-name class)))))
    (let ((sd (first sds)))
      (when (or (eq slots t) (memq (clos:slot-definition-name sd) slots))
        (when-bind (sql-column (sql-column sd class))
          (setq alist (acons sd (if qualifier
                                    (sql-expression :column (name sql-column)
                                                    :table qualifier)
                                    sql-column)
                             alist)))))))

(defun set-slot-values-from-db (instance sds data)
 ;;; Called by: find-all, update-instance-from-records, update-slot-from-record
  (do ((sds sds (rest sds))
       (values data (rest values)))
      ((null sds)
       instance)
    (let ((sd (first sds)))
      (set-slot-value-from-datum (first values) instance (clos:slot-definition-name sd)
                                 (clos:slot-definition-type sd) sd))))

;;; Symbolic SQL syntax

(defsql sql-slot-value (:symbol "SLOT-VALUE") (class-name slot-name)
  (let* ((class (if (symbolp class-name) (find-class class-name) class-name))
	 (sd (clos:find-slot-definition slot-name class t)))
    (if sd
        (sql-column sd class)
        ;(slot-missing class instance slot-name 'slot-value
        (error 'sql-reader-error :format-control "Slot ~A is missing from class ~A"
               :format-arguments (list slot-name
                                       (if (symbolp class-name) class-name
                                           (class-name class)))))))

;(defsql sql-view-class (:symbol "view-class") (class-name)
;  (sql-ident-table (find-class class-name)))

(defmethod join-where ((class standard-db-class) join-slot &optional join-class sc-alist)
 ;;; Args: sc-alist  ((slot-description . sql-column) ...) correspoinding to join-class.
  ;; Values:
  ;;  1) List of conjuncts that can be used only in where
  ;;  2) List of conjuncts that can be used in either INNER/OUTER join or WHERE
  ;; Used by: Filling in join slot with :retrieval :immediate when no instance is given
  ;; CAUTION: On redefining class, esd instances can diverge. Due to cdr-assoq, this
  ;;	      code begin producing spurious IsNull(...) or ... IS NULL conjuncts!
  (let ((db-info (db-info join-slot))
        (where ())
        (on ()))
    (unless join-class (setq join-class (find-class (getf db-info :join-class))))
    (loop for hk in (listify (getf db-info :home-key))
          and jk in (listify (getf db-info :join-key))
          and join = t
          for arg1 = (if (and hk (symbolp hk))
                         (if (eq hk :not-null)
                             (setq join nil)
                             (sql-slot-value class hk))
                         ;(sql-column (clos:find-slot-definition hk class t) class)
                         (progn (setq join nil)
                           hk))
          and arg2 = (if (and jk (symbolp jk))		; substitute column reference
                         (let ((sd (clos:find-slot-definition jk join-class t)))
                           (if sc-alist
                               (cdr-assoq sd sc-alist)
                               (sql-slot-value join-class jk)))
                                    ;(sql-column sd join-class)))
                         (progn (setq join nil)
                           jk))
          for conjunct = (if arg1
                             (if arg2
                                 (sql-operation '= arg1 arg2)
                                 (sql-operation 'null arg1))
                             (sql-operation (if (eq hk :not-null) :not-null 'null) arg2))
          do (if join
                 (push conjunct on)
                 (push conjunct where)))
    (values where on)))
    ;(apply #'sql-and conjuncts)))

(defmethod join-where ((instance standard-db-object) join-slot &optional join-class
									 db-info)
 ;;; Values: (both in reverse order!)
  ;;  1) List of conjuncts that are to be used only in where;
  ;;     NIL if some of home key slots is unbound or null.
  ;;  2) List of values that the home key consists of.
  (unless db-info (setq db-info (db-info join-slot)))
  (unless join-class (setq join-class (find-class (getf db-info :join-class))))
  (do (hk jk arg1 arg2
       (where ())
       (tuple ())			; values of home key
       (hks (listify (getf db-info :home-key)) (rest hks))
       (jks (listify (getf db-info :join-key)) (rest jks)))
      ((not (and hks jks))
       (values where tuple))
    (setq arg1 (if (and (setq hk (first hks)) (symbolp hk))
                   (cond ((eq hk :not-null) nil)
                         ((and (slot-boundp instance hk) (slot-value instance hk)))
                         (t (return-from join-where ())))
                   hk)
          arg2 (if (symbolp (setq jk (first jks)))	; substitute column reference
                   (sql-column (clos:find-slot-definition jk join-class t) join-class)
                   jk)
          tuple (cons arg1 tuple)
          where (cons (if arg1
                          (sql-operation '= arg1 arg2)
                          (sql-operation (if (eq hk :not-null) :not-null 'null) arg2))
                      where))))
  #|(loop for hk in (listify (getf db-info :home-key))
          and jk in (listify (getf db-info :join-key))
          for arg1 = (if (and hk (symbolp hk))
                         (cond ((eq hk :not-null) nil)
                               ((and (slot-boundp instance hk) (slot-value instance hk)))
                               (t (return-from join-where ())))
                         hk)
          and arg2 = (if (symbolp jk)		; substitute column reference
                         (sql-column (clos:find-slot-definition jk join-class t)join-class)
                         jk)
          collect (if arg1
                      (sql-operation '= arg1 arg2)
                      (sql-operation (if (eq hk :not-null) :not-null 'null) arg2)))))|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; DDL

(defmacro def-view-class (class-name superclasses slots &rest options)
 `(dspec:def (def-view-class ,class-name)
    (defclass ,class-name ,superclasses ,slots
      ,@(dolist (option '((:metaclass standard-db-class)) options)
                          ;(:optimize-slot-access nil))  ; enable slot-value-using-class
          (setf options (adjoin option options :key #'first))))
    (let ((class (find-class ',class-name)))
      (unless (clos:class-finalized-p class)
        (clos:finalize-inheritance class))
      class)))

(dspec:define-dspec-alias def-view-class (class-name) `(defclass ,class-name))

(defgeneric db-create-class-view (database class)
 (:method ((database database) (class standard-db-class))
 ;;; Is not key-slots obligatory?
  ;; TODO: In constraints, translate referenced class and slot names to table and columns.
  (let* ((columns (mapcar (lambda (sd)
                            (if-bind (constraints (getf (db-info sd) :constraints))
                              (cons (sql-column sd class)
                                    (mappend (lambda (constraint)
                                               (if (consp constraint)	; with name
                                                   (cons :constraint constraint)
                                                   (list constraint)))
                                             (listify constraints)))
                              (sql-column sd class)))
                          (base-slots class)))
         (constraints 
          `(,@(when-bind (key-slots (key-slots class))
                `((,(sql-expression :identifier (string-append "PK_"
                                                 (substitute #\_ #\-
                                                  (string (name(sql-ident-table class))))))
                   :primary-key ,(mapcar (lambda (sd) (sql-column sd class)) key-slots))))
            ,@(constraints class))) )
    (unless columns
      (error 'sql-user-error :message (format nil "Class ~S has no :base slots" class)))
    (create-table (sql-ident-table class) columns
                  :constraints constraints
                  :database database)))
 (:method :after ((database database) (class standard-db-class))
  (push class (view-classes database)))
 (:method ((database database) (class-name symbol))
  (db-create-class-view database (find-class class-name))) )
  ;(if-bind (class (find-class class-name))
  ;  (error 'sql-user-error :message (format nil "Class ~S not found." class-name)))) )

(defun create-view-from-class (class-name &key (database *default-database*))
 "Create a table as defined by the view class CLASS-NAME in DATABASE."
  (db-create-class-view database class-name) ;:transactions transactions
  (values))


(defgeneric db-drop-class-view (database class)
 ;;; TODO: Invoke drop-view instead of drop-table if needed?
 (:method ((database database) (class standard-db-class))
  (drop-table (sql-ident-table class) :database database))
 (:method :after ((database database) (class standard-db-class))
  (lw:removef (view-classes database) class))
 (:method ((database database) (class-name symbol))
  (db-drop-class-view database (find-class class-name)))
 (:method :around ((database database) (class-name symbol))
  (when (table-exists-p (sql-ident-table (find-class class-name)) :database database)
    (call-next-method))) )

(defun drop-view-from-class (class-name &key (database *default-database*))
 "Remove a table defined by the view class CLASS-NAME from DATABASE."
  (db-drop-class-view database class-name)
  (values))


(defun list-classes (&key (root-class (find-class 'standard-db-object))
 		          (test #'identity)
		          (database *default-database*))
 "Returns a list of all the view classes that are connected to DATABASE and
 - descend from the ROOT-CLASS given,
 - satisfy the TEST predicate."
  (when-bind (view-classes (and database (view-classes database)))
    (remove-if-not (lambda (class) (and (funcall test class)
                                        (memq root-class
                                              (clos:class-precedence-list class))))
                   view-classes)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; DQL
;;; Join slot triple ::= (join-class sql-ident-table . sc-alist)

;; Assigned T by slot-unbound, fetch-objects, and update-objects-joins
;; to prevent recursive entering to join-slot-value-fault
(defvar %db-reading% nil)

#+deprecated
(defgeneric instance-refreshed (instance &optional slots)
 ;;; Args: slots  List of slot names that have changed; empty if cannot be obtained.
  ;; Q: Is it worth passing a kind of source or update origin argument?
 (:method ((instance standard-db-object) &optional slots)
  (declare (ignore instance slots))))

(defgeneric instance-updated (instance &key changed &allow-other-keys)
 ;;; Hook on select and standard commit action defined when
  ;; - the refresh argument is true,
  ;; - and the instance have just been fetched, updated or inserted.
  ;; Args: changed If eq :new, record corresponding the instance just have been inserted
  ;;		   by means of the running application (:new cannot come from fetch!).
  ;;		   Otherwise an alist ((slot-name . old-slot-value) ...) is passed
  ;;		   moreover, join slots can be listed.
  ;;	   slots   List of slot names or definitions that have changed;
  ;;	   	   empty if cannot be obtained.
  ;;		   The specialization method itself must tell slot names from definitions.
  ;;       source  The update origin, nil if comes from the database.
  ;; Values: None (ignored).
  ;; Called by: (select ... :refresh t) -> (db-fetch-into cursor instance :refresh t)
  ;;		or (update-records-from-instance/slot ... :refresh t)
  ;; NB: Detecting whether some slots were actually refreshed is rather tough during fetch
  ;;     and can be performed
  ;;	 - either before in the specialization on ensure-db-object,
  ;;	 - or later in the specialization on instance-updated itself by means of
  ;;	   comparing to some local copy.
 (:method ((instance standard-db-object) &key changed)
  (declare (ignore changed))) )


(defgeneric instance-deleted (instance &key &allow-other-keys)
 ;;; - The generics is either invoked by delete-instance-records when the refresh argument
  ;;   is true and the instance have just been deleted or explicitly.
  ;; - The condition is signaled by update-record(s)-from-slot/instance and
  ;;   update-slot/instance-from-record(s)
  ;;   In the case of lock, the (subcondition of) sql-temporary-error should be signaled.
  ;; Args: source  The update origin, nil if comes from the database.
  ;; Values: None (ignored).
 (:method ((instance standard-db-object) &key)))
 ;(:method :after ((instance standard-db-object) &key) (setf (database instance) nil))

(define-condition instance-deleted (sql-user-error)
 ((instance :initarg :instance))
 (:report (lambda (c stream)
            (format stream "Object ~A was deleted (or locked) by another user~@[.~%~A~]"
                    (slot-value c 'instance) (message c)))))


(declaim (inline allocate-db-object))
(defun allocate-db-object (class database)
 ;;; Create an instance for an existing DB record in a zero-initialization manner
  (let ((instance (allocate-instance class)))
    (setf (database instance) database)
    instance))

(defgeneric db-class-to-allocate (instance cursor start-column)
 ;;; Determine the actual class of a new instance to allocate.
  ;; Args: instance  Usually class-prototype
  ;; Value: Class metaobject chosen from the column value, which could be read this way:
  ;;	(sql:fetch-item cursor <column-name>)
  ;;  or, to avoid searching column position from the column name at run-time,
  ;;	(sql:fetch-item cursor
  ;;	 (+ start-column (load-time-value (sql.sys:base-slot-position
  ;;					   '<slot-name> '<standard-db-class-name>))))
 (:method ((instance standard-db-object) cursor start-column)
  (class-of instance)))

(defun base-slot-position (slot-name class)
  (position slot-name (base-slots (if (symbolp class) (find-class class) class))
            :key #'clos:slot-definition-name))

(defgeneric ensure-db-object (cursor instance
                              &key ;start-column slots columns database refresh
                              &allow-other-keys)
 ;;; Make a new instance of standard-db-object or return an existing object from cache.
  ;; Args: instance     Class prototype - must not be mutated in any way.
  ;;	   database	When defaulted, called by db-fetch-into to fill from a db record.
  ;;	   start-column Number of the first descriptor record
  ;;			(corresponding base slots usually occupy adjacent records).
  ;;	   slots	List of all base slot definitions.
  ;;	   refresh	True (specifically, :join if instance is automatically joined)
  ;;			or NIL.
  ;; Values: 1) New or extant instance
  ;;	     2) Non-empty ((slot-name . old-value)...) alist or any true value
  ;;		if found in cache and instance-updated should be invoked,
  ;;		or NIL if not present in cache/RAM.
  ;; Called by: (select ... :refresh t) -> (db-fetch-into cursor instance :refresh t)
  ;; NB: Key values to examine can be obtained from the cursor by means of
  ;;	   (fetch-item cursor <column> :current t)
  ;;	 where <column> can be found in one of the following ways:
  ;;     1) (+ start-column
  ;;           (load-time-value (base-slot-position <key-slot-name> <class-name>))
  ;;	 2) (+ start-column
  ;;	       (position <key-slot-name> slots :key #'clos:slot-definition-name))
  ;;	 Passing <slot-name> instead of column number is unsafe due to potential aliases.
  ;; NB2: Specialization can initialize specific or only standard slots according
  ;;	  to initforms as follows:
  ;;      (shared-initialize instance (mapcar #'clos:slot-definition-name
  ;;                                          (remove :standard (clos:class-slots class)
  ;;                                                  :key #'db-kind :test-not #'eq)))
 (:method (cursor (instance standard-db-object) &key (database (database cursor))
                                                     start-column)
  ;; Default method always allocates a new instance.
  (allocate-db-object (if start-column
                          (db-class-to-allocate instance cursor start-column)
                          (class-of instance))
                      database)) )
  ;(make-instance (class-of instance) :database database)))

(defun object-order-by (order-by class)
  (mapcar (lambda (elt &aux sd)
            (if (and (symbolp elt) (setq sd (clos:find-slot-definition elt class t)))
                (sql-column sd class)
                elt))
          (listify order-by)))

(defun object-select (names selections key-args)
 ;;; If a single view class is being queried, allow plain symbols in :order-by
  ;; and implicitly qualify them with the corresponding table alias.
  ;; As an alternative, write down :order-by [slot-value 'class 'slot]
  (let ((order-by (getf key-args :order-by)))
    (when (and order-by (null (rest names)))
      (setq order-by (object-order-by order-by (find-class (first names)))))
    (apply #'fetch-objects names selections
           :order-by order-by
           (remove-properties key-args '(:order-by)))))

(defun collect-join-triple-lists (join-slot-lists &optional classes)
 ;;; Collect selections and from-tables with aliases if needed
  ;; Args: join-slot-lists ::= ((sd11 ... sd1m) (sd21 ... sd2n) ...)
  ;;			Is result of (mapcar #'join-slots classes).
  ;;       classes	Selected classes themselves; checked to switch aliasing on.
  ;; Values: List of lists of triples:
  ;;		(((join-class sql-ident-table (sd111 . col111) (sd112 . col112)) ...) ...)
  ;;	     where table and col can take foreign slot name as a primary table alias.
  ;; NB:   ,    join-    
  ;;	   :referenced . ,    division  chief,
  ;;	  accountant (:join-classs employee),    !
  ;; TODO:         
  ;;	   join-.
  (do* ((flatten (flatten join-slot-lists))		; Loop over classes of selections
        (acc ())
        (rest join-slot-lists (rest rest)))
       ((null rest)
        (nreverse acc))
    (do* ((triples ())		    			; Loop over slot in join-slots,
          (join-slots (first rest) (rest join-slots)))	; i.e. over join-classes
         ((null join-slots)
          (push (nreverse triples) acc))		; retain join slots order!
      (let* ((join-slot (first join-slots))
             (class-name (join-class join-slot))
             (join-class (find-class class-name))
             (sql-ident-table (sql-ident-table join-class))
             (alias (when (or (> (count class-name flatten :key #'join-class) 1)
                              (memq join-class classes))
                      (clos:slot-definition-name join-slot))))
        (push (list* join-class
                     (if alias				; :type?
                         (sql-expression :table (name sql-ident-table) :alias alias)
                         sql-ident-table)
                     (collect-base-sc-alist join-class
                       ;(if-bind (target-slot (getf (db-info join-slot) :target-slot))
                       ;  (list target-slot)		; is target-slot base?
                       t
                       alias))
              triples)))))


(defun fetch-objects (names selections &rest args
		      &key from where (database *default-database*) flatp refresh
                           (joins t) 
                      &allow-other-keys)		; distinct order-by result-types
 ;;; Fetch object query results including all or none of immediately joined objects
  ;; Args:
  ;;  names	 List of view class names.
  ;;  selections SQL-expressions of "ordinary" selections from 'from';
  ;;		 corresponding data are fetched at the tail of every resulting tuple.
  ;;  from	 An "ordinary" table/view or a list of such tables that are being
  ;;		 theta-joined with the "master" ones according to where condition.
  ;;  refresh	 If true, causes for all instances to be fetched in "refresh mode"
  ;;		 except for the following distinct values:
  ;;	   :join - means applying "refresh mode" only to joined objects but not to home;
  ;;		     this value is useful when the home object definitely does
  ;;		     not exist and for avoiding infinite recursion.)
  ;;	   :home - means applying "refresh mode" only to home objects.
  ;;  joins   	 If true, join corresponding tables the SELECT statement:
  ;;		 - if a list of slot names, for all the join slots mentioned
  ;;		   (no matter what retrieval kind was specified in def-view-class),
  ;;		 - otherwise, if eq T, all immediate join slots;
  ;;		 Additionally, this parameter provides the value passed as the refresh
  ;;		 argument to fetch-into for each joined object.
  ;;		 If false, do not fetch joined objects at all.
  ;; Called by:
  ;;  - select when the view class names are passed,
  ;;  - join-slot-value-fault.
  ;; NB: We generate LEFT JOIN instead of WHERE to treat the situation where
  ;;	 foreign keys are null.
  ;(declare (ignore distinct order-by result-types))
  #-debug (declare (optimize (debug 3) (speed 1)))
  (let* ((%db-reading% t)
         (refresh-home (if (eq refresh :join) nil refresh))
         (refresh-join (if (and refresh (neq refresh :home)) joins nil))
         (classes (mapcar #'find-class names))
         (sc-alists (mapcar #'collect-base-sc-alist classes))
         (join-slot-lists (cond ((consp joins)			; explicit list - collect
                                 (mapcar (lambda (class)	; all retrieval kinds
                                           (delete joins (join-slots class t)
                                                   :test-not #'xmemq
                                                   :key #'clos:slot-definition-name))
                                         classes))
                                (joins
                                 (mapcar #'join-slots classes))
                                (t (make-list (length classes)))))
         (triple-lists (collect-join-triple-lists join-slot-lists classes))
         ;; plain selection sc-alist with class and join class columns enterleaved
         (fullsels (reduce #'append (mapcar (lambda (sc-alist triples)
                                              (append sc-alist (mappend #'cddr triples)))
                                            sc-alists triple-lists)))
         ;; CAUTION: collect-table-refs puts alias in the name slot - do we need this?
         #|(tables (nconc (mapcar 'sql-ident-table classes)
                        (mapcan (lambda (triples) (mapcar #'cadr triples))
                                triple-lists)))|#
         ;(join-from ())
         (join-where ())
         (tuples ())
        )
    (setq from (listify from))
    ;; Create from and join-where, append it to WHERE specified
    (loop for class in classes
          and join-slots in join-slot-lists
          and triples in triple-lists
          for join-from = (sql-ident-table class)
          do
          (loop for join-slot in join-slots
                and (join-class sql-ident-table . sc-alist) in triples
                do (multiple-value-bind (where on)
                       (join-where class join-slot join-class sc-alist)
                     (if on
                         (setq join-from (|LEFT OUTER JOIN| join-from sql-ident-table
                                                            :on (apply #'sql-and on)))
                         (lw:push-end sql-ident-table from))
                     (lw:appendf join-where where)))
          (lw:push-end join-from from))
    (when join-where
      (setq where (apply #'sql-and (append (listify where) join-where))))

    (with-cursor (cursor (apply #'sql-query-exp
                                (nconc (mapcar #'cdr fullsels)
                                       (copy-list selections)
                                       (list :from from :where where)
						;:result-types result-types
                                       (remove-properties args
                                                         '(:where :flatp :result-types))))
                  :database database)
      (loop				; over tuples in the result set
       (let ((tuple ()))
         (loop with start-column = 1 ; then (1+ end-column)
                    ;end-column = (length sc-alist) then (+ end-column (length sc-alist))
               for class in classes	; over classes
               and sc-alist in sc-alists
               and join-slots in join-slot-lists
               and triples in triple-lists
               and current = nil then t
               for length = (length sc-alist)
               for instance = (db-fetch-into cursor (clos:class-prototype class)
                                             :slots (mapcar #'car sc-alist)
                                             :refresh refresh-home
                                             :current current
                                             :start-column start-column
                                             :end-column (+ start-column length -1))
               unless instance do (return nil)
               do ;; Fill in join slots with retrieval :immediate
                  ;; if the (compound) home key is not null, othewise set to nil.
                  (incf start-column length)
                  (loop for join-slot in join-slots
                        and (join-class nil . sc-alist) in triples
                        for length = (length sc-alist)
                        do
                        (setf (slot-value instance (clos:slot-definition-name join-slot))
                              (if (every (lambda (hk) (imp (and hk (symbolp hk)
                                                                (neq hk :not-null))
                                                           (slot-value instance hk)))
                                         (listify (getf (db-info join-slot) :home-key)))
                                  (db-fetch-into cursor (clos:class-prototype join-class)
                                          :slots (mapcar #'car sc-alist)
                                          :refresh refresh-join
                                          :current t
                                          :start-column start-column
                                          :end-column (+ start-column length -1))
                                  nil))
                        (incf start-column length))
                  (cond (flatp (return (setq tuple instance)))
                        ((setq tuple (cons instance tuple))))
               finally
                  ;; Fill in ordinary selections if any
                  (dolist (expression selections)
                    (setq tuple (cons (db-fetch-item cursor start-column :current t)
                                      tuple))
                    (incf start-column))
                  (setq tuple (nreverse tuple)))
         (if tuple
             (setq tuples (cons tuple tuples))
             (return)))))		; end of result set
    (nreverse tuples)))
                      
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; DML
;;; Methods db-... can be combined within a transaction.
;;; Hooks instance-deleted, instance-updated, and instance-delete are triggered
;;;  1) only if the database is outside an explicit transaction,
;;;  2) basing on the value(s) returned by the effective method.

(defgeneric db-delete-instance-records (database instance &key &allow-other-keys))
(defgeneric db-update-record-from-slot (database instance slot-name &key
                                                                    &allow-other-keys))
(defgeneric db-update-records-from-instance (database instance &key &allow-other-keys))

(defmethod db-delete-instance-records ((database database) (instance standard-db-object)
                                       &key where) ;refresh in-bulk
 ;;; Value: True on success
  (delete-records :from (sql-ident-table (class-of instance))
                  :where (or where (make-pk-where database instance))
                  :database database)
  (setf (database instance) nil)
  t)
  ;(setf (record-caches database) nil)
  ;(when refresh
  ;  (sql:instance-deleted instance :source refresh))
  ;(values))

(defun delete-instance-records (instance &rest args &key refresh &allow-other-keys)
 ;;; Args: refresh  If true, add instance-deleted as a standard commit action.
  (if-bind (database (database instance))
    (when (and (apply #'db-delete-instance-records database instance args)
               refresh)
      (if (memq database *active-transactions*) ;*commit-actions*
          (define-commit-action instance
           `(sql:instance-deleted ,instance ,:source ,refresh))
          (sql:instance-deleted instance :source refresh)))
    (signal-no-database-error))
  (values))

(defmethod db-update-record-from-slot ((database database) (instance standard-db-object)
                                       slot-name
                                       &key where (old *unbound-value*) refresh)
 ;;; Args: slot-name Base (possibly, key, but not auto-unique-valued!) or join slot.
  ;;       old	     Old slot-value passed to instance-updated.
  ;; Values: 1) changed alist or :new
  ;;	     2) slot-name or slot-definition
  (declare (ignore refresh))
  (let* ((class (class-of instance))
         (sql-ident-table (sql-ident-table class))
         ;(sd (clos:find-slot-definition slot-name class t))
         ;(value (%sd-value-sql instance sd database))
         ;(sql-column (sql-column sd)))
         (sds (clos:find-slot-definition (sql.sys:home-slot-names slot-name class)class t))
         (columns (mapcar #'sql-column sds))
         (values (mapcar (lambda (sd) (%sd-value-sql instance sd database)) sds)))
      (cond ((and sds (database instance))
             (multiple-value-bind (result count)
                 (update-records sql-ident-table
                                 :columns columns  :values values  :database database
                                 :where (or where (make-pk-where database instance)))
               (declare (ignore result))
               ;(when refresh
               (if (eql count 1)
                   (values (if (eq old *unbound-value*) nil (acons slot-name old nil))
                           slot-name)
                   (error 'instance-deleted :instance instance))))
                     ;:deleted))))
                     ;(instance-updated instance :slots (list slot-name) :source refresh
                     ;        :changed (if (eq old *unbound-value*)
                     ;                     nil
                     ;                     (acons slot-name old nil)))
                     ;(instance-deleted instance)))))
            ((and sds (not (database instance)))
             (insert-records :into sql-ident-table
                             :columns columns  :values values :database database)
             (setf (database instance) database)
             (values :new slot-name))
             ;(when refresh
             ;  (instance-updated instance :changed :new  :source refresh)))
            ((error 'sql-user-error :message
                    (format nil "Unable to update record from slot '~A'." slot-name))))))

(defun update-record-from-slot (instance slot-name &rest args
                                                   &key (database *default-database*)
                                                        refresh &allow-other-keys)
                                                   ;where refresh (old *unbound-value*))
 ;;; Args: slot-name Base (possibly, key, but not auto-unique-valued!) or join slot.
  ;;       refresh   If true, add instance-updated as a standard commit action.
  ;;       old	     Old slot-value passed to instance-updated.
  (multiple-value-bind (changed slots)
      (apply #'db-update-record-from-slot (or (database instance) database)
             instance slot-name
             (remove-properties args '(:database)))
    (when refresh
      (cond ;((eq changed :deleted)
            ; (error 'instance-deleted :instance instance)) ;(instance-deleted instance))
            ((or changed slots)
             (if (memq (database instance) *active-transactions*) ;*commit-actions*
                 (define-commit-action instance
                  `(instance-updated ,instance ,:changed ,changed
                                               ,:slots ,(listify slots)
                                               ,:source ,refresh))
                 (instance-updated instance :changed changed  :slots (listify slots)
                                   :source refresh))))))
  (values))


(defmethod db-update-records-from-instance ((database database)
                                            (instance standard-db-object)
                                            &key where (slots t) refresh changed)
 ;;; Args: slots (non-standard)
  ;;		T     means all base bound slots excluding keys for extant instance,
  ;;		:key  means all base bound slots including keys),
  ;;		list  explicit list of slot names possibly including:
  ;;			base		   - are taken as is,
  ;;			auto-unique-valued - are simply excluded,
  ;;		        standard and join  - are replaced by corresponding home keys.
  ;;	   refresh If true, try to insert via a cursor, refetch auto-unique-valued slots
  ;;	   changed Alist ((slot-name . old-slot-value) ...) pinpoints to what was
  ;;		   actually modified and is passed to instance-updated as is.
  ;; Values: 1) The value of changed argument as is or :new
  ;;	     2) List of slot names coinciding with slots when it is a non-empty list,
  ;;		or slot definitions when slots is atom.
  ;; NB: Only updates the records from the base slots of instance without looking
  ;;	 at its join slots.
  (let* ((class (class-of instance))
         (sds (cond ((or (eq slots :key) (null (database instance)))
                     (remove-if (lambda (sd)
                                  (or (not (slot-boundp instance ; skip unbound
                                                        (clos:slot-definition-name sd)))
                                      (auto-unique-valued sd)))	; skip auto-unique-valued
                                (base-slots class)))
                    ((eq slots t)
                     (remove-if (lambda (sd)
                                  (or (not (slot-boundp instance
                                                        (clos:slot-definition-name sd)))
                                      (auto-unique-valued sd)
                                      (eq (db-kind sd) :key)))	; skip key slots as well
                                (base-slots class)))
                    (t						; collect base slots from
                     (do (slot					; the explicitly specified.
                          (acc ())				; set of sds accumulated
                          (rest slots (rest rest)))
                         ((null rest)
                          acc)
                       (when (slot-boundp instance (setq slot (first rest)))
                         (case (db-kind (setq slot (clos:find-slot-definition
                                                    slot class t)))
                           ((:base :key)			; base-slot-p
                            (when (not (auto-unique-valued slot))
                              (setq acc (adjoin slot acc :test #'eq))))
                           ((:join :standard)			; a la home-slot-names
                            (cond ((null (setq slot (getf (db-info slot) :home-key))))
                                  ((consp slot)
                                   (when (loop for slot-name in slot
                                             always (imp (symbolp slot-name)
                                                         (slot-boundp instance slot-name)))
                                     (dolist (slot-name slot)
                                       (when (symbolp slot-name)
                                         (setq acc (adjoin (clos:find-slot-definition
                                                            slot-name class t)
                                                           acc :test #'eq))))))
                                  ((and (symbolp slot) (slot-boundp instance slot))
                                   (setq acc (adjoin (clos:find-slot-definition
                                                      slot class t)
                                                     acc :test #'eq))) ))))))))
                     #|(remove-if (lambda (sd)
                                         &aux (slot-name (clos:slot-definition-name sd)))
                                   (or (not (slot-boundp instance slot-name))
                                       (auto-unique-valued sd)	; leave auv?
                                       (not (memq slot-name slots))))))
                                   base-slots))|#
         auv-sds
         (sql-ident-table (sql-ident-table class))
         (columns (mapcar #'sql-column sds))
         (values (mapcar (lambda (sd) (%sd-value-sql instance sd database))
                           ;(slot-value-to-sql instance (clos:slot-definition-name sd)
                           ;                   (clos:slot-definition-type sd) sd))
                           #|(let ((value (slot-value instance
                                                    (clos:slot-definition-name sd))))
                             (check-slot-type sd value)
                             (db-value-from-slot database value sd)))|#
                         sds)))
    ;; Exclude auto-unique-valued slots from update.
    (cond ((null sds)
           (cerror "Do not update records from this instance."
              'sql-user-error :message (format nil "No settable slots in ~S." instance)))
          ((database instance)
           (multiple-value-bind (result count)
               (update-records sql-ident-table
                               :columns columns :values values
                               :where (or where (make-pk-where database instance))
                               :database database)
             (declare (ignore result))
             ;(when refresh
             (if (eql count 1)
                 (values changed (if (consp slots) slots sds))
                 (error 'instance-deleted :instance instance))))
                   ;:deleted))))
                   ;(instance-updated instance :source refresh
                   ;                  :slots sds ;(mapcar #'clos:slot-definition-name sds)
                   ;                  :changed changed)
                   ;(instance-deleted instance)))))
          ((and refresh
                (setq auv-sds (remove-if-not #'auto-unique-valued
                                             (clos:class-slots class))))
           (let ((auv-columns (mapcar #'sql-column auv-sds)))
             (with-cursor (cursor (apply #'sql-query-exp
                                         (append columns auv-columns
                                                 (list :from sql-ident-table)))
						;:result-types result-types
                                :cursor-type :keyset-driven
                                :concurrency :optimistic ;:lock
                                :rowset-size 1
                                :start-column 0
                                :database database)
               (let ((bookmark (insert-records :into cursor
                                               :columns columns :values values)))
                 ;(setf (database instance) database)	; DI 2008-Oct-09
                 ;(refetch-record cursor record)
                 (when (integerp bookmark)
                   (set-cursor-position cursor :bookmark :offset bookmark)
                   (fetch-into cursor instance
                               :current t
                               :slots auv-sds :columns auv-columns)
                   (values :new (if (consp slots) slots auv-sds)))))))
                   ;(instance-updated instance ;(mapcar 'clos:slot-definition-name auv-sds)
                   ;                  :slots auv-sds
                   ;                  :changed :new  :source refresh))))))
          (t
           (insert-records :into sql-ident-table
                           :columns columns :values values
                           :database database)
           ;(setf (database instance) database)		; DI 2008-Oct-09
           (values :new (if (consp slots) slots nil))) )))
           ;(when refresh
           ;  (instance-updated instance :changed :new  :source refresh))) ))

(defun update-records-from-instance (instance &rest args &key (database *default-database*)
                                                              refresh &allow-other-keys)
                                                   ;where (slots t) refresh changed)
  ;;; Args: refresh  If true, add instance-updated as a standard commit action.
  (multiple-value-bind (changed slots)
      (apply #'db-update-records-from-instance (or (database instance) database) instance
             (remove-properties args '(:database)))
    (when (eq changed :new)				; set database slot only on commit
      (if (memq database *active-transactions*)		;*commit-actions*
          (define-commit-action instance
            `(,#'(setf database) ,database ,instance))
          (setf (database instance) database)))
    (when refresh
      (cond ;((eq changed :deleted)
            ; (error 'instance-deleted :instance instance)) ;(instance-deleted instance))
            ((or changed slots)
             (if (memq (database instance) *active-transactions*) ;*commit-actions*
                 (define-commit-action instance
                  `(instance-updated ,instance ,:changed ,changed
                                     ,:slots ,slots
                                     ,:source ,refresh))
                 (instance-updated instance :changed changed  :slots slots
                                   :source refresh))))))
  (values))

(defmethod update-slot-from-record ((instance standard-db-object) slot-name
                                    &key refresh (database *default-database*))
 ;;; Value: instance
  (let* ((class (class-of instance))
         (database (or (database instance) database))
         (sd (clos:find-slot-definition slot-name class t))
         (sds (sql.sys:home-slot-names sd class))
         (tuple (first (if sds
                           (apply #'select
                                  (nconc (mapcar #'sql-column sds)
                                         (list :from (sql-ident-table class)
                                               :where (make-pk-where database instance)
                                               :limit 1)))
                           (error 'sql-user-error :message
                                  (format nil "Unable to update instance slot '~A'."
                                          slot-name))))))
    (cond (tuple
           (if (eq (db-kind sd) :join)
               ;; For join slot, retrieve correspoing instance only if home key vales
               ;; were changed
               (let ((values (mapcar (lambda (sd)
                                       (let ((slot-name (clos:slot-definition-name sd)))
                                         (if (slot-boundp instance slot-name)
                                             (slot-value instance slot-name)
                                             *unbound-value*)))
                                     sds))
                     (%db-reading% t))
                 (set-slot-values-from-db instance sds tuple)
                 (unless (equal values tuple)
                   (setf (slot-value instance slot-name)
                         (join-slot-value-fault instance sd)))
                 instance)
               (set-slot-values-from-db instance sds tuple)))
          ((and refresh (database instance))
           (error 'instance-deleted :instance instance)))))
           ;(instance-deleted instance)))))

(defmethod update-instance-from-records ((instance standard-db-object)
                                         &key (slots t) refresh
                                              (database *default-database*))
 ;;; Args: refresh If true, invoke instance-deleted on failure.
  ;; Value: instance on success; signals instance-deleted if cannot be fetched this time.
  ;; NB: The update is not recursive on joins. Join slots (that is, slots with
  ;;	 :db-kind :join) are updated, but the joined objects are not updated.
  ;; Implementation note:
  ;;	We simply make join slots unbound for further deferred retrieval.
  (let* ((class (class-of instance))
         (database (if (database instance)
                       (if (eq (database instance) database)
                           database
                           (error 'sql-user-error :message
                            (format nil
                                    "Instance ~S has associated with different database ~S"
                                    instance database)))
                       database))
         (alist (collect-base-sc-alist class slots))
         (tuple (apply #'select
                       (nconc (mapcar #'cdr alist)
                              (list :from (sql-ident-table class)
                                    :where (make-pk-where database instance)
                                    :limit 1	;:result-types nil
                                    :database database)))))
    (cond (tuple
           (unless (database instance)
             (setf (database instance) database))
           (dolist (join-slot (join-slots class t))
             (slot-makunbound instance (clos:slot-definition-name join-slot)))
           (set-slot-values-from-db instance (mapcar #'car alist) (first tuple)))
          ((and refresh (database instance))
           (error 'instance-deleted :instance instance)))))
           ;(instance-deleted instance)
           ;nil))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Remote Joins

(defgeneric find-join-object (standard-db-object clos:class-prototype &optional arg
                                                                      &rest args)
 ;;; Application specific search of cached instance of class-name
  ;; Args: instance	Instance from the home side of a subclass of standard-db-object.
  ;;	   (arg . args) Represents the tuple of (compound) home key values;
  ;;	    the order of the values corresponds to the one specified in def-view-class.
  ;; Value: An instance of standard-db-object, a custom query can be executed to fetch it;
  ;;	or  NIL if not found and no further standard fetch is required.
  ;;    or  *unbound-value* if the standard fetch query should be executed.
 (:method (instance clos:class-prototype &optional arg &rest args)
  (declare (ignore clos:class-prototype arg args))
  *unbound-value*))

(defun join-slot-value-fault (instance join-slot &optional (refresh :unbound))
 ;;; Args: refresh  NIL or :join (explicitly comes only from update-objects-joins). 
  ;; Value: NIL  If the record(s) correspoinding to join-slot cannot be found.
  ;;		 Should we signal instance-deleted?
  ;; ASSUMPTION: %db-reading% is bound to T.
  ;; TODO:     target-slot  inner-join
  ;;        (db-info (clos:find-slot-definition target-slot join-class t))
  ;;	        (and target-slot set)
  ;; CommonSQL non-conformance:
  ;;	When set is t and target-slot is defined, the join slot will contain a list of
  ;;	   dotted pairs (target-value . join-instance)
  ;;	   not pairs (target-value join-instance) as defined by CommonSQL standard
  ;;	where target-value is the value of the target slot
  ;;	and join-instance is the corresponding instance of the join class.
  ;; Q:	      join-instance?
  ;; A:     !
  (let* ((db-info (db-info join-slot))
	 (target-slot (getf db-info :target-slot))	; target slot name
	 (set (getf db-info :set))
	 (class-name (getf db-info :join-class))
         (join-class (find-class class-name)))
    (multiple-value-bind (where tuple) (join-where instance join-slot join-class db-info)
      (when where					; home slots are bound and not null
        (unless (or set target-slot)
          (let ((result (apply #'find-join-object
                               instance (clos:class-prototype join-class)
                               (nreverse tuple))))
            (unless (eq result *unbound-value*)
              (return-from join-slot-value-fault result))))
        (let ((result (fetch-objects (list class-name) ()
                                     :where (apply #'sql-and where)
                                     :order-by (object-order-by (getf db-info :order-by)
                                                                join-class)
                                     :refresh refresh
                                     :flatp t :database (database instance))))
          (if target-slot
              (cond (set	; may recurse in (slot-value-using-class join-class ...)?!
                     (mapcar (lambda (object) (cons (slot-value object target-slot)object))
                             result))
                    (result
                     (slot-value (first result) target-slot)))
              (if set
                  result
                  (first result))))))))

(defmethod slot-unbound (class (instance standard-db-object) slot-name)
  (declare (optimize (speed 3)))
  (let ((sd #+lispworks (clos:find-slot-definition slot-name class)
            #-lispworks (prog1 slot-name
                          (setq slot-name (slot-definition-name slot-name)))))
    (if (eq (db-kind sd) :join)
        (if (and (not %db-reading%) (database instance))
            (let ((%db-reading% t))
              (setf (slot-value instance slot-name) (join-slot-value-fault instance sd)))
            nil)
        (call-next-method))))		; signal error

(defvar *default-update-objects-max-len* nil
  "The default value to use for the MAX-LEN keyword argument to UPDATE-OBJECT-JOINS.")

(defun update-objects-joins (objects &key (slots t) (force t) class-name
                                          (max-len *default-update-objects-max-len*)
                                          refresh)
 ;;; Updates join slots from the records of the associated database
  ;; Args: objects    Instance or list of instances whose slots are being updated;
  ;;		      multiple instances must be of the same class (class-name if given).
  ;;	   slots      List of slot names or T meaning that all join slots
  ;;		      with :retrieval :deferred are updated.
  ;;	   class-name The view class of all the objects;
  ;;	 	      when null, the class of the first of the objects is used.
  ;;	   force      T	  = all join slots are updated
  ;;		      NIL = only unbound join slots are updated.
  ;;	   max-len    If integer, update-object-joins may issue a query with a condition
  ;;		      corresponding to maximum MAX-LEN objects, i.e. specifies
  ;;		      how many disjuncts are allowed in
  ;;		         WHERE (join-key1=(slot-value object1 home-key1)
  ;;				AND join-key2=(slot-value object1 home-key2)...)
  ;;			    OR (join-key1=(slot-value object2 home-key1)...) ...
  ;; Value: The objects list itself or (list object) if the argument is a singleton.
  ;; Implementation Notes:
  ;; 1) The efficiency improvement reducing the number of queries is not impelemented.
  ;;    For each slot, we query the database using the home key value of each object
  ;;	separately, not all the objects in bulk,
  ;;	and then assigns the resulting value to its home slot.
  ;; 2)	CLSQL implementation does not work for compound keys.
  ;; NB: fetch-object-joins would be a better name.
  ;(assert (or (null max-len) (plusp max-len)))
  (declare (ignore max-len))
  (when (setq objects (listify objects))
    (do* ((%db-reading% t)				; Loop over slots
          (class-name (or class-name (class-name (class-of (first objects)))))
          (class (find-class class-name))
          (sds (if (eq slots t)
                   (join-slots class :deferred)		; all deferred slots by default
                   (clos:find-slot-definition slots class t))
               (rest sds)))
         ((null sds)
          objects)
      (do* ((join-slot (first sds))			; Loop over objects
            (slot-name (clos:slot-definition-name join-slot))
            (instance)
            (rest objects (rest rest)))
           ((null rest))
        (setq instance (first rest))
        (when (or force (not (slot-boundp instance slot-name)))
          (setf (slot-value instance slot-name) (join-slot-value-fault
                                                 instance join-slot refresh)))))))


#||;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; EVALUATION ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
#+deprecated
(defgeneric value-to-sql (value type &optional sd)
 (:method (value type &optional sd)
  (declare (ignore type sd))
  value)
 (:method (value (type (eql 'boolean)) &optional sd)
  (declare (ignore type sd))
  (cond ((numberp value) value)
          ;((column-nullable column) (case value ((:false nil) value) (otherwise :true)))
        (value :true)
        (:false)))
 (:method ((value integer) (type (eql 'universal-time)) &optional sd)
  ;; Used without local-time
  (declare (ignore type sd))
  (timestring value))
)

(defun slot-value-from-db (instance sd datum)
    (cond ((and value (null slot-reader))
           (setf (slot-value instance slot-name)
                 (read-sql-value value (delistify slot-type)
                                 (database instance)
				 (database-underlying-type (database instance)))))
	  ((null value)
           (update-slot-with-null instance slot-name slotdef))
          ((typep slot-reader 'string)
           (setf (slot-value instance slot-name)
                 (format nil slot-reader value)))
          ((typep slot-reader '(or symbol function))
           (setf (slot-value instance slot-name)
                 (apply slot-reader (list value))))
          (t
           (error "Slot reader is of an unusual type.")))))

(defun db-value-from-slot (database value sd)
  (let ((dbwriter (view-class-slot-db-writer slotdef))
	(dbtype (specified-type slotdef)))
    (typecase dbwriter
      (string (format nil dbwriter val))
      ((and (or symbol function) (not null)) (apply dbwriter (list val)))
      (t
       (database-output-sql-as-type (first-or-self dbtype) value database
                                    (database-underlying-type database))))))

(defmethod key-value-from-db (database value sd)
  (declare (optimize (speed 3) #+cmu (extensions:inhibit-warnings 3)))
  (let ((slot-reader (db-reader slotdef))
        (slot-type (specified-type slotdef)))
    (cond ((and value (null slot-reader))
           (read-sql-value value (first-or-self slot-type) database
			   (database-underlying-type database)))
          ((null value)
           nil)
          ((typep slot-reader 'string)
           (format nil slot-reader value))
          ((typep slot-reader '(or symbol function))
           (apply slot-reader (list value)))
          (t
           (error "Slot reader is of an unusual type.")))))

||#
