;;; -*- Mode: Lisp; -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; YstokSQL - Common Lisp ODBC interface 
;;; SQL broadcast streams for tracing
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :sql-sys)

(defvar *record-process* t "Include process information into trace.")

(defun start-sql-recording (&key (type :command) (database *default-database*))
 "Begin recording SQL command or result traffic onto stream. By
default the broadcast stream is just *standard-output*.
TYPE determines whether SQL command or result traffic is recorded, or both;
  must be either :command, :result, :all or :both, and defaults to :command.
DATABASE defaults to *default-database*"
  (case type
    ((:all :both :command) (setf (command-recording-stream database)
                                  (make-broadcast-stream *standard-output*))))
  (case type
    ((:all :both :result) (setf (result-recording-stream database)
                                 (make-broadcast-stream *standard-output*)))))

(defun stop-sql-recording (&key (type :command) (database *default-database*))
  "Stops recording of SQL command or result traffic.
TYPE determines whether to stop SQL command or result traffic, or both;
 must be either :command, :result, :all or :both, defaulting to :command.
DATABASE defaults to *default-database*"
  (case type ((:all :both :command) (nilf (command-recording-stream database))))
  (case type ((:all :both :result)  (nilf (result-recording-stream database)))))


(defun sql-recording-p (&key (type :command) (database *default-database*))
  "Returns t if recording of TYPE of SQL interaction specified is
enabled.  TYPE must be either :command, :result, :all or :any.
DATABASE defaults to *default-database*."
  (case type
    (:command	  (command-recording-stream database))
    (:result	  (result-recording-stream database))
    ((:all :both) (and (result-recording-stream database)
                       (command-recording-stream database)))
    (:any	  (or (result-recording-stream database)
		      (command-recording-stream database))) ))

(defun add-sql-stream (stream &key (type :command) (database *default-database*))
 "Add the given STREAM as a component stream for the recording
broadcast stream for the given SQL interaction TYPE.
TYPE must be either :command, :result, or :all, defaulting to :command.
DATABASE defaults to *default-database*"
  (case type
    ((:all :both :command)
     (unless (member stream (list-sql-streams :type :command :database database))
       (setf (command-recording-stream database)
              (apply #'make-broadcast-stream stream
                     (list-sql-streams :type :command :database database))))))
  (case type
    ((:all :both :result)
     (unless (member stream (list-sql-streams :type :result :database database))
       (setf (result-recording-stream database)
              (apply #'make-broadcast-stream stream
                     (list-sql-streams :type :result :database database))))))
  stream)
			      
(defun delete-sql-stream (stream &key (type :command) (database *default-database*))
 "Removes the given STREAM from the recording broadcast stream for the given TYPE
 of SQL interaction.
TYPE must be either :command, :result, or :all, defaulting to :command.
DATABASE defaults to *default-database*"
  (case type
    ((:all :both :command)
     (setf (command-recording-stream database)
	    (apply #'make-broadcast-stream
		   (remove stream (list-sql-streams :type :command
						    :database database))))))
  (case type
    ((:all :both :result)
     (setf (result-recording-stream database)
	    (apply #'make-broadcast-stream
		   (remove stream (list-sql-streams :type :result
						    :database database)))))))

(defun list-sql-streams (&key (type :command) (database *default-database*))
 "Returns the set of streams which the recording broadcast stream
send SQL interactions of the given TYPE sends data."
  (ecase type
    (:command
     (when (command-recording-stream database)
       (broadcast-stream-streams (command-recording-stream database))))
    (:result
     (when (result-recording-stream database)
       (broadcast-stream-streams (result-recording-stream database))))))


(defun record-sql-command (expr database)
  (when-bind (command-recording-stream (command-recording-stream database))
    (fresh-line command-recording-stream)
    (when *record-process*
      (format command-recording-stream "(~A) "
              (mp:process-name mp:*current-process*)))
    (format command-recording-stream "~A ~A >> ~A;"
            #-cmu (princ-to-string (get-universal-time))
            #+cmu (ext:format-universal-time nil (get-universal-time)
                                             :print-timezone nil
                                             :print-weekday nil)
            (database-name database)
            expr))
  expr)

(defun record-sql-result (res database)
  (when-bind (result-recording-stream (result-recording-stream database))
    (fresh-line result-recording-stream)
    (when *record-process*
      (format result-recording-stream "(~A) "
              (mp:process-name mp:*current-process*)))
    (format result-recording-stream "~A ~A << ~A"
            #-cmu (princ-to-string (get-universal-time))
            #+cmu (ext:format-universal-time nil (get-universal-time)
                                             :print-timezone nil
                                             :print-weekday nil)
            (database-name database)
            res))
  res)
  
(defmethod execute-command :around (expression &key (database *default-database*)
                                               &allow-other-keys)
  (record-sql-command expression database)
  (record-sql-result (call-next-method) database))

(defmethod query :around (query-expression &key (database *default-database*)
                                           &allow-other-keys)
  (record-sql-command query-expression database)
  (record-sql-result (call-next-method) database))

  
