;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2008 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; SQL-schema: Roles, users, and grant/revoke privileges
;;; NB: There is no point in defining special subclasses of sql-expressions.
#|
Exceprt from SQL-2003 (ISO/IEC 9075-2:2003)
-------------------------------------------
An SQL-session context contains a time-varying sequence of cells,
known as the _authorization stack_, each cell of which contains
either a user identifier, a role name, or both.

The term _current authorization identifier_ denotes an authorization identifier
in the top cell of the authorization stack.

Each privilege is represented by a _privilege descriptor_.
A privilege descriptor contains:
 The identification of the base table, view, column, domain, character set, 
  collation, transliteration, userdefined type, table/method pair, trigger,
  SQL-invoked routine, or sequence generator that the descriptor describes.
 The <authorization identifier> of the grantor of the privilege.
 The <authorization identifier> of the grantee of the privilege.
 Identification of the <action> that the privilege allows.
 An indication of whether or not the privilege is grantable.
 An indication of whether or not the privilege has the WITH HIERARCHY OPTION specified.

A role, identified by a <role name>, is a set of privileges defined by
the union of 
 - the privileges defined by the privilege descriptors whose grantee is that <role name>
   (..    )
 - the sets of privileges for the <role name>s defined by the role authorization
   descriptors whose grantee is the first <role name>. (???)

A role may be granted to <authorization identifier>s with a <grant role statement>.
No cycles of role grants are allowed.

The term _enabled authorization identifiers_ denotes 
the set of authorization identifiers whose members are 
 - the current user identifier, 
 - the current role name, 
 - and every role name that is contained in the current role name.

The term _applicable privileges_ for an authorization identifier A denotes the union of 
 - the set of privileges whose grantee is PUBLIC 
 - the set of privileges whose grantees are A and, if A is a role name,
   every role name contained in A.

The term _current privileges_ denotes the union of 
 - the applicable privileges for the current user identifier
 - the applicable privileges for the current role name.

NB: "Contains" is treated as a "usual" relationship among sets of privileges.
    E.g. after
	GRANT R1 TO R3; GRANT R2 TO R3;
    we say that R1 is contained in R3 and R2 is contained in R3.
|#

(in-package :ystok.sql.sys)

;;; SQL-2003:
;;;  CREATE ROLE <role name> [ WITH ADMIN <grantor> ]
;;;  Commands defining users should be specified by a specific database

(defun create-role (name &rest args &key (database *default-database*) ;if-exists
                                    &allow-other-keys)
 ;;; Create a group-like role.
  (apply #'db-create-role database name ;:if-exists if-exists
         (remove-properties args '(:database)))) ;:if-exists
 
(defmethod db-create-role (database name &rest args &key) ;admin ;if-exists
 ;;; Args: by ::= NIL | :user (means ADMIN CURRENT_USER) | :role (means ADMIN CURRENT_ROLE)
  (execute-command 
   (with-sql-string-stream ()
     (write-string "CREATE ROLE " *sql-stream*)
     (output-sql-identifier name database t)
     (when args
       (write-string " WITH " *sql-stream*)
       (%output-plist args database)))
       ;(sublis '((:by . "ADMIN") (:user . "CURRENT_USER") (:role . "CURRENT_ROLE"))
       ;        args)
   :database database))

(defun drop-role (name &key (database *default-database*))
  (db-drop-role database name)) ;:if-does-not-exists if-does-not-exists

(defmethod db-drop-role (database name) ;if-does-not-exists
  (execute-command (write-drop-exp-to-string :role name database)
                   ;(string-append "DROP ROLE "
                   ;              (write-sql-identifier-to-string name database))
                   :database database))


(defun create-user (name &rest args &key (database *default-database*)
                                         ;if-exists group password
                                    &allow-other-keys)
  ;; Args: group  A role the new user should be added to
  ;;       LOGIN means create user in PostgreSQL
  (apply #'db-create-user database name ;:if-exists if-exists
         (remove-properties args '(:database)))) ;:if-exists

(defun drop-user (name &key (database *default-database*))
  (db-drop-user database name)) ;:if-does-not-exists if-does-not-exists

(defmethod db-drop-user (database name)
  (db-drop-role database name))

;;; GRANT <role granted> [ { <comma> <role granted> }... ]
;;;	TO <grantee> [ { <comma> <grantee> }... ]
;;;	[ WITH ADMIN OPTION ]
;;;	[ GRANTED BY <grantor> ]
;;; REVOKE [ ADMIN OPTION FOR ] <role revoked> [ { <comma> <role revoked> }... ]
;;;	FROM <grantee> [ { <comma> <grantee> }... ]
;;;	[ GRANTED BY <grantor> ]
;;;	<drop behavior>
;;; <drop behavior> ::= CASCADE or RESTRICT

(defun grant-role (role to &rest args &key (database *default-database*)
                           &allow-other-keys)
  (apply #'db-grant-role database role to (remove-properties args '(:database))))

(defmethod db-grant-role (database role to &rest args)
 ;;; TODO: Convert keyword args: :admin -> WITH ADMIN OPTION, :by -> GRANTED BY
  (execute-command 
   (with-sql-string-stream ()
     (write-string "GRANT " *sql-stream*)
     (if (consp role)
         (output-sql-identifiers role database t nil)
         (output-sql-identifier role database t))
     (write-string " TO " *sql-stream*)
     (if (consp to)
         (output-sql-identifiers to database t nil)
         (output-sql-identifier to database t))
     (%output-plist args database))
   :database database))

(defun revoke-role (role from &rest args &key (database *default-database*)
                               &allow-other-keys)
  (apply #'db-revoke-role database role from (remove-properties args '(:database))))

(defmethod db-revoke-role (database role from &rest args &key admin &allow-other-keys)
 ;;; TODO: Convert keyword args: :by -> GRANTED BY
  (execute-command 
   (with-sql-string-stream ()
     (write-string "REVOKE " *sql-stream*)
     (when admin
       (write-string "ADMIN OPTION FOR " *sql-stream*))
     (if (consp role)
         (output-sql-identifiers role database t nil)
         (output-sql-identifier role database t))
     (write-string " FROM " *sql-stream*)
     (if (consp from)
         (output-sql-identifiers from database t nil)
         (output-sql-identifier from database t))
     (%output-plist (remove-properties args '(:admin)) database))
   :database database))


#|
GRANT <privileges> TO <grantee> [ { <comma> <grantee> }... ]
	[ WITH HIERARCHY OPTION ]
	[ WITH GRANT OPTION ]
	[ GRANTED BY <grantor> ]

<privileges> ::= <object privileges> ON <object name>
<object name> ::= [ TABLE ] <table name>
	| DOMAIN <domain name>
	| COLLATION <collation name>
	| CHARACTER SET <character set name>
	| TRANSLATION <transliteration name>
	| TYPE <schema-resolved user-defined type name>
	| SEQUENCE <sequence generator name>
	| <specific routine designator>

<object privileges> ::= ALL PRIVILEGES | <action> [ { <comma> <action> }... ]
<action> ::= SELECT [ <left paren> { <privilege column list> | <privilege method list> }
		     <right paren> ]
	| DELETE
	| INSERT [ <left paren> <privilege column list> <right paren> ]
	| UPDATE [ <left paren> <privilege column list> <right paren> ]
	| REFERENCES [ <left paren> <privilege column list> <right paren> ]
	| USAGE			; on domain, u-d type, charset, collation, or sequence
	| TRIGGER		; create triggers
	| UNDER			; creation subtype of structured type
	| EXECUTE		; execute an SQL-invoked routine
<privilege method list> ::= <specific routine designator>
                          [ { <comma> <specific routine designator> }... ]
<privilege column list> ::= <column name list>
<grantee> ::= PUBLIC | <authorization identifier>
<grantor> ::= CURRENT_USER | CURRENT_ROLE

REVOKE [ <revoke option extension> ] <privileges>
	FROM <grantee> [ { <comma> <grantee> }... ]
	[ GRANTED BY <grantor> ]
	<drop behavior>

<revoke option extension> ::= GRANT OPTION FOR | HIERARCHY OPTION FOR
|#

;;; Only one action privilege can be granted/revoked per call
;;; Args: privilege ::= keyword | (keyword . components)
;;;	  on ::= table-name | (schema-object-type . name)
;;;       options ::= ([:option {:grant | :hierarchy}]] [:by <grantor>]) - not implemented

(defun grant (privilege on to &rest args &key (database *default-database*)
                              &allow-other-keys)
  (apply #'db-grant database (first-or-self privilege) on to
         (nconc (when (consp privilege) (list :components (cdr privilege)))
                (remove-properties args '(:database)))))

(defun revoke (privilege on from &rest args &key (database *default-database*)
                                 &allow-other-keys)
  (apply #'db-revoke database (first-or-self privilege) on from
         (nconc (when (consp privilege) (list :components (cdr privilege)))
                (remove-properties args '(:database)))))

(defun %output-grant-revoke (database operation privilege components on to options)
 ;;; Args: operation ::= :grant | :revoke
  (with-sql-string-stream ()
    (write-string operation *sql-stream*)
    (write-char #\Space *sql-stream*)
    (output-sql privilege database)
    (when components
      (output-sql-identifiers components database t))
    (when on
      (write-string " ON " *sql-stream*)
      (when (consp on)
        (output-sql (car on) database)
        (write-char #\Space *sql-stream*)
        (setq on (cdr on)))
      (output-sql-identifier on database))
    (write-char #\Space *sql-stream*)
    (if (consp to)
        (output-sql-identifiers to database t nil)
        (output-sql-identifier to database t))
    (%output-plist (remove-properties options '(:components)) database)))

(defmethod db-grant (database privilege on to &rest options &key components
                                              &allow-other-keys)
  (execute-command (%output-grant-revoke database :grant privilege components on to 
                                         options)
                   :database database))

(defmethod db-revoke (database privilege on from &rest options &key components
                                                 &allow-other-keys)
  (execute-command (%output-grant-revoke database :revoke privilege components on from
                                         options)
                   :database database))
