;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2005 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; SQL square bracket syntax
;;; Reader macro and operators implementation

(in-package :ystok.sql.sys)

(defvar *sql-op-table* (make-hash-table :test #'equal)
  "Hashtable for mapping symbol names to sql generator functions.")

;;; Hashtable for unique identification and coalescing of SQL references.
;;; Bind to NIL in run-time dynamic context when the uniqueness undesirable.
;;;
;;; Convenience macro in-identable defaults to
;;; - Compile time:		   :clear = Clear if set or create if not.
;;; - Loading static module:	   T = don't change if has been set, otherwise create
;;; - Lazy-loading dynamic module: Do nothing because identifiers are already coalesced.

(defvar *sql-identable* nil)

(defmacro in-identable (&optional (table nil specified-p))
  `(progn
     (eval-when (:compile-toplevel)
       (setq *sql-identable* ,(if (eq (if specified-p table :clear) :clear)
                                  `#1=(if (hash-table-p *sql-identable*)
                                          (clrhash *sql-identable*)
                                          #2=(make-hash-table :test #'equal
                                              #+lispworks :weak-kind #+lispworks :value))
                                  table)))
     (eval-when (:load-toplevel :execute)
       (setq *sql-identable* ,(case (cond (specified-p table)
                             #+ys-product (ys::*compile-for-lazyload* '*sql-identable*)
                                          (t))
                                (:clear #1#)
                                ((t) `(if (hash-table-p *sql-identable*)
                                          *sql-identable* #2#))
                                (otherwise table))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Making Expressions

(defun sql-operator (operator)
 ;;; Return the proper SQL operator.
  (typecase operator
    (symbol (gethash (string-upcase (symbol-name operator)) *sql-op-table*))))
    ;(string (gethash operator *sql-op-table*))))

(defun sql-operation (operator &rest args)
 ;;; Return the proper SQL operation instance.
  (if-bind (sql-operator (sql-operator operator))
    (case sql-operator
      ((sql-comparison-exp sql-function-exp sql-aggregate-function-exp
        sql-case-exp sql-set-query-exp)
       ; sql-backend-function)
       (apply sql-operator operator args))
      (otherwise
       (apply (symbol-function sql-operator) args)))
    (error 'sql-reader-error
           :format-control "~A is not a recognized SQL operator."
           :format-arguments (list operator))))

(defun sql-expression (&rest rest &key verbatim string identifier table alias
                                       column attribute
                                       (coalescep (hash-table-p *sql-identable*))
                                  &allow-other-keys)		; type
  (cond ((or verbatim string)
         (make-instance 'sql-verbatim :text (or verbatim string)))
        ((and coalescep (hash-table-p *sql-identable*)
              (progn (remf rest :coalescep)
                (gethash rest *sql-identable*))))
        (identifier				; single indentifier of whatever kind
         (let ((value (cond ((and (symbolp identifier)
                                  (char= (schar (symbol-name identifier) 0) #\?))
                             (make-instance 'sql-parameter))
                            ((or (symbolp identifier) (stringp identifier))
                             (make-instance 'sql-ident :name identifier))
                            (t (error 'sql-reader-error
                                      :format-control "Illegal SQL identifier name ~S"
                                      :format-arguments (list identifier))))))
           (if (hash-table-p *sql-identable*)
               (setf (gethash rest *sql-identable*) value)
               value)))
        (t
         (let ((value
                (without-properties (plist rest :table :alias :column :attribute
                                                :coalescep)
                  (cond ((setq column (or column attribute))
                         (cond ((and (symbolp column)
                                     (char= (schar (symbol-name column) 0) #\?))
                                (make-instance 'sql-parameter :plist plist))
                               ((or (symbolp column) (stringp column))
                                (let ((qualifier (or table alias)))
                                  (if (imp qualifier
                                           (or (symbolp qualifier) (stringp qualifier)))
                                      (make-instance 'sql-column :name column
                                                     :qualifier qualifier :plist plist)
                                      (error 'sql-reader-error
                                 :format-control "Illegal qualifier in SQL reference ~S"
                                 :format-arguments (list qualifier)))))
                               (t (error 'sql-reader-error
                                         :format-control "Illegal SQL column name ~S"
                                         :format-arguments (list column)))))
                        ((and table (not column))
                         (cond ((not (or (symbolp table) (stringp table)))
                                (error 'sql-reader-error
                                 :format-control "Illegal SQL table name ~S"
                                 :format-arguments (list table)))
                               ((and alias (not (or (symbolp alias) (stringp alias))))
                                (error 'sql-reader-error
                                 :format-control "Illegal SQL table alias ~S."
                                 :format-arguments (list alias)))
                               (t (make-instance 'sql-ident-table :name table
                                                 :alias alias :plist plist)) ))))))
           (if (and coalescep (hash-table-p *sql-identable*))
               (setf (gethash rest *sql-identable*) value)   ; :coalescep already removed
               value))) ))

(defparameter *reference-keywords* '(:type :name :schema :catalog :as :asc :desc))

(defun generate-sql-reference (&rest args)
 ;;; Called by: sql-open-reader only
  (cond ((null args)
	 (error 'sql-reader-error :format-control "Bad SQL expression syntax."))
        ((null (rest args))	; string or single identifier (table or column)
         (let ((first (first args)))
           (if (stringp first)
               (sql-expression :verbatim first)
               (sql-expression :identifier (keep-symbol-names first)))))
        (t
         (flet ((plist (args)
                  ;; Place implicitely given column type ahead (if any)
                  ;; and complete :asc, :desc key pairs.
                  (let (end (type nil))
                    (cond ((member (first args) *reference-keywords*)
                           (setq end 0))
                          ((setq end (position-if #'keywordp (cdr args)))
                           (if (zerop end)
                               (setq type (first args) end 1)
                               (setq type (subseq args 0 (incf end)))))
                          ((rest args)
                           (setq type args))
                          (t (setq type (first args))))
                    (cond ((null end) (setq args nil))
                          ((not (zerop end)) (setq args (nthcdr end args))))
                    (when (oddp (length args)) (setq args (append args '(t))))
                    (if type (list* :type type args) args))))
           (cond ((stringp (second (keep-symbol-names args)))	    ; [foo "bar"]
                  (apply #'sql-expression :table (first args)
                                          :alias (second args)
                                          (cddr args)))		    ; no type allowed
                 ((keywordp (second args))			  
                  (if (member (second args) '(:alias :schema :catalog))
                      ;; can't schema be used in an column?
                      (apply #'sql-expression :table (first args)
                                              (cdr args))	     
                      (apply #'sql-expression :column (first args)  ; [foo :integer 5]
                                              (plist (cdr args))))) ; or [foo :desc]
                 (t
                  (apply #'sql-expression :column (second args)	    ; [foo bar :integer]
                                          :table (first args)	    ; ["foo" bar]
                                          (plist (cddr args)) )))))))
     
    
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Reader control

(defparameter *sql-macro-open-char* #\[)
(defparameter *sql-macro-close-char* #\])

(defun sql-open-reader (stream char)
 ;;; ys::*compile-for-lazyload* pinpoints during compile time how to macroexpand:
  ;;  t -> #<sql-ident ...>
  ;;  nil -> (load-time-value (generate-sql-reference '...))
  ;; Backend function can only be called via [funcall 'fn] or (sql-operation 'fn)
  (declare (ignore char))
  (let* ((expr-list (read-delimited-list *sql-macro-close-char* stream t))
         (sql-operator (sql-operator (first expr-list))))
    (cond (*read-suppress*				; suppress odd errors when parsing
           nil)						; code under conditional notations
          (sql-operator
           ;; compose a list evaluated at execution time
           (case sql-operator
             ((sql-comparison-exp sql-function-exp sql-aggregate-function-exp
                                  sql-case-exp sql-set-query-exp)
              (list* sql-operator (list 'quote (first expr-list)) (rest expr-list)))
             ;(sql-call-procedure-exp
             ; (list* sql-operator (list 'quote (first expr-list)) nil (rest expr-list)))
             ;(sql-backend-function
             ; (apply #'generate-sql-reference expr-list))
             (otherwise
              (cons sql-operator (rest expr-list)))))
           ((and #+ys-product ys::*compile-for-lazyload*)
            (apply #'generate-sql-reference expr-list))	; post-delivery lazy-loaded module
           (t
            (list 'load-time-value			; pre-delivery loaded module
                  (list* 'generate-sql-reference	; fsl-file size increase by 3-4%
                         (mapcar (lambda (expr) (list 'quote expr)) expr-list)))))))

(defun setup-sql-readmacro (&optional (readtable *readtable*))
 ;;; Set up SQL reader macros on readtable
  ;; Value: the modified readtable.
  (set-macro-character *sql-macro-open-char* 'sql-open-reader nil readtable)
  (set-macro-character *sql-macro-close-char* (get-macro-character #\))
                       nil readtable)
  readtable)

;;; NB: As SQL syntax can be intermixed with other, e.g. #L"", we makes a copy
;;;	of the current readtable, not the standard CL syntax readtabel
;;; CAUTION: The below approach could set *original-readtable* incompatibly by
;;;	     another process if LOAD is ongoing from two processes at the same time.
(defvar *sql-readtable* (setup-sql-readmacro (copy-readtable)))
(defvar *original-readtable* nil)
(defvar *global-reader-state* nil "Global state checked by restore: t if reader is on.")

(defmacro locally-enable-sql-reader-syntax ()
  `(eval-when (:compile-toplevel :load-toplevel :execute)
     (unless *original-readtable*
       (setq *original-readtable* *readtable*))
     (setq *readtable* *sql-readtable*)))

(defmacro locally-disable-sql-reader-syntax ()
  `(eval-when (:compile-toplevel :load-toplevel :execute)
     (when *original-readtable*
       (setq *readtable* *original-readtable*
             *original-readtable* nil))))

(defmacro restore-sql-reader-syntax-state ()
  `(eval-when (:compile-toplevel :load-toplevel :execute)
     (if *global-reader-state*
         (locally-enable-sql-reader-syntax)
         (locally-disable-sql-reader-syntax))))

(defun enable-sql-reader-syntax ()
 "Globally enable SQL bracket syntax."
  (locally-enable-sql-reader-syntax)
  (setq *global-reader-state* t)
  #+lispworks		 ; why can't it be that elegant in MCL and ACL?
  (progn (editor::set-vector-value
          (slot-value editor::*default-syntax-table* 'editor::table) #\[ 2)
         (editor::set-vector-value 
          (slot-value editor::*default-syntax-table* 'editor::table) #\] 3)))

(defun disable-sql-reader-syntax ()
 "Globally disable SQL bracket syntax."
  (locally-disable-sql-reader-syntax)
  (setq *global-reader-state* nil)
  #+lispworks
  (progn (editor::set-vector-value
          (slot-value editor::*default-syntax-table* 'editor::table) #\[ 7)
         (editor::set-vector-value
          (slot-value editor::*default-syntax-table* 'editor::table) #\] 7)))
#|
(defvar *original-open-reader* nil "Current local state: true if SQL reader is on.")
(defvar *original-close-reader* nil)

(defun locally-enable-sql-reader-syntax (&optional (readtable *readtable*))
  (unless *original-open-reader*
    (setq *original-open-reader* (get-macro-character *sql-macro-open-char* readtable)
          *original-close-reader* (get-macro-character *sql-macro-close-char*
                                                       readtable))
    (set-macro-character *sql-macro-open-char* 'sql-open-reader nil readtable)
    (set-macro-character *sql-macro-close-char* (get-macro-character #\))
                         nil readtable)))

(defun locally-disable-sql-reader-syntax (&optional (readtable *readtable*))
  (when *original-open-reader*
    (set-macro-character *sql-macro-open-char* *original-open-reader* nil readtable)
    (set-macro-character *sql-macro-close-char* *original-close-reader* nil readtable)
    (setq *original-open-reader* nil
          *original-close-reader* nil)))

 |#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; SQL operation definitions 

(defmacro defsql (maker (&key (symbol maker) (keep nil keep-specified)) &body body)
 "Define the same maker function as a handler for the symbol or list of symbols."
  `(progn
     ,(when body
        `(defun ,maker ,@body))
     (keep-symbol-names ',(mapcar (lambda (symbol)(intern (string symbol) #.*package*))
                                  (listify (if keep-specified keep symbol)))
                        ',maker)
     ,@(mapcar (lambda (symbol)
                 `(setf (gethash (string-upcase ',symbol) *sql-op-table*) ',maker))
               (listify symbol))))

(defsql sql-query-exp (:symbol "SELECT" :keep nil))
(defsql sql-insert-exp (:symbol "INSERT" :keep nil))
(defsql sql-update-exp (:symbol "UPDATE" :keep nil))
(defsql sql-delete-exp (:symbol "DELETE" :keep nil))
(defsql sql-create-table-exp (:symbol "CREATE-TABLE" :keep nil))

(defsql sql-any (:symbol ("ANY" "SOME")) (&rest rest)
  (make-instance 'sql-value-exp :modifier 'ANY :components rest))

(defsql sql-all (:symbol "ALL") (&rest rest)
  (make-instance 'sql-value-exp :modifier 'ALL :components rest))

(defsql sql-not (:symbol "NOT") (&rest rest)
  (make-instance 'sql-value-exp :modifier 'NOT :components rest))

;; Oracle8/10: INTERSECT == SQL-99:INTERSECT DISTINCT
;; Oracle8:  MINUS == SQL-99:EXCEPT ALL
;; Oracle10: MINUS == SQL-99:EXCEPT DISTINCT

(defun sql-set-query-exp (operator &rest args)
 "Handle cases like [union :all ...] or [intersect :distinct ...]"
  (let ((modifier (if (memq (first args) '(:all :distinct))
                      (prog1 (first args) (setq args (rest args)))
                      nil)))
    (make-instance 'sql-set-query-exp :operator operator :components args
                   :modifier modifier)))

(defsql sql-set-query-exp (:symbol ("EXCEPT" "INTERSECT" "UNION")))

;(defsql sql-group-by (:symbol "GROUP-BY") (&rest rest)
;  (make-instance 'sql-value-exp :modifier 'GROUP-BY :components rest))
;(defsql sql-having (:symbol "HAVING") (&rest rest)
;  (make-instance 'sql-value-exp :modifier 'HAVING :components rest))
;(defsql sql-limit (:symbol "LIMIT") (&rest rest)
;  (make-instance 'sql-value-exp :modifier 'LIMIT :components rest))

(defsql sql-null (:symbol "NULL" :keep (IS Null)) (&rest args)
  (if args
      (make-instance 'sql-relational-exp :operator 'IS        ; treat as predicate IS
                                         :components (list (first args) nil))
      (make-instance 'sql-value-exp :components 'Null)))      ; treat as constant Null

(defsql |NOT NULL| (:symbol "NOT-NULL" :keep |NOT NULL|) (&rest args)
  (if args
      (make-instance 'sql-relational-exp :operator 'IS        ; treat as predicate IS
                                         :components (list (first args) '|NOT NULL|))
      (make-instance 'sql-value-exp :components '|NOT NULL|)))

(defsql sql-exists (:symbol "EXISTS") (&rest rest)
  (make-instance 'sql-value-exp :modifier 'EXISTS :components rest))

(defvar sql-* (make-instance 'sql-ident :name '*))

(defsql sql-* (:symbol "*") (&rest args)
  (if (rest args)
      (make-instance 'sql-relational-exp :operator '* :components args) ; times
      sql-*)) ;(make-instance 'sql-ident :name '*)			; all cols

(defsql sql-+ (:symbol "+") (&rest args)
  (if (rest args)
      (make-instance 'sql-relational-exp :operator '+ :components args) ; n-ary
      (make-instance 'sql-value-exp :modifier '+ :components args)))         ; unary

(defsql sql-/ (:symbol "/") (&rest args)
  (make-instance 'sql-relational-exp :operator '/ :components args))

(defsql sql-- (:symbol "-") (&rest args)
  (if (rest args)
      (make-instance 'sql-relational-exp :operator '- :components args)	     ; n-ary
      (make-instance 'sql-value-exp :modifier '- :components args)))         ; unary

(defsql sql-between (:symbol "BETWEEN") (&rest args)
  (if (= (length args) 3)
      (make-instance 'sql-between-exp :operator 'BETWEEN :components args)
      (error 'sql-reader-error
             :format-control "Wrong number of arguments for 'BETWEEN'.")))
      ;(error 'maisql-sql-syntax-error
      ;       :reason "Wrong number of arguments for 'BETWEEN'.")))

(defsql sql-like (:symbol "LIKE") (arg pattern &rest plist &key escape case)
 ;;; Args: case ::= :upcase or :ignore
  (declare (ignore escape case))
  (make-instance 'sql-like-exp :operator 'LIKE
                 :components (list arg pattern) :plist plist))

;(defsql sql-upcase-like (:symbol "UPLIKE") (&rest rest)
;  (make-instance 'sql-upcase-like :components rest))

(defsql sql-and (:symbol "AND") (&rest rest)
  (make-instance 'sql-relational-exp :operator 'AND :components rest))

(defsql sql-or (:symbol "OR") (&rest rest)
  (make-instance 'sql-relational-exp :operator 'OR :components rest))

(defsql sql-in (:symbol "IN") (&rest rest)
  (make-instance 'sql-relational-exp :operator 'IN :components rest))

;; CONCAT is also an ODBC scalar function invoked by {fn CONCAT(str1,str2)} 
(defsql sql-concat-exp (:symbol ("||" "" "CONCAT") :keep '\|\|) (&rest args)
  (make-instance 'sql-concat-exp :operator '\|\| :components args))

(defsql sql-is (:symbol "IS") (&rest rest)
  (make-instance 'sql-relational-exp :operator 'IS :components rest))

;; USQL: = and == are interchanged
(defsql sql-== (:symbol "==" :keep =) (&rest rest)
  (make-instance 'sql-assignment-exp :operator '= :components rest))

(defsql sql-comparison-exp (:symbol ("=" "<" ">" "<=" ">=") :keep (= < > <= >=))
                           (operator arg1 arg2 &key case)
  (let ((components (list arg1 arg2)))
    (if case                           
        (make-instance 'sql-comparison-exp :operator operator :components components
                       :case case)
        (make-instance 'sql-relational-exp :operator operator :components components))))

(defsql sql-<> (:symbol ("/=" "<>" "!=")) (arg1 arg2 &key case)
  (let ((components (list arg1 arg2)))
    (if case                           
        (make-instance 'sql-comparison-exp :operator '<> :components components :case case)
        (make-instance 'sql-relational-exp :operator '<> :components components))))

#|
(defsql sql-= (:symbol "=") (&rest rest)
  (make-instance 'sql-relational-exp :operator '= :components rest))

(defsql sql-< (:symbol "<") (&rest rest)
  (make-instance 'sql-relational-exp :operator '< :components rest))

(defsql sql-> (:symbol ">") (&rest rest)
  (make-instance 'sql-relational-exp :operator '> :components rest))

(defsql sql->= (:symbol ">=") (&rest rest)
  (make-instance 'sql-relational-exp :operator '>= :components rest))

(defsql sql-<= (:symbol "<=") (&rest rest)
  (make-instance 'sql-relational-exp :operator '<= :components rest))|#

(defsql sql-cast-exp (:symbol "CAST" :keep nil) (arg &key as)
  (make-instance 'sql-cast-exp :modifier as :components arg))
;(defsql sql-the (:symbol "the") (&rest rest)
;  (make-instance 'sql-typecast-exp :modifier (first rest) :components (second rest)))

(defsql |CROSS JOIN| (:symbol "CROSS-JOIN" :keep |CROSS JOIN|) (arg1 arg2)
  (make-instance 'sql-join-exp :operator '|CROSS JOIN|
                               :components (list arg1 arg2)))

;;; Args: USING can be equal to :natural

(defsql |INNER JOIN| (:symbol "INNER-JOIN" :keep |INNER JOIN|)
	             (arg1 arg2 &key on using)
  (make-instance 'sql-join-exp :operator '|INNER JOIN|
                               :components (list arg1 arg2) :on on :using using))

(defsql |LEFT OUTER JOIN| (:symbol "LEFT-JOIN" :keep |LEFT OUTER JOIN|)
			  (arg1 arg2 &key on using)
  (make-instance 'sql-join-exp :operator '|LEFT OUTER JOIN|
                               :components (list arg1 arg2) :on on :using using))
(defsql |RIGHT OUTER JOIN| (:symbol "RIGHT-JOIN" :keep |RIGHT OUTER JOIN|)
			   (arg1 arg2 &key on using)
  (make-instance 'sql-join-exp :operator '|RIGHT OUTER JOIN|
                               :components (list arg1 arg2) :on on :using using))
(defsql |FULL OUTER JOIN| (:symbol "FULL-JOIN" :keep |FULL OUTER JOIN|)
			  (arg1 arg2 &key on using)
  (make-instance 'sql-join-exp :operator '|FULL OUTER JOIN|
                               :components (list arg1 arg2) :on on :using using))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; SQL Functions
;;; The 'registered SQL functions', which are defined via defsql-function,
;;; has priority over column names. Example: [count] => COUNT() 
;;; Oracle: Within a SQL statement, the names of database columns take precedence
;;;         over the names of functions with no parameters.

(defsql sql-funcall (:symbol ("FUNCALL" "APPLY" "SQL-FUNCTION") :keep nil) (fn &rest args)
 ;;; Used to call either
  ;;  - standard SQL or back-end-specific function computed on-the-fly, 
  ;;    which can either be declared with defsql-function beforehand or not.
  ;;  - or user-defined procedure or function stored in the back-end.
  ;; NB: If fn is a symbol, its name may not be kept
  (cond ((sql-operator fn)
         (apply #'sql-operation fn args))
        ((typep fn 'sql-ident)
         ;; User-defined stored procedure or function, possibly qualified.
         (apply #'sql-call-procedure-exp fn nil args))
        ;((typep fn 'sql-call-procedure-exp)	; already formed procedure expr.
        ; (setf (slot-value fn 'plist) (nconc (slot-value fn 'plist) args)
        ; (setf (slot-value fn 'args) args)	; - substitute actual args
        ; fn)
        (t
         ;(error 'sql-reader-error
         ;      :format-control "SQL function ~S is not defined with defsql-function."
         ;      :format-arguments (list fn))
         (warn 'sql-undefined-function-warning :name fn :args args)
         (apply #'sql-function-exp fn args))))

;;; Do we need the following?
(defsql sql-=call (:symbol "=CALL" :keep nil) (result fn &rest args)
  (apply #'sql-call-procedure-exp fn result args))

(defun sql-function-exp (fn &rest args)
 "Standard or semi-standard scalar (single-row) function used with traditional syntax."
  (make-instance 'sql-function-exp :name fn :args args))

(defun sql-aggregate-function-exp (fn &rest args)
 "Handle cases like [count [foo] :distinct]"
  (case (second args)
   ((:all :distinct)
    (make-instance 'sql-aggregate-function-exp :name fn :modifier (second args)
                                               :args (list (first args))))
   (otherwise
    (make-instance 'sql-aggregate-function-exp :name fn :args args))))

;(defun sql-backend-function (fn &rest args)
; "Scalar (single-row) functions and stored-procedures."
;  (make-instance 'sql-backend-function-exp :name fn :args args))

(defun sql-call-procedure-exp (fn result &rest args)
 "User-defined function or stored procedure call."
  ;; If the expression is static, keep function name.
  (when (sys:featurep :deliver) (keep-symbol-names fn 'sql-call-procedure-exp))
  (make-instance 'sql-call-procedure-exp :name fn :result result :args args))

;;; NB:           
;;;      funcall/apply?
(defmacro defsql-function ((&key aggregate procedure case) &body symbols)
 "Define a standard SQL or user-defined function resulting in a hard-wired SQL code."
  (let ((fn (cond (aggregate 'sql-aggregate-function-exp)
                  (procedure 'sql-call-procedure-exp)
                  (case 'sql-case-exp)
                  (t 'sql-function-exp))))
    `(progn
       (keep-symbol-names ',symbols ',fn)
       (setf ,@(mapcan (lambda (symbol)
                         `((gethash (string-upcase ',symbol) *sql-op-table*) ',fn))
                       symbols)))))


(defsql-function (:aggregate t) AVG COUNT FIRST LAST MAX MIN SUM)

;;; These are SQL standard; nevertheless some of them are redefined in YSQL.ODBC as scalar.
(defsql-function ()
  current-date current-time current-timestamp
  localtime localtimestamp
  substring upper lower)

(defun sql-case-exp (operator &rest args)
  (let* ((name (first (member operator '(CASE COND COALESCE) :test #'string-equal)))
         (args (case name
                 (COND (do ((acc ())
                            (rest args (cddr rest)))
                           ((endp rest) (nreverse acc))
                         (push (list (first rest) (second rest)) acc)))
                 (CASE (do ((acc (list (first args)))
                            (rest (rest args) (cddr rest)))
                           ((endp rest) (nreverse acc))
                         (push (list (first rest) (second rest)) acc)))
                 (otherwise args))))
    (make-instance 'sql-case-exp :name (or name operator) :args args)))

(defsql-function (:case t) CASE COND COALESCE)
