;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface 
;;; Copyright (c) 2002-2005 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Transaction support 

(in-package :ystok.sql.sys)

(defun start-transaction (&key (database *default-database*) access isolation)
 "Starts a transaction.
 Arguments (if unspecified or null, current connection-related values are not unchanged):
  ACCESS    One of :read-write or :read-only; 
  ISOLATION One of :uncommitted, :committed, :repeatable, or :serializable."
 ;;; CAUTION: Manipulate *commit-actions* and *rollback-actions* lists if used manually
  ;;	      outside a with-transaction form!
  (db-start-transaction database access isolation))

(defun commit (&key (database *default-database*) autocommit access isolation)
 "Commits pending transactions and restore default attribute values."
  (db-end-transaction (or database *default-environment*) :commit
                      autocommit access isolation))

(defun rollback (&key (database *default-database*) autocommit access isolation)
 "Rolls back pending transactions and restore default attribute values."
  (db-end-transaction (or database *default-environment*) :rollback
                      autocommit access isolation))

(defparameter *commit-nested-transactions-p* nil
  "If is false, transaction are committed/rolled back on exit from the outermost WITH-TRANSACTION form only")

;; Dynamic variable bound by with-transaction containing the transactions to be committed 
;; inside a with-transaction form and is only visible inside a with-transaction form.
(defvar *active-transactions* () "A list of active transactions")

;; Trasaction completion action lists
(defvar *commit-actions* ())
(defvar *rollback-actions* ())

(defun %execute-action (data)
  (if (consp data)
      (apply (car data) (cdr data))
      (funcall data)))			; or (apply data args)?

(defun %action-executer (action-list args &key post-process &allow-other-keys)
 ;;; LW analog: (lw:with-action-list-mapping (action-list item data post-process) ...)
  (declare (ignore args post-process))
  (mapc #'%execute-action action-list))

;;; Add the action EXPR within or outside the dynamic scope of with-transaction
;;; Args: name  Must be unique within the transaction
;;;		 (unused, specify the instance or the expr itself as a last resort)
;;;	  expr ::= function | symbol | (function . args)
;;;	  args  The spec arguments of lw:define-action (unused)
;;; LW analog:  `(lw:define-action *commit-actions* ,name ,expr ,@args)
(defmacro define-commit-action (name expr &rest args)
  (declare (ignore name args))
  (with-gensyms (data)
    `(let ((,data ,expr))
       (if *active-transactions*
           (setq *commit-actions* (cons ,data *commit-actions*))
           (%execute-action ,data)))))

(defmacro define-rollback-action (name expr &rest args)
  (declare (ignore name args))
  `(when *active-transactions*
     (setq *rollback-actions* (cons ,expr *rollback-actions*))))

(defmacro with-transaction ((&key (database '*default-database*) access isolation)
                                  ;#+lispworks (sort-time :define-action)
                                  ;#+lispworks dummy-actions)
                            &body body)
"If the forms in BODY execute without error, then on exit
 - first, all active transactions pertaining to DATABASE are committed,
 - second, *commit-actions* list is executed.
 If an error occurs,
 - first, all the active transactions are rolled back,
 - second, *rollback-actions* list is executed."
  ;; NB: Rebinding *active-transactions*, *commit-actions*, and *rollback-actions*
  ;;	 is actually per-thread.
  ;; LW analog:	(lw:make-unregistered-action-list
  ;;		   :execution-function '%action-executer
  ;;		   :sort-time ,sort-time
  ;;		   ,@(when dummy-actions `(:dummy-actions ,dummy-actions))
  (with-gensyms (db successp autocommit old-access old-isolation g-body)
   `(let ((,db ,database))
      (flet ((,g-body () ,@body))
        (if (and (not *commit-nested-transactions-p*)	;    
                 (find ,db *active-transactions*))	;     
            (,g-body)
            (let ((,successp nil)
                  (*active-transactions* (cons ,db *active-transactions*))
                  (*commit-actions* ())
                  (*rollback-actions* ()))
              (multiple-value-bind (,autocommit ,old-access ,old-isolation)
                  (db-start-transaction ,db ,access ,isolation)
                (unwind-protect
                    (multiple-value-prog1 (,g-body)
                      (setf ,successp t))
                  (db-end-transaction ,db (if ,successp :commit :rollback)
                                      ,autocommit ,old-access ,old-isolation)))))))))

(defmethod db-start-transaction :before (database &optional access isolation)
  (declare (ignore access isolation))
  (ensure-connected database))

;;; Can anything be checked for environment?
(defmethod db-end-transaction :before ((db-object null) completion-type
                                       &optional autocommit access isolation)
  (declare (ignore db-object completion-type autocommit access isolation))
  (signal-no-database-error))

(defmethod db-end-transaction :before ((database database) completion-type
                                       &optional autocommit access isolation)
  (declare (ignore completion-type autocommit access isolation))
  (ensure-connected database))
  ;(unless (connected database))
  ;(signal-closed-database-error database))

;;; At the end of transaction, execute actions in the order of definition
;;; LW analog: (SYSTEM::EXECUTE-ACTIONS-INTERNAL *commit-actions* ())))
;;;	      ;(lw:execute-actions (*commit-actions*))); => (load-time-value *commit-actions*)!

(defmethod db-end-transaction :after ((database database) (completion-type (eql :commit))
                                      &optional autocommit access isolation)
  (declare (ignore completion-type autocommit access isolation))
  (when *commit-actions*
    (%action-executer (nreverse *commit-actions*) ())
    (setq *commit-actions* ())))

(defmethod db-end-transaction :after ((database database) (completion-type (eql :rollback))
                                      &optional autocommit access isolation)
  (declare (ignore completion-type autocommit access isolation))
  (when *rollback-actions*
    (%action-executer (nreverse *rollback-actions*) ())
    (setq *rollback-actions* ())))

#||;;;;;;;;;;;;;;;;;;;;;;;;;;;; EVALUATION ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(setq *commit-actions* (lw:make-unregistered-action-list
                        :execution-function '%action-executer))

(lw:define-action *commit-actions* '(cons 'a 'b) '(cons 'a 'b))
(lw:define-action *commit-actions* '(cons 1 2) '(cons 1 2))
(lw:define-action *commit-actions* '(conss 1 2) '(conss 1 2))
(lw:undefine-action *commit-actions* nil)
(lw:print-actions *commit-actions*)

(lw:execute-actions (*commit-actions* :post-process :collect))

(defmethod database-start-transaction ((database closed-database))
  (signal-closed-database-error database))

(defmethod database-start-transaction (database)
  (unless database
    (error 'maisql-nodb-error))
  (with-accessors ((level transaction-level))
    database
    (incf level)
    (when (= level 1)
      (execute-command "BEGIN" :database database))))

(defmethod database-commit-transaction ((database closed-database))
  (signal-closed-database-error database))

(defmethod database-commit-transaction (database)
  (with-accessors ((level transaction-level))
    database
    (if (< 0 level)
        (progn
          (decf level)
          (when (= level 0)
            (execute-command "COMMIT" :database database)))
        (warn "Continue without commit."
              'maisql-simple-error
              :format-control "Cannot commit transaction against ~A because there is no transaction in progress."
              :format-arguments (list database)))))

(defmethod database-abort-transaction ((database closed-database))
  (signal-closed-database-error database))

(defmethod database-abort-transaction (database)
  (with-accessors ((level transaction-level))
    database
  (if (< 0 level)
      (progn
        (setf level 0)
        (execute-command "ROLLBACK" :database database))
      (warn "Continue without abort."
              'maisql-simple-error
              :format-control "Cannot abort transaction against ~A because there is no transaction in progress."
              :format-arguments (list database)))))

#+unused
(defvar *inside-transaction-p* nil "True if we are inside a with-transaction form")

(defvar *transaction-level* 0)
(defvar *transaction-id* nil)

(defvar *transaction-aborts* (make-hash-table))
(defvar *transaction-completes* (make-hash-table))

(defun on-txn-abort (fn)
  (push (cons *transaction-level* fn) (gethash *transaction-id* *transaction-aborts*)))

(defun on-txn-complete (fn)
  (if (> *transaction-level* 0)
      (push fn (gethash *transaction-id* *transaction-completes*))
      (warn "Cannot define on-txn-complete actions outside of transactions.")))

(defun run-abort-hooks ()
  (let ((remainder (remove-if (lambda (hook)
                                (< (car hook) *transaction-level*))
                              (gethash *transaction-id* *transaction-aborts*))))
    (mapcar #'(lambda (hook)
                (funcall (cdr hook)))
            (gethash *transaction-id* *transaction-aborts*))
    (setf (gethash *transaction-id* *transaction-aborts*) remainder)))
    

(defmacro with-transaction ((&key database)
			    &rest body)
  (let ((dbsym (gensym "db"))
	(transym (gensym "tran")))
    `(let ((,dbsym (or ,database *default-database*))
	   (,transym nil)
           (*transaction-id* (or *transaction-id*
                                 (gensym "txn")))
           (*transaction-level* (1+ *transaction-level*)))
      (unwind-protect
        (progn
          (database-start-transaction ,dbsym)
          (setf ,transym t)
          ,@body
          (database-commit-transaction ,dbsym)
          (setf ,transym nil))
        (if ,transym
            (progn                      ; was aborted
              (database-abort-transaction ,dbsym)
              ;; (format t "~&;; Transaction Abort, level ~d~%" *transaction-level*)
              (run-abort-hooks)
              (when (= 1 *transaction-level*)
                (remhash  *transaction-id* *transaction-aborts*)))
            (when (= 1 *transaction-level*)
              (let ((completes (gethash *transaction-id* *transaction-completes*)))
                ;; (format t "~&;; Running ~d post actions.~%" (length completes))
                (mapcar #'funcall completes)
                (remhash  *transaction-id* *transaction-completes*))))))))
||#              
