;;; -*- Mode: Lisp; -*-
;;; YstokSQL - Common Lisp ODBC interface - Test suite - Basic definitions and top-level
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;   -  ,      
;;;   ,      .

(in-package :sql.test)

(defvar *test-start-utime*)
(defvar *test-user-name* nil)

(defvar *test-skips* ())	; list of test names skipped due to known shortcomings

;; ACL tester controls
(setq test:*break-on-test-failures* nil
      test:*error-protect-tests* t
      test::*announce-test* nil)

;; *config-pathname* specifies a file containing lists of the form:
;;	(env-type {be-type | nil} connection-spec . connect-args), 
;; where connect-args is a plist of keyword arguments passed to the connect function;
;;       among them (:create-destroy T) means using the same as connection-spec.
(defvar *config-pathname*
  #+lispworks (lw:current-pathname "test" "config")
  #-lispworks (make-pathname :name "test" :type "config" :defaults (user-homedir-pathname))
)

(defgeneric test-connect (env-type be-type spec &rest args
                                                &key create-destroy &allow-other-keys)
 (:documentation "Destroy and create if indicated, set *test-user-name*, and connect"))
(defgeneric test-connection (type database))
(defgeneric test-functional (env-type database))
(defgeneric test-initialise (env-type database)
 (:documentation "Creating schema objects in the database and populating it"))
(defgeneric test-object (env-type database))
(defgeneric test-syntax (env-type database))
(defgeneric test-types (env-type database))

(defmacro incf-test (form)
 `(multiple-value-bind (result errors successes failures skips) ,form
    (declare (ignore result))
    (incf *test-errors* (or errors 0))
    (incf *test-successes* (or successes 0))
    (incf *test-unexpected-failures* (or failures 0))
    (setf *test-skips* (nconc *test-skips* (listify skips))))) ; if some return number!

(defun test-database (env-type be-type database)
  (let ((*test-skips* ())
        (*test-start-utime* (get-universal-time)))
    (format *error-output* "~2%**** Time ~A,  User ~S"
            (timestring *test-start-utime*) *test-user-name*)
    (with-tests (:name (format nil "YSQL connection ~A~@[ ~A~] ~A"
                               env-type be-type database)
                 :values (*test-skips*))
      (test-connection (or be-type env-type) database)	 ; the most specific type
      (test-types env-type database)
      (test-syntax env-type database)
      (test-initialize env-type database)
      (test-functional env-type database)
      (test-object env-type database)
      #+(or Russian) (test-collate env-type database)
      (test-mp env-type database)
      (when (eq env-type :odbc)
        (test-cursor database))

      (let ((test-end-utime (get-universal-time)))
        (format *error-output* "~&**** Skips: ~S ~:S~%**** Time ~A, seconds elapsed ~A~%"
                (length *test-skips*) *test-skips*
                (timestring test-end-utime)
                (- test-end-utime *test-start-utime*)) ))))

    ;(values (nreverse test-forms) (nreverse skip-tests))
    ;(list *test-errors* *test-successes* *test-unexpected-failures* *test-skips*)

(defun do-tests (&optional (pathname *config-pathname*))
  (allf 0 *test-errors* *test-successes* *test-unexpected-failures*)
  (setq *test-start-utime* (get-universal-time)
        *test-user-name* nil
        *test-skips* nil)
  (with-open-file (in pathname)
    (loop for tuple = (read in nil :eof)
          until (eq tuple :eof)
          do (destructuring-bind (env-type be-type spec &rest args) tuple
               (when (apply 'test-connect env-type be-type spec args)
                 (incf-test (test-database env-type be-type sql:*default-database*))
                 (disconnect)))))
  (values (and (= 0 *test-errors* *test-unexpected-failures*)
               (/= 0 *test-successes*))
          *test-errors* *test-unexpected-failures* *test-successes*
          *test-skips*
          (- (get-universal-time) *test-start-utime*)))



#||;;;;;;;;;;;;;;;;;;;;;;;;;;;;; EVALUATION ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(do-tests)

(test:test-error (parse-uri "<foo")
                 :condition-type 'parse-error :include-subtypes t)

(test:test-no-error (uri:parse-uri "http://foo.com/bar?a=zip|zop"))
(setq *test-user-name* "admin")


#+unused
  (let ((test-forms '())
	(skip-tests '()))
    (dolist (test-form (append *rt-internal* *rt-connection* *rt-basic* *rt-fddl* *rt-fdml*
			       *rt-ooddl* *rt-oodml* *rt-syntax*))
      (let ((test (second test-form)))
	(cond
	  ((and (null (clsql-sys:db-type-has-views? db-underlying-type))
		(clsql-sys:in test :fddl/view/1 :fddl/view/2 :fddl/view/3 :fddl/view/4))
	   (push (cons test "views not supported") skip-tests))
	  ((and (null (clsql-sys:db-type-has-boolean-where? db-underlying-type))
		(clsql-sys:in test :fdml/select/11 :oodml/select/5))
	   (push (cons test "boolean where not supported") skip-tests))
	  ((and (null (clsql-sys:db-type-has-subqueries? db-underlying-type))
		(clsql-sys:in test :fdml/select/5 :fdml/select/10
                              :fdml/select/32 :fdml/select/33))
	   (push (cons test "subqueries not supported") skip-tests))
	  ((and (null (clsql-sys:db-type-transaction-capable? db-underlying-type
						    *default-database*))
		(clsql-sys:in test :fdml/transaction/1 :fdml/transaction/2 :fdml/transaction/3 :fdml/transaction/4))
	   (push (cons test "transactions not supported") skip-tests))
	  ((and (null (clsql-sys:db-type-has-fancy-math? db-underlying-type))
		(clsql-sys:in test :fdml/select/1))
	   (push (cons test "fancy math not supported") skip-tests))
	  ((and (eql *test-database-type* :sqlite)
		(clsql-sys:in test :fddl/view/4 :fdml/select/10
				:fdml/select/21 :fdml/select/32 
                                :fdml/select/33))
	   (push (cons test "not supported by sqlite") skip-tests))
	  ((and (eql *test-database-type* :sqlite3)
		(clsql-sys:in test :fddl/view/4 :fdml/select/10
			      :fdml/select/21 :fdml/select/32 
			      :fdml/select/33))
	   (push (cons test "not supported by sqlite3") skip-tests))
	  ((and (not (clsql-sys:db-type-has-bigint? db-type))
		(clsql-sys:in test :basic/bigint/1))
	   (push (cons test "bigint not supported") skip-tests))
	  ((and (eql *test-database-underlying-type* :mysql)
		(clsql-sys:in test :fdml/select/26))
	   (push (cons test "string table aliases not supported on all mysql versions") skip-tests))
	  ((and (eql *test-database-underlying-type* :mysql)
		(clsql-sys:in test :fdml/select/22 :fdml/query/5 
				:fdml/query/7 :fdml/query/8))
	   (push (cons test "not supported by mysql") skip-tests))
	  ((and (null (clsql-sys:db-type-has-union? db-underlying-type))
		(clsql-sys:in test :fdml/query/6 :fdml/select/31))
	   (push (cons test "union not supported") skip-tests))
	  ((and (eq *test-database-type* :oracle)
		(clsql-sys:in test :fdml/query/8 :fdml/select/21 
                              :fddl/table/6))
	   (push (cons test "syntax not supported") skip-tests))
          ((and (eq *test-database-type* :odbc)
		(eq *test-database-underlying-type* :postgresql)
		(clsql-sys:in test :fddl/owner/1))
           (push (cons test "table ownership not supported by postgresql odbc driver") skip-tests))
	  ((and (not (member *test-database-underlying-type* 
                             '(:postgresql :oracle)))
		(clsql-sys:in test :fddl/owner/1))
           (push (cons test "table ownership not supported") skip-tests))
          ((and (null (clsql-sys:db-type-has-intersect? db-underlying-type))
		(clsql-sys:in test :fdml/query/7))
	   (push (cons test "intersect not supported") skip-tests))
          ((and (null (clsql-sys:db-type-has-except? db-underlying-type))
		(clsql-sys:in test :fdml/query/8))
	   (push (cons test "except not supported") skip-tests))
          ((and (eq *test-database-underlying-type* :mssql)
                (clsql-sys:in test :fdml/select/9))
           (push (cons test "mssql uses integer math for AVG") skip-tests))
	  (t
	   (push test-form test-forms))))))

||#
